#include "CollectDelayedHitsProcessor.h"

using namespace marlin;
using namespace lcio;

namespace marlintpc
{
 
CollectDelayedHitsProcessor aCollectDelayedHitsProcessor;
  
CollectDelayedHitsProcessor::CollectDelayedHitsProcessor() : Processor( "CollectDelayedHitsProcessor" )  
{
  _description = "Find Tracks with Delayed Hits, creates two output Cols: with and without delayed hits";
    
  registerInputCollection( LCIO::TRACK,
    "InputCollectionName" , 
    "Name of the input collection"  ,
    _inputColName ,
    std::string("TPCTracks") ) ;
				
  registerOutputCollection( LCIO::TRACK,
    "OutputCollectionNameWithoutDelayedHits" , 
    "Name of the output collection"  ,
    _outputColNameWithoutDelayedHits ,
    std::string("TracksWOdelayedHits") ) ;
		      
  registerOutputCollection( LCIO::TRACK,
    "OutputCollectionNameWithDelayedHits" , 
    "Name of the output collection"  ,
    _outputColNameWithDelayedHits ,
    std::string("TracksWithDelayedHits") ) ;
				  
  registerProcessorParameter( "SetOutputTransient" , 
    "if not 0 the output hits collection is set transient (default: 0)",
    _outputIsTransient, 
    int(0));
       
  registerProcessorParameter( "IntrinsicResolution" , 
    "Intrinsic Detector Resolution in Drift Direction [mu]",
    _intrinsicRes, 
    float(60));   
    
  registerProcessorParameter( "longDiff" , 
    "longitudinal Diffusion during drift [mu/sqrt(cm)] - cut is applied at (cutlevel x sigma)",
    _longDiff, 
    float(128)); 
    
  registerProcessorParameter( "CutLevel" , 
    "cut is applied at (cutlevel x sigma)",
    _CutLevel, 
    float(6.0));
}
  
CollectDelayedHitsProcessor::~CollectDelayedHitsProcessor() {}
  
  
void CollectDelayedHitsProcessor::init()
{

  //usual a good idea to...
  printParameters();     
   
  // a residual is at most the full drift length  
  _histoRange = 270; 
  // 100mu bins should do; residuals are positive or negative gives a factor 2 + bin 0
  _histoBin = 2*_histoRange*10+1; 
    
  _OffendingZResiduals = new TH1F("OffendingZResiduals", "OffendingZResiduals", 1000, 0, 300);
  _meanZofTracksWithOffendingZResiduals = new TH1F("_Z0ofTracksWithOffendingZResiduals", 
    "_Z0ofTracksWithOffendingZResiduals", 1000, 0, 300);
  _eventDisplayCreator = new TimePixEventDisplayCreator( 0.00275, 14.07725, 0.00275, 
    14.07725, 0.0, 265.0, 256, 256);  
}  // end init
  
  
void CollectDelayedHitsProcessor::processRunHeader(LCRunHeader *run)
{    
  run->parameters().setValue( _processorName + "_revision", "$Rev$");
  
  for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) 
  {
    if ( ! i->second->isOptional() || i->second->valueSet() ) 
    {
      run->parameters().setValue( _processorName + "_" + i->second->name(), 
        i->second->value() );
    }
  }
}  // end processRunHeader
  
  
void CollectDelayedHitsProcessor::processEvent(LCEvent *evt)
{
  // get the input collection from the event
  LCCollection* inputCol = 0;
  LCCollectionVec* outputColWithoutDelayedHits = new LCCollectionVec(LCIO::TRACK);
  LCCollectionVec* outputColWithDelayedHits = new LCCollectionVec(LCIO::TRACK);
    
  // set the transient flag as definded by the processor parameter
  outputColWithoutDelayedHits->setTransient( (_outputIsTransient != 0) );
  // the output collection is a subset of the input collection
  // only the pointers are stored, the track is not copied
  outputColWithoutDelayedHits->setSubset(true);
     
  outputColWithDelayedHits->setTransient( (_outputIsTransient != 0) );
  outputColWithDelayedHits->setSubset(true);
    
  try
  {      
    inputCol = evt->getCollection( _inputColName ) ;
    //streamlog_out(DEBUG) << "INPUT : " << inputCol->getNumberOfElements() << std::endl;
  } //end of try {
  catch(DataNotAvailableException) 
  {
    // if there is not input collection give a warning to the user
    streamlog_out(WARNING4)<<" Input collection " <<_inputColName<<" cannot be found in event "
      <<evt->getEventNumber()<<std::endl;
	
    //just return, no data to process
    return;
  }
    
  AIDAProcessor::tree(this);		// enter correct AIDA tree
	
  //int keepHistosFlag = 0;    
  int eventHasATrackWithADelayedHit = 0;   
  for(int iTrack=0; iTrack< inputCol->getNumberOfElements() ; iTrack++) 
  {
    /* getElement returns only a LCObject you have to cast it to the get the right type
     * please use dynamic_cast which will return 0 if the object cannot be casted
     */
    Track* track = dynamic_cast<Track*>( inputCol->getElementAt( iTrack ) ) ;
      
    if ( track != 0 ) 
    {
      // get a track fitter instance
      TrackFitterBase *trackFitter = TrackFitterFactory::getTrackFitter( 
        TrackFitterBase::FitterTypes::SIMPLEMINIMIZER, 0);
	
      // get hands on the hits of the track
      TrackerHitVec HitVec = track->getTrackerHits();	
	
      TH1F *histoNFitresiduals = new TH1F("histoNFitresiduals", "histoNFitresiduals",
        _histoBin, - _histoRange, _histoRange );
	
      for ( TrackerHitVec::iterator hitUnderInvestigation = HitVec.begin(); //fill histoNFitresiduals
        hitUnderInvestigation < HitVec.end(); hitUnderInvestigation++ ) 
      { 
        //calculate the distance from each hit to the track and fill in histogram
	try 
        {	    
	  EVENT::DoubleVec residuals = trackFitter->calculateResiduals(*hitUnderInvestigation, track);	    
	  histoNFitresiduals->Fill( residuals[1] );			// fills histo for all hits included
	}	    
	catch ( IntersectionCalculator::NoIntersectionException &e )
        {
	  std::cout << e.what() << std::endl;
	}	   	   	  	  
      }	// histoNFitresiduals is filled by now  
	
      // std::cout << "histoNFitresiduals is filled by now" << std::endl;
	
      // *****************************************************************************************************
      // calculate residuals for N-1 Tracks
      EVENT::DoubleVec offendingZresidualsInEvent;
      EVENT::DoubleVec MeanZofTracksWitoffendingZresidual;
      double MeanZofNminus1Track;
	
      for ( TrackerHitVec::iterator hitUnderInvestigation = HitVec.begin(); 
        hitUnderInvestigation < HitVec.end(); hitUnderInvestigation++ ) // this loops all hits on the track
      { 
        MeanZofNminus1Track=0;
	// z Pos of hitUnderInvestigation
	float zPos = (dynamic_cast<TrackerHit*>(*hitUnderInvestigation))->getPosition()[2];
	  
	// construct a new track with the hitUnderInvestigation exluded
	IMPL::TrackImpl* nMinus1Track = new IMPL::TrackImpl();
	nMinus1Track->setTypeBit(LCIO::TRBIT_HITS);   
	// fill all hits into the seed track exept the one under investigation
	for (EVENT::TrackerHitVec::iterator Hit = HitVec.begin(); Hit < HitVec.end(); Hit++) 
        {
	  if ( Hit != hitUnderInvestigation )
          {
	    nMinus1Track->addHit(*Hit);
	    MeanZofNminus1Track = MeanZofNminus1Track + (dynamic_cast<TrackerHit*>(*Hit))->getPosition()[2];
	  }
	}
	MeanZofNminus1Track = MeanZofNminus1Track / (nMinus1Track->getTrackerHits()).size();

	//std::cout << "n-1 track is build" << std::endl;
	  
	// set N-1 Track parameters  
	//if track is parameterized by y ( x = ay + b )
	if ( fabs( track->getPhi() ) > M_PI/4 || fabs( track->getPhi() ) == M_PI/4 ) 
        {
	  LinearTrackRegression::StraightLine straightlineX = LinearTrackRegression::doRegressionX(nMinus1Track);
	    
	  streamlog_out(DEBUG4)<<"straigtline: a = "<<straightlineX.slope1 <<" b = "
            <<straightlineX.intersept1<<std::endl;
	    
	  // now set the estimated values in the seed track
	  nMinus1Track->setOmega( 0 );	      
	  nMinus1Track->setPhi( M_PI/2 - atan(straightlineX.slope1) );	      
	  nMinus1Track->setD0( - straightlineX.intersept1 * sin( nMinus1Track->getPhi())  ); 
	  nMinus1Track->setTanLambda( tan( straightlineX.slope2 / fabs( straightlineX.slope2 ) 
            * acos( sqrt( straightlineX.slope1*straightlineX.slope1 + 1 ) 
            / sqrt( straightlineX.slope1*straightlineX.slope1 + 1 
            + straightlineX.slope2*straightlineX.slope2) ) ) );
	  nMinus1Track->setZ0( - (straightlineX.slope1*straightlineX.intersept1*straightlineX.slope2) 
            / (straightlineX.slope1*straightlineX.slope1 + 1) + straightlineX.intersept2);	            
        }
	//if track is parameterized by x ( y = ax + b )
	else 
        { 
	  LinearTrackRegression::StraightLine straightlineY = LinearTrackRegression::doRegressionY(nMinus1Track);
	    
	  // now set the estimated values in the seed track
	  nMinus1Track->setOmega( 0 );	
	  nMinus1Track->setPhi( atan(straightlineY.slope1) );
	  nMinus1Track->setD0( straightlineY.intersept1 * cos( nMinus1Track->getPhi())  );
	  nMinus1Track->setTanLambda( tan( straightlineY.slope2 / fabs( straightlineY.slope2 ) 
            * acos( sqrt( straightlineY.slope1*straightlineY.slope1 + 1 ) 
	    / sqrt( straightlineY.slope1*straightlineY.slope1 + 1 
            + straightlineY.slope2*straightlineY.slope2) ) ) );
	  nMinus1Track->setZ0( - (straightlineY.slope1*straightlineY.intersept1*straightlineY.slope2) 
            / (straightlineY.slope1*straightlineY.slope1 + 1) + straightlineY.intersept2);		
	}
	  
	// check if the hit is more than 5 sigma away from the corresponding track 
        //-> in this case it is most likly an outlier
	// get hands on the hits of the N-1 track
	EVENT::DoubleVec residuals;
	TrackerHitVec hitsOnNminus1Track = nMinus1Track->getTrackerHits();
	try 
        {  
	  residuals = trackFitter->calculateResiduals(*hitUnderInvestigation, nMinus1Track);
	}
	catch ( IntersectionCalculator::NoIntersectionException &e )
        {
	  std::cout << e.what() << std::endl;
	}
	if ( residuals[1]*1000 >  (_CutLevel * sqrt((_longDiff*_longDiff) * zPos/10.0 
          + _intrinsicRes*_intrinsicRes )))  // delayed hits have a positive residual
        {
	  eventHasATrackWithADelayedHit++;
	  offendingZresidualsInEvent.push_back(residuals[1]);
	  MeanZofTracksWitoffendingZresidual.push_back(MeanZofNminus1Track);
	}
      } // end for (TrackerHitVec::iterator hitsOnNminus1TrackIter = hitsOnNminus1Track.begin(); 
        //hitsOnNminus1TrackIter < hitsOnNminus1Track.end(); hitsOnNminus1TrackIter++)	
      
      //std::cout << "track was investigated" << std::endl;
      if (false && eventHasATrackWithADelayedHit > 0) 
      {
        histoNFitresiduals->Write();
      }
      delete histoNFitresiduals;
      
      if ( eventHasATrackWithADelayedHit > 0 )
      {
        double maxRes = -1;
        double ZofcorrespondingNminus1Track=-1;
        int i=0;

        for(DoubleVec::iterator offendingZresidualsInEventIterator = offendingZresidualsInEvent.begin(); 
          offendingZresidualsInEventIterator<offendingZresidualsInEvent.end(); 
          offendingZresidualsInEventIterator++)
        {
	  if (*offendingZresidualsInEventIterator > maxRes) 
          {
	    maxRes=*offendingZresidualsInEventIterator;	
	    ZofcorrespondingNminus1Track = MeanZofTracksWitoffendingZresidual[i];
	  }
	}
        _OffendingZResiduals->Fill(maxRes);
        _meanZofTracksWithOffendingZResiduals->Fill(ZofcorrespondingNminus1Track);
      }
    } // end if ( track != 0 )
  }   // end for(int iTrack=0; iTrack< inputCol->getNumberOfElements() ; iTrack++)
    
  // Sort Events to the two output collections with and without delayed hits
    
  // add tracks to the corresponding collection
  if ( eventHasATrackWithADelayedHit > 0 )
  {
    for(int iTrack=0; iTrack< inputCol->getNumberOfElements() ; iTrack++)
    {
      outputColWithDelayedHits->addElement( inputCol->getElementAt(iTrack)  );	
    }     
    //streamlog_out (DEBUG) << "track added -delayed" << std::endl;
  }
  else
  {
    for(int iTrack=0; iTrack< inputCol->getNumberOfElements() ; iTrack++)
    {
      outputColWithoutDelayedHits->addElement( inputCol->getElementAt(iTrack) );	
    }    
    //streamlog_out (DEBUG) << "track added -not delayed" << std::endl;
  }

  //add the output collection to the event
  if( !outputColWithoutDelayedHits->empty() )
  {
    evt->addCollection(outputColWithoutDelayedHits , _outputColNameWithoutDelayedHits );
    //streamlog_out(DEBUG) << "elements - not delayed " << 
    //  outputColWithoutDelayedHits->getNumberOfElements ()<< std::endl; 
  }
  else
  {
    delete outputColWithoutDelayedHits;
  }
    
  if( !outputColWithDelayedHits->empty() )
  {
    evt->addCollection(outputColWithDelayedHits , _outputColNameWithDelayedHits );
    //streamlog_out(DEBUG) << "elements - delayed " << 
    //  outputColWithDelayedHits->getNumberOfElements ()<< std::endl; 
  }
  else
  {
    delete outputColWithDelayedHits;
  }
    
    
  // ******************************************************* create eventdisplays for checking purposes
    
  if ( eventHasATrackWithADelayedHit > 0 )
  {
    TH2F *eventDisplay;
      
    std::cout << "event " << evt->getEventNumber() 
      <<" HasATrackWithADelayedHit - creating eventdisplay" << std::endl;
    
    AIDAProcessor::tree(this);
    // use an LCCollectionVec instead of LCCollection so we can use iterators later
    LCCollectionVec* inputColVec = 0;
      
    // do a try / catch in case there is an empty event
    try
    {
      inputColVec = dynamic_cast<LCCollectionVec *>(evt->getCollection( _inputColName ) );
    }
    catch (DataNotAvailableException &e)
    {
      // no input data, just return and process the next event
      return;
    }
    
    /*
    eventDisplay = _eventDisplayCreator->GetTH2F_TracksXY( inputColVec  , 
      evt->getEventNumber(), evt->getRunNumber(), 1);
    eventDisplay->Write();
    delete eventDisplay;
    eventDisplay = _eventDisplayCreator->GetTH2F_TracksXY( inputColVec  , 
      evt->getEventNumber(), evt->getRunNumber(), 0);
    eventDisplay->Write();
    delete _eventDisplay;
    */

    eventDisplay = _eventDisplayCreator->GetTH2F_TracksZX( inputColVec  , 
      evt->getEventNumber(), evt->getRunNumber() );
    eventDisplay->Write();
    delete eventDisplay;
    
    /*
    eventDisplay = _eventDisplayCreator->GetTH2F_TracksZY( inputColVec  , 
      evt->getEventNumber(), evt->getRunNumber() );
    eventDisplay->Write();
    delete eventDisplay;
    */
  }   
}  // end processEvent
  
  
void CollectDelayedHitsProcessor::check(LCEvent *evt)
{
  //std::cout << "CollectDelayedHitsProcessor::check" << evt->getEventNumber()<< std::endl;
} // end check
  
  
void CollectDelayedHitsProcessor::end()
{
  //std::cout << "CollectDelayedHitsProcessor::end" << std::endl;
  
  AIDAProcessor::tree(this);		// enter correct AIDA tree
  _OffendingZResiduals->Write();
  delete _OffendingZResiduals;   
    
  _meanZofTracksWithOffendingZResiduals->Write();
  delete _meanZofTracksWithOffendingZResiduals;
    
  delete _eventDisplayCreator;
} // end end
   
  
} // end namespace marlintpc
