#include "HitDistancesProcessor.h"

using namespace marlin ;
namespace marlintpc
{

// define the indices for the template parameter
// they are for convenience and better code readability
const unsigned int HitDistancesProcessor::_distanceIndex   = 0;

HitDistancesProcessor aHitDistancesProcessor;
    
HitDistancesProcessor::HitDistancesProcessor(): ZBinTemplateProcessor<N_1D_HISTOS_HIT_DISTANCES_PROCESSOR , 
  N_2D_HISTOS_HIT_DISTANCES_PROCESSOR>("HitDistancesProcessor"), _distanceHisto(0)
{  

  // modify processor description
  _description = "Measures the Distance of Hits in an Event" ;
  
  // register steering parameters: name, description, class-variable, default value
  registerInputCollection(    LCIO::TRACK,
    "InputTracks" , 
    "Name of the input tracks collection (default: TPCTracks)"  ,
    _tracksColName ,
    std::string("TPCTracks") ) ;
    
  registerInputCollection(    LCIO::TRACKERHIT,
    "InputHits" , 
    "Name of the input hits collection (default: TPCHits)"  ,
     _hitsColName ,
     std::string("TPCHits") ) ;

  registerProcessorParameter( "MaxDistance" , 
    "Maximum value in Distance  histogram (in mm; default: 20 )"  ,
    _maxDistance,
    double(20.0) ) ;

  // set correct names for the z binned histograms
  _histo1DBaseNames[_distanceIndex] = "hitDistancesHisto";
  _histo1DTitles[_distanceIndex] = "Distance Between two hits of an Event";
}


HitDistancesProcessor::~HitDistancesProcessor()
{
}


void HitDistancesProcessor::init()
{ 
  // usually a good idea to
  printParameters() ;

  // get the not-z-binned histograms from the AIDA processor
  _distanceHisto = AIDAProcessor::histogramFactory(this)->createHistogram1D( "hitDistancesHisto_noZBinning", 
    "Distance Between two hits of an Event", 
    (int)(_maxDistance * 1000), 0, _maxDistance);

  // set histo range and one bin per int for templated histograms
  // according to Processor parameters
  _histoRanges1D[_distanceIndex].max = _maxDistance;
  _nBins1D[_distanceIndex] = (int) _maxDistance * 1000;

  // now call ZBinTemplateProcessor's init which initialises the histograms for us
  ZBinTemplateProcessor<N_1D_HISTOS_HIT_DISTANCES_PROCESSOR, N_2D_HISTOS_HIT_DISTANCES_PROCESSOR>::init();
}


void HitDistancesProcessor::processRunHeader( EVENT::LCRunHeader* run )
{ 
  run->parameters().setValue( _processorName + "_revision", "$Rev: 3380 $");  

  for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) 
  {
    if ( ! i->second->isOptional() || i->second->valueSet() ) 
    {
      run->parameters().setValue( _processorName + "_" + i->second->name(), 
        i->second->value() );
    }
  }


}
    
void HitDistancesProcessor::processEvent( LCEvent * evt ) 
{ 
  //read input data

  // we need a LCCollectionVec to use iterators, as LCCollectio is not a std::vector
  LCCollectionVec* tracksCollection=0;
  try
  {
    tracksCollection = dynamic_cast<LCCollectionVec*>(evt->getCollection( _tracksColName )) ;
  }
  catch(DataNotAvailableException &e)
  {
    // there are no tracks
    // nevermind, just catch the exception and continue 
    streamlog_out(DEBUG0) << "no input data returning"<<std::endl;
  }

  typedef std::vector< std::pair<TrackerHit*, double> > HitDouble;
	
  // if there are tracks count them and the hits per track
  if( tracksCollection != 0 )
  {
    TrackerHitVec allHitsInEvent;

    // loop over tracks in event	
    for(LCCollectionVec::iterator it = tracksCollection->begin() ; it < tracksCollection->end(); it ++ )
    {
      Track * track = dynamic_cast<Track *>(*it) ;
      TrackerHitVec hitVector = track->getTrackerHits();
	    
      for (TrackerHitVec::iterator hitIter = hitVector.begin(); hitIter <  hitVector.end(); hitIter ++) 
      {
        allHitsInEvent.push_back( *hitIter );
      }
    } // end loop over tracks in event

    float xdistance, ydistance, distance;

    for (TrackerHitVec::iterator hitIter = allHitsInEvent.begin(); hitIter <  allHitsInEvent.end(); hitIter ++) 
    {
      TrackerHit* referenceHit = *hitIter;

      for (TrackerHitVec::iterator hitIter2 = allHitsInEvent.begin(); hitIter2 <  allHitsInEvent.end(); 
        hitIter2 ++)
      {
        TrackerHit* secondaryHit = *hitIter2;
	    
	xdistance = (secondaryHit->getPosition()[0] - referenceHit->getPosition()[0]);
	ydistance = (secondaryHit->getPosition()[1] - referenceHit->getPosition()[1]);
	    
	distance = sqrt( (xdistance*xdistance) + (ydistance*ydistance) );
	   
	if (distance !=0 )
        {
	  // fill the distance into the histogram, use z of referenceHit as z pos
	  fillHistogram( _distanceIndex, distance , referenceHit->getPosition()[2] );
	  _distanceHisto->fill( distance );
	}
      }
    }
  } // end if tracksCollection != 0
}

}// namespace 
