/** Processor to sort amplified electrons into voxel
 */
#include "ChargeDistributionProcessor.h"


//MarlinTPC
#include "libgem.h"
#include "LCObjectCopier.h"

//tpcconddata
#include "TPCConditions.h"

//lccd
#include <lccd.h>
#include <lccd/LCConditionsMgr.hh>
#include <lccd/DataFileHandler.hh>



//C++
#include <iostream>
#include <cmath>
#include <cstring>

//LCIO
#include <IO/LCWriter.h>
#include <IMPL/LCRunHeaderImpl.h>
#include <IMPL/LCEventImpl.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/SimTrackerHitImpl.h>
#include <EVENT/LCCollection.h>

//GEAR
#include <gear/TPCParameters.h>

//Marlin
#include <marlin/Global.h>



using namespace lcio ;
using namespace marlin ;
using namespace std;
using namespace tpcconddata;

namespace marlintpc{

std::list< std::pair< VoxelTPC::VoxelIndex, int > > * ChargeDistributionProcessor::_fetchedList = 0;


ChargeDistributionProcessor aChargeDistributionProcessor ;


ChargeDistributionProcessor::ChargeDistributionProcessor() : Processor("ChargeDistributionProcessor"),
							     _trackerData(0), _isFirstEvent(true)
{
  
  _description = "ChargeDistributionProcessor maps charge to padplane ..." ;
  
  
  // register steering parameters: name, description, class-variable, default value


  registerInputCollection(    LCIO::SIMTRACKERHIT,
                              "InputCollectionName" , 
			      "Name of the Input collection with drifted electrons (default: AmplifiedElectrons)"  ,
			      _inputCollectionName ,
			      string("AmplifiedElectrons") ) ;

  registerOutputCollection(   LCIO::LCGENERICOBJECT ,
                              "OutputCollectionName" ,
                              "Name of the Output GEM collection (default: DistributedElectrons)" ,
                              _outputCollectionName ,
			      string("DistributedElectrons") );

  registerProcessorParameter( "ETrans1" , 
			      "Value of Transfer Field 1 in V/cm (default: 2500)"  ,
			      _ETrans1 ,
			      float(2500) ) ;

  registerProcessorParameter( "ETrans2" , 
			      "Value of Transfer Field 2 in V/cm (default: 2500)"  ,
			      _ETrans2 ,
			      (float)2500) ;

  registerProcessorParameter( "EInd" , 
			      "Value of Induction Field in V/cm (default: 5000)"  ,
			      _EInd ,
			      (float)5000) ;

  registerProcessorParameter( "Gas" , 
			      "DriftGas (default: TDR)"  ,
			      _gas ,
			      std::string("TDR")) ;

  registerProcessorParameter( "BunchIntervall",
			      "Intervall between two bunch crossings in ns (default: 369)",
			      _bunchIntervall,
			      (double) 369);

  registerProcessorParameter("WriteOutputToStorage",
                               "If true the output collection is written (default: true)",
                               _outputIsPersistent,
                               true);

  registerProcessorParameter( "zBinLength" , 
			      "Length of a zBin in ns (default: 1)"  ,
			      _zBinLength,
			      (double) 1) ;

  registerOptionalParameter(  "TPCConditions" ,
			      "Name of collection with TPCConditions" ,
			      _tpcConditionsColName,
			      tpcconddata::TPCConditions::getDefaultColName() );

  registerOptionalParameter(  "AlternativTPCConditionsName",
			      "This is the name of newly created TPCConditions, if the original file specified by \"TPCConditions\" is not writeable (default: TPCConditionsChargeDistr)" ,
			      _altTPCConditionsName,
			      string("TPCConditionsChargeDistr") ) ;

  registerProcessorParameter( "IonisationEnergy" ,
			      "The energy in GeV needed to ionize a gas particle (default: 26.E-9 = Argon) ",
			      _ionisationEnergy,
			      26.E-9); //default value is for argon

  registerOptionalParameter(  "WriteAfterNEvents" ,
			      "Force readout of the full TPC after overlaying n events (optional)" ,
			      _writeAfterNEvents,
			      312 );

  registerProcessorParameter( "ImproveMemoryFootprint",
			      "Flag whether to use fetchVoxelCollection instead of the output collection (default:false)",
			      _improveMemoryFootprint,
			      false );
}

void ChargeDistributionProcessor::init() { 

  streamlog_out(MESSAGE2) << _description << endl;

  _tpcParameters = &(Global::GEAR->getTPCParameters());

  //read TPC parameters from gear
  _BField = Global::GEAR->getTPCParameters().getDoubleVal("BField");

  //this is for a GEM setup with 2 mm between the GEMs
  _sigma = 1./1000*sqrt( (0.2*pow( CalculateTransverseDiffusion(_ETrans1) ,2)
                         +0.2*pow( CalculateTransverseDiffusion(_ETrans2) ,2)
                         +0.2*pow( CalculateTransverseDiffusion(_EInd)    ,2)) ); // in mm

  streamlog_out(MESSAGE1) << "Diffusion (transversal): " << _sigma << endl;

  //Get a VoxelTPC
  if( _trackerData ) delete _trackerData;
  _trackerData = new marlintpc::VoxelTPC( _tpcParameters, _zBinLength );

  _isFirstEvent = true;

  printParameters();
}


 
void ChargeDistributionProcessor::processRunHeader( LCRunHeader* run) {

  run->parameters().setValue( _processorName + "_revision", "$Rev: 3124 $" );

  for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {
    if ( ! i->second->isOptional() || i->second->valueSet() ) {
      run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
    }
  }

} 

 
void ChargeDistributionProcessor::processEvent( LCEvent * evt ) {

  if( _isFirstEvent){

    LCCollectionVec *conditionsCol = 0;

    try{
      //First try to get the conditions collection with TPCConditions from 
      //the current event.
      conditionsCol = dynamic_cast<LCCollectionVec*>( evt->getCollection( _tpcConditionsColName ) );
      TPCConditions tpcCond( conditionsCol->getElementAt(0) );
    
      //Check whether object is writeable 
      try{
	//This thing throws an exception if the object is write protected
	//tpcCond.obj()->checkAccess();

	//Add amplification to TPCCondtions if collection is writeable
	tpcCond.setTransDefocussing( _sigma );

      }catch(EVENT::ReadOnlyException &){
	//If collection is not writeable an new writeable collection has
	//to be created with a new also writeable conditions object. The
	//new conditions object needs to be a deep copy of the original 
	//conditions object 
	
	conditionsCol  = new LCCollectionVec( LCIO::LCGENERICOBJECT );
	TPCConditions* newtpcCond = new TPCConditions;
	newtpcCond->setTransDiffusionCoef( tpcCond.getTransDiffusionCoef() );
	newtpcCond->setLongDiffusionCoef(  tpcCond.getTransDiffusionCoef() );
	newtpcCond->setDriftVelocity(  tpcCond.getDriftVelocity() );
	
	//Add mean amplification to the new conditions object
	newtpcCond->setAmplification( _sigma );

	//Add conditions object to collection and collection to event
	conditionsCol->addElement( newtpcCond );
	evt->addCollection( conditionsCol, _altTPCConditionsName);

	//Create a lccd::DataFileHandler 
	lccd::LCConditionsMgr::instance()->registerHandler( _altTPCConditionsName , new lccd::DataFileHandler( _altTPCConditionsName ) );
  
	// call updateEvent only for the TPCConditions handler! Don't run it on LCConditionsMgr!!!
	lccd::LCConditionsMgr::instance()->getHandler( _altTPCConditionsName )->updateEvent(  evt ) ;
    
	// The updateEvent performs an LCEvent::takeCollection, which sets the collection to transient
	// In this special case we want to write it to the file, so we have to remove the transient flag
	conditionsCol->setTransient(false);

      }// catch ReadOnlyException

    }catch( DataNotAvailableException &e ){
      //If there is no TPCConditions collection create a new one and register it 
      //with lccd.

      conditionsCol  = new LCCollectionVec( LCIO::LCGENERICOBJECT );

      TPCConditions *tpcCond = new TPCConditions;
      tpcCond->setAmplification( _sigma );

      conditionsCol->addElement( tpcCond );
      evt->addCollection( conditionsCol, _tpcConditionsColName );

      //Create a lccd::DataFileHandler 
      lccd::LCConditionsMgr::instance()->registerHandler( _tpcConditionsColName , new lccd::DataFileHandler( _tpcConditionsColName ) );

      // call updateEvent only for the TPCConditions handler! Don't run it on LCConditionsMgr!!!
      lccd::LCConditionsMgr::instance()->getHandler( _tpcConditionsColName )->updateEvent(  evt ) ;
      
   
      // The updateEvent performs an LCEvent::takeCollection, which sets the collection to transient
      // In this special case we want to write it to the file, so we have to remove the transient flag
      conditionsCol->setTransient(false);

    }// catch DataNotAvailableException

    _isFirstEvent = false;
  }// if isFirstEvent


  LCCollection* inputCol = 0; 
  int objects = 1;

  try{
    inputCol = evt->getCollection( _inputCollectionName ); 
    objects = inputCol->getNumberOfElements();
  
    streamlog_out(DEBUG4) << "object number: " << objects << std::endl;
  }catch(EventException &exception) {
    streamlog_out(DEBUG4) << "Exception: " << exception.what() << endl;
    return;
  }

  //time of this event in ns
  double eventTime = evt->getEventNumber()* _bunchIntervall;

  if( _bunchIntervall < 0 ) eventTime = 0;

  for( int i = 0; i < objects; i++)
  {  
    // prepare the sim hit for the voxelTPC
    // copy the hit before modification, the input collection should not be changed
    SimTrackerHitImpl *hit_in = LCObjectCopier::copy( dynamic_cast<SimTrackerHit*>( inputCol->getElementAt(i) ));

    if (hit_in->getPosition()[2] < 0 )
    {
      streamlog_out(DEBUG2) << "negative half tpc, skipping " << std::endl;
      delete hit_in;
      continue;
    }

    //The time of the hit is its drifttime [ns].
    hit_in->setTime( hit_in->getTime() + eventTime );

    streamlog_out(DEBUG4) << "hit has dEdx of " << hit_in->getEDep() << std::endl;

    // check whether charge is in GeV or electrons
    if (hit_in->getEDep() < 1) // obviously not in electrons
    {
      hit_in->setEDep( round( hit_in->getEDep() / _ionisationEnergy ));
    }
    // else: everything above 1 is assumed to be number of electrons. 1 GeV per hit is too much

    // does the whole smearing and management of the charges
    streamlog_out(DEBUG4) << "now put charge In Voxels..." << std::endl;
    _trackerData->putChargeInVoxels(hit_in, _sigma, 0);
 
    // delete the temporary hit we created
    delete hit_in;
  }

  // filling of the map complete, give debug output
  streamlog_out(DEBUG4) << "Event " << evt->getEventNumber()
			<< " : Size of voxel map is " << _trackerData->getSize() << std::endl;

  // check if writing out after a fixed number of events is activated
  if (parameterSet("WriteAfterNEvents"))
  {
    // if the reqested number of events is reached read out the full TPC 
    if ( evt->getEventNumber() % _writeAfterNEvents == (_writeAfterNEvents - 1) )
    {
      eventTime=-1; // negative times read out the whole map
    }
    else
    {
      return; // continue with next event, do not read out anything
    }
  }

  if(_bunchIntervall<0)
  {
    eventTime=-1;
  }

  if (  _improveMemoryFootprint )
  {
    // It is safe to do a delete since fetchVoxelList sets this to 0.
    // In case the list is not fetched this avoids a memory leak
    delete _fetchedList;
    _fetchedList = _trackerData->fetchVoxelList( eventTime );
  }
  else // no memory saving, write the output collection
  {
    LCCollectionVec* voxelVec = _trackerData->fetchVoxelCollection( eventTime );

    if (!voxelVec)
    {
      streamlog_out(DEBUG2) << "Empty voxel collection while input was not empty." << std::endl;
      return;
    }

    streamlog_out(DEBUG2) << "voxel collection has size " <<voxelVec->size() << std::endl;

    if( _outputIsPersistent )   
    {
      voxelVec->setTransient( false );
    }
    else
    {
      voxelVec->setTransient( true );
    }
  
    voxelVec->parameters().setValue("zBinLength",static_cast<float>(_zBinLength));
    evt->addCollection( voxelVec , _outputCollectionName );
  }  
}



double ChargeDistributionProcessor::CalculateTransverseDiffusion(double EField){
  const char* Gas = _gas.c_str();

  double a=0.0, b=0.0, c=0.0, d=0.0, e=0.0, f=0.0, g=0.0,h=0.0;
  if(strcmp(Gas,"TDR")==0){

    //determine fitparameter dependend on B
    a=CalculateDiffTransBParam( -106.046108535431,
				-7925.61400273791,
				6686.92350165348 ,
				-1736.42761722403,
				146.611430723786);
    b=CalculateDiffTransBParam( 1970.33307314786,
				8539.99540224613,
				-8165.5604789030,
				2185.06605798116,
				-186.64910130239);
    c=CalculateDiffTransBParam( -1310.89853174629,
				-3859.80438359763,
				4067.95823611071,
				-1112.18595370353,
				95.7332916666454);
    d=CalculateDiffTransBParam( 359.575535714393,
				906.582285713412,
				-1055.48393749909,
				294.561624999709,
				-25.5341874999711);
    e=CalculateDiffTransBParam( -47.8321031745661,
				-121.309271693408,
				156.069644444719,
				-44.4134370371213,
				3.87508333334145);
    f=CalculateDiffTransBParam( 3.08740047617949,
				9.46675825402465,
				-13.3412950000488,
				3.86233555556974,
				-0.338861666667986);
    g=CalculateDiffTransBParam( -0.0802989230166763,
				-0.406835040606432,
				0.616783563887318,
				-0.181050885184737,
				0.0159530999999581);
    h=CalculateDiffTransBParam( 0.000202925869483748,
				0.00751979708915846,
				-0.0119714449267601,
				0.00355000019899887,
				-0.000313752165998087);
  }
  if(strcmp(Gas,"P5")==0){

    a=CalculateDiffTransBParam( 1717.55091269849 ,
				-11862.4161798943,
				7645.31818055561 ,
				-1711.3409351852 ,
				131.273458333334);
    b=CalculateDiffTransBParam( -1187.37916666688,
				15041.8556111118 ,
				-10250.1270416673,
				2349.85980555573 ,
				-183.043375000017);
    c=CalculateDiffTransBParam( 749.58074206344,
				-8386.30572089888,
				5915.78454861053,
				-1385.18655092575,
				109.588270833316);
    d=CalculateDiffTransBParam( -277.356452380926,
				2393.05499206331,
				-1741.8060416665,
				415.234138888839,
				-33.2723749999953);
    e=CalculateDiffTransBParam( 58.7842686507959,
				-378.769551719536,
				284.079220138844,
				-68.8185782407264,
				5.57579791666526);
    f=CalculateDiffTransBParam( -6.89263670634272,
				33.816481256577,
				-26.0894200694113,
				6.41412967591606,
				-0.524940624999067);
    g=CalculateDiffTransBParam( 0.411706769841139,
				-1.59868090608373,
				1.26615805555469,
				-0.315568351851591,
				0.0260675833333084);
    h=CalculateDiffTransBParam( -0.00974956896824944,
				0.031175398716901,
				-0.0252933602777505,
				0.00638410037036229,
				-0.00053190499999924);
  }
  if(strcmp(Gas,"P10")==0){

    a=CalculateDiffTransBParam( 5275.68159650525 ,
				-16880.2740313074,
				12867.7142585106 ,
				-3363.51052406353,
				292.460617554278);
    b=CalculateDiffTransBParam(-5706.28654761803,
			       20937.2623412812 ,
			       -16759.4706250131,
			       4464.23652778203 ,
			       -392.358958333753);
    c=CalculateDiffTransBParam(3047.80333333317,
			       -11063.7030555512,
			       9116.48541666194,
			       -2460.36027777624,
			       217.967916666513);
    d=CalculateDiffTransBParam(-899.991821429039,
			       3046.22859523997,
			       -2579.98914583488,
			       705.574708333773,
			       -63.0282291667072);
    e=CalculateDiffTransBParam(154.97143253985,
			       -473.445552646092,
			       412.008873611581,
			       -114.223045370503,
			       10.2919291666788);
    f=CalculateDiffTransBParam(-15.3491762500078,
			       41.9985363334113,
			       -37.5006902292425,
			       10.53647404169,
			       -0.957505645835577);
    g=CalculateDiffTransBParam(0.805057706349255,
			       -1.98744670661367,
			       1.81693607777767,
			       -0.516984125925895,
			       0.0473617166666638);
    h=CalculateDiffTransBParam(-0.0172623238095152,
			       0.0389708056348575,
			       -0.0363897941666099,
			       0.0104751738888722,
			       -0.000966802499998446);
  }
  //now d_t dependend on E
  return CalculateGasParameter(a,b,c,d,e,f,g,h,0,0,0, EField);
}


double ChargeDistributionProcessor::CalculateGasParameter(double a, double b ,double c, 
                                                          double d,double e, double f,
                                                          double g, double h, double i, 
                                                          double j,double k, double EField){

  double logE=std::log(EField);

  return a+b*logE+c*pow(logE,2)+d*pow(logE,3)+e*pow(logE,4)+f*pow(logE,5)+g*pow(logE,6)
                               +h*pow(logE,7)+i*pow(logE,8)+j*pow(logE,9)+k*pow(logE,10);
}

double ChargeDistributionProcessor::CalculateDiffTransBParam(double a, double b, double c, double d,double e) {

  return a + b*_BField + c*pow(_BField,2) + d*pow(_BField,3) + e*pow(_BField,4);
}

std::list< std::pair< VoxelTPC::VoxelIndex, int > > * ChargeDistributionProcessor::fetchVoxelList()
{
  std::list< std::pair<VoxelTPC::VoxelIndex, int > > * retval = _fetchedList;
  _fetchedList = 0;
  return retval;
}

}//end of namespace marlintpc
