#include <HitDistributor.h>

//C++
#include <cmath>
#include <complex>

//MARLIN
#include <streamlog/streamlog.h>


namespace marlintpc{

  //Constructor
  HitDistributor::HitDistributor(const gear::TPCParameters* tpcParameters, 
                                 double zBinLength) {

    _tpcParameters = tpcParameters;
    _zBinLength    = zBinLength;
    _padLayoutType = _tpcParameters->getCoordinateType();
  }

  HitDistributor::~HitDistributor() {}

// Steering function. The only function accessible from the outside.
// Takes a hit for input and calls all other functions for the smearing
// depending on the pad layout of the readout plane.
std::list<TPCVoxel> HitDistributor::distributeHit(SimTrackerHit* hit, 
						          double sigmaTrans, 
                                                          double sigmaLong)
{
    //Hit(coordinate of electron) values in cartesian coordinates
    _xHit = hit->getPosition()[0];
    _yHit = hit->getPosition()[1];

    //Distribute charge of the hit according to the pad layout type (polar or rectangular geometry)
    switch( _padLayoutType ){

    case gear::PadRowLayout2D::POLAR: {
    streamlog_out(DEBUG4) << "distributeHit: POLAR" << std::endl;

    std::list<GlobalPadIndex> padIndexList = getPadsPolar( sigmaTrans );
    streamlog_out(DEBUG4) << "padIndexList size: " << padIndexList.size() << std::endl;

    if( padIndexList.empty() )
    { 
       streamlog_out(DEBUG4) << "list empty!!!" << std::endl;
       return std::list<TPCVoxel>();
    }
    else{
       return distributeHitOnPadsPolar( padIndexList, hit, sigmaTrans, sigmaLong );
    }

      break;
    }

    case gear::PadRowLayout2D::CARTESIAN: {
      std::list<GlobalPadIndex> padIndexList = getPadsRect( sigmaTrans );
      if( padIndexList.empty() )
	return std::list<TPCVoxel>();
      else
        return distributeHitOnPadsRect( padIndexList, hit, sigmaTrans, sigmaLong );
      break;
    }

    default:
      throw gear::Exception("Unknown pad layout type");
    };

    return std::list<TPCVoxel>();    
  }

//Gets all pads within the smearing radius and stores their indices 
//in list for circular pad layout.
std::list<GlobalPadIndex> HitDistributor::getPadsPolar( double sigmaTrans )
{
    //the list of all pads which has charged
    std::list<GlobalPadIndex> padIndexList;

    //put hit to polar coordinate
    std::complex<double> compHit( _xHit, _yHit );
    double rHit   = std::abs( compHit );
    double phiHit = std::arg( compHit );
    if( phiHit < 0 ) phiHit = ( 2 * M_PI )+phiHit;

    //get global index of the nearest pad
    GlobalPadIndex hitIndex(-1,-1);
    try{
       gear::GlobalPadIndex ghitIndex = _tpcParameters->getNearestPad(rHit, phiHit);
       hitIndex.setPadIndex(ghitIndex.getPadIndex());
       hitIndex.setModuleID(ghitIndex.getModuleID());

       streamlog_out(DEBUG4) << "rHit: " << rHit << "  phiHit:" << phiHit
                             << "  mod:" << hitIndex.getModuleID()
                             << "  padindex:" << hitIndex.getPadIndex() << std::endl;
    }catch(std::out_of_range& e){
      return padIndexList;
    }

    //Checks whether the distance between the given pad and the hit is lower than the smearing range
    if( checkDistancePolar( hitIndex, sigmaTrans ) ){
      padIndexList.push_back( hitIndex );
      getPadsOnTheLeftPolar ( padIndexList, hitIndex, sigmaTrans );
      getPadsOnTheRightPolar( padIndexList, hitIndex, sigmaTrans );
      streamlog_out(DEBUG4) << "push back a hitIndex, the list size is: " << padIndexList.size() << std::endl;
    }else{
      return padIndexList;
    }

    //step up
    GlobalPadIndex cursor = hitIndex;
    const gear::TPCModule & tpcModule = getModuleOfPad(hitIndex);
    int   row = getRowNumberOfPad(hitIndex);

    while(1) {
      row++;

      if(row < tpcModule.getNRows()){
	try{
          double newrHit = tpcModule.getPadCenter(tpcModule.getPadIndex(row,0))[0];
	  cursor.setPadIndex(tpcModule.getNearestPad(newrHit, phiHit));
	}catch(std::out_of_range& e){
	  break;
	}

	if(checkDistancePolar(cursor, sigmaTrans)) {
          padIndexList.push_back( cursor );
	  getPadsOnTheLeftPolar ( padIndexList, cursor, sigmaTrans );
	  getPadsOnTheRightPolar( padIndexList, cursor, sigmaTrans );
      streamlog_out(DEBUG4) << "going up, now padIndex list size: " << padIndexList.size() << std::endl;
	}
        else
	  break;
      }
      else
	break;
    }

    //step down 
    cursor = hitIndex;
    //tpcModule = getModuleOfPad(hitIndex);
    row = getRowNumberOfPad(hitIndex);

    while(1){
      row--;
      if(row >= 0){
	try{
          double newrHit = tpcModule.getPadCenter(tpcModule.getPadIndex(row,0))[0];
          cursor.setPadIndex(tpcModule.getNearestPad(newrHit, phiHit));
	}catch(std::out_of_range& e){
	  break;
	}

	if(checkDistancePolar( cursor, sigmaTrans )){
          padIndexList.push_back( cursor );
	  getPadsOnTheLeftPolar ( padIndexList, cursor, sigmaTrans );
	  getPadsOnTheRightPolar( padIndexList, cursor, sigmaTrans );
          streamlog_out(DEBUG4) << "going down, now padIndex list size: " << padIndexList.size() << std::endl;
	}else
	  break;
      }
      else
	break;
    }

    return padIndexList;
  }

// Gets all pads within the smearing radius and stores their indices 
// in list for rectangular pad layout. 
std::list<GlobalPadIndex> HitDistributor::getPadsRect( double sigmaTrans )
{
    std::list<GlobalPadIndex> padIndexList;
    GlobalPadIndex hitIndex(-1,-1);

    try{
       gear::GlobalPadIndex ghitIndex = _tpcParameters->getNearestPad( _xHit , _yHit );
       hitIndex.setPadIndex(ghitIndex.getPadIndex());
       hitIndex.setModuleID(ghitIndex.getModuleID());
    }catch(std::out_of_range& e){
      return padIndexList;
    }

    if( checkDistanceRect( hitIndex, sigmaTrans ) ){

      padIndexList.push_back( hitIndex );
      getPadsOnTheLeftRect(  padIndexList, hitIndex, sigmaTrans );
      getPadsOnTheRightRect( padIndexList, hitIndex, sigmaTrans );

    }else
      return padIndexList;

    // step up
    GlobalPadIndex cursor = hitIndex;
    const gear::TPCModule & tpcModule = getModuleOfPad(hitIndex);
    int row = getRowNumberOfPad(hitIndex);

    while(1) {
      row++;
      if( row < tpcModule.getNRows() ) {
	try{
          double new_xHit = tpcModule.getPadCenter(tpcModule.getPadIndex(row,0))[1];
          cursor.setPadIndex(tpcModule.getNearestPad(_xHit, new_xHit));
	}catch(std::out_of_range& e){
	  break;
	}
	if( checkDistanceRect( cursor, sigmaTrans) ) {
          padIndexList.push_back( hitIndex );
	  getPadsOnTheLeftRect(  padIndexList, hitIndex, sigmaTrans );
	  getPadsOnTheRightRect( padIndexList, hitIndex, sigmaTrans );
	}else
	  break;
      }else
	break;
    }

    cursor = hitIndex;

    // step down
    while(1) {
      row--;
      if( row >= 0 ){
	try{
          double new_xHit = tpcModule.getPadCenter(tpcModule.getPadIndex(row,0))[1];
          cursor.setPadIndex(tpcModule.getNearestPad(_xHit, new_xHit));
	}catch(std::out_of_range& e){
	  break;
	}
	if( checkDistanceRect( cursor, sigmaTrans ) ){
	  padIndexList.push_back( cursor );
	  getPadsOnTheLeftRect(  padIndexList, cursor, sigmaTrans );
	  getPadsOnTheRightRect( padIndexList, cursor, sigmaTrans );
	}else
	  break;
      }else
	break;
    }

    return padIndexList;
  }

  // Gets the indices of all pads in the row, which are left from the given pad 
  // and wthin the smearing radius, and adds them to the list which contains
  // all pad indices of hit. (for circular pad layout)
  void HitDistributor::getPadsOnTheLeftPolar(std::list<GlobalPadIndex>& padList, 
						        GlobalPadIndex  centerIndex, 
							               double sigmaTrans)
  {
    GlobalPadIndex np = centerIndex;

    while(1){
      const gear::TPCModule & tpcModule = getModuleOfPad(np);
 
      int index;
      try
      {
        index = tpcModule.getLeftNeighbour(np.getPadIndex());
      }
      catch(std::exception& e)
      {
        streamlog_out(DEBUG4) << "left pad is not available..." << std::endl;
        return;
      }

      np.setPadIndex( index );
      if( checkDistancePolar(np, sigmaTrans) && !(np == centerIndex) )
      {
        streamlog_out(DEBUG4) << "left is OK ..." << std::endl;
	padList.push_back(np);   
      }  
      else 
      {
        streamlog_out(DEBUG4) << "left is not OK ..." << std::endl;
        return;
      }
    }
  }

  // Gets the indices of all pads in the row, which are left from the given pad 
  // and wthin the smearing radius, and adds them to the list which contains
  // all pad indices of hit. (for rectangular pad layout)
  void HitDistributor::getPadsOnTheLeftRect( std::list<GlobalPadIndex> & padList, 
                                             GlobalPadIndex centerIndex, double sigmaTrans){
    GlobalPadIndex np = centerIndex;

    while(1){
      const gear::TPCModule & tpcModule = getModuleOfPad(np);
      int index;
      
      try{
        index = tpcModule.getLeftNeighbour( np.getPadIndex() );
      }catch(std::exception& e){   //exception handling has to be adjusted when the new gear version is installed!
	return;
      }

      np.setPadIndex( index );
      if( checkDistanceRect( np, sigmaTrans ) )
	padList.push_back( np );
      else
	return;
    }
  }

  // Gets the indices of all pads in the row, which are right from the given pad 
  // and wthin the smearing radius, and adds them to the list which contains
  // all pad indices of hit. (for circular pad layout)
  void HitDistributor::getPadsOnTheRightPolar( std::list<GlobalPadIndex>& padList, 
							   GlobalPadIndex centerIndex, 
							                 double sigmaTrans)
  {
    GlobalPadIndex np = centerIndex;

    while(1){
      const gear::TPCModule & tpcModule = getModuleOfPad(np);
      int index = 0;

      try
      {
        index = tpcModule.getRightNeighbour(np.getPadIndex());
      }
      catch(std::exception& e)
      {
        streamlog_out(DEBUG4) << "right pad is not available..." << std::endl;
        return;
      }

      np.setPadIndex( index );

      if( checkDistancePolar(np, sigmaTrans) && !(np == centerIndex) )
      {
        streamlog_out(DEBUG4) << "right OK ..." << std::endl;
        padList.push_back(np);
      }
      else
      {
        streamlog_out(DEBUG4) << "right not OK ..." << std::endl;
        return;
      }
    }
  }

  //  Gets the indices of all pads in the row, which are right from the given pad 
  //  and wthin the smearing radius, and adds them to the list which contains
  //  all pad indices of hit. (for rectangular pad layout)
  void HitDistributor::getPadsOnTheRightRect(std::list<GlobalPadIndex>& padList, GlobalPadIndex centerIndex, double sigmaTrans){
    GlobalPadIndex np = centerIndex;

    while(1){
      const gear::TPCModule & tpcModule = getModuleOfPad(np);
      int index = 0;

      try{
        index = tpcModule.getRightNeighbour(np.getPadIndex());  
      }catch(std::exception& e){
	return;
      }

      np.setPadIndex( index );

      if( checkDistanceRect( np, sigmaTrans) )
	padList.push_back( np );
      else
	return;
    }
  }

  const gear::TPCModule & HitDistributor::getModuleOfPad(GlobalPadIndex globalPadIndex)
  {
    return _tpcParameters->getModule( globalPadIndex.getModuleID() );
  }

  int HitDistributor::getRowNumberOfPad(GlobalPadIndex globalPadIndex)
  {
    const gear::TPCModule & tpcModule = getModuleOfPad(globalPadIndex);
    return tpcModule.getRowNumber(globalPadIndex.getPadIndex());
  }

  // Checks whether the distance between the given pad and the hit is lower than the smearing 
  // range. Returns true when any part of the pad is within the smearing radius, even when it 
  // is only a little corner of the pad. (for circular pad layout)
  bool HitDistributor::checkDistancePolar(GlobalPadIndex globalPadIndex, double sigmaTrans )
  {
    static const double SMEAR = 4;

    //FIXME::
    //every time should calculate!
    //
    //global coordinate of hit
    std::complex<double> hit(_xHit, _yHit);
    double rHit = std::abs(hit);
    double phiHit = std::arg(hit);

    const gear::TPCModule & tpcModule = getModuleOfPad(globalPadIndex);
    streamlog_out(DEBUG4) << "rHit: " << rHit << "  phiHit: " << phiHit 
                          << "  distance:" 
                          << tpcModule.getDistanceToPad(rHit, phiHit, globalPadIndex.getPadIndex())
                          << std::endl;

    // add 1e-6 mm = 1 nm for rounding errors in case of sigmaTrans = 0
    if( tpcModule.getDistanceToPad(rHit, phiHit, globalPadIndex.getPadIndex() ) 
         <= (SMEAR*sigmaTrans + 1e-6) )
    {
      streamlog_out(DEBUG4) << "< Distance" << std::endl;
      return true;
    }
    else
    {
      streamlog_out(DEBUG4) << "> Distance" << std::endl;
      return false;
    }

    return false;
  }

  // Checks whether the distance between the given pad and the hit is lower than the smearing 
  // range. Returns true when any part of the pad is within the smearing radius, even when it 
  // is only a little corner of the pad. (for rectangular pad layout)
  bool HitDistributor::checkDistanceRect(GlobalPadIndex globalPadIndex, double sigmaTrans)
  {
    static const double SMEAR = 4;

    const gear::TPCModule & tpcModule = getModuleOfPad(globalPadIndex);

    // add 1e-6 mm = 1 nm for rounding errors in case of sigmaTrans = 0
    if ( tpcModule.getDistanceToPad(_xHit, _yHit, globalPadIndex.getPadIndex()) <= (SMEAR*sigmaTrans + 1e-6) )
    {
      return true;
    }

    return false;
  }

  //Smears the distributes the charge of the given hit in voxels according to the given pads, 
  //transversal and longitudinal diffusion for circular pad layout. 
  std::list<TPCVoxel> HitDistributor::distributeHitOnPadsPolar( std::list<GlobalPadIndex> padIndexList, 
                                                                                 SimTrackerHit* hit, 
								                         double sigmaTrans,
                                                                                         double sigmaLong )
  {
    streamlog_out(DEBUG4) << std::endl << "In distributeHitOnPadsPolar()..." << std::endl;
    std::list<TPCVoxel> voxelList;

    if( hit->getEDep() > 0 ) {

      while(padIndexList.begin()!=padIndexList.end()) {

        GlobalPadIndex gpadIndex = padIndexList.front();

        const gear::TPCModule & tpcModule = getModuleOfPad(gpadIndex);

        //calculate the x,y coordinates of the pad center
        //tpcModule.getPadCenter(tpcModule.getPadIndex(row,0))[0]

        gear::Vector2D position = tpcModule.getPadCenter(gpadIndex.getPadIndex());
        double sinPhi = sin(position[1]);
        double cosPhi = cos(position[1]);
        double xPad = position[0]*cosPhi;
        double yPad = position[0]*sinPhi;

	streamlog_out( DEBUG4 ) << "get pad coordinates " << std::endl;

        //rotate coordinate system so that the new x,y axis are parallel to pad widht,height
        //FIXME::rotation?

        double xHitRot =   _xHit*sinPhi 
                         - _yHit*cosPhi;

        double yHitRot =    _xHit*cosPhi 
                          - _yHit*sinPhi;

        double xPadRot =    xPad*sinPhi
                          - yPad*cosPhi;

        double yPadRot = +  xPad*cosPhi
                          - yPad*sinPhi;

	streamlog_out( DEBUG4 ) << "performed coordinate transformation " << std::endl;
      
        double halfPadPitch  = position[0]*tpcModule.getPadPitch(gpadIndex.getPadIndex());
        double halfRowHeight = tpcModule.getRowHeight(tpcModule.getRowNumber(gpadIndex.getPadIndex()))/2;

	streamlog_out(DEBUG4) << "calculated pad heigth and width: " << halfRowHeight << "  " << halfPadPitch << std::endl;

       //start and end values for the integration
	double xEnd   = xPadRot + halfPadPitch;
	double xStart = xPadRot - halfPadPitch;
	double yEnd   = yPadRot + halfRowHeight;
	double yStart = yPadRot - halfRowHeight;

        // integration over a 2d gaussian
        double charge = hit->getEDep()/4 *((erf((xEnd  - xHitRot)/(sqrt(2.)*sigmaTrans))
					   -erf((xStart- xHitRot)/(sqrt(2.)*sigmaTrans)))
					  * (erf((yEnd  - yHitRot)/(sqrt(2.)*sigmaTrans))	
					     -erf((yStart- yHitRot)/(sqrt(2.)*sigmaTrans))));
	//	if (charge < 0.5) streamlog_out(MESSAGE) << "Charge is " <<charge << ". Integration will result in 0" << std::endl;

	streamlog_out(DEBUG4) << "performed integration" << std::endl;

	if ( charge > 0.5 )
	{
	  int zBin = static_cast<int>( hit->getTime()  /_zBinLength ); 

	  if( sigmaLong ){

	    const double tDiff = 4;

	    int minBin = static_cast<int>( (hit->getTime() - (tDiff*sigmaLong)) / _zBinLength  );
	    int maxBin = static_cast<int>( ceil( (hit->getTime() + (tDiff*sigmaLong)) / _zBinLength  ) );

	    if( minBin < 0 ) minBin = 0;

	    for(int j = minBin; j <= maxBin; j++){ 
	      
	    int tDiffCharge = static_cast<int> ( round( charge / 2  * ( erf(((j+1)*_zBinLength-hit->getTime())/(sqrt(2.)*sigmaLong))-
									erf(( j   *_zBinLength-hit->getTime())/(sqrt(2.)*sigmaLong))  )));
	    
	    voxelList.push_back( TPCVoxel(gpadIndex.getPadIndex(), j, tDiffCharge, gpadIndex.getModuleID()) );
	    }
	  }// if sigmaLong
	  else{ 
	    voxelList.push_back(TPCVoxel(gpadIndex.getPadIndex(), zBin,static_cast<int>(round(charge)),
					 gpadIndex.getModuleID() ));
            streamlog_out(DEBUG4) << "current voxellist size: " << voxelList.size() << std::endl;
	  }
	}// if charge > 0.5

	// always erase the iterator, even if no charge has been deposited
	padIndexList.erase( padIndexList.begin() );
      }
    }

    return voxelList;
  }

  // Smears the distributes the charge of the given hit in voxels according to the given pads, 
  //transversal and longitudinal diffusion for rectangular pad layout. 
  std::list<TPCVoxel> HitDistributor::distributeHitOnPadsRect( std::list<GlobalPadIndex> padIndexList, 
                                                               SimTrackerHit* hit, 
                                                               double sigmaTrans, 
                                                               double sigmaLong ) {
    std::list<TPCVoxel> voxelList;

    if( hit->getEDep() > 0 ){

      //integration for all the stored pad indices
      while(padIndexList.begin()!=padIndexList.end()) {
        GlobalPadIndex gpadIndex = padIndexList.front();
        const gear::TPCModule & tpcModule = getModuleOfPad(gpadIndex);
   
	//calculate the width of the pad (with pitch!)
	int row = tpcModule.getRowNumber( gpadIndex.getPadIndex() );
	double halfPadPitch = tpcModule.getPadPitch( gpadIndex.getPadIndex() )/2;
	if( tpcModule.getPadsInRow( row ).size() >= 2) {
	  int one = tpcModule.getPadIndex( row, 0);
	  int two = tpcModule.getPadIndex( row, 1);
 	  halfPadPitch = fabs( tpcModule.getPadCenter(one)[0] - tpcModule.getPadCenter(two)[0] ) /2.;
	}

        //start and end values for the integration
        double xStart, xEnd, yStart, yEnd; 
        if( tpcModule.getPadCenter( gpadIndex.getPadIndex() )[0] < _xHit ) {
          xStart = tpcModule.getPadCenter( gpadIndex.getPadIndex() )[0]+ halfPadPitch;
    	  xEnd   = tpcModule.getPadCenter( gpadIndex.getPadIndex() )[0]- halfPadPitch;
        }else {
          xStart = tpcModule.getPadCenter( gpadIndex.getPadIndex() )[0]- halfPadPitch;
	  xEnd   = tpcModule.getPadCenter( gpadIndex.getPadIndex() )[0]+ halfPadPitch;
        }  
        if( tpcModule.getPadCenter( gpadIndex.getPadIndex() )[1]<hit->getPosition()[1]) {
          yStart = tpcModule.getPadCenter( gpadIndex.getPadIndex() )[1]+tpcModule.getRowHeight(tpcModule.getRowNumber(gpadIndex.getPadIndex()))/2;
	  yEnd   = tpcModule.getPadCenter(gpadIndex.getPadIndex())[1]-tpcModule.getRowHeight(tpcModule.getRowNumber(gpadIndex.getPadIndex()))/2;
        }else {
	  yStart = tpcModule.getPadCenter(gpadIndex.getPadIndex())[1]-tpcModule.getRowHeight(tpcModule.getRowNumber(gpadIndex.getPadIndex()))/2;
	  yEnd   = tpcModule.getPadCenter(gpadIndex.getPadIndex())[1]+tpcModule.getRowHeight(tpcModule.getRowNumber(gpadIndex.getPadIndex()))/2;
        }

        // integration over a 2d gaussian
        double charge= hit->getEDep()/4 *((erf((xEnd  - _xHit)/(sqrt(2.)*sigmaTrans))
					   -erf((xStart- _xHit)/(sqrt(2.)*sigmaTrans)))
					  * (erf((yEnd  - _yHit)/(sqrt(2.)*sigmaTrans))	
					     -erf((yStart- _yHit)/(sqrt(2.)*sigmaTrans))));


	if ( charge > 0.5 ) // only perform integration if there can be charge after integration and rounding
	{
	  int zBin = static_cast<int>( hit->getTime()  /_zBinLength ); 
	  
	  if( sigmaLong )
	  {
	    const double tDiff = 4;

	    int minBin = static_cast<int>( (hit->getTime() - (tDiff*sigmaLong)) / _zBinLength  );
	    int maxBin = static_cast<int>( ceil( (hit->getTime() + (tDiff*sigmaLong)) / _zBinLength  ));

	    if( minBin < 0 )
	      minBin = 0;

	    for(int j = minBin; j <= maxBin; j++)
	    { 
	      
	      int tDiffCharge = static_cast<int> ( round (charge * ( erf(((j+1)*_zBinLength-hit->getTime())/(sqrt(2.)*sigmaLong))-
								     erf(( j   *_zBinLength-hit->getTime())/(sqrt(2.)*sigmaLong))  )));


	      voxelList.push_back( TPCVoxel(gpadIndex.getPadIndex(), j, tDiffCharge, gpadIndex.getModuleID() ) );
	    }
	  }
	  else // if sigmalong
	  { 
	    voxelList.push_back(TPCVoxel(gpadIndex.getPadIndex(), zBin, static_cast<int>(round(charge)),
					 gpadIndex.getModuleID() )); 
	  }
	} // if ( charge > 0.5 ) 

	padIndexList.erase(padIndexList.begin());
      }// while  padIndexList
    }
    return voxelList;
  }


} //namespace marlintpc
