/** Processor to calculate
 */
#include "InhEFieldCalculationProcessor.h"

//C++
#include "iostream"
#include "cmath"
#include "libgem.h"
#include "set"

//LCIO
#include "IO/LCWriter.h"
#include "IMPL/LCRunHeaderImpl.h"
#include "IMPL/LCEventImpl.h"
#include "IMPL/LCCollectionVec.h"
#include "EVENT/LCCollection.h"

//GEAR
#include "gear/TPCParameters.h"

//MARLIN
#include "marlin/Global.h"

//MARLINTPC
#include "TPCVoxel.h"


using namespace lcio ;
using namespace marlin ;
using namespace std;

namespace marlintpc
{

  InhEFieldCalculationProcessor aInhEFieldCalculationProcessor ;


  InhEFieldCalculationProcessor::InhEFieldCalculationProcessor() : Processor("InhEFieldCalculationProcessor"), _rectPL(0), _circPL(0)
  {

    _description = "InhEFieldCalculationProcessor maps charge to padplane ..." ;


    // register steering parameters: name, description, class-variable, default value


    registerInputCollection(LCIO::LCGENERICOBJECT,
                            "InputCollectionName" ,
                            "Name of the Input collection with drifted electrons"  ,
                            _inputCollectionName ,
                            string("IonOccupancy")) ;

    registerOutputCollection(LCIO::LCGENERICOBJECT ,
                             "OutputCollectionName" ,
                             "Name of the Output GEM collection" ,
                             _outputCollectionName ,
                             string("InhEFieldMap"));

    registerProcessorParameter("Transient" ,
                               "switch off writing of output collection"  ,
                               _transient,
                               (int) 0) ;

    registerProcessorParameter("xBins" ,
                               "Number of bins in x direction"  ,
                               _xBins,
                               (int) 2) ;

    registerProcessorParameter("yBins" ,
                               "Number of bins in y direction"  ,
                               _yBins,
                               (int) 2) ;

    registerProcessorParameter("zBins" ,
                               "Number of bins in z "  ,
                               _zBins,
                               (int) 2) ;

  }

  void InhEFieldCalculationProcessor::init()
  {

    streamlog_out(MESSAGE2) << _description << endl;

    printParameters() ;

    _padLayout = &(Global::GEAR->getTPCParameters().getPadLayout());

    switch (_padLayout->getPadLayoutType())
      {
        case gear::PadRowLayout2D::CARTESIAN:
        {
          _rectPL = dynamic_cast<gear::RectangularPadRowLayout const*>(_padLayout);
          break;
        }
        case gear::PadRowLayout2D::POLAR:
        {
          _circPL = dynamic_cast<gear::FixedPadSizeDiskLayout const*>(_padLayout);
          break;
        }
        default:
          throw gear::Exception("Unknown pad layout type");
      };

    /** Determine the length of the x,y,z-bins
     */
    double tpcOuterRadius = Global::GEAR->getTPCParameters().getDoubleVal("TPCRadius");
    double maxDriftLength = Global::GEAR->getTPCParameters().getMaxDriftLength();

    _xBinLength = (2 * tpcOuterRadius) / _xBins;
    _yBinLength = (2 * tpcOuterRadius) / _yBins;
    _zBinLength = maxDriftLength / _zBins;
  }




  void InhEFieldCalculationProcessor::processRunHeader(LCRunHeader* run)
  {

    run->parameters().setValue(_processorName + "_revision", "$Rev: 3308 $");

    for (ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++)
      {
        if (! i->second->isOptional() || i->second->valueSet())
          {
            run->parameters().setValue(_processorName + "_" + i->second->name(), i->second->value());
          }
      }

  }


  void InhEFieldCalculationProcessor::processEvent(LCEvent * evt)
  {

    LCCollection* ions = NULL;

    try
      {
        ions = evt->getCollection(_inputCollectionName);
      }
    catch (EventException &exception)
      {
        streamlog_out(WARNING3) << "Exception: " << exception.what() << endl
        << "Event will not be processed!"    << endl;
        return;
      }

    //set output collection
    LCCollectionVec* inhEField = new LCCollectionVec(LCIO::LCGENERICOBJECT);

    //set write mode on/off
    if (_transient)
      inhEField->setTransient(true);

    set<EFieldVector*> gridPoints;
    set<EFieldVector*>::iterator iter = gridPoints.begin();

    for (int i = 0; i < _xBins; i++)
      {

        for (int j = 0; j < _yBins; j++)
          {

            for (int k = 0; k < _zBins; k++)
              {

                inhEField->addElement(calculateEField(i, j, k, ions));
              }
          }
      }


    //add output collection to event
    evt->addCollection(inhEField , _outputCollectionName);

    //if(evt->getEventNumber()%100==0)
    streamlog_out(MESSAGE3) << "Event " << evt->getEventNumber() << " done!" << endl;

  }


  void InhEFieldCalculationProcessor::check(LCEvent * evt)
  {

  }


  void InhEFieldCalculationProcessor::end()
  {

  }

  /** Calculates the EField in the TPC at the given grid point in units of
   *  V/cm.
   */
  EFieldVector* InhEFieldCalculationProcessor::calculateEField(int x, int y, int z, LCCollection* ions)
  {

    const double epsilonNull = 8.854187817 * pow(10. , -12.);
    const double eCharge = 1.60217733 * pow(10., -19.);

    const float constant = static_cast<float>((eCharge / (4 * M_PI * epsilonNull)) * pow(10., 4.)) ;

    double xPos = x * _xBinLength;
    double yPos = y * _yBinLength;
    double zPos = z * _zBinLength;

    float Ex = 0;
    float Ey = 0;
    float Ez = 0;

    for (int i = 0; i < ions->getNumberOfElements(); i++)
      {

        TPCVoxel* myVoxel = dynamic_cast<TPCVoxel*>(ions->getElementAt(i));

        //voxel coordinates
        double xVoxel;
        double yVoxel;
        double zVoxel = myVoxel->getTimeIndex() * 2.5;

        // setting of x,y coordinates of the voxel depending on the pad layout type
        switch (_padLayout->getPadLayoutType())
          {
            case gear::PadRowLayout2D::CARTESIAN:
            {
              xVoxel = _padLayout->getPadCenter(myVoxel->getPadIndex())[0];
              yVoxel = _padLayout->getPadCenter(myVoxel->getPadIndex())[1];
              break;
            }
            case gear::PadRowLayout2D::POLAR:
            {
              double radius = _padLayout->getPadCenter(myVoxel->getPadIndex())[0];
              double phi    = _padLayout->getPadCenter(myVoxel->getPadIndex())[1];
              xVoxel = radius * cos(phi);
              yVoxel = radius * sin(phi);
              break;
            }
            default:
              throw("Unknown pad layout type!");
          };

        double dist = sqrt(pow((xPos - xVoxel), 2) + pow((xPos - xVoxel), 2) + pow((zPos - zVoxel), 2));

        Ex += (myVoxel->getCharge() / pow(dist, 3)) * (xPos - xVoxel);
        Ey += (myVoxel->getCharge() / pow(dist, 3)) * (yPos - yVoxel);
        Ez += (myVoxel->getCharge() / pow(dist, 3)) * (zPos - zVoxel);
      }

    EFieldVector* myEFieldVec = new EFieldVector(x, y, z, constant*Ex, constant*Ey, constant*Ez);

    return myEFieldVec;
  }
}
