//*****************************************************
// Part of program to simulate a TPC with GEM readout *
//                                                    *
// Author: Astrid Muennich                            *
//                                                    *
//   Astrid Muennich                                  *
//   RWTH Aachen, III. Phys. Inst. B                  * 
//   Office: 28 A 112, Phone: +49/(0)241/8            *
//   Email: muennich@physik.rwth-aachen.de            * 
//                                                    *
//*****************************************************

/** Programm to apply electronic effects like shaping and binning of chareg into ADCCounts.
 */

#include "TPCElectronicsProcessor.h"
#include "ChargeDistributionProcessor.h"

//MarlinTPC
#include "TPCVoxel.h"
#include "VoxelTPC.h"

//tpcconddata
#include "Pedestal.h"

//LCIO
//#include <IO/LCWriter.h>
#include <EVENT/LCCollection.h>
#include <IMPL/TrackerRawDataImpl.h>
#include <IMPL/LCEventImpl.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/LCRunHeaderImpl.h>
#include <IMPL/LCFlagImpl.h>


//Marlin
#include <marlin/Global.h>
#include <marlin/Exceptions.h>

//lccd
#include <lccd.h>
#include <lccd/LCConditionsMgr.hh>
#include <lccd/DataFileHandler.hh>

//C++
#include <iostream>
#include <cmath>
#include <set>
#include <cstdlib>


using namespace lcio ;
using namespace marlin ;
using namespace std;
using namespace tpcconddata;

namespace marlintpc{


TPCElectronicsProcessor aTPCElectronicsProcessor;

TPCElectronicsProcessor::TPCElectronicsProcessor() : Processor("TPCElectronicsProcessor") 
{
  
  _description = "TPCElectronicsProcessor takes care of binning the times and charges of electrons into an ADC.";
  
  registerInputCollection(    LCIO::LCGENERICOBJECT,
                              "InputCollectionName" , 
			      "Name of the Input collection with TPCVoxels (default: DistributedElectrons)"  ,
			      _inputColName,
			      std::string("DistributedElectrons") ) ;

  registerOutputCollection(   LCIO::TRACKERRAWDATA,
                              "OutputCollectionName",
                              "Name of the output collection with adc values (default: TPCRawData)",
                              _outputColName,
                              std::string("TPCRawData") ),
  
  registerProcessorParameter( "ReadoutFrequency" , 
			      "Readout Frequency in Mhz (default: 40.)"  ,
			      _frequency ,
			      40.) ;
  
  registerProcessorParameter( "RiseTime" , 
			      "Rise Time of Shaper in ns (default: 60)"  ,
			      _riseTime ,
			      60.) ;
  
  registerProcessorParameter( "ADCMaxValue" , 
			      " Dynamic range of the ADC in ADC counts (default 1023)"  ,
			      _ADCMaxValue ,
			      1023) ;
  
   registerProcessorParameter( "NElectronsDynamicRange",
			       "Dynamic range of the ADC in electrons (default 1e6)",
			       _nElectronsDynamicRange,
			       int(1e6));
  
  registerProcessorParameter( "ADCThreshold",
			      "Threshold for ADC in ADC bins (default: 1, just zero suppression)",
			      _ADCThreshold,
			      1);

  registerProcessorParameter( "MinimumPulseHeight",
			      "Minimal value of the highest sample in a pulse (default: 1)",
			      _minimumPulseHeight,
			      1);

//   registerProcessorParameter( "IntCutOff",
//			       "Integration Cut Off of Gaussian Shape (default: 3.0)",
//			       _cutOff,
//			       (double)3.0);

   registerProcessorParameter( "VoxelZLength",
			       " Length of a voxel in ns (default 1.)",
			       _voxelZLength,
			       1.);

    registerProcessorParameter("WriteOutputToStorage",
                               "Output collection will be written if set to true (default: true)",
                               _outputIsPersistent,
                               true);

    registerProcessorParameter( "ImproveMemoryFootprint",
				"Flag whether to use ChargeDistributionProcessor::fetchVoxelCollection instead of the input collection (default:false)",
				_improveMemoryFootprint,
				false );
}

TPCElectronicsProcessor::~TPCElectronicsProcessor()
{
  // delete the pulse which was allocated in init()
  delete pulse;
}

void TPCElectronicsProcessor::init()
{
  streamlog_out(MESSAGE2) << _description << endl;
  printParameters() ;

  pulse = new ADCPulse(_frequency, _riseTime, _nElectronsDynamicRange, _ADCMaxValue, _ADCThreshold);
}


void TPCElectronicsProcessor::processRunHeader( LCRunHeader* run) 
{ 
  run->parameters().setValue( _processorName + "_revision", "$Rev: 3055 $" );

  for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {
    if ( ! i->second->isOptional() || i->second->valueSet() ) {
      run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
    }
  }
} 


void TPCElectronicsProcessor::processEvent( LCEvent * evt ) {

  streamlog_out(DEBUG2) << "Hi, it's me again "<< evt->getEventNumber() << std::endl;

  //Sort TPCVoxels by padIndex first, then by time index
  std::list< std::pair<VoxelTPC::VoxelIndex, int> >* padSortedVoxels;

  if (_improveMemoryFootprint)
  {
    streamlog_out(DEBUG4) << "fetching voxel list" << std::endl;
    padSortedVoxels = ChargeDistributionProcessor::fetchVoxelList();

    if (padSortedVoxels == 0)
    { // there was no voxel list in this event
      streamlog_out(DEBUG4) << "No voxel list in ChargeDistributionProcessor in event "<<
	evt->getEventNumber() << std::endl;
      return;
    }

    streamlog_out(DEBUG4) << "size of voxel list is " << padSortedVoxels->size()
			  << std::endl;    
  }
  else
  {
    LCCollectionVec* inputColVec = 0;
    
    try {
      inputColVec = dynamic_cast<LCCollectionVec *>(evt->getCollection( _inputColName ) );
    }catch (DataNotAvailableException &e) {

      // the collection of electron clouds with that name wasn't in the LCIO event file, skip simulating this event
      //    throw marlin::SkipEventException( this );
   
      return; // just return, otherwise the data from the previous processors is not written since the 
      //  OutputProcessor is also skipped
    }
    
    streamlog_out(DEBUG4) << "Size of input collection is "<< inputColVec->size()  << std::endl;
    if ( inputColVec->empty() )
    {
      streamlog_out(WARNING) << "Input collection exists, but is empty!" << std::endl;
      return;
    }

    padSortedVoxels = new  std::list< std::pair<VoxelTPC::VoxelIndex, int> >;
    
    for( LCCollectionVec::iterator voxelIter = inputColVec->begin() ; 
	 voxelIter < inputColVec->end() ; voxelIter++ )
    {
      TPCVoxel voxel(*voxelIter);

      padSortedVoxels->push_back( std::pair< VoxelTPC::VoxelIndex, int> (
				  VoxelTPC::VoxelIndex(voxel.getPadIndex(), voxel.getTimeIndex(), voxel.getModuleID() ),
				  voxel.getCharge() ) );
    }
  }

  
  //Initialise and set transient for output collection
  LCCollectionVec* outputColVec = new LCCollectionVec( LCIO::TRACKERRAWDATA )  ;
  if( !_outputIsPersistent )
  {
    outputColVec->setTransient(true);
  }

  LCFlagImpl flag(0);
  flag.setBit( LCIO:: TRAWBIT_ID1);
  outputColVec->setFlag( flag.getFlag() );

  // As the list in a non-sorted container we have to sort manually.
  // The output of the VoxelTPC is sorted by time, so we have to resort
  padSortedVoxels->sort( voxelComp );

  // initialise the pulse with the first pad index
  if( !padSortedVoxels->empty()) pulse->clear( padSortedVoxels->begin()->first.padIndex,
					       padSortedVoxels->begin()->first.moduleID );

  int nPulsesWhichHaveBeenSplit = 0;
  int nPulsesAfterSplitting = 0;

  for( std::list< std::pair<VoxelTPC::VoxelIndex, int> >::iterator voxelIter 
	 = padSortedVoxels->begin(); voxelIter != padSortedVoxels->end() ; voxelIter++)
  {
    if ( ( voxelIter->first.padIndex != pulse->getChannelNumber() )
	 || ( voxelIter->first.moduleID != pulse->getModuleID() )
	 // The charge in this voxel belongs to a new pulse, since it is on another channel

	 || ( (pulse->getStartBin()>=0) && 
	      (voxelIter->first.zBin*_voxelZLength > pulse->getEndTime() ) ) )
         // The pulse is not empty and the voxel time is after the end time of the pulse,
         // so they don't overlap
    {
      // Write the old pulse, then initialise a new one
      std::list<TrackerRawData *> lcioPulses = pulse->getLCIORawData();
      for ( std::list<TrackerRawData *>::iterator pulseIter = lcioPulses.begin();
	    pulseIter != lcioPulses.end(); pulseIter++ ) // loop the list
      {
	//check for minimum pulse height
	short pulseMaxium = 0;
	for ( ShortVec::const_iterator sampleIter = (*pulseIter)->getADCValues().begin();
	      sampleIter < (*pulseIter)->getADCValues().end(); sampleIter++ )
	{
	  if ( *sampleIter > pulseMaxium )
	  {
	    pulseMaxium =  *sampleIter;
	  }
	}

	if ( pulseMaxium >=  _minimumPulseHeight)
	{
	  outputColVec->addElement( *pulseIter );
	  streamlog_out(DEBUG2) << "Pulse length is " << (*pulseIter)->getADCValues().size() 
			      << std::endl;
	}
	else
	{
	  streamlog_out(DEBUG3) << "Rejected pulse on channel " << (*pulseIter)->getCellID0()
				<< " in module "<< (*pulseIter)->getCellID1()
				<< ", pulse height only " << pulseMaxium
				<< std::endl;
	}
      }

      if ( lcioPulses.empty() )
      {
	streamlog_out(DEBUG2) << "empty pulse arrived in processor" << std::endl;
      }

      if ( lcioPulses.size() >1  )
      {
	nPulsesAfterSplitting += lcioPulses.size();
	nPulsesWhichHaveBeenSplit++;
      }

      pulse->clear( voxelIter->first.padIndex, voxelIter->first.moduleID );
    }
    
    pulse->add( voxelIter->first.zBin*_voxelZLength ,  voxelIter->second );

  }// for voxelIter

  delete padSortedVoxels;

  //Add ouput collection to event
  if ( outputColVec->empty() ) 
    delete outputColVec;
  else
    evt->addCollection( outputColVec , _outputColName) ;

  streamlog_out(DEBUG4) << nPulsesWhichHaveBeenSplit << " pulses have been split in event "
			<< evt->getEventNumber() << ", resulting in "
			<< nPulsesAfterSplitting << " pulses." << std::endl;
}


}//end of namespace marlintpc
