//CLASS
#include <VoxelTPC.h>

//MARLIN
#include "streamlog/streamlog.h"


using namespace marlin ;

namespace marlintpc{

  VoxelTPC::VoxelTPC(gear::TPCParameters const * tpcParameters , double lengthOfZBin){

    _lengthOfZBin     = lengthOfZBin;
    _tpcParameters    = tpcParameters;

    // Instantiate HitDistributor.
    _distributeHitPtr = new HitDistributor( _tpcParameters, _lengthOfZBin );
  }

  VoxelTPC::~VoxelTPC(){
  
    delete _distributeHitPtr;
  }

  LCCollectionVec* VoxelTPC::fetchVoxelCollection( double outputLength ) {

    streamlog_out(DEBUG4) << "In fetchVoxelCollection()" << std::endl;
    if(_voxelMap.empty()){
      streamlog_out(DEBUG4) << "no entries in voxel map!" << std::endl;
      return 0;
    }
    else
    {
      streamlog_out(DEBUG4) << "voxelmap size: " << _voxelMap.size() << std::endl;
    }

    /* Instanciate the output collection.
     */
    float zBinLength=_lengthOfZBin;
    LCCollectionVec* voxelPtr = new LCCollectionVec(LCIO::LCGENERICOBJECT);
    voxelPtr->parameters().setValue("zBinLength",zBinLength);

    /* All voxels with a zBin smaller or equal to this bin will be added to the collection.
     */
    int outputZBin;

    /* Get the zBin according to the given output length. If output length is set
     *  to zero set outputZBin to the highest zBin in the map in order to get all
     *  entries. 
     */
    if( outputLength >= 0 )
      outputZBin = getZBin( outputLength );      
    else 
      outputZBin = _voxelMap.rbegin()->first.zBin;   
    
    streamlog_out(DEBUG4) << "outputZBin is: " << outputZBin << std::endl;
    /* Add all voxels with zBin smaller or equal to the output zBin to the output collection.
     *  Delete all voxels added to the collection from the map.
     */
    for(std::map<VoxelIndex, int>::iterator iter = _voxelMap.begin();
	(iter->first.zBin <= outputZBin) && iter != _voxelMap.end(); 
	/*do not increase here, iterator is already erased and invalid*/ )
      {
	TPCVoxel* myVoxel = 
	  new TPCVoxel(iter->first.padIndex, iter->first.zBin, iter->second, iter->first.moduleID);
        streamlog_out(DEBUG4) << "a Voxel!" << "  ZBin: " << iter->first.zBin << std::endl;
        
	 voxelPtr->addElement( myVoxel);
	// the iter++ is evaluated befor the erase function call, so the iterator is still valid
	_voxelMap.erase(iter++);
      }

        streamlog_out(DEBUG4) << "end of fetchVoxelCollection()" << std::endl;
    return voxelPtr;
  }

  LCCollectionVec* VoxelTPC::getVoxelCollection( double outputLength )  {

    if(_voxelMap.empty()){
      streamlog_out(DEBUG4) << "no entries in voxel map!" << std::endl;
      return 0;
    }

    /* Instanciate the output collection.
     */
    LCCollectionVec* voxelPtr = new LCCollectionVec(LCIO::LCGENERICOBJECT);

    float zBinLength=_lengthOfZBin;
    voxelPtr->parameters().setValue("zBinLength",zBinLength);

    /* All voxels with a zBin smaller or equal to this bin will be added to the collection.
     */
    int outputZBin;

    /* Get the zBin according to the given output length. If output length is set
     *  to zero set outputZBin to the highest zBin in the map in order to get all
     *  entries. 
     */
    if(outputLength >= 0)
      outputZBin = getZBin( outputLength );
    else 
      outputZBin = _voxelMap.rbegin()->first.zBin;

    /* Add all voxels with zBin smaller or equal to the output zBin to the output collection.
     */
    for(std::map<VoxelIndex, int>::iterator iter = _voxelMap.begin();
	(iter->first.zBin <= outputZBin) && iter != _voxelMap.end();
	++iter)
      {
	TPCVoxel* myVoxel = new TPCVoxel(iter->first.padIndex, 
                                         iter->first.zBin, iter->second, iter->first.moduleID);
	voxelPtr->addElement( myVoxel);	
      }
    return voxelPtr;
  }

  std::list< std::pair< VoxelTPC::VoxelIndex, int > > * VoxelTPC::getVoxelList( double  outputLength , bool erase )
  {
    std::map< VoxelIndex, int >::iterator endOfList;

    if ( outputLength >= 0 )
    {
      // the bin corresponding to outputLength is included.
      // So for the endZBin, which is not included, we have to add 1
      int endZBin = getZBin( outputLength ) + 1;

      // the lower_bound works because the voxels are sorted in time and then in pads
      endOfList = _voxelMap.lower_bound( VoxelIndex(0, endZBin, 0));
    }
    else
    {
      endOfList = _voxelMap.end();
    }

    // create the list and fill it
    std::list< std::pair< VoxelIndex, int > > *returnList = new std::list< std::pair< VoxelIndex, int > >;

    for (  std::map< VoxelIndex, int >::iterator mapIter =  _voxelMap.begin(); 
	   mapIter != endOfList; /* do not increase here because of the erase */)
    {
      returnList->push_back( *mapIter );

      // To keep the memory footprint low erase each item immediately after copying.
      // The perfomance penalty should be small compared to swapping ;-)
      if (erase)
      {
	_voxelMap.erase( mapIter++ );
      }
      else
      {
	++mapIter;
      }
    }

    return returnList;
  }

  std::list< std::pair< VoxelTPC::VoxelIndex, int > > *  VoxelTPC::fetchVoxelList( double  outputLength )
  {
    return getVoxelList( outputLength, true );
  }

  /* Puts charges in voxels according to a the input SimTrackerHit. Smearing is done in the x,y-plane
   *  according to transversal diffusion and if wanted also in z according to longitudinal diffusion
   */
  void VoxelTPC::putChargeInVoxels(SimTrackerHit *hit, double sigmaTrans, double sigmaLong ) {

    std::list<TPCVoxel> voxelList = _distributeHitPtr->distributeHit( hit, sigmaTrans, sigmaLong);
    streamlog_out(DEBUG4) << "putChargeInVoxels: voxelList size:" << voxelList.size() << std::endl;

    for ( std::list<TPCVoxel>::iterator voxelIter = voxelList.begin() ;
	  voxelIter != voxelList.end() ; ++voxelIter ) 
    {
      putChargeInVoxel( *voxelIter );
    }
  }


  void VoxelTPC::putChargeInVoxel(TPCVoxel const &voxel){

    if(voxel.getCharge()) // only perform this if the charge is not 0 to save memory
    {
      std::map<VoxelIndex, int>::iterator voxelIter = 
	_voxelMap.find( VoxelIndex( voxel.getPadIndex(), voxel.getTimeIndex(), voxel.getModuleID() ) );
      
      if ( voxelIter == _voxelMap.end() )
      { 
        // the voxel does not exist, assign it (ints are not initialised to 0,so adding won't work)
	_voxelMap[ VoxelIndex( voxel.getPadIndex(), 
                               voxel.getTimeIndex(),
			       voxel.getModuleID())
                 ] = voxel.getCharge();
      }
      else 
      {
	// voxel exists, add the charge
	voxelIter->second += voxel.getCharge();
      }
    }

  }

  void VoxelTPC::putChargeInVoxel(int padIndex, double zLength, int charge, int moduleID) {
    int zBin = getZBin( zLength );
    if(charge)
      putChargeInVoxel( TPCVoxel(padIndex, zBin, charge, moduleID) );
  }

  TPCVoxel VoxelTPC::getVoxel(int padIndex, int zBin, int moduleID) {
    std::map<VoxelIndex, int>::iterator voxelIter 
      = _voxelMap.find( VoxelIndex(padIndex, zBin, moduleID) );

    if ( voxelIter!=_voxelMap.end() )
    {  
      return  TPCVoxel( padIndex, zBin, voxelIter->second, moduleID ) ;
    }
    else 
    {
      // there is no voxel in the map, return an empty voxel
      return TPCVoxel( padIndex, zBin , 0, moduleID );
    }
  }


  std::vector<TPCVoxel> VoxelTPC::getVoxelsOnPad(int padIndex, int moduleID)  {

    streamlog_out(DEBUG2)<<"----------------------------"<<std::endl;
    streamlog_out(DEBUG2)<<"VoxelTPC::getVoxelsOnPad "<<std::endl;

    std::vector<TPCVoxel> aTPCVoxelVec;

    int maxTimeBin = _voxelMap.rbegin()->first.zBin;

    streamlog_out(DEBUG1) << "Maximum zBin: " << maxTimeBin << std::endl;

    streamlog_out(DEBUG1) << " map.size() = " << _voxelMap.size()<<std::endl;
    
    streamlog_out(DEBUG1) <<" zBin begin: "<< _voxelMap.begin()->first.zBin
			  <<" end: "<< _voxelMap.rbegin()->first.zBin<<std::endl;

    streamlog_out(DEBUG1) <<" padIndex begin: "<< _voxelMap.begin()->first.padIndex
			  <<" end: "<< _voxelMap.rbegin()->first.padIndex<<std::endl;
    
    for(int i = _voxelMap.begin()->first.zBin; 
	i <= _voxelMap.rbegin()->first.zBin && _voxelMap.begin() != _voxelMap.end(); 
	i++ ){

      std::map<VoxelIndex,int>::iterator it = _voxelMap.find( VoxelIndex(padIndex,i, moduleID) );
      
      if ( i%10000 == 0 ) 
	  streamlog_out(DEBUG0)<<"i = "<<i<<std::endl;

      if(it != _voxelMap.end())
        aTPCVoxelVec.push_back( TPCVoxel(padIndex,i,it->second, moduleID) );
    }

    streamlog_out(DEBUG2)<<"VoxelTPC::getVoxelsOnPad  done!! "<<std::endl;
    streamlog_out(DEBUG2)<<"----------------------------"<<std::endl;

    return(aTPCVoxelVec);
  }


  std::vector<TPCVoxel> VoxelTPC::getVoxelsWithZBin(int zBin){

    std::vector<TPCVoxel> aTPCVoxelVec;

    for(std::map<VoxelIndex,int>::iterator iter = _voxelMap.lower_bound(VoxelIndex(0,zBin,0));
	*iter < *_voxelMap.lower_bound(VoxelIndex(0, zBin+1, 0)); ++iter)
      aTPCVoxelVec.push_back(TPCVoxel(iter->first.padIndex, zBin, iter->second, iter->first.moduleID));
      
    return(aTPCVoxelVec);
  }

  void VoxelTPC::addCollection(LCCollection* voxelCol){
    if(voxelCol==0)
      return;

    double colZBinLength=voxelCol->getParameters().getFloatVal("zBinLength");
  
    for(int i =0; i < voxelCol->getNumberOfElements(); i++) {
      TPCVoxel myVoxel(dynamic_cast<LCGenericObject*>( voxelCol->getElementAt(i) ));
      double zLength = myVoxel.getTimeIndex()*colZBinLength;
      putChargeInVoxel(myVoxel.getPadIndex(),zLength,myVoxel.getCharge(),myVoxel.getModuleID());
    }
  }


  void VoxelTPC::removeVoxel(TPCVoxel const &voxel) {

    removeVoxel(voxel.getPadIndex() , voxel.getTimeIndex(), voxel.getModuleID());    
  }


  void VoxelTPC::removeVoxel(int padIndex, int zBin, int moduleID) {

    _voxelMap.erase(VoxelIndex(padIndex, zBin, moduleID));
  }    


  void VoxelTPC::removeVoxels(std::vector<TPCVoxel> const &voxelVec){
    for(unsigned int i=0; i<voxelVec.size(); i++) 
      removeVoxel(voxelVec[i].getPadIndex(), voxelVec[i].getTimeIndex(), voxelVec[i].getModuleID());
  }


  void VoxelTPC::removeVoxelsOnPad(int padIndex, int moduleID) {

    for( std::map<VoxelIndex, int>::iterator iter = _voxelMap.lower_bound(VoxelIndex(padIndex,0,moduleID)) ; 
	 *iter < *_voxelMap.lower_bound(VoxelIndex(padIndex+1,0,moduleID));
	 /* do not increment here, iterater already erased and invalid*/ )
    {
      // it's save to increment here:
       _voxelMap.erase(iter++);
    }
  }


  void VoxelTPC::removeVoxelsWithZBin(int zBin) {

    for( std::map<VoxelIndex, int>::iterator iter = _voxelMap.begin(); 
         iter != _voxelMap.end(); /* do not increase here because of the erase */) 
    {
      // the iterator is always counted up. Either in the post-incerement operator 
      // of the erase, or in the else branch. 
      if(iter->first.zBin == zBin) 
      {
	_voxelMap.erase(iter++);
      }
      else
      {
	++iter;
      }
    }//for iter
  }


  int VoxelTPC::getZBin( double zLength ){

    int zBin = static_cast<int>( zLength/_lengthOfZBin );
    return zBin;
  }


  size_t VoxelTPC::getSize()
  {
    return _voxelMap.size();
  }

} //end of namespace marlintpc

