#ifndef ChargeDistributionProcessor_h
#define ChargeDistributionProcessor_h 1

#include "VoxelTPC.h"

//LCIO
#include "lcio.h"

//MARLIN
#include "marlin/Processor.h"

//C++
#include <string>

namespace marlintpc{




/* The following comment is tranfered to doxygen documentation
 * which begins with a second asterix, or in case of on line comment an extra third slash 
 */

/** This processor takes the coordinates of the drifted electrons 
 *  and the charge from amplification and smears the charge 
 *  on the padplane. Broadening of the charge cloud due to 
 *  diffussion between the GEMs is included as well as charge 
 *  transfer parameters in dependence on the GEM settings.
 *  The processor adds the transversale defocussing in the GEM stack 
 *  to the conditions data. If the conditions data is write
 *  protected and therefor the defocussing can not be added 
 *  a new writeable TPCConditions object is added to the data stream.
 *  The new object is a copy of the original TPCConditions and 
 *  additionally holds the value for the mean amplification of an single
 *  electron. Later processors should only use the newly created object.
 *
 *  Do not use the marlin::ConditionsProcessor to manage TPCConditions
 *  set or modified by the GEMProcessor!!!
 *
 *  The WriteAfterNEvents parameter is optional. It is inteded for a CLIC-like bunch train 
 *  structure where the bunch train length is much shorter than the readout time of the TPC.
 *  The TPC is read out completely if ( eventNr % WriteAfterNEvents == (WriteAfterNEvents - 1) ).
 *  All events up to this point are empty, the whole TPC picture is written to one event.
 *
 *  The \c ImproveMemoryFootprint flag disables the output collection. Instead the static
 *  fetchVoxelMap() function has to be called by the subsequent processors.
 *
 *  <h4>Input - Prerequisites</h4>
 *  Collection of SimTrackerHits with charge values set as dEdx
 *
 *  <h4>Output</h4> 
 *  Collection with TPCVoxels containing charges on pads
 *
 * @param InputCollectionName Name of input collection, SimTrackerHits
 * @param OutputCollectionName Name of output collection, TPCVoxel
 * @param WriteOutputToStorage If \c true the output collection is written (default: true)",
 * @param xBinLength Length of the bins in z direction in ns
 * @param ETrans1 Value of first transfer field in V/cm
 * @param ETrans2 Value of second transfer field in V/cm
 * @param EInd Value of Induction field in V/cm
 * @param Gas Gas mixture
 * @param BunchIntervall Time between two bunch crossings in ns
 * @param TPCConditions Name of the collection with TPCConditons to be read in 
 * @param AlternativeTPCConditionsName Name of newly created TPCCondtions collection if original object is writeable
 * @param IonisationEnergy Mean ionisation energy of the chambers gas in [GeV] (default: 26E-9 (argon))
 * @param WriteAfterNEvents Force readout of the full TPC after overlaying n events (optional)
 * @param ImproveMemoryFootprint Flag whether to use fetchVoxelCollection instead of the output collection (default:false)
 *
 *
 * @author Thorsten Krautscheid
 */




class ChargeDistributionProcessor : public marlin::Processor {
  
 public:
  
  virtual Processor*  newProcessor() { return new ChargeDistributionProcessor ; }
  
  
  ChargeDistributionProcessor() ;
  virtual ~ChargeDistributionProcessor(){}
  
  /** Called at the begin of the job before anything is read.
   *  Use to initialize the processor, e.g. book histograms.
   */
  virtual void init() ;
  
  /** Called for every run.
   */
  virtual void processRunHeader( EVENT::LCRunHeader* run ) ;
  
  /** Called for every event - the working horse.
   */
  virtual void processEvent( EVENT::LCEvent * evt ) ; 
  
  
  double CalculateTransverseDiffusion(double  EField);

  double CalculateGasParameter(double a, double b ,double c, double d, double e, 
                               double f, double g, double h, double i, double j, 
                               double k, double EField);
  
  double CalculateDiffTransBParam(double a, double b, double c, double d, double e);

  /** A function to directly get the voxels from the internal mal. Like this the data does not go to 
   *  the event and thus reduces the memory footprint. Sort of a hack which bypasses the processor
   *  interface, but necessary due to the huge memory consumption.
   *  Make sure to delete it in order not to leak memory.
   */
  static std::list< std::pair< VoxelTPC::VoxelIndex, int > > * fetchVoxelList();

protected:

  /** Input/output collection names.
   */
  std::string _inputCollectionName;///<Name of the input collection
  std::string _outputCollectionName;///<Name of the output collection
  std::string _tpcConditionsColName;///<Name of collection containing the TPCConditions
  std::string _altTPCConditionsName;///<Name of the newly created TPCConditonsCollection if the original one is not writeable
  std::string _gas;///<Name of the gas mixture

  float _ETrans1;///<First transfer field
  float  _ETrans2;///<Second transfer field
  float  _EInd;///<Induction field
  bool _outputIsPersistent;/// If true the output collection is written
  double _sigma;///<Value of the transverse defocussing
  float _BField;///<Strength of the magnetic field in T
  double _bunchIntervall;///<Time between two bunch crossings in ns
  double _zBinLength;///<Length of a z-bin in ns

  int _writeAfterNEvents; ///< Number of overlayed events after which the TPC is read out

  const gear::TPCParameters* _tpcParameters;

  marlintpc::VoxelTPC* _trackerData;

  //For internal bookkeepin
  bool _isFirstEvent; ///< Test if this is the first event to be processed at all

  double _ionisationEnergy;///<Ionisation energy of the gas 
  
  /** A subset of voxels which has been removed (fetched) from the VoxelTPC. It is to be fetched
   *  by the subsequent processor using fetchVoxelList.
   */
  static std::list< std::pair< VoxelTPC::VoxelIndex, int > >* _fetchedList;

  bool _improveMemoryFootprint; ///< Flag whether to use the memory saving shotcut.
};

}//end of namespace marlintpc

#endif 
