#ifndef GEMProcessor_h
#define GEMProcessor_h 1

//C++
#include <string>
#include <vector>

//LCIO
#include <IMPL/LCCollectionVec.h>
#include <marlin/Processor.h>
#include <lcio.h>

namespace marlintpc{

/* The following comment is tranfered to doxygen documentation
 * which begins with a second asterix, or in case of on line comment an extra third slash 
 */

/** This Processor takes drifted electrons and uses a triple 
 *  GEM structure for amplification. The amplified electrons 
 *  are safed as SimTrackerHits with charge safed in dEdx. 
 *  ATTENTION!This processor checks also for electrons arriving 
 *  on padplane. Those not on padplane are dumped and
 *  not amplified. Correctly this should be done in the 
 *  digitisation part but it is done here for reasons of 
 *  performance. This has to be kept in mind when the processor 
 *  is not used in combination with the ChargeDistributionProcessor. 
 *  The processor adds the mean amplification value of an single 
 *  electron to the conditions data. If the conditions data is write
 *  protected and therefor the amplification value can not be added 
 *  a new writeable TPCConditions object is added to the data stream.
 *  The new object is a copy of the original TPCConditions and 
 *  additionally holds the value for the mean amplification of an single
 *  electron. Later processors should only use the newly created object.
 *
 *  Do not use the marlin::ConditionsProcessor to manage TPCConditions
 *  set or modified by the GEMProcessor!!!
 *
 *  <h4>Input - Prerequisites</h4>
 *  Collection of SimTrackerHits with drifted electrons
 *
 *  <h4>Output</h4> 
 *  Collection of SimTrackerHits with dEdx set as charge
 *
 * @param InputCollectionName Name of input collection (default: DriftedElectrons)
 * @param OutputCollectionName Name of output collection (default: AmplifiedElectrons)
 * @param Transient If not 0 the output collection is set transient (default: 0)
 * @param ParamFile Name of the file containing GEM-parametrisation (default: parameterfile)
 * @param GainFactor Correction factor for the gain (default: 1)
 * @param DriftField Value of the drift field (default: 240)
 * @param ETrans1 Value of transfer field 1 in V/cm
 * @param ETrans2 Value of transfer field 2 in V/cm
 * @param EInd Value of the induction field in V/cm
 * @param UGEM1 Voltage on GEM 1 in V (default: 315)
 * @param UGEM2 Voltage on GEM 2 in V (default: 315)
 * @param UGEM3 Voltage on GEM 3 in V (default: 315)
 * @param Gas Name of the gas in the TPC (default: TDR)
 * @param IonisationEnergy Mean ionisation energy of the chambers gas in [GeV] (default: 26E-9 (argon))
 * @param FullCurlerBackwardCompatibility Turn on for data from PrimaryIonisation < r1876
 *        and FullCurler turned off (default: false)<br>
 *        Without FullCurler the old versions of PrimaryIonisation grouped 10 electrons 
 *        but did not set the charge currectly.
 * @param TPCConditions Name of the collection with TPCConditons to be read in 
 * @param AlternativeTPCConditionsName Name of newly created TPCCondtions collection if original object is writeable
 *
 * @author Astrid Muennich and Thorsten Krautscheid
 */


class GEMProcessor : public marlin::Processor {
  
 public:
  
  virtual Processor *newProcessor(){ return new GEMProcessor ; }
   
  /** C'tor
   */
  GEMProcessor() ;

  /** D'tor
   */
  virtual ~GEMProcessor(){}
  
  /** Called at the begin of the job before anything is read. 
   *  Use to initialize the processor, e.g. book histograms.
   */
  virtual void init() ;
  
  /** Called for every run.
   */
  virtual void processRunHeader( EVENT::LCRunHeader* run ) ;
  
  /** Called for every event - the working horse.
   */
  virtual void processEvent( EVENT::LCEvent * evt ) ; 
    
  /** Reads parameter file with GEM parametrisations
   *  and sets collection and extraction efficiencies
   *  as well as the gain for the single GEMs
   */
  void LoadGEMSettings();

  /** Gives the Collection into GEM i=GEMNr for enumber of 
   *  electrons in front of GEM hole.
   */
  int GEM_Collection(int GEMNr, int enumber);

  /** Gives the gain in GEM i=GEMNr using an exponential 
   *  distribution for the gain in one GEM. 
   */
  double GEM_Gain(int GEMNr, int enumber);

  /** Gives the Extraction out of GEM i=GEMNr for enumber 
   *  of electrons indside GEM hole.
   */
  int GEM_Extraction(int GEMNr, int enumber);

  /** Claculates the attachment between GEMs
   */
  double CalculateAttachment(double E);

  double AttachmentParam(double E, double a0, double a1, double a2, 
			 double a3, double a4, double a5);

  /** Function to claculate number of electrons after 
   *  passing through the GEM structure. This function 
   *  makes use of the parametrisation of the charge 
   *  transfer through the GEMs.
   */
  int CalculateNrElectrons(double attachment1, double attachment2, 
			   double attachment3, int charge);


  /** Called after data processing for clean up.
   */
  virtual void end() ;

 
protected:

  std::string _inputCollectionName; ///<Name of input collection
  std::string _outputCollectionName;///<Name of output collection
  std::string _tpcConditionsColName;///<Name of collection containing the TPCConditions
  std::string _altTPCConditionsName;///<Name of the newly created TPCConditonsCollection if the original one is not writeable

  std::string paramfile;///<Name of file containig GEM parametrization
  std::string gas; ///<Name of gas in TPC (P5, P10 or TDR)
  const char* Gas;

  float UGEM1; ///<Voltage of GEM 1 in [V]
  float UGEM2; ///<Voltage of GEM 2 in [V]
  float UGEM3; ///<Voltage of GEM 3 in [V]

  float ETrans1; ///<Strength of first transversal field in [V/cm]
  float ETrans2; ///<Strength of second transversal field in [V/cm]
  float EInd;    ///<Strength of induction field in [V/cm]

  float _gainFactor;///<Correction factor for GAIN

  int _transient;///<Sets output collection to transient when set to another value than 0.

  float _driftField;///<Strength of the drift field in [V/cm]

  double Extr[3];
  double Coll[3];
  double Gain[3];

  /* Attachment 
   */
  double A1, A2, A3;

  //For internal bookkeeping
  int _nRun;
  int _nEvt;

  double _meanAmp;///<Mean amplification in GEMs

  double _ionisationEnergy;///<Ionisation energy of the gas 
  bool _fullCurlerBackwardCompatibility; ///< Flag whether to multiply hits in curlers with 10


}; //end of class definition

} //end of namespace marlintpc

#endif 
