#ifndef HITDISTRIBUTOR_H
#define HITDISTRIBUTOR_H 1

//MARLINTPC
#include "TPCVoxel.h"

//C++
#include <list>
#include <stdexcept>

//LCIO
#include <EVENT/SimTrackerHit.h>

//GEAR
#include <gearimpl/RectangularPadRowLayout.h>
#include <gearimpl/FixedPadSizeDiskLayout.h>
#include <gearimpl/GlobalPadIndex.h>
#include <gear/TPCParameters.h>
#include <gear/TPCModule.h>


namespace marlintpc{

class GlobalPadIndex : public gear::GlobalPadIndex
{
public:
       GlobalPadIndex(int padIndex, int moduleID)
       :gear::GlobalPadIndex(padIndex, moduleID)
       {}

       bool operator == (const  GlobalPadIndex & cmp) const
       {
          return (_moduleID==cmp._moduleID && _padIndex==cmp._padIndex);
       }
   
       inline void setPadIndex(int padindex){_padIndex = padindex;}
       inline void setModuleID(int moduleID){_moduleID = moduleID;}
};


class HitDistributor{

 public:

  /** Constructors for the class. 
   */
  HitDistributor(const gear::TPCParameters*, double zBinLength);

  /** Destructor
   */
  ~HitDistributor();

  /** Only function accessible from the outside. Takes a hit and the diffusion parameters 
   *  for input. Calls all the over function necessary, depending on the type of the pad layout.
   */
  std::list<TPCVoxel> distributeHit(SimTrackerHit* hit, double sigmaTrans, double sigmaLong = 0);

 protected:
   const gear::TPCModule & getModuleOfPad(GlobalPadIndex globalPadIndex);

   int   getRowNumberOfPad(GlobalPadIndex globalPadIndex);

  /** Methods to determine the pads on which the charge of the hit is distributed on. On for each geometry type.
   */
  std::list<GlobalPadIndex> getPadsPolar( double sigmaTrans );
  std::list<GlobalPadIndex> getPadsRect(  double sigmaTrans );

  /** Called by getPads(). Add all pad indices to the list which are within the 
   *  distribution range and left/right from the input pad index. Again one function 
   *  for each pad layout.
   *  For performance reasons a reference is passed and the list in the argument is modified
   */
  void getPadsOnTheLeftPolar( std::list<GlobalPadIndex> & padList, GlobalPadIndex centerIndex, double sigmaTrans);
  void getPadsOnTheRightPolar( std::list<GlobalPadIndex> & padList, GlobalPadIndex centerIndex, double sigmaTrans );

  void getPadsOnTheLeftRect( std::list<GlobalPadIndex> & padList, GlobalPadIndex centerIndex, double sigmaTrans);
  void getPadsOnTheRightRect( std::list<GlobalPadIndex> & padList, GlobalPadIndex centerIndex, double sigmaTrans );

  /** Check if the distance between the input pad and the hit is within the distribution range.
   */
  bool checkDistancePolar( GlobalPadIndex globalPadIndex, double sigmaTrans );
  bool checkDistanceRect(  GlobalPadIndex globalPadIndex, double sigmaTrans );

  /** Distributes the charge of the hit on the pads in the input list by integration 
   *  over a 2D gaussian.
   */
  std::list<TPCVoxel> distributeHitOnPadsPolar( std::list<GlobalPadIndex> padIndexList, SimTrackerHit* hit, double sigmaTrans, double sigmaLong );
  std::list<TPCVoxel> distributeHitOnPadsRect(  std::list<GlobalPadIndex> padIndexList, SimTrackerHit* hit, double sigmaTrans, double sigmaLong );

 protected:

  /** Pointer on the gear pad layout.
   */
  const gear::TPCParameters* _tpcParameters;

  /** Initilazation parameter. 
   */
  double _zBinLength;

  /** Diffusion parameters. Come with the hit to
   *  smear and may change from hit to hit.
   */
  double _sigmaTrans,_sigmaLong;

  /** Coordinates of the hit. Are newly set for each hit.
   */
  double _xHit, _yHit;

  /** Type of the pad layout. Decides which smearing 
   *  decides which smearing function is called.
   */
  int _padLayoutType;

  //std::list<int> padIndexList;

};

} //end of namespace marlintpc



#endif // HITSMEARER_H
