#ifndef IonsInVoxelsProcessor_h
#define IonsInVoxelsProcessor_h 1

//MARLINTPC
#include "VoxelTPC.h"

//LCIO
#include <lcio.h>

//MARLIN
#include <marlin/Processor.h>

//C++
#include <string>

//GEAR
#include <gear/PadRowLayout2D.h>

#include <TTree.h>
#include <TFile.h>


namespace marlintpc{

class VoxelTPC;

/* The following comment is tranfered to doxygen documentation
 * which begins with a second asterix, or in case of on line comment an extra third slash 
 */

/** This processor takes all ions from the primary ionisation
 *  and stores there charges in voxels. The binning of the voxels 
 *  is given in the x,y-plane by the pads of the readout plane. The 
 *  length of the voxel in the z-direction can be chosen via the
 *  steering file. For storing and management of the voxels the 
 *  helper class VoxelTPC is used. Output collections contain
 *  the charges of all the ions which emerged to the given event or
 *  all ions which appeared up to the given events in the TPC.
 *
 *  A root tree for debugging can be written out. This happens for every event 
 *  if OutputType==0 or once at the end if OutputType==1.
 *  I case WriteAfterNEvents is turned on, the tree is also written after n events.
 *
 *  <h4>Input - Prerequisites</h4>
 *  Collection of SimTrackerHits with charge values set as dEdx
 *
 *  <h4>Output</h4> 
 *  Collection containing the charges of the primary ions in
 *  TPCVoxels. 
 *
 * @param InputCollectionName Name of input collection, SimTrackerHits
 * @param OutputCollectionName Name of output collection, TPCVoxel
 * @param WriteOutputToStorage If true the output collection is written (default: true)
 * @param OutputType If 0 only ions from one event are written, else all events are summed up (default 0)
 * @param BinLength Length of the bins in z direction ( default: 2.5 mm)
 * @param IonisationEnergy The energy in GeV needed to ionize a gas particle (default: 26.E-9 = Argon)
 * @param WriteAfterNEvents Force readout of the full TPC after overlaying n events (optional)
 *        Forses the OutputType to 1
 * @param WriteRootFile File name for a root tree with the charge per voxel for debugging (optional)
 *
 * @author Thorsten Krautscheid (University of Bonn)
 * @author Martin Killenberg (CERN)
 */


class IonsInVoxelsProcessor : public marlin::Processor {

 
 public:
  
  virtual Processor*  newProcessor() { return new IonsInVoxelsProcessor ; }  
  
  IonsInVoxelsProcessor() ;
  virtual ~IonsInVoxelsProcessor();
  
  /** Called at the begin of the job before anything is read.
   *  Use to initialize the processor, e.g. book histograms.
   */
  virtual void init() ;
  
  /** Called for every run.
   */
  virtual void processRunHeader( EVENT::LCRunHeader* run ) ;
  
  /** Called for every event - the working horse.
   */
  virtual void processEvent( EVENT::LCEvent * evt ) ; 

  /**
   * Write the root file and close it porperly
   */
  virtual void end();
    

  /** Function to get the pad corresponding to a hit for 
   *  the different pad layouts
   */
  int getPad(gear::PadRowLayout2D const* padLayout, SimTrackerHit* hit);
  
  
protected:

  //input and output collections
  std::string _inputIonsColName;
  std::string _outputVoxelColName;

  //padLayout pointer
  gear::PadRowLayout2D const * _padLayout;

  //the pad layout type (cartesian or polar)
  int _padLayoutType;

  //pointer to the VoxelTPC which stores the ions
  marlintpc::VoxelTPC* _voxelTPC;

  //input parameters from steering file
  bool _outputIsPersistent; ///< Flag whether to write the output to disk
  int _outputType;///< Output type: event based or total.
  int _nEvt;
  double _binLength;///< Length of the zBins.
  double _ionisationEnergy;///< The mean ionisation energy in the gas
  int _writeAfterNEvents;///< Only write after n events (and summ up until then)
  std::string _rootFileName;///<Name of the file for the debug tree

  //TPC parameters
  double _maxDriftLength;

  // Parameters for the root tree
  std::vector <double> _leaf_chargePerVolume; ///< the charge per volume in the voxel (in C/mm^2)
  std::vector <int> _leaf_charge;///< charge per voxel
  std::vector <int> _leaf_padIndex; ///< the gear pad index
  std::vector <int> _leaf_zIndex; ///< the z index
  std::vector <double> _leaf_x,_leaf_y,_leaf_z; ///< the coordinates of the voxel centre

  TTree *_chargeInVoxelsTree; // the root tree
  TFile * _rootFile; // the root file

  /// a helper function to avoid code duplication.
  void fillTree(LCCollectionVec* outputVoxelCollection);
};

}//end of namespace marlintpc

#endif 
