#ifndef MokkaToVoxelProcessor_h
#define MokkaToVoxelProcessor_h 1

//MarlinTPC
#include "VoxelTPC.h"

//Marlin
#include "marlin/Processor.h"

//LCIO
#include "lcio.h"

//C++
#include <string>


namespace marlintpc{

/** The MokkaToVoxelProcessor is a fast digitisation processor. It takes hits from Mokka
 *  simulations for input and does the whole digitisation in the TPC in one step. The 
 *  energy deposition of the Mokka hit is transformed into the corresponding charge. The 
 *  charge is then multiplied by a user given amplification factor. The broadening of the 
 *  amplified charges is calculated from transverse diffusion and transverse defocussing. 
 *  The time binning of the TPCVoxels is done by the frequency of the readout electronics.
 *  The broadening of the charge in longitudinal direction is done by the longitudinal 
 *  diffusion and width of the electronics shaping, which is assumed to be gaussian. No 
 *  event overlay is done.  
 *
 *
 *  <h4>Input - Prerequisites</h4>
 *  \li A Collection of SimTrackerHits (from Mokka)
 *  \li GEAR TPCParameters with a pad layout and maximum drift length
 *  \li tpcconddata::TPCConditions for the drift velocity, transverse diffusion and transverse defocussing
 *
 *  <h4>Output</h4>
 *  A collection of TPCVoxel which contain charges detected on the corresponding pad at the corresponding time
 *
 *  @param InputCollectionName Name of input SimTrackerHit collection (default: MokkaHits)
 *  @param OutputCollectionName Name of output TPCVoxel collection (default: MokkaHitsInVoxels)
 *  @param TPCConditionsCollection Name of the input collection containing the TPC 
           conditions data (default: TPCConditions::getDefaultColName())
 *  @param Transient If not 0 the output collection is set transient (default: 0)
 *  @param ReadoutFrequency Frequency of readout electronics in [MHz] (default: 40)
 *  @param Amplification Total amplification factor of primary electrons (default: 1)
 *  @param IonisationEnergy Mean ionisation energy of the chambers gas in [GeV] (default: 26E-9 (argon))
 *  @param SigmaElectronics Width of the electronics shaping, which is assumed to be gaussian (default: 0)
 *
 *  @author Thorsten Krautscheid, University of Bonn
 *  @author Martin Killenberg, CERN
 */


  class MokkaToVoxelProcessor : public marlin::Processor {
  
  public:
  
    virtual Processor*  newProcessor() { return new MokkaToVoxelProcessor ; }
  
    /**C'tor
     */
    MokkaToVoxelProcessor() ;

    /**D'tor
     */
    virtual ~MokkaToVoxelProcessor(){}
  
    /** Called at the begin of the job before anything is read.
     *  Use to initialize the processor, e.g. book histograms.
     */
    virtual void init() ;
    
    /** Called for every run.
     */
    virtual void processRunHeader( EVENT::LCRunHeader* run ) ;
    
    /** Called for every event - the working horse.
     */
    virtual void processEvent( EVENT::LCEvent * evt ) ; 
    
    
    virtual void check( EVENT::LCEvent * evt ) ; 
  
  
    /** Called after data processing for clean up.
     */
    virtual void end() ;

  protected:

    std::string _inputCollectionName; ///<Name of input collection
    std::string _outputCollectionName;///<Name of output collection
    std::string _negativeOutputCollectionName;///<Name of output collection for the negative half TPC
    std::string _conditionsColName;   ///<Name of collection with conditions data

    //VoxelTPC
    marlintpc::VoxelTPC  *_MokkaHitsInVoxelTPC_positive, *_MokkaHitsInVoxelTPC_negative;
    //GEAR pad layout
    gear::PadRowLayout2D const* _padLayout;


    int _transient;///<If not 0 output collection is set transient
    double _ionisationEnergy;///<Ionisation energy of the gas 
    double _readoutFrequency;///<Frequency of the readout electronics
    double _amplification;///<Total amplification of primary electrons
    double _diffTrans;///<Transversal diffusion coefficient, comes from TPCConditions or as processor parameter.
    double _diffLong;///<Longitudinal diffusion coefficient, comes from TPCConditions or as processor parameter.
    double _defocusTrans;///<Transversal defocussing, comes from TPCConditions  or as processor parameter.
    double _vDrift;///<Electron drift velocity, comes from TPCConditions or as processor parameter.
    double _sigmaElectronics;///<Charge broadening due to electronics shaping, which is assumed to be Gaussian.
    //    double _maxDriftLength;///<Maximum drift length in the TPC.
    double _zAnode;///<The z position of the anode (edge of the readout terminating the drift volume)

};

}//end of namespace marlintpc

#endif 





