#ifndef PixelClockFinderProcessor_h
#define PixelClockFinderProcessor_h 1

// C++
#include <string>
#include <iostream>
#include <sstream>
#include <fstream>   // needed to write log File

// LCIO
#include "lcio.h"
#include <EVENT/TrackerData.h>

// Marlin
#include "marlin/Processor.h"
#include "TimePixPixelMode.h"
#include "lccd/ConditionsMap.hh"
#include <TimePixMapListener.h>
#include <TimePixPixelMode.h>
#include <marlin/AIDAProcessor.h>
#include <marlin/Global.h>

// Aida
#ifdef MARLIN_USE_AIDA
#include <AIDA/AIDA.h>
#include <marlin/AIDAProcessor.h>
#include <AIDA/IHistogram2D.h>
#endif

//root
#include "TH2F.h"

//GEAR
#include <gear/TPCModule.h>
#include <gear/TPCParameters.h>


namespace AIDA {
  class IHistogram1D;
  class IHistogram2D;


}

namespace marlintpc{

   /* the following comment is tranfered to doxygen documentation
     * which begins with a second asterix, or in case of on line comment an extra third slash
     */
    /**  
     *  PixelClockFinderProcessor creates a ToT or ToA spectrum of all pixels on the chip
     *  and two occupancy plots: One for all pixel above a given threshold and one for all pixel
     *  below. The processor also creates a eventdisplay of events with bad counts (counts in a
     *  given range) 
     *  
     *
     *  <h4>Input - Prerequisites</h4>
     *  a collection of TrackerData objects, zeros suppreed
     *
     *  <h4>Output</h4>
     *  
     *
     * @param InputCollectionName: Name of input collection, TrackerData (defualt: TimePixZeroSuppressedData) 
     * @param Mode: mode that should be plotted (TOT, TIMEPIX, MEDIPIX, ONEHIT or ANY), (default: ANY)
     * @param Minclock: Minimum z value of lowest bin in case of z binning (default: 2500)
     * @param Maxclock: Maximum z value of highest bin in case of z binning (default: 4000)
     * @param nBins: Number ob bins in z, 0 means one Histo without z cut (default: 100)
     * @param LogFile: Position of logfile
     * @param HighTH: High threshold from wich pixel are collected in _myBadPixelLocation2dHisto (default: 1800)
     * @param LowTH: Low threshold from wich pixel are collected in _myBadPixelLocation2dHisto (default: 285)
     * @param LowerBadCountsLimit: Minimal Number of Pixels with Bad Counts to create Event Display (default: 0)
     *
     * @author Martin Schultens (started the processor) & C. Brezina (changed a lot) (RFWUniversity of Bonn)
     */


    class PixelSpectrumProcessor : public marlin::Processor {
       
    public:
       
        /* This method will be called by the marlin package
         * It returns a processor of the currend type (here TimePixHitCenterCalculatorProcessor)
         */
        virtual Processor*  newProcessor() { return new PixelSpectrumProcessor ; }
       
        /* the default constructor
         * here the processor parameters are registered to the marlin package
         * other initialisation should be placed in the init method
         */
        PixelSpectrumProcessor() ;
       
        /* Called at the beginning of the job before anything is read.
         * Use to initialize the processor, e.g. book histograms
         */
        virtual void init() ;
       
        /* Called for every run.
         * in this processor it is not used
         */
        virtual void processRunHeader( lcio::LCRunHeader* run ) ;
       
        /* Called for every event - the working horse.
         * Here the real conversion take place
         */
        virtual void processEvent( lcio::LCEvent * evt ) ;
	
	virtual void check( lcio::LCEvent *evt );

	virtual void end();
       


       

       
    protected:
       
        /* normally the doxygen documentation is palced before the method/variable
         * but for one line comments in case of variables it may be more convenied for you
         * to place it after the variable. In this case you use an arrow to tell doxygen
         * to which variable the commet belongs to.
         */
       
        std::string _inputColName ;    ///<- the name of the input collection
        std::string _modeMapColName ;  ///<- the name of the mode map collection
        std::string _modeMapCorrName ; ///<- the name of the correction to mode map collection
	std::string _mode;             ///<- the mode that should be plottet
	std::string _logFile;          ///<-positon of logfile

        tpcconddata::TimePixMap * _timePixMap;
        AIDA::IHistogram1D * _mySpectrumHisto; ///<- Histo used to create a ToT or ToA spectrum of the chip 
	
        /* Histo used to store a occupancy plot of all 
         * pixels below a given threshold (_LowTH)
         */
	AIDA::IHistogram2D * _myPixelAboveTH_Location2dHisto;

        /* Histo used to store a occupancy plot of all 
         * pixels above a given threshold (_HighTH)
         */ 
	AIDA::IHistogram2D * _myPixelBelowTH_Location2dHisto; 
	
   
        double _shutter; ///<- length of shutter signal in us
        double _readoutFrequency; ///<- readout frequency of TimePix chip in MHz, optional, should come from gear
        
        bool _multiShutter; ///<- true if multiple shutters are allowed during acquisition
       
	int _nBins;        ///<- Number of z bins
	int _HighTH; 	   ///<- High Threshold from which Pixels are collected in _myBadPixelLocation2dHisto
	int _LowTH;	   ///<- Low Threshold from which Pixels are collected in _myBadPixelLocation2dHisto
        float _min_clock;  ///<- Minimum z value of lowest bin in case of z binning
        float _max_clock;  ///<- Maximum z value of highest bin in case of z binning
	int _lowerBadCountsLimit; /// <- Minimal Number of Pixels with Bad Counts to create Event Display

    private:
//      AIDA::IHistogram2D * histoCluster;
//      AIDA::IHistogram2D * histoSP;  
//      AIDA::IHistogram3D * histo3D;  

    } ;
   
}

#endif
   
   
