#ifndef TPC_VOXEL_H
#define TPC_VOXEL_H 1

//C++
#include "iostream"
#include "string"

//LCIO
#include "lcio.h"
#include "UTIL/LCFixedObject.h"

#define TPCVoxelNINTVals 4 
#define TPCVoxelNFLOATVals 0
#define TPCVoxelNDOUBLEVals 0

namespace marlintpc{


/** An LCFixedObject class to store charge in voxels of a TPC.
 *  Also provides comparison operators for sorting, e.g. in 
 *  STL containers.
 */
class TPCVoxel : public UTIL::LCFixedObject<TPCVoxelNINTVals,
			  TPCVoxelNFLOATVals,TPCVoxelNDOUBLEVals> {

public:
    //Constructror
    TPCVoxel(int padIndex, int timeIndex, int charge, int moduleID=0);

    TPCVoxel(EVENT::LCObject* obj) : UTIL::LCFixedObject<TPCVoxelNINTVals,
								  TPCVoxelNFLOATVals,
								  TPCVoxelNDOUBLEVals>(obj) { } 

    TPCVoxel(const TPCVoxel &original)
      : UTIL::LCFixedObject<TPCVoxelNINTVals,
                            TPCVoxelNFLOATVals,
                            TPCVoxelNDOUBLEVals>(original)
    {
      if( _createdObject )  _obj = new LCGenericObjectImpl(*(original._obj));
    }
    
    virtual ~TPCVoxel();

    /// Returns the GEAR ModuleID
    int getModuleID() const;

    /// Returns the GEAR pad index
    int getPadIndex() const;

    /// Returns the time index of the voxel
    int getTimeIndex() const;

    /// Returns the charge in the voxel (in electrons/ions)
    int getCharge() const;

    /// Set the charge if the voxel (in electrons/ions)
    void setCharge(int charge);

    /** Two voxels are equal if their pad index and time index are equal.
     *  The charge is not considered.
     */
    bool operator==(const TPCVoxel &right) const;

    /** < operator for e.g.\ sorting in STL containers. 
     *  They are sorted in pad index, and within this in time index.
     *  Charge is not considdered.
     */
    bool operator<(const TPCVoxel &right) const; // needed for STL containers

    TPCVoxel& operator=(const TPCVoxel& that);

    /// returns the svn revision of the code
    static std::string getRevision();

    /// Print the voxel information
    void print ( std::ostream& os = std::cout ) const;

    // -------- need to implement abstract methods from LCGenericObject
    /// Implementation of the getTypeName method of LCGenericObject
    virtual const std::string getTypeName() const{
	return std::string("TPCVoxel");
    }

    /// Implementation of the getDataDescripton method of LCGenericObject
    /// PadIndex, TimeIndex and Charge are all ints.
    virtual const std::string getDataDescription() const{
	   return std::string("i:PadIndex,i:TimeIndex,i:Charge,i:ModuleID");	
    }

protected:
    /// Set the module, only called from the constructor
    void setModuleID(int moduleID);

    /// Set the pad index, only called from the constructor
    void setPadIndex(int padIndex);

    /// Set the time index, only called from the constructor
    void setTimeIndex(int timeIndex);
};

    /** Operator to stream the voxel information to an outstream:
     *  \code std::cout << myVoxel << std::endl; \endcode
     */
    std::ostream &operator<<(std::ostream &os, const TPCVoxel &tv);
}//namespace marlintpc

#endif // TPC_VOXEL2_H
