#ifndef ChannelMappingProcessor_h
#define ChannelMappingProcessor_h 1

// LCIO
#include "lcio.h"

// Marlin
#include "marlin/Processor.h"
#include "marlin/Exceptions.h"

// MarlinTPC
#include "ADCChannelMapping.h"
#include "ChannelMappingListener.h"

// C++
#include <string>


/// the MarlinTPC Package: reconstruction of TPC data
namespace marlintpc
{

  /** Channel Mapping processor of the MarlinTPC package.
   *
   *  This processor translates the channel mapping from hardware channel and readout group to gear pad ID and module ID.
   *  Mapping is usually provided via LCCD (default: database, also simple file possible).
   *
   *  <h4>Input - Prerequisites</h4>
   *  \li A collection of TrackerPulses
   *  \li A collection with the mapping (ADCChannelMapping)
   *  <h4>Output</h4>
   *  Collection of TrackerPulses (with different name).
   *
   * @param InputCollectionName Name of input data collection (default: TPCPulses)
   * @param OutputCollectionName Name of output data collection (default: TPCPulsesMapped)
   * @param WriteOutputToStorage:bool output collection will be written if set to true (which is the default)
   * @param ChannelMappingCollectionName Name of the collection containing the channel mapping (default: TPCChannelMapping)
   * @param RemoveChannelsWithoutMapping If "true", found signals without mapping in the conditions data will be removed from collection,
   *   otherwise a tpcconddata::CondDataNotAvailableException exception is thrown an processing will stop
   * @param MissingConditionsDataOverrideTimespan \c optional  Sets the time in seconds that missing conditions data will be ignored by using the last valid conditions data collection (optional parameter, if not set, processing stops if no conditions data available for event time).
  **/
  /**
  *
  * @author originally M. Killenberg, CERN; repackaged by R. Diener, DESY; modified by Ch. Rosemann, DESY
  */

  class ChannelMappingProcessor : public marlin::Processor
  {

    public:

      virtual Processor*  newProcessor()
      {
        return new ChannelMappingProcessor ;
      }

      ChannelMappingProcessor() ;

      virtual void init() ;

      virtual void processRunHeader(lcio::LCRunHeader* run) ;

      virtual void processEvent(lcio::LCEvent * evt) ;

      virtual void check(lcio::LCEvent * evt) ;

      virtual void end() ;


    protected:

      /// the name of the input collection
      std::string _inputColName ;
      /// the name of the output collection
      std::string _outputColName ;
      /// the name of the ChannelMapping collection
      std::string _channelMappingColName ;

      /// set the persistency flag of the output collection
      bool _outputIsPersistent;

      /** The timepan in seconds for which the last valid collection will be used if no conditions data are available via LCCD.
       */
      int _missingConditionsDataOverrideTimespan;
      bool _removeChannelsWithoutMapping;
    private:

      ChannelMappingListener* _ChannelMappingListener;

      /// Internal member to hold the time for the missing conditions overide. Unit: nanoseconds.
      lcio::long64 _missingConditionsDataOverrideLCIOTime;

      //fg: add a map of channels that have been removed from the collection as
      //    their hardware address is not mapped to any known software address (padID)
      typedef std::map< std::pair< int, int > , unsigned > MAP ;
      MAP _cMap ;
  } ;
}
#endif
