#ifndef COUNTSTOPRIMARYELECTRONSPROCESSOR_H
#define COUNTSTOPRIMARYELECTRONSPROCESSOR_H 1

// MarlinTPC
#include "TPCConditionsListener.h"

// Marlin
#include <marlin/Processor.h>

// LCIO
#include "lcio.h"

// stl
#include <string>

namespace marlintpc
{

/** CountsToPrimaryElectronsProcessor: converts FADC counts into primary electrons (assumes a linear gain needing no corrections)
 *  <h4>Input - Prerequisites</h4>
 *  Collection of TrackerPulse
 *
 *  <h4>Output</h4>
 *  Collection of TrackerPulse
 *
 * @param InputCollectionName Name of input data collection
 * @param OutputCollectionName Name of output data collection (optional, if not set, input collection is overwritten)
 * @param WriteOutputToStorage:bool output collection will be written if set to true (which is the default)
 * @param TPCConditionsCollectionName Name of the collection containing the tpc conditions (default: "TPCConditions")
 * @param AmplificationValueOverride Amplification Value of the gas amplification and read-out electronics to convert ADC counts to primary electrons. If this parameter is set it overides condition information and no condition listener is used (optional parameter, if not set conditions data from LCCD are used).
 * @param MissingConditionsDataOverrideTimespan Sets the time in seconds that missing conditions data will be ignored by using the last valid conditions data collection (optional parameter, if not set, processing stops if no conditions data available for event time)
 **/
/**
 *
 * @author R. Diener (DESY), J. Abernathy (UVic)
 */

class CountsToPrimaryElectronsProcessor : public marlin::Processor
{

    public:

        CountsToPrimaryElectronsProcessor() ;

        virtual void init() ;

        virtual Processor*  newProcessor()
        {
            return new CountsToPrimaryElectronsProcessor();
        }

        virtual void processRunHeader(lcio::LCRunHeader *run);

        virtual void processEvent(lcio::LCEvent *evt);

        virtual void check(lcio::LCEvent *evt);

        virtual void end();

    private:

        /** the name of the collection the input pulses will be stored under
         */
        std::string _inputColName;

        /** the name of the collection the output pulses will be stored under
         */
        std::string _outputColName;

        /** the name of the collection the TPC conditions are stored with
         */
        std::string _tpcConditionsCollectionName;

        /** The timepan in seconds for which the last valid collection will be used if no conditions data are available via LCCD.
         */
        int _missingConditionsDataOverrideTimespan;
        /// Internal member to hold the time for the missing conditions overide. Unit: nanoseconds.
        lcio::long64 _missingConditionsDataOverrideLCIOTime;

        /// set the persistency flag of the output collection
        bool _outputIsPersistent;

        // how many runs / events have been processed sofar
        int _number_of_runs;
        int _number_of_events;

        /** Conditions listener for the TPC conditions
         */
        TPCConditionsListener *_tpcConditionsListener;
};

}
#endif                           // COUNTSTOPRIMARYELECTRONSPROCESSOR_H
