#ifndef TPCConditionsListener_h
#define TPCConditionsListener_h 1

//C++
#include <iostream>
#include <utility>
#include <stdexcept>

//LCIO
#include "lcio.h"
#include "EVENT/LCCollection.h"
#include "IMPL/LCCollectionVec.h"
#include "UTIL/LCTime.h"

//LCCD
#include "lccd.h"
#include "lccd/DBInterface.hh"
#include "lccd/IConditionsChangeListener.hh"
#include "lccd/IConditionsHandler.hh"

//Marlin
#include "marlin/ConditionsProcessor.h"

//TPCCondData
#include "TPCConditions.h"

// the package schould be capsulated in a name space
namespace marlintpc
{

/** Helper Class which provides access to the TPC conditions for every event.
 *  It is derived from IConditionsChangeListener,
 *  so the TPC Conditions are updated every time the conditions data change.
 * @author R. Diener (DESY)
 */

class TPCConditionsListener : public lccd::IConditionsChangeListener
{
    public:

        /** Constructor
         * @param ColName name of the collection in which the tpc conditions are stored
         */
        TPCConditionsListener(std::string ColName = tpcconddata::TPCConditions::getDefaultColName());

        /// Destructor
        virtual ~TPCConditionsListener();

        /// Returns the default name of the conditions data collection.
        static std::string getDefaultColName()
        {
            return  tpcconddata::TPCConditions::getDefaultColName();
        };

        /// Updates stored pedestal values if conditions have change.
        /** Called by the Marlin ConditionsProcessor, if the condition change
         *  inplemetentation of the IConditionsChangeListener interface
         */
        void conditionsChanged(EVENT::LCCollection* col);

        /// Implementation of the listener class, takes care of default collection settings and registration of the handler.
        void registeredWithHandler(lccd::IConditionsHandler* condHandler);

        /// Implementation of the listener class, cleans up after deregistration and gives warning message.
        void deRegisteredWithHandler(lccd::IConditionsHandler* condHandler);

        /// Get TPCConditions Object
        /** Returns the stored \link tpcconddata::TPCConditions TPCConditions object \endlink
         *  @return \link tpcconddata::TPCConditions TPCConditions object \endlink
         */
        tpcconddata::TPCConditions* getTPCConditions();

        /// return the drift velocity (in mm/us)
        float getDriftVelocity() const;

        /// return the diffusion coefficent for the longitudinal diffusion (in sqrt(mm))
        float getLongDiffusionCoef() const;

        /// return the defocussion constant for the longitudinal diffusion (in mm)
        float getLongDefocussing() const;

        /// return the diffusion coefficent for the transversal diffusion (in sqrt(mm))
        float getTransDiffusionCoef() const;

        /// return the defocussion constant for the transversal diffusion (in mm)
        float getTransDefocussing() const;

        /// return the amplification of the gas amplification device (e.g. GEMs)
        float getAmplification() const;

        /// return the conversion factor from ADC counts to primary electrons (ADC * conv.Factor = # primary e)
        float getADCtoPrimaryElectronsFactor() const;

        /// Get the validity start time of the current conditions collection or default collection.
        lcio::long64 collectionValidSince();
        /// Get the validity end time of the current conditions collection or default collection.
        lcio::long64 collectionValidTill();

        /// Set if an exception should the thrown in case of no data available in the database.
        void useLastValidIfNoDataAvailable(bool useLastValid);
        /// Return if currently data is available in the database/file or not.
        bool noConditionsDataAvailable();

    private:

        /// Holds the name of the conditions data collection
        std::string _collectionName;

        /// Default collection that is registered and will be returned from the handler if no data is available.
        lcio::LCCollectionVec* _thisEmptyCollection;

        /// Map of registered default collections for differnet handlers. Will hold here only one entry, since this listener registers itself with only one handler.
        std::map<lcio::LCCollection* , lccd::IConditionsHandler* > _handlerDefaultCollectionMap;

        /// Internal member to hold the "since" of the validity range of the current conditions collection.
        lcio::long64 _timeStampSince;
        /// Internal member to hold the "till" of the validity range of the current conditions collection.
        lcio::long64 _timeStampTill;

        /// Used to set if the last valid collection should be used if no current data are in the database.
        bool _useLastValidIfNoDataAvailable;
        /// Holds the indormation, if there are no current data available in the database.
        bool _noDataAvailable;

        /// Pointer to hold the address of the handler this listener is registered with.
        lccd::IConditionsHandler* _condHandler;

        /// Pointer to the TPC conditions object
        tpcconddata::TPCConditions* _tpcConditions;
};

}
#endif
