// Marlin/MarlinTPC
#include <marlin/Exceptions.h>
#include <LCObjectCopier.h>
#include "GainCorrectorProcessor.h"
#include "ChannelCorrectionListener.h"

// C++
#include <iostream>

// LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/TrackerPulse.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerPulseImpl.h>
#include <IMPL/LCFlagImpl.h>

// tpcconddata
#include "TPCCondDataExeptions.h"

using namespace lcio ;
using namespace marlin ;
using namespace std;
using namespace tpcconddata;

namespace marlintpc
{

GainCorrectorProcessor::GainCorrectorProcessor() : Processor("GainCorrectorProcessor")
{
    _description = "Applies a gain correction to LCIO::TrackerPulse collection by dividing the stored charge values by the factor given in the ChannelCorrection conditions collection.";

    /* register steering parameters: name, description, class-variable, default value
     * the type will by definded by the type of the default value
     * string, double, float and int are possible
     * plaese note to get a float you must cast in C++: (float) 1.3
     */
    registerInputCollection(LCIO::TRACKERPULSE ,
                            "InputCollectionName",
                            "Name of the input collection as stored in the LCIO file",
                            _inputColName,
                            std::string("TPCPulses"));

    registerOptionalParameter("OutputCollectionName",
                              "Optional: Name of output collection",
                              _outputColName,
                              std::string("TPCDataGainCorrected"));

    registerProcessorParameter("WriteOutputToStorage",
                               "if true the output collection is written",
                               _outputIsPersistent,
                               true);

    registerProcessorParameter("ChannelCorrectionCollectionName",
                               "Name of the collection with the channel corrections",
                               _channelCorrectionColName,
                               tpcconddata::ChannelCorrection::getDefaultColName());

    registerProcessorParameter("TypeOfIndex",
                               "Type of the cellID0 and cellID1 indices (options are 'readout' or 'gear')",
                               _indexType,
                               std::string("gear"));

    registerOptionalParameter("MissingConditionsDataOverrideTimespan",
                              "The timespan in seconds for which the last valid collection will be used if no conditions data are available via LCCD.",
                              _missingConditionsDataOverrideTimespan,
                              (int)0);
}

void GainCorrectorProcessor::init()
{
    // usually a good idea to
    printParameters();

    _number_of_runs = 0;
    _number_of_events = 0;

    // create the object for handling the channel corrections
    _channelCorrectionListener = new ChannelCorrectionListener(_channelCorrectionColName);
    // give warning message if optional maxADCValue parameter overrides conditions data
    if (parameterSet("MissingConditionsDataOverrideTimespan"))
        {
            _missingConditionsDataOverrideLCIOTime = 1000000000 * (lcio::long64)(_missingConditionsDataOverrideTimespan);
            m_out(WARNING) << _processorName << ": Processing will go on using the last valid conditions collection even if there are no data available in the database for " << _missingConditionsDataOverrideTimespan << " seconds!" << std::endl;
            //tell the listeners that we want to override
            _channelCorrectionListener->useLastValidIfNoDataAvailable(true);
        }
    if (_indexType != "gear" && _indexType != "readout")
        {
            m_out(ERROR) << _processorName << ": Processor parameter TypeOfIndex has only two options: 'readout' or 'gear', but value " << _indexType << " was set." << std::endl;
            throw marlin::ParseException(_processorName + " : Processor parameter TypeOfIndex has only two options: 'readout' or 'gear'");
        }
}

void GainCorrectorProcessor::processRunHeader(LCRunHeader* run)
{
    _number_of_runs++ ;
    run->parameters().setValue(_processorName + "_revision", "$Rev: 2598 $");
    for (marlin::ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++)
        {
            if (! i->second->isOptional() || i->second->valueSet())
                {
                    run->parameters().setValue(_processorName + "_" + i->second->name(), i->second->value());
                }
        }
}

void GainCorrectorProcessor::processEvent(LCEvent * evt)
{
    // Get input collection and check if there are data in it
    // Cast the input to LCCollectionVec if we want to use the iterator or modify it
    IMPL::LCCollectionVec* inputColl = 0;
    try
        {
            inputColl = dynamic_cast<IMPL::LCCollectionVec*>(evt->getCollection(_inputColName));
            if (inputColl->size() == 0)
                throw DataNotAvailableException(std::string("Empty TrackerPulses collection")
                                                + _inputColName);
        }
    catch (EVENT::DataNotAvailableException& e)
        {
            std::cout << _processorName << ": Empty event " << evt->getEventNumber() << std::endl;
            throw marlin::SkipEventException(this);
        }

    // The TrackerData collection is the "output".
    // If the input collection is writeable and no output collection name
    // is given we use the input collection
    IMPL::LCCollectionVec* outputColl = 0;

    // Test if the collection is writeable
    try
        {
            dynamic_cast<IMPL::TrackerPulseImpl*>(inputColl->getElementAt(0))->setTime(
                dynamic_cast<EVENT::TrackerPulse*>(inputColl->getElementAt(0))->getTime());
        }
    catch (lcio::ReadOnlyException& e)
        {
            // Check if output collection name is given
            if (! parameterSet("OutputCollectionName"))
                {
                    m_out(ERROR) << "Error: Collection is not writeable." << std::endl;
                    m_out(ERROR) << "You have to specify an OutputCollectionName." << std::endl;
                    m_out(ERROR) << "Remember to adapt the input collection name of the next processor!" << std::endl;
                    throw lcio::ReadOnlyException(_inputColName + " not writeable");
                }
        }

    // Clone input if OutputCollectionName is given
    // (note that this is mandatory if input is writeOnly)
    if (parameterSet("OutputCollectionName"))
        {
            outputColl = new LCCollectionVec("TrackerPulse");
            for (IMPL::LCCollectionVec::iterator it = inputColl->begin(); it < inputColl->end() ; it++)
                {
                    outputColl->addElement(lcio::LCObjectCopier::copy(dynamic_cast<IMPL::TrackerPulseImpl*>(*it)));
                }

            // set the transient flag as definded by the processor parameter
            outputColl->setTransient(!_outputIsPersistent);
        }
    else
        {
            outputColl = inputColl;
        }
    // set the flags that cellID1 should be stored
    lcio::LCFlagImpl trkFlag(0) ;
    trkFlag.setBit(lcio::LCIO::TRAWBIT_ID1) ;
    outputColl->setFlag(trkFlag.getFlag());

    // correct the gain
    for (IMPL::LCCollectionVec::iterator it = outputColl->begin(); it < outputColl->end(); it++)
        {
            IMPL::TrackerPulseImpl* trackerPulse = dynamic_cast<IMPL::TrackerPulseImpl*>(*it);
            float oldCharge = trackerPulse->getCharge();
            float calibrationFactor = 0;
            if (_indexType == "gear")
                {
                    calibrationFactor = _channelCorrectionListener->getCalibrationFactorFromGearID(std::make_pair(trackerPulse->getCellID0(), trackerPulse->getCellID1()));
                    m_out(DEBUG2) << _processorName << " : pad:" << trackerPulse->getCellID0() << ", module:" << trackerPulse->getCellID1();

                }
            else if (_indexType == "readout")
                {
                    calibrationFactor = _channelCorrectionListener->getCalibrationFactorFromHardwareID(std::make_pair(trackerPulse->getCellID0(), trackerPulse->getCellID1()));
                    m_out(DEBUG2) << _processorName << " : channel:" << trackerPulse->getCellID0() << ", readoutGroup:" << trackerPulse->getCellID1();
                }
            else //should never come to here
                {
                    throw marlin::ParseException(_processorName + " : Processor parameter TypeOfIndex has only two options: 'readout' or 'gear'.");
                }
            trackerPulse->setCharge(oldCharge / calibrationFactor);
            m_out(DEBUG2) << " - charge value: before=" << oldCharge << ", after=" << oldCharge / calibrationFactor << " calibration" << std::endl;
        }

    // if the pulses collection has been created by cloning
    // we have to add it to the event
    if (parameterSet("OutputCollectionName"))
        {
            evt->addCollection(outputColl, _outputColName);
        }

    _number_of_events++;
}

void GainCorrectorProcessor::check(LCEvent * evt)
{
    // nothing to check here - could be used to fill checkplots in reconstruction processor
}

void GainCorrectorProcessor::end()
{
    // nothing to do here
}

}
