#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerDataImpl.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/LCFlagImpl.h>
#include <UTIL/LCFixedObject.h>

// Marlin
#include <marlin/Exceptions.h>
#include <LCObjectCopier.h>

// MarlinTPC
#include "ChannelCorrectionListener.h"
#include "TimeShiftCorrectorProcessor.h"
#include "TPCCondDataExeptions.h"

namespace marlintpc
{

TimeShiftCorrectorProcessor::TimeShiftCorrectorProcessor() : marlin::Processor("TimeshiftCorrector")
{

    // the processor description
    _description = "TimeShiftCorrectorProcessors applies time offset corrections to individual channels." ;

    // register steering parameters: name, description, class-variable, default value
    registerInputCollection(LCIO::TRACKERDATA,
                            "InputCollectionName",
                            "Name of the input collection",
                            _inputColName,
                            std::string("TPCData"));

    registerOptionalParameter("OutputCollectionName",
                              "Optional: Name of output collection",
                              _outputColName,
                              std::string("TPCDataTimeShiftCorrected"));

    registerProcessorParameter("WriteOutputToStorage",
                               "if true the output collection is written",
                               _outputIsPersistent,
                               true);

    registerProcessorParameter("ChannelCorrectionCollectionName",
                               "Name of the collection with the channel corrections",
                               _channelCorrectionColName,
                               tpcconddata::ChannelCorrection::getDefaultColName());

    registerProcessorParameter("TypeOfIndex",
                               "Type of the cellID0 and cellID1 indices (options are 'readout' or 'gear')",
                               _indexType,
                               std::string("gear"));

    registerOptionalParameter("MissingConditionsDataOverrideTimespan",
                              "The timespan in seconds for which the last valid collection will be used if no conditions data are available via LCCD.",
                              _missingConditionsDataOverrideTimespan,
                              (int)0);

}

void TimeShiftCorrectorProcessor::processRunHeader(LCRunHeader* run)
{
    run->parameters().setValue(_processorName + "_revision", "$Rev: 2711 $");

    for (marlin::ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++)
        {
            if (! i->second->isOptional() || i->second->valueSet())
                {
                    run->parameters().setValue(_processorName + "_" + i->second->name(), i->second->value());
                }
        }
}

void TimeShiftCorrectorProcessor::init()
{

    // usually a good idea to
    printParameters();

    // create the object for handling the channel corrections
    _channelCorrectionListener = new ChannelCorrectionListener(_channelCorrectionColName);
    // give warning message if optional maxADCValue parameter overrides conditions data
    if (parameterSet("MissingConditionsDataOverrideTimespan"))
        {
            _missingConditionsDataOverrideLCIOTime = 1000000000 * (lcio::long64)(_missingConditionsDataOverrideTimespan);
            m_out(WARNING) << _processorName << ": Processing will go on using the last valid conditions collection even if there are no data available in the database for " << _missingConditionsDataOverrideTimespan << " seconds!" << std::endl;
            //tell the listeners that we want to override
            _channelCorrectionListener->useLastValidIfNoDataAvailable(true);
        }
    if (_indexType != "gear" && _indexType != "readout")
        {
            m_out(ERROR) << _processorName << ": Processor parameter TypeOfIndex has only two options: 'readout' or 'gear', but value " << _indexType << " was set." << std::endl;
            throw marlin::ParseException(_processorName + " : Processor parameter TypeOfIndex has only two options: 'readout' or 'gear'");
        }
}

void TimeShiftCorrectorProcessor::end()
{
    if (_channelCorrectionListener)
        {
            delete _channelCorrectionListener;
            _channelCorrectionListener = NULL;
        }
}

void TimeShiftCorrectorProcessor::processEvent(LCEvent * evt)
{
    // Get input collection and check if there are data in it
    // Cast the input to LCCollectionVec if we want to use the iterator or modify it
    IMPL::LCCollectionVec* inputColl = 0;
    try
        {
            inputColl = dynamic_cast<IMPL::LCCollectionVec*>(evt->getCollection(_inputColName));
            if (inputColl->size() == 0)
                throw DataNotAvailableException(std::string("Empty TrackerData collection")
                                                + _inputColName);
        }
    catch (EVENT::DataNotAvailableException& e)
        {
            std::cout << _processorName << ": Empty event " << evt->getEventNumber() << std::endl;
            throw marlin::SkipEventException(this);
        }

    // The TrackerData collection is the "output".
    // If the input collection is writeable and no output collection name
    // is given we use the input collection
    IMPL::LCCollectionVec* trackerDataColl = 0;

    // Test if the collection is writeable
    try
        {
            dynamic_cast<IMPL::TrackerDataImpl*>(inputColl->getElementAt(0))->setTime(
                dynamic_cast<EVENT::TrackerData*>(inputColl->getElementAt(0))->getTime());
        }
    catch (lcio::ReadOnlyException& e)
        {
            // Check if output collection name is given
            if (! parameterSet("OutputCollectionName"))
                {
                    m_out(ERROR) << "Error: Collection is not writeable." << std::endl;
                    m_out(ERROR) << "You have to specify an OutputCollectionName." << std::endl;
                    m_out(ERROR) << "Remember to adapt the input collection name of the next processor!" << std::endl;
                    throw lcio::ReadOnlyException(_inputColName + " not writeable");
                }
        }

    // Clone input if OutputCollectionName is given
    // (note that this is mandatory if input is writeOnly)
    if (parameterSet("OutputCollectionName"))
        {
            trackerDataColl = new LCCollectionVec("TrackerData");
            for (IMPL::LCCollectionVec::iterator it = inputColl->begin(); it < inputColl->end() ; it++)
                {
                    trackerDataColl->addElement(lcio::LCObjectCopier::copy(dynamic_cast<IMPL::TrackerDataImpl*>(*it)));
                }

            // set the transient flag as definded by the processor parameter
            trackerDataColl->setTransient(!_outputIsPersistent);
        }
    else
        {
            trackerDataColl = inputColl;
        }
    // set the flags that cellID1 should be stored
    lcio::LCFlagImpl trkFlag(0) ;
    trkFlag.setBit(lcio::LCIO::TRAWBIT_ID1) ;
    trackerDataColl->setFlag(trkFlag.getFlag());

    // correct the time shift
    for (IMPL::LCCollectionVec::iterator it = trackerDataColl->begin(); it < trackerDataColl->end(); it++)
        {
            IMPL::TrackerDataImpl* trackerData = dynamic_cast<IMPL::TrackerDataImpl*>(*it);
            float oldtime = trackerData->getTime();
            float timeshift = 0;
            if (_indexType == "gear")
                {
                    timeshift = _channelCorrectionListener->getTimeOffsetFromGearID(std::make_pair(trackerData->getCellID0(), trackerData->getCellID1()));
                }
            else if (_indexType == "readout")
                {
                    timeshift = _channelCorrectionListener->getTimeOffsetFromHardwareID(std::make_pair(trackerData->getCellID0(), trackerData->getCellID1()));
                }
            else //should never come to here
                {
                    throw marlin::ParseException(_processorName + " : Processor parameter TypeOfIndex has only two options: 'readout' or 'gear'.");
                }
            trackerData->setTime(oldtime - timeshift);
        }

    // if the pulses collection has been created by cloning
    // we have to add it to the event
    if (parameterSet("OutputCollectionName"))
        {
            evt->addCollection(trackerDataColl, _outputColName);
        }
}

}  // namespace marlintpc

