#ifndef HITSPLITTERPROCESSOR_H
#define HITSPLITTERPROCESSOR_H

#include <marlin/Processor.h>
#include <lcio.h>
#include <string>
#include <set>

#include <EVENT/TrackerPulse.h>


namespace marlintpc
{

/**  The HitSplitterProcessor searches for minima in the pad response and splits the hit.
 *   It uses the HitSplitter, which assumes a gaussian pad response.
 *
 *   The main output collection has the subSet flag set. In case the hit is not split the
 *   hit from the input collection is stored. The splits hits are stored in the SplitHitsCollection
 *   (which is not a subSet)  and  in the OutputCollection.
 *   As pulses might be split during hit splitting there is also a collection for the split hits.
 *
 *   @param OutputCollectionName Name of the output collection containing all hits
 *          (split or unsplit), default TPCSplitHits
 *   @param OnlySplitHitsCollectionName Name of the collection containing only the split hits,
 *          default OnlySplitHits
 *   @param SplitPulsesInSplitHitsCollectionName Name of the collection containing split pluses
 *          in split hits, default SplitPulsesInSplitHits
 *   @param TPCConditonsCollectionName Name of the TPCConditons collection (for vDrift), default: TPCConditions
 *   @param VDriftOverride \c optional Use to overwrite the drift velocity from conditions data,
 *          or in case there is no conditions data (in mm / us )
 *
 *    @author M. Killenberg, CERN
 *
 */


class HitSplitterProcessor : public marlin::Processor
{
    public:

        virtual Processor*  newProcessor()
        {
            return new HitSplitterProcessor;
        }

        HitSplitterProcessor();

        virtual void init();

        virtual void processRunHeader(EVENT::LCRunHeader* run);

        virtual void processEvent(EVENT::LCEvent* evt);

        virtual void check(EVENT::LCEvent* evt);

        virtual void end();


    protected:
        /* the place for protected and private member data and functions */
        std::string _inputColName; ///< Name of the input collection
        std::string _outputColName; ///< Name of the output collection
        std::string _tpcConditionsColName; ///< Name of the TPCConditions collection (for vDrift)
        std::string _onlySplitHitsColName; ///< Name of the collection with split hits
        std::string _splitPulsesInSplitHitsColName; //// Name of the collection with split pulses in split hits

        float _vDrift; ///< The drift velocity

};
} // namespace marlintpc
#endif // HITSPLITTERPROCESSOR_H
