#ifndef ROWBASEDHITFINDERPROCESSOR_H
#define ROWBASEDHITFINDERPROCESSOR_H

// MARLIN
#include "marlin/Processor.h"
#include "marlin/Global.h"

//LCIO
#include "lcio.h"
#include "EVENT/LCCollection.h"
#include "IMPL/LCCollectionVec.h"
#include "EVENT/TrackerPulse.h"

//GEAR
#include "gear/TPCParameters.h"

// C++
#include <string>
#include <vector>
#include <map>

// MARLINTPC
#include "ChannelCorrectionListener.h"
#include "TPCConditionsListener.h"

namespace marlintpc
{

  /** A hit finder for row based (pad) geometries
  *  It searches for connected pulses in a row and combines them into a hit.
  *  The position inside the row is calculated by a center-of-gravity w.r.t. to the charge of the individual pulses.
  *  The position on the row is fixed by the r/y coordinate of the row.
  *  The z-position is inherited from the pulse with the highest charge within the hit.
  *
  *
  *  @author Christoph Rosemann, DESY
  *
  *  <h4>Input - Prerequisites</h4>
   *  \li TrackerPulse collection
   *  \li tpcconddata::TPCConditions for the drift velicity (optionally provided as processor parameter)
   *  \todo Hit quality flag (including inheritance of pulse flags)
   *  \todo Handling of dead and noisy channels.
   *
   *  <h4>Output</h4>
   *  \li TrackerHit collection
   *
   *  @param InputTrackerPulses Name of input TrackerPulses collection (default: TPCPulses)
   *  @param OutputTrackerHits Name of output TrackerHits collection (default: TPCHits)
   *  @param InputTPCConditions \c OPTIONAL: Name of the input collection containing the TPC
   *         conditions data (default: tpcconddata::TPCConditions::getDefaultColName() )
   *  @param WriteOutputToStorage:bool If true the output collection is written to file (default: true)
   *  @param MaxEmptyPadsOverride Maximum number of consecutive empty pads in a row (default: 1)
   *  @param MinHitSizeOverride Minimum size of hit in number of pads (default: 1)
   *  @param MinMaxPulseSliceHeightOverrid Minimum size of the maximum pulse slice of the highest pulse in the hit (default: 12.)
   *  @param MaxTimeSpreadOverride Maximum time between pulses in a hit [in ns] (default: 200)
   *  @param VDriftOverride \c OPTIONAL: Set drift velocity in case there is no conditions data [in mm/us] (default: 0)
   *  @param DeadPadsOverride \c OPTIONAL: a vector of geometry indices of dead channels, default should be conditions data
   *  @param ChargeConversionFactorOverride \c OPTIONAL The factor to convert adc counts into number of primary electrons (-> should be conditions data)
   *  @param UseGlobalCoordinates \c OPTIONAL If set to true, the output coordinate system will be forced to global/cartesian; otherwise it is the local coordinate system of the pad plan (either polar or cartesian) (default: false)
   *
   * List of open issues/questions:
   * - error calculation/storage
   * - hit quality flag: definition of bit map (?) and LCIO implementation
   * - extend debug output
   * - incorporate damaged pad info
   * - incorporate inheritance of pulse quality
   * - separation of hits?
   * +++ needs testing!
   */

  class RowBasedHitFinderProcessor : public marlin::Processor
  {
    public:

      virtual Processor*  newProcessor()
      {
        return new RowBasedHitFinderProcessor;
      }

      RowBasedHitFinderProcessor() ;

      /** Called at the begin of the job before anything is read.
       * Use to initialize the processor, e.g. book histograms.
       */
      virtual void init() ;

      /** Called for every run.
       */
      virtual void processRunHeader(EVENT::LCRunHeader* run) ;

      /** Called for every event - the working horse.
       */
      virtual void processEvent(EVENT::LCEvent * evt) ;


      virtual void check(EVENT::LCEvent * evt) ;


      /** Called after data processing for clean up.
       */
      virtual void end() ;

    private:
      /** the name of the collection the input pulses will be stored under
       */
      std::string _inputTrackerPulsesCollectionName ;

      /** the name of the collection the output hits will be stored under
       */
      std::string _outputTrackerHitsCollectionName ;

      /** the name of the collection the TPC conditions are stored with
       */
      std::string _inputTPCConditionsCollectionName;

      ///<- the name of the pedestal collection
      std::string  _channelCorrectionCollectionName;

      /** a list of dead channels, given in the GearPadIndex,
       *  for now: just a single module
       * this is deprecated from the beginning,
       * just a workaround until the conditions object is there
       */
      std::vector<int> _deadChannelIndicesOverride;

      /** Set the persistency flag of the output collection.
       */
      bool _outputHitsPersistent;

      /** deprecated var to store some unused value -- will be removed in the next release
       */
      int _deprecated;

      /** maximum number of empty pads (noisy/dead) in a row
       */
      int _maxEmptyChannelsOverride;

      /** maximum number of consecutive dead channels in a hit
       */
      int _maxDeadChannelsOverride;

      /** minimum size of hit in number of pads
       */
      int _minHitSizeOverride;

      /** minimum height for the maximal pulse in a hit to be accepted
       */
      float _minMaxPulseSliceHeightOverride;

      /** maximum time between pulses belonging to a single hit [in ns]
       */
      float _maxTimeSpreadOverride;

      /** drift velocity in [mm/mu s]
       */
      float _driftVelocityOverride;

      /** conversion factor from ADC counts to number of primary electrons
       */
      float _chargeConversionFactorOverride;

      /** conversion output coordinate system [local or global, default is local]
       */
      bool _useGlobalCoord;

      /** the standard way to access pad/channel based data: corrections, noisy/dead flag, etc.
      */
      ChannelCorrectionListener* _channelCorrectionListener;

      /** the standard way to access TPC conditions: the TPCConditions object via listening
      */
      TPCConditionsListener* _tpcConditionsListener;

      // a helper class -- does it make sense to externalise it from the HitFinder?
      class hitCandidate
      {
        public:
          hitCandidate(EVENT::TrackerPulse*);

          void addPulse(EVENT::TrackerPulse*);
          void addDeadChannel();

          int qualityFlag;

          const EVENT::TrackerPulse* getMaximumPulse() const
          {
            return _maxPulse;
          };
          double getCharge() const;
          double getMaxPulseHeight();
          const std::vector<EVENT::TrackerPulse*>& getPulses() const
          {
            return _thePulses;
          };
          unsigned int getWidth() const;

        private:
          hitCandidate()
          {
            ;
          };
          std::vector<EVENT::TrackerPulse*> _thePulses;
          EVENT::TrackerPulse* _maxPulse;
          double _maxPulseCharge;
          int _numberOfDeadChannels;
      };

      /** helper functions:
       */
      void _sortPulses(EVENT::LCCollection*, const gear::TPCParameters&,
                       std::map<int, std::map<int, std::vector<EVENT::TrackerPulse*>* >* >&);
      void _findHitCandidates(std::map<int, std::map<int, std::vector<EVENT::TrackerPulse*>* >* >&,
                              const gear::TPCParameters&,
                              std::map<int, std::vector<RowBasedHitFinderProcessor::hitCandidate> >&);
      void _createHitCollection(std::map<int, std::vector<RowBasedHitFinderProcessor::hitCandidate> >&,
                                const gear::TPCParameters&, IMPL::LCCollectionVec*);

      bool _pulseComparator(EVENT::TrackerPulse*, EVENT::TrackerPulse*);
      EVENT::TrackerPulse* _findHighestPulseInRow(std::vector<EVENT::TrackerPulse*>&);
      bool _maxPulseHeightCheck(EVENT::TrackerPulse*);
      void _addNeighbourPulses(hitCandidate&, const gear::TPCParameters&, std::vector<EVENT::TrackerPulse*>&);
      std::vector<EVENT::TrackerPulse*>::iterator _findBestNeighbourPulseMatch
      (std::vector<std::vector<EVENT::TrackerPulse*>::iterator>&, float);

      bool _hitIsValid(const RowBasedHitFinderProcessor::hitCandidate&);
      double _calculateFixedCoordinate(const RowBasedHitFinderProcessor::hitCandidate&, const gear::TPCParameters&);
      double _calculateFreeCoordinate(const RowBasedHitFinderProcessor::hitCandidate&, const gear::TPCParameters&);
      double _calculateZCoordinate(const RowBasedHitFinderProcessor::hitCandidate&);
      double _calculateFixedError(const RowBasedHitFinderProcessor::hitCandidate&, const gear::TPCParameters&);
      double _calculateFreeError(const RowBasedHitFinderProcessor::hitCandidate&, double, const gear::TPCParameters&);
      double _calculateZError(const RowBasedHitFinderProcessor::hitCandidate&);
      double _convertCharge(const RowBasedHitFinderProcessor::hitCandidate&);

      unsigned int _getMaxDeadChannels();
      int _getMinHitSize();
      double _getMinMaxPulseSliceHeight();
      double _getMaxTimeSpread();
      double _getDriftVelocity();
      bool _padIsDead(const int, const int);
      bool _padIsNoisy(const int, const int);
      double _getChargeConversionFactor();
  };
}//namespace marlintpc
#endif//ROWBASEDHITFINDERPROCESSOR_H
