#include "HitSplitterProcessor.h"
#include "HitSplitter.h"
#include "TPCConditions.h"

// include what you need from lcio
#include <EVENT/LCEvent.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerHitImpl.h>
#include <IMPL/TrackerPulseImpl.h>

using namespace lcio;
using namespace marlin;
using namespace std;


namespace marlintpc
{

HitSplitterProcessor aHitSplitterProcessor;


HitSplitterProcessor::HitSplitterProcessor() : Processor("HitSplitterProcessor")
{
    // the processor description
    _description = "HitSplitterProcessor searches for minima in the pad response and splits the hit.";

    /* register steering parameters: name, description, class-variable, default value

      for example:
    */

    registerInputCollection(LCIO::TRACKERHIT,
                            "InputCollectionName" ,
                            "Name of the input hits to be split, default: TPCHits"  ,
                            _inputColName ,
                            string("TPCData")) ;

    registerOutputCollection(LCIO::TRACKERHIT,
                             "OutputCollectionName",
                             std::string("Name of the output collection containing all hits") +
                             std::string(" (split or unsplit), default TPCSplitHits"),
                             _outputColName,
                             "TPCSplitHits");

    registerOutputCollection(LCIO::TRACKERHIT,
                             "OnlySplitHitsCollectionName",
                             std::string("Name of the collection containing only") +
                             std::string(" the split hits, default OnlySplitHits"),
                             _onlySplitHitsColName,
                             "OnlySplitHits");

    registerOutputCollection(LCIO::TRACKERPULSE,
                             "SplitPulsesInSplitHitsCollectionName",
                             std::string("Name of the collection containing split pluses") +
                             std::string(" in split hits, default SplitPulsesInSplitHits"),
                             _splitPulsesInSplitHitsColName,
                             "SplitPulsesInSplitHits");

    registerProcessorParameter("TPCConditonsCollectionName",
                               "Name of the TPCConditons collection (for vDrift), default: TPCConditions",
                               _tpcConditionsColName,
                               tpcconddata::TPCConditions::getDefaultColName());

    registerOptionalParameter("VDriftOverride" ,
                              "Use VDriftOverride to overwrite the drift velocity from conditions data (in mm / us )",
                              _vDrift,
                              float(50.0));

    /*
          registerOutputCollection(   LCIO::TRACKERDATA,
            "OutputCollectionName" ,
            string("Name of the output TrackerData collection")+
            string("only the data related to a pulse will be stored")+
            string(\, default: TPCDataInPulses),
            _outputCollectionName ,
            string("TPCDataInPulses")) ;

          // It's nive to have the default value in the comment
          registerProcessorParameter( "MinPulsesPerHit" ,
             "Minimum number of pulses for a valid hit, default: 3",
             _minPulsesPerHit,
             int(3));

          // Optional parameters should only be used when set
          // in the steering file.
          // The default value is there only for technical reasons,
          // it is meaningless.
          //
          // If properly implemented the cut is not applied when the parameter is not set:
          //   if (parameterSet("ChageCut")) { //apply the cut }
          //
          registerOptionalParameter( "ChargeCut" ,
              "If the charge is below this value the data will be ignored",
              _chargeCut,
             float(17.8));
    */
}


void HitSplitterProcessor::init()
{
    m_out(DEBUG) << " init() called  " << endl;

    // you might want to list the used parameters by invoking
    printParameters();

}


void HitSplitterProcessor::processRunHeader(LCRunHeader* run)
{
    // do something for each run
}


void HitSplitterProcessor::processEvent(LCEvent* evt)
{
    // do something for each event, usually access the input collection:

    // use an LCCollectionVec instead of LCCollection so we can use iterators later
    LCCollectionVec* inputColVec = 0;

    // do a try / catch in case there is an empty event
    try
        {
            inputColVec = dynamic_cast<LCCollectionVec *>(evt->getCollection(_inputColName));
        }
    catch (DataNotAvailableException &e)
        {
            // no input data, just return and process the next event
            return;
        }

    // check whether the drift velocity is stored as conditions data
    try
        {
            LCCollection *TPCConditionsCollection = evt->getCollection(_tpcConditionsColName);
            tpcconddata::TPCConditions myTPCConditions(TPCConditionsCollection->getElementAt(0));

            // conditions data are available, but can be
            // overwritten by using processor parameter
            // give a warning in this case
            if (parameterSet("VDriftOverride"))
                {
                    streamlog_out(WARNING4) << _processorName << ": WARNING: drift velicity from conditions data is"
                    << " overwritten by processor parameter!" << std::endl
                    << "Remove parameter \"VDriftOverride\" from steering file to use vDrift from TPCCondData."
                    << std::endl;
                }
            else
                {
                    _vDrift = myTPCConditions.getDriftVelocity();
                }
        }
    catch (DataNotAvailableException &e)
        {
            // no conditions data available, try to use processor parameter
            if (!parameterSet("VDriftOverride"))
                {
                    throw(DataNotAvailableException(_processorName +
                                                    std::string(" no drift velocity available, provide either as TPCCondData or processor parameter")));
                }
        }

    // The collection containing the hits which have not been split and the split results
    // This collection is a subset so the non-split hits (the majority) is not duplicated.
    // Unfortunately this means that the splits hits have to go into a separate collection.
    LCCollectionVec * outputCollection = new LCCollectionVec(LCIO::TRACKERHIT);
    outputCollection->setSubset();

    // The collection containing the split hits, and the pulses they contain
    LCCollectionVec * splitHitsCollection = new LCCollectionVec(LCIO::TRACKERHIT);
    LCCollectionVec * pulsesInSplitHitsCollection = new LCCollectionVec(LCIO::TRACKERPULSE);

    // initialise the splitter class
    HitSplitter hitSplitter(splitHitsCollection, pulsesInSplitHitsCollection, _vDrift);

    // loop the data collection and do something
    for (LCCollectionVec::iterator inputHitIter = inputColVec->begin() ;
            inputHitIter < inputColVec->end() ;
            inputHitIter++)
        {
            // dynamic cast the input object (LCObject) to what it really is e.g.
            TrackerHit * inputHit = dynamic_cast<TrackerHit *>(*inputHitIter);

            TrackerHitVec splitHitsVec = hitSplitter.splitHit(inputHit);

            // fill the output collection
            if (splitHitsVec.size() == 0) // the hit has not been split
                { // add the original hit
                    outputCollection->addElement(inputHit);
                }
            else // fill the split hits into the output collection
                {
                    for (TrackerHitVec::iterator hitIter = splitHitsVec.begin();
                            hitIter < splitHitsVec.end(); ++hitIter)
                        {
                            outputCollection->addElement(*hitIter);
                        }// for hitIter
                }// else splitHitsVec.size()==0

        }// for inputHitIter

    // fill the event
    if (splitHitsCollection->getNumberOfElements() > 0)
        {
            evt->addCollection(splitHitsCollection, _onlySplitHitsColName);

            if (pulsesInSplitHitsCollection->getNumberOfElements() > 0)
                {
                    evt->addCollection(pulsesInSplitHitsCollection, _splitPulsesInSplitHitsColName);
                }
        }
    evt->addCollection(outputCollection, _outputColName);
}


void HitSplitterProcessor::check(LCEvent* evt)
{
    // check what is needed here
}


void HitSplitterProcessor::end()
{
    // do something at the end of the processing, e.g. cleaning up
}


} // close namespace brackets
