#include "PulseFinderProcessor.h"

// the exception by TPCCondData in MarlinTPC
#include "TPCCondDataExeptions.h"
// the common flag word definitions in MarlinTPC
#include "FlagwordDefinitions.h"

// Marlin
#include <marlin/ConditionsProcessor.h>

// gsl
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_multifit.h>

//LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/TrackerRawData.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerDataImpl.h>
#include <IMPL/TrackerPulseImpl.h>
#include <IMPL/LCFlagImpl.h>

//C++
#include <iostream>
#include <algorithm>
#include <iterator>
#include <cmath>

using namespace lcio;
using namespace marlin;
using namespace std;
using namespace tpcconddata;

namespace marlintpc
{
  PulseFinderProcessor aPulseFinderProcessor;

  PulseFinderProcessor::PulseFinderProcessor() : Processor("PulseFinderProcessor")
  {

    _description = "PulseFinderProcessor: finds pulses out of the raw data and calculate the time and charge (in FADC) information";

    registerInputCollection(LCIO::TRACKERDATA,
                            "InputCollectionName",
                            "Name of the input TrackerData collection",
                            _inputColName,
                            string("TPCData"));

    registerOutputCollection(LCIO::TRACKERDATA,
                             "OutputTrackerDataCollectionName",
                             string("Name of the output TrackerData collection ") +
                             string("only the data related to a pulse will be stored"),
                             _outputDataColName,
                             string("TPCDataInPulses"));

    registerOutputCollection(LCIO::TRACKERPULSE,
                             "OutputTrackerPulseCollectionName",
                             "Name of the output TrackerPulse collection",
                             _outputPulseColName,
                             string("TPCPulses"));

    registerProcessorParameter("WriteOutputToStorage",
                               "if true the output collection is written",
                               _outputIsPersistent,
                               true);

    registerProcessorParameter("ForceSpectrumSave",
                               "if true the ADC spectrum will be stored for all pulses",
                               _forceSpectrumSave,
                               true);

    registerProcessorParameter("PulseStartThreshold",
                               string("A pulse starts if the ADC value is the given value times the pedestal width."),
                               _pulseStartThreshold,
                               (float)7.0);

    registerProcessorParameter("PulseEndThreshold",
                               string("A pulse ends if the ADC value drops below this value times the pedestal width."),
                               _pulseEndThreshold,
                               (float)3.0);

    registerProcessorParameter("MinimumPulseHeight",
                               string("Declares the minimal height of the maximum entry of a pulse in ADC counts."),
                               _minimumPulseHeight,
                               (float)12.0);

    registerProcessorParameter("MinimumPulseLength",
                               string("Declares the minimal length of a pulse in time bins above the thresholds."),
                               _minimumPulseLength,
                               5);

    registerProcessorParameter("SaveNBinsBeforeStart",
                               "# bins saved before start threshold (except there is another pulse)",
                               _saveNBinsBeforeStart,
                               2);

    registerProcessorParameter("SaveNBinsAfterEnd",
                               "# bins saved after the end threshold (except there is another pulse)",
                               _saveNBinsAfterEnd,
                               8);

    registerProcessorParameter("PedestalWidthScale",
                               "Sets the scaling factor for the noise value calculation: noise_value=scale*pedestal_width (default: 3)",
                               _pedestalWidthScale,
                               (float)3.0);

    registerProcessorParameter("ElectronicsParametersObjectName" ,
                               "Name of the conditions object in which the parameters of the read out electronics is stored",
                               _electrParaObjName,
                               string("GenericADCElectronicsParameters")
                              );

    registerProcessorParameter("ElectronicsParametersCollectionName" ,
                               "Name of the conditions collection in which the parameters of the read out electronics is stored",
                               _electrParaColName,
                               tpcconddata::GenericADCElectronicsParameters::getDefaultColName()
                              );

    registerProcessorParameter("PedestalCollectionName" ,
                               "Name of the conditions collection in which the pedestals are stored",
                               _pedestalColName,
                               tpcconddata::Pedestal::getDefaultColName()
                              );

    registerOptionalParameter("UseParabolicFitAmplitudeAsCharge?" ,
                              "Set this flag to use the amplitude calculated by fitting a parabola to the log of the pulse data.",
                              _use_fit_charge,
                              false
                             );

    registerOptionalParameter("UseParabolicFitMaxAsTime?",
                              "Set this flag to use the time caclulated by fitting a parabola to the log of the pulse data.",
                              _use_fit_time,
                              false);

    // now some parameters for overriding the default values obtained by listeners/conditions data
    registerOptionalParameter("MaximumADCValueOverride",
                              "Maximum number of the adc spectrum; to determine wether a pulse is maxed out. This single value overrides any value, from any readout group, which may have been taken from LCCD.",
                              _maximumADCValueOverride,
                              (float)255.);

    registerOptionalParameter("PedestalValueOverride",
                              "The pedestal value to be used for every hardware channel in every readout group. This overrides any information provided by LCCD.",
                              _pedestal_value_override,
                              (float)0.);


    registerOptionalParameter("PedestalWidthOverride",
                              "The pedestal width to be used for every hardware channel in every readout group. This overrides any information provided by LCCD.",
                              _pedestal_width_override,
                              (float)0.);

    registerOptionalParameter("IsPedestalSubstractedOverride",
                              "Sets, if the data is already pedestal substracted in the read-out electronics. This overrides any information provided by LCCD.",
                              _isPedestalSubstracted_override,
                              (bool)0.);


    registerOptionalParameter("SignalPolarityOverride",
                              "The polarity of the signals (1 for positive, -1 for negative) produced by all readout groups. This overrides any information provided by LCCD.",
                              _polarity_override,
                              1);

    registerOptionalParameter("ReadoutFrequencyOverride",
                              "The readout frequency of every readout group (in Hz). This overrides any information provided by LCCD.",
                              _readout_frequency_override,
                              (float)1.E9);
    // end of the override variables

    _pedestalListener = NULL;
    _adcElectronicsListener = NULL;
  }



  void PulseFinderProcessor::init()
  {

    // usually a good idea to
    printParameters();

    _nRun = 0;
    _nEvt = 0;

    // determine if the pedestal information is supplied by LCCD
    //If both the pedestal width and value are overridden no listener is necessary
    if ( !(parameterSet("PedestalWidthOverride")  && parameterSet("PedestalValueOverride")) )
      {
        _pedestalListener = new PedestalListener(_pedestalColName);
      }

    // determine if the polarity, maxADC value and readout frequency are supplied by LCCD
    // no listener needed if all three values overridden
    if (!(parameterSet("ReadoutFrequencyOverride") && parameterSet("SignalPolarityOverride") && parameterSet("MaximumADCValueOverride")))
      {
        _adcElectronicsListener = new ADCElectronicsListener(_electrParaObjName, _electrParaColName);
      }
  }



  void PulseFinderProcessor::processRunHeader(LCRunHeader* run)
  {
    run->parameters().setValue(_processorName + "_revision", "$Rev: 2755 $");
    for (ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++)
      {
        if (! i->second->isOptional() || i->second->valueSet())
          {
            run->parameters().setValue(_processorName + "_" + i->second->name(), i->second->value());
          }
      }

    _nRun++;
  }



  void PulseFinderProcessor::processEvent(LCEvent* evt)
  {
    // give a life sign if you are in debug mode
    _nEvt++;
    m_out(MESSAGE) << "Finding the pulses in " << _inputColName << "." << endl;

    try
      {
        LCCollection* InputCollection = evt->getCollection(_inputColName);

        // create the output collections for pulses and zero suppressed data (ADC spectra)
        IMPL::LCCollectionVec* outputColPulse = new IMPL::LCCollectionVec(LCIO::TRACKERPULSE);
        IMPL::LCCollectionVec* outputColData  = new IMPL::LCCollectionVec(LCIO::TRACKERDATA);

        // set the persistency flag as defined by the processor parameter
        outputColPulse->setTransient(!_outputIsPersistent);
        outputColData->setTransient(!_outputIsPersistent);

        // set the flags that cellID1 should be stored
        lcio::LCFlagImpl trkFlag(0);
        trkFlag.setBit(lcio::LCIO::TRAWBIT_ID1);
        trkFlag.setBit(lcio::LCIO::TRAWBIT_CM);
        outputColPulse->setFlag(trkFlag.getFlag());
        outputColData->setFlag(trkFlag.getFlag());

        /* temporary data object, to be filled by /raw/ pulse finding
             first pair object is the actual tracker data,
           second pair object is the quality
        */
        vector<pair<TrackerDataImpl*, int> > tempPulses;
        tempPulses.clear();

        unsigned int nData = InputCollection->getNumberOfElements();

        // first loop over all channels within the event to search for the /raw/ pulses
        for (unsigned int iData = 0; iData < nData; ++iData)
          {
            TrackerData* thisData = dynamic_cast<TrackerData*>(InputCollection->getElementAt(iData));

            //FIXME? If cast is not OK notify somehow or throw an exception
            if (thisData)    // cast is OK
              {
                m_out(DEBUG) << "Looking at Pad:" << thisData->getCellID0()
                << " module:" << thisData->getCellID1() << endl;

                /*
                 * I use the internal function GetPedestalWidth that automatically outputs
                 * the overridden value if the override is set or the pedestal width taken from the
                 * pedestal collection if that exist otherwise throws an exception
                 * This way the user MUST define an override if the pedestal collection is not taken from LCCD
                 */
                _weightedPulseStartThr = getPedestalWidth(thisData->getCellID0(), thisData->getCellID1()) * _pulseStartThreshold;
                _weightedPulseEndThr = getPedestalWidth(thisData->getCellID0(), thisData->getCellID1()) * _pulseEndThreshold ;




                //float noiseValue = -1.;

/*                if (_pedestalListener != NULL)
                  {
                    m_out(DEBUG) << "Pedestal Value for pad:" << thisData->getCellID0()
                    << " module:" << thisData->getCellID1() << " = "
                    << _pedestalListener->getPedestalValue(thisData->getCellID0(), thisData->getCellID1()) << endl;

                    m_out(DEBUG) << "Pedestal Width for pad:" << thisData->getCellID0()
                    << " module:" << thisData->getCellID1() << " = "
                    << _pedestalListener->getPedestalWidth(thisData->getCellID0(), thisData->getCellID1()) << endl;
                    _weightedPulseStartThr = _pedestalListener->getPedestalWidth(thisData->getCellID0(), thisData->getCellID1()) * _pulseStartThreshold;
                    _weightedPulseEndThr = _pedestalListener->getPedestalWidth(thisData->getCellID0(), thisData->getCellID1()) * _pulseEndThreshold;
                  }*/


/*                if (_weightedPulseStartThr < 0.)
                    _weightedPulseStartThr = _pulseStartThreshold;

                if (_weightedPulseStartThr < 0.)
                    _weightedPulseStartThr = _pulseEndThreshold;*/





                /*
                 *  this is the master call: search in the spectrum for pulses and
                 *  store them into the collection vector that is passed to the call
                 *  The third parameter is used to set the threshold for the pulse splitting
                 */

                findPulses(thisData, &tempPulses, getPedestalWidth(thisData->getCellID0(), thisData->getCellID1())*_pedestalWidthScale);
              }
          }

        // use the found pulses to calculate time and charge information and store the data
        for (vector<pair<TrackerDataImpl*, int> >::const_iterator pulseCandidate = tempPulses.begin();
             pulseCandidate < tempPulses.end(); ++pulseCandidate)
          {
            TrackerDataImpl* trackerData = pulseCandidate->first;
            int quality                  = pulseCandidate->second;

            //save Pulse
            TrackerPulseImpl* newPulse = new TrackerPulseImpl();
            newPulse->setCellID0(trackerData->getCellID0());
            newPulse->setCellID1(trackerData->getCellID1());

            _pulse myPulse = calculatePulseTimeAndCharge(trackerData);
            newPulse->setCharge(myPulse.getCharge());
            newPulse->setTime(myPulse.getTime());
            newPulse->setQuality(quality);
            vector<float> covarianceMatrix; // the "newly" added error information concerning the pulse
            covarianceMatrix.push_back(myPulse.getChargeError()*myPulse.getChargeError());
            covarianceMatrix.push_back(0.);
            covarianceMatrix.push_back(myPulse.getTimeError()*myPulse.getTimeError());
            newPulse->setCovMatrix(covarianceMatrix);

            if (streamlog::out.write< streamlog::DEBUG >())
              {
                m_out(DEBUG) << "[***] Found a pulse on [Pad/Module]: [" << trackerData->getCellID0()
                << "/" << trackerData->getCellID1() << "], at the time " << newPulse->getTime() << " +- " << covarianceMatrix.at(2) << " of length "
                << (trackerData->getChargeValues()).size()  << " containing charge " << newPulse->getCharge() << " +- " << covarianceMatrix.at(0)
                << " with quality flag " << quality
                << " (MultiplePulseCandidate:" << pulseflag::isMultiplePulseCandidate(quality) << ", "
                << "AnomalousShape:" << pulseflag::isAnomalousShape(quality) << ", "
                << "Underflow:" << pulseflag::isUnderflow(quality) << ", "
                << "Overflow:" << pulseflag::isOverflow(quality) << ")."
                << endl;

                m_out(DEBUG) << "[***] The pulse spectrum is : ";

                vector<float> theADCSpectrum = trackerData->getChargeValues();
                for (vector<float>::const_iterator adcVal = theADCSpectrum.begin(); adcVal < theADCSpectrum.end(); ++adcVal)
                  {
                    m_out(DEBUG) << *adcVal << " ";
                  }
                m_out(DEBUG) << endl;
              }

            // store the adc spectrum connected to the pulse
            // this is implicitly done by incorporating /_forceSpectrumSave/ into /quality/
            if (quality != 0 || _forceSpectrumSave)
              {
                // make a copy of the TrackerData
                TrackerDataImpl* newData = new TrackerDataImpl();
                newData->setCellID0(trackerData->getCellID0());
                newData->setCellID1(trackerData->getCellID1());
                newData->setTime(trackerData->getTime());
                vector<float> thisDataChargeVal = trackerData->getChargeValues();
                for (vector<float>::const_iterator iSlice = thisDataChargeVal.begin(); iSlice != thisDataChargeVal.end(); ++iSlice)
                  (newData->chargeValues()).push_back((*iSlice));

                // add the spectrum to the <tracker data> collection
                outputColData->addElement(newData);
                // and add the tracker data to the <tracker pulse> collection
                newPulse->setTrackerData(newData);
              }
            outputColPulse->addElement(newPulse);
#if 1
            delete trackerData;
#endif
          }

        evt->addCollection(outputColPulse, _outputPulseColName);
        evt->addCollection(outputColData, _outputDataColName);

        m_out(MESSAGE) << "Found " << outputColPulse->getNumberOfElements() << " pulses." << endl;
      }
    catch (DataNotAvailableException &e)
      {
        // if there is not input collection give a warning to the user
        m_out(ERROR) << "Warning of PulseFinderProcessor::processEvent: input collection "
             << _inputColName << " cannot be found in event " << _nEvt << endl;
      }
  }



  void PulseFinderProcessor::check(LCEvent* evt)
  {
  }



  void PulseFinderProcessor::end()
  {
    if (_pedestalListener != NULL)
      delete _pedestalListener;
    if (_adcElectronicsListener != NULL)
      delete _adcElectronicsListener;
  }



  void PulseFinderProcessor::findPulses(TrackerData* thisData, vector<pair<TrackerDataImpl*, int> >* tempPulses, float NoiseThreshold)
  {
    /** the basic idea to find pulses is split into three steps:
       (1) simply look for candidate pulses: above values between start and end threshold
          due to the high frequency of this already check here for a minimum length
       (2) search the candidate pulses for possible multiple pulse and split, if necessary
       (3) finally check the pulse boundaries and assemble the pulse out of the bins before and after the pulse, if possible
    */

    // get the maximum ADC value
    const float maximumADCValue = getMaxADCValue(thisData->getCellID1());

    // determine the polarity of the data
    const int polarity = getPolarity(thisData->getCellID1());

    float startTime  = thisData->getTime();

    vector<float> ADCSpectrum = thisData->getChargeValues();

    /* this is a help construct for the data (including a typedef)
       simply store the start and end position of the pulse
    */
    typedef std::vector<float>::iterator timeIter;
    vector<pair<timeIter, timeIter> > theCandidatePulses;

    /* do the first task:
       simply look for the values above the start threshold and below end threshold
    */
    brutePulseSearch(ADCSpectrum, &theCandidatePulses, polarity);

    /* do the check for multiple pulses,
       technically: divide the start- and end-iterator into a new pair
       TODO: in addition to pulse splitting adjust the charge at the boundaries
    */
    vector<internalSplitPulse> theExtendedCandidatePulses = splitMultiplePulses(&theCandidatePulses, NoiseThreshold);

    /* as last part determine the possible boundaries of the pulse;
       - check wether the storing of the time bins before and after the end is possible
       - set the flags accordingly
    */

    // this is the supposed minimum time between two pulses
    int nominalSeparation =  _saveNBinsBeforeStart + _saveNBinsAfterEnd;

    /* now prepare the pulse candidates for storage:
     - assign quality flags
     - check boundaries at the beginning and the end
     - return the data construct for storage
    */

    for (vector<internalSplitPulse>::const_iterator valueAddedPulseBoundary = theExtendedCandidatePulses.begin();
         valueAddedPulseBoundary < theExtendedCandidatePulses.end(); ++valueAddedPulseBoundary)
      {
        int quality = 0;

        // check if we're on the first or last pulse boundaries
        timeIter previous_pulse_end, next_pulse_start;

        // first special case: first pulse in question -> set previous end to the beginning of the ADCspectrum
        if (valueAddedPulseBoundary == theExtendedCandidatePulses.begin())
          previous_pulse_end = ADCSpectrum.begin();
        else
          previous_pulse_end = (valueAddedPulseBoundary - 1)->endBin;
        // second special case: the end pulse -> set next start to the end of the ADCSpectrum
        if (valueAddedPulseBoundary == (theExtendedCandidatePulses.end() - 1))
          next_pulse_start = ADCSpectrum.end();
        else
          next_pulse_start = (valueAddedPulseBoundary + 1)->startBin;

        // check for a double-pulse candidate
        // this is a misnomer? this specifies if a pulse has a too close neighbour
        if ((distance(previous_pulse_end, valueAddedPulseBoundary->startBin) < nominalSeparation)
            || (distance(valueAddedPulseBoundary->endBin, next_pulse_start) < nominalSeparation))
          {
            pulseflag::setMultiplePulseCandidate(quality);
          }

        // check for anomalous pulse shape -- TODO: how to determine this?
        if (false)
          {
            pulseflag::setAnomalousShape(quality);
          }

        // check for underflow
        //TODO: what is underflow? surely there will be some bin values of "zero" charge in simulations. perhaps this only makes sense when noise is included in the channel
        if (false)
          {
            pulseflag::setUnderflow(quality);
          }

        // check for overflow
        const timeIter max = max_element(valueAddedPulseBoundary->startBin, valueAddedPulseBoundary->endBin);

        float maxBinValue = maximumADCValue;
        // if data has already been pedestal substracted, take this into account for overflow value maxBinValue ( = maximumADCValue)
        if (isPedestalSubstracted(thisData->getCellID1()))
          {
            maxBinValue = maximumADCValue - floor(getPedestalValue(thisData->getCellID0(), thisData->getCellID1()));
          }

        if (*max >= maxBinValue) // overflow in at least one time bin
          {
            pulseflag::setOverflow(quality);;
          }

        if (valueAddedPulseBoundary->isSplit)
          {
            pulseflag::setSplit(quality);
          }

        // now copy the data to a TrackerData object
        // determine the start and end positions for the copy:
        timeIter copy_start = valueAddedPulseBoundary->startBin;
        timeIter copy_end = valueAddedPulseBoundary->endBin;

        // check wether the start point can be moved in front to the nonimal position
        const int start_displacement = std::max(-_saveNBinsBeforeStart, static_cast<int>(distance(copy_start, previous_pulse_end)));
        advance(copy_start, start_displacement);
        // check wether the end point can be moved the nonimally requested position
        advance(copy_end, min(_saveNBinsAfterEnd, static_cast<int>(distance(copy_end, next_pulse_start))));

        /* calculate the start time of the pulse to be stored:
        * this is calculated from the time of the pulse (see far above) and needs to be corrected for the modifications
        * within the pulse from its calculation
        */
        float currentTime = startTime + (valueAddedPulseBoundary->startBin - ADCSpectrum.begin()) + start_displacement;

        TrackerDataImpl* toStoreTrackerData;
        toStoreTrackerData = new TrackerDataImpl();
        toStoreTrackerData->setCellID0(thisData->getCellID0());
        toStoreTrackerData->setCellID1(thisData->getCellID1());
        toStoreTrackerData->setTime(currentTime);
        toStoreTrackerData->chargeValues().resize(distance(copy_start, copy_end) + 1);

        copy(copy_start, copy_end, toStoreTrackerData->chargeValues().begin());

        pair<TrackerDataImpl*, int> tempPair(toStoreTrackerData, quality);
        tempPulses->push_back(tempPair);
      }

    return;
  }



  void PulseFinderProcessor::brutePulseSearch(vector<float>& ADCSpectrum, vector<pair<vector<float>::iterator, vector<float>::iterator> >* theCandidatePulses, const int polarity)
  {
    typedef vector<float>::iterator timeIter;
    timeIter candStart, candEnd;

    // determine if the pulse is negative
    const bool negativePolarity = (polarity == -1);

    // also look if the pulse candidate meets the minimum requirements of length and height
    // loop through the pulse data looking for continuous areas of charge

    vector<float>::iterator current_value = ADCSpectrum.begin();

    int candidate_length = 0;
    float candidate_max_height = 0.;
    bool inside_pulse = false;

    while (current_value < ADCSpectrum.end())
      {
        if (!inside_pulse)
          {
            // we're not inside a pulse, look for the start
            const bool above_start_threshold = negativePolarity ? (*current_value <= _weightedPulseStartThr) : (*current_value >= _weightedPulseStartThr);

            if (above_start_threshold)
              {
                inside_pulse = true;
                candidate_length = 1;
                candidate_max_height = *current_value;
                candStart = current_value;
              }
          }
        else
          {
            // we're inside a pulse, look for the end
            const bool below_end_threshold = negativePolarity ? (*current_value > _weightedPulseEndThr) : (*current_value < _weightedPulseEndThr);
            const bool end_of_data = (current_value + 1 == ADCSpectrum.end());

            if (!below_end_threshold)
              {
                ++candidate_length;
                candidate_max_height = negativePolarity ? min(*current_value, candidate_max_height) : max(*current_value, candidate_max_height);
              }

            if (below_end_threshold || end_of_data)
              {
                // determine if this pulse is a candidate
                const bool meets_height_requirement = negativePolarity ? (candidate_max_height <= _minimumPulseHeight) : (candidate_max_height >= _minimumPulseHeight);

                const bool meets_length_requirement = (candidate_length >= _minimumPulseLength);

                if (meets_height_requirement && meets_length_requirement)
                  {
                    m_out(DEBUG1) << "Candidate accepted. Reason?: max height " << candidate_max_height << ", length " << candidate_length << "." << endl;
                    theCandidatePulses->push_back(pair<timeIter, timeIter>(candStart, current_value));
                  }
                else
                  {
                    // this feels like fishing, the pulse wasn't big enough so toss it back
                    m_out(DEBUG1) << "Candidate rejected. Reason?: below height " << !meets_height_requirement << ", below length " << !meets_length_requirement << "." << endl;
                  }

                // reset the candidate
                inside_pulse = false;
              }
          }

        ++current_value;
      }
  }



  vector<PulseFinderProcessor::internalSplitPulse> PulseFinderProcessor::splitMultiplePulses(vector<pair<vector<float>::iterator, vector<float>::iterator> >* theCandidatePulses, float NoiseThreshold)
  {
    typedef vector<float>::iterator timeIter;

    /* now check the candidates for multiple pulses
       since the iterators would be invalidated do this in two steps:
       first find the positions where the candidate needs to be split
        for this: construct the derivative vector for the pulse in question
       then do the actual splitting
    */

    // create a temporary return object for the "new" separated pulses
    vector<internalSplitPulse> tempPulseStorage;
    bool needsSplitting = false;

    for (vector<pair<timeIter, timeIter> >::iterator aPulse = theCandidatePulses->begin(); aPulse < theCandidatePulses->end(); ++aPulse)
      {
        vector<float> derivativeVec;
        vector<unsigned int> splitPositions;

        /* for each candidate in the lot:
           loop from start to end of pulse and check wether
           a change in the derivative from positive to negative occurs
        */
        for (timeIter aBin = aPulse->first; aBin < (aPulse->second - 1); ++aBin)
          {
            double derivative = *(aBin + 1) - *aBin;
            derivativeVec.push_back(derivative);
          }

        // now take a closer look at the candidate
        for (vector<float>::const_iterator dVal = derivativeVec.begin(); dVal + 1 < derivativeVec.end() ; ++dVal)
          {
            // check the sign change and if the absolute change is above the noise threshold
            // store the position of the /end/ of the split pulse + 1 (this is the split value +2)
            if ((*dVal < 0) && (*(dVal + 1) > 0) && (fabs(*dVal - *(dVal + 1)) > NoiseThreshold))
              {
                splitPositions.push_back(dVal - derivativeVec.begin() + 2);
                needsSplitting = true;
                /* TODO!!:
                   for a more exact charge evaluation the charge needs to be spread between the split pulses
                   double factor_change = *(dVal+1) / (*dVal - *(dVal + 1) );
                */
              }
          }

        if (splitPositions.empty())
          tempPulseStorage.push_back(internalSplitPulse(aPulse->first, aPulse->second, false));
        else
          {
            timeIter currentStart = aPulse->first;

            // do the splittings
            for (vector<unsigned int>::const_iterator split = splitPositions.begin(); split < splitPositions.end(); ++split)
              {
                // the current split pulse starts the currentStart position and ends after the given length
                tempPulseStorage.push_back(internalSplitPulse(currentStart, aPulse->first + *split , true));
                currentStart   = aPulse->first + *split;
              }
            // and store the trailing end
            tempPulseStorage.push_back(internalSplitPulse(currentStart, aPulse->second, true));
          }
      }
    return tempPulseStorage;
  }



  PulseFinderProcessor::_pulse PulseFinderProcessor::calculatePulseTimeAndCharge(const EVENT::TrackerData *pulse_data)
  {
      _pulse returnValue;

    // check if the fit method is used
    if (_use_fit_charge || _use_fit_time)
        returnValue = (calculatePulseWithFit(pulse_data));

    if (!_use_fit_charge)
      returnValue.add(calcPulseCharge(pulse_data));

    if (!_use_fit_time)
      returnValue.add(calcPulseTime(pulse_data));

    return returnValue;
  }



  PulseFinderProcessor::_pulse PulseFinderProcessor::calculatePulseWithFit(const EVENT::TrackerData *pulse_data)
  {
    // find the polarity of the pulse data
    const int polarity = getPolarity(pulse_data->getCellID1());

    const vector<float> &charge_values = pulse_data->getChargeValues();

    // fit a parabola to the log of the data around the maximum

    // find the maximum (or minimum)
    vector<float>::const_iterator extreme_point = (polarity == -1.) ? min_element(charge_values.begin(), charge_values.end()) : max_element(charge_values.begin(), charge_values.end());

    m_out(DEBUG) << "Spectrum is ";

    // use 3 points on each side (at most)

    for (vector<float>::const_iterator i_bin = charge_values.begin(); i_bin != charge_values.end(); i_bin++)
      {
        m_out(DEBUG) << ", " << *i_bin;
      }

    m_out(DEBUG) << endl;

    // check the begining pulses for non-negativity
    vector<float>::const_iterator fit_start = extreme_point;
    const int max_reverse = min(3, static_cast<int>(distance(charge_values.begin(), fit_start)));
    for (int i = 0; i < max_reverse; i++)
      {
        const float previous_value = *(fit_start - 1);
        if ((previous_value * polarity) > 0)
          {
            fit_start = fit_start - 1;
          }
        else
          {
            break;
          }
      }

    // and check the ending ones
    vector<float>::const_iterator fit_end = extreme_point;
    const int max_advance = min(3, static_cast<int>(distance(fit_end, charge_values.end())));
    for (int i = 0; i < max_advance; i++)
      {
        const float next_value = *(fit_end + 1);
        if ((next_value * polarity) > 0)
          {
            fit_end = fit_end + 1;
          }
        else
          {
            break;
          }
      }

    // initialize the fit
    const int point_count = distance(fit_start, fit_end) + 1;

    // make sure that the distance is greater than 3
    if (point_count == 1)
      {
        const int bin_time = static_cast<int>(pulse_data->getTime() + distance(charge_values.begin(), fit_start) + 0.5);
        const float time_bin_width = 1.E03 / getReadoutFrequency(pulse_data->getCellID1());
        const float fit_time = bin_time * time_bin_width;

        const float fit_amplitude = *fit_start;

        m_out(DEBUG0) << "Time = 0.5 + " << distance(charge_values.begin(), fit_start) << " + " << pulse_data->getTime() << " = " << bin_time << "bins" << endl;
        m_out(DEBUG0) << "Bin width = " << time_bin_width << endl;

        _pulse returnValue;
        returnValue.setCharge(fit_amplitude);
        returnValue.setTime(fit_time);

        return returnValue;
      }


    gsl_matrix *X = gsl_matrix_alloc(point_count, 3);
    gsl_vector *y = gsl_vector_alloc(point_count);
    gsl_vector *c = gsl_vector_alloc(3);
    gsl_matrix *covariance = gsl_matrix_alloc(3, 3);
    double minimum = 0.;

    m_out(DEBUG0) << "Calculating pulse time using the points ";

    vector<float>::const_iterator current_charge = fit_start;

    for (int row = 0; current_charge <= fit_end; current_charge++, row++)
      {
        // prevent bad log values
        if ((*current_charge * polarity) > 0.001)
          {
            gsl_vector_set(y, row, std::log(*current_charge * polarity));
          }
        else
          {
            continue;
          }

        // may as well transform the fit in the t-direction
        gsl_matrix_set(X, row, 0, 1.);
        gsl_matrix_set(X, row, 1, row);
        gsl_matrix_set(X, row, 2, row*row);

        m_out(DEBUG0) << '(' << row << ", " << *current_charge << ") ";
      }

    gsl_multifit_linear_workspace *workspace = gsl_multifit_linear_alloc(point_count, 3);

    // perform the fit
    gsl_multifit_linear(X, y, c, covariance, &minimum, workspace);

    m_out(DEBUG0) << ". Got: " << gsl_vector_get(c, 2) << "x^2 + " << gsl_vector_get(c, 1) << "x + " << gsl_vector_get(c, 0) << endl;

    // calculate the time
    // convert the result to nanoseconds
    const double translated_bin = -0.5 * gsl_vector_get(c, 1) / gsl_vector_get(c, 2);
    const double bin_time = pulse_data->getTime() + translated_bin + distance(charge_values.begin(), fit_start) + 0.5;

    const double time_bin_width = 1.E03 / getReadoutFrequency(pulse_data->getCellID1());
    const double fit_time = bin_time * time_bin_width;

    m_out(DEBUG0) << "Time = " << translated_bin << " + 0.5 + " << distance(charge_values.begin(), fit_start) << " + " << pulse_data->getTime() << " = " << bin_time << "bins" << endl;
    m_out(DEBUG0) << "Bin width = " << time_bin_width << endl;

    // calculate the amplitude
    const float max_log = gsl_vector_get(c, 0) - (0.25 * gsl_vector_get(c, 1) * gsl_vector_get(c, 1) / gsl_vector_get(c, 2));
    const float fit_amplitude = exp(max_log);

    // clean up the fit
    gsl_multifit_linear_free(workspace);
    gsl_matrix_free(X);
    gsl_vector_free(y);
    gsl_vector_free(c);

    _pulse returnValue;
    returnValue.setCharge(fit_amplitude);
    returnValue.setTime(fit_time);

    return returnValue;
  }



  PulseFinderProcessor::_pulseTime PulseFinderProcessor::calcPulseTime(const TrackerData* thisPulse)
  {
    vector<float> chargeVec = thisPulse->getChargeValues();

    // determine the polarity of the pulse from the electronics parameters
    const int polarity = getPolarity(thisPulse->getCellID1());

    // calculate the noise
    float noise_value = _pedestalWidthScale * getPedestalWidth(thisPulse->getCellID0(), thisPulse->getCellID1());

    /*
     * If the pedestal width is not defined and
     */
    // if pedestal width is not defined its default value is -1
    // then set pedestal width equal to 1
    if (noise_value <= 0.)
      noise_value = _pedestalWidthScale;

    float PulseTime      = 0.0;
    float PulseTimeError = 0.0;
    float Derivative     = 0.0;
    float Norm           = 0.0;
    float SliceTime      = thisPulse->getTime();

    // determine the mean value of the positive derivative == the best estimate for the pulse time
    if (chargeVec.size() > 1)
      {
        for (vector<float>::const_iterator timeBin = chargeVec.begin() + 1; timeBin < chargeVec.end(); ++timeBin)
          {
            Derivative  = (*timeBin) - (*(timeBin - 1));

            // make derivative sign insensitive
            Derivative *= polarity;

            SliceTime += 1.;

            // break out of calculation if decrease is larger than the noise
            if (Derivative < (-noise_value))
              break;

            Norm      += Derivative;
            PulseTime += Derivative * SliceTime;
          }
        if (Norm > 0.)
          PulseTime /= Norm;
        else // no iteration means a single slot that counts!
          PulseTime = SliceTime;
      }
    else
      PulseTime = SliceTime;

    // now determine the pulse time error -- which needs a second loop
    const float TimeSum = (PulseTime - thisPulse->getTime()) * Norm;

    if (chargeVec.size() > 1)
      {
        for (vector<float>::const_iterator timeBin = chargeVec.begin() + 1; timeBin < chargeVec.end(); ++timeBin)
          {
            Derivative  = (*timeBin) - (*(timeBin - 1));
            unsigned int index = timeBin - chargeVec.begin();

            // break out of calculation if decrease is larger than the noise
            if (Derivative < (-noise_value))
              break;

            double DerivativeVariance = (*timeBin) * (*timeBin) + (*(timeBin - 1)) * (*(timeBin - 1));
            PulseTimeError += (index * Norm - TimeSum) * (index * Norm - TimeSum) * DerivativeVariance;
          }
        PulseTimeError = sqrt(PulseTimeError) / (Norm * Norm);
      }
    else
      PulseTimeError = -1.;

    const float bin_width = 1.E03 / getReadoutFrequency(thisPulse->getCellID1());
    PulseTime = PulseTime * bin_width; //conversion in ns

    _pulseTime returnValue;
    returnValue.setTime(PulseTime);
    returnValue.setTimeError(PulseTimeError);

    return  returnValue;
  }



  PulseFinderProcessor::_pulseCharge PulseFinderProcessor::calcPulseCharge(const TrackerData* thisPulse)
  {

    const int polarity = getPolarity(thisPulse->getCellID1());
    const FloatVec charge_values = thisPulse->getChargeValues();

    float         PulseCharge     = 0.0;
    float         InvChargeError  = 0.0;

    // let's assume poissonian errors for the values in each bin (and cross our fingers, that it's true)
    for (vector<float>::const_iterator adcVal = charge_values.begin(); adcVal < charge_values.end(); ++adcVal)
      {
        PulseCharge += (*adcVal);
        InvChargeError += abs(*adcVal);
      }

    /// flip sign if polarity is negative
    PulseCharge *= polarity;

    _pulseCharge returnValue;
    returnValue.setCharge(PulseCharge);
    if (InvChargeError != 0.)
      returnValue.setChargeError(1. / sqrt(InvChargeError));
    else
      returnValue.setChargeError(0.);

    return returnValue;
  }



  /** Returns the pedestal value for a given TrackerData / TrackerPulse object taking into consideration whether it should be overwritten by a processor parameter
  */
  float PulseFinderProcessor::getPedestalValue(const int hardware_channel, const int readout_group)
  {
    if (parameterSet("PedestalValueOverride"))
      {
        m_out(DEBUG) << "Pedestal Value for channel: " << readout_group << ":" << hardware_channel
                << " overridden to " << _pedestal_value_override << std::endl;
        return _pedestal_value_override;
      }
    else
      {
        if (_pedestalListener != NULL)
        {
            m_out(DEBUG) << "Pedestal Value for channel: " << readout_group << ":" << hardware_channel
                    << " = " << _pedestalListener->getPedestalValue(hardware_channel, readout_group) << std::endl;
            return _pedestalListener->getPedestalValue(hardware_channel, readout_group);
        }
        else
          {
            throw tpcconddata::ConditionsObjectException(string("[PulseFinderProcessor]::getPedestalValue-Exception: The Pedestal value is not defined or can not be found!"));
            return 0.;
          }
      }
  }



  /** Returns the pedestal width for a given TrackerData / TrackerPulse object taking into consideration whether it should be overwritten by a processor parameter
   */
  float PulseFinderProcessor::getPedestalWidth(const int hardware_channel, const int readout_group)
  {
    if (parameterSet("PedestalWidthOverride"))
      {
        m_out(DEBUG) << "Pedestal Width for channel: " << readout_group << ":" << hardware_channel
                << " overridden to " << _pedestal_width_override << std::endl;
        return _pedestal_width_override;
      }
    else
      {
        if (_pedestalListener != NULL)
        {
            m_out(DEBUG) << "Pedestal Width for channel: " << readout_group << ":" << hardware_channel
                    << " = " << _pedestalListener->getPedestalWidth(hardware_channel, readout_group) << std::endl;
            return _pedestalListener->getPedestalWidth(hardware_channel, readout_group);
        }
        else
          {
            throw tpcconddata::ConditionsObjectException(string("[PulseFinderProcessor]::getPedestalWidth-Exception: The Pedestal width is not defined or can not be found!"));
            return 0.;
          }
      }
  }



  /** Returns if the data has been pedestal sibstracted in the read-out electronics or not
   */
  bool PulseFinderProcessor::isPedestalSubstracted(const int readout_group)
  {
    if (parameterSet("IsPedestalSubstractedOverride"))
      {
        return _isPedestalSubstracted_override;
      }
    else
      {
        if (_adcElectronicsListener != NULL)
          return _adcElectronicsListener->isPedestalSubstracted(readout_group);
        else
          {
            throw tpcconddata::ConditionsObjectException(string("[PulseFinderProcessor]::isPedestalSubstracted-Exception: If data is pedestal substracted or not is either not defined or can not be found!"));
            return 0;
          }
      }
  }


  /** Returns the polarity of a TrackerData / TrackerPulse object taking into consideration whether it should be overwritten by a processor parameter
   */
  int PulseFinderProcessor::getPolarity(const int readout_group)
  {
    if (parameterSet("SignalPolarityOverride"))
      {
        return _polarity_override;
      }
    else
      {
        if (_adcElectronicsListener != NULL)
          return _adcElectronicsListener->getPolarity(readout_group);
        else
          {
            throw tpcconddata::ConditionsObjectException(string("[PulseFinderProcessor]::getPolarity-Exception: The polarity is either not defined or can not be found!"));
            return 0;
          }
      }
  }



  /** Returns the readout frequency of a readout group (in Mhz) taking into consideration whether it should be overwritten by a processor parameter
   */
  float PulseFinderProcessor::getReadoutFrequency(const int readout_group)
  {
    if (parameterSet("ReadoutFrequencyOverride"))
      {

        return 1.E-06 * _readout_frequency_override;
      }
    else
      {
        if (_adcElectronicsListener != NULL)
            //In the DB the frequency is supposed to be stired in MHz
          return _adcElectronicsListener->getFrequency(readout_group);
        else
          {
            throw tpcconddata::ConditionsObjectException(string("[PulseFinderProcessor]::getReadoutFrequency-Exception: The Readout Frequency is not defined or can not be found!"));
            return 0.;
          }
      }
  }



  /** Returns the maximum ADC value of a readout group taking into consideration whether it should be overwritten by a processor parameter
   */
  float PulseFinderProcessor::getMaxADCValue(const int readout_group)
  {
    if (parameterSet("MaximumADCValueOverride"))
      {
        return _maximumADCValueOverride;
      }
    else
      {
        if (_adcElectronicsListener != NULL)
          return (float)(_adcElectronicsListener->getMaxADCValue(readout_group));
        else
          {
            throw tpcconddata::ConditionsObjectException(string("[PulseFinderProcessor]::getMaxADCValue-Exception: The maximum ADC Value is not defined or can not be found!"));
            return 0.;
          }
      }
  }
}
