#include "RowBasedHitFinderProcessor.h"

//C++
#include <iostream>
#include <algorithm>

//LCIO
#include <EVENT/LCFloatVec.h>
#include <IMPL/TrackerDataImpl.h>
#include <IMPL/TrackerPulseImpl.h>
#include <IMPL/TrackerHitImpl.h>
#include <IMPL/LCFlagImpl.h>

// GEAR
#include <gear/TPCParameters.h>
#include <gear/TPCModule.h>
#include <gearimpl/GlobalPadIndex.h>

//MARLIN
#include <marlin/Global.h>

//lccd
#include <lccd_exceptions.h>

// the exception by TPCCondData in MarlinTPC
#include "TPCCondDataExeptions.h"
#include "TPCConditions.h"

// the common flag word definitions in MarlinTPC
#include "FlagwordDefinitions.h"

using namespace std;
using namespace lcio;
using namespace gear;
using namespace marlin;
using namespace tpcconddata;

namespace marlintpc
{

  RowBasedHitFinderProcessor aRowBasedHitFinderProcessor;


  RowBasedHitFinderProcessor::RowBasedHitFinderProcessor() : Processor("RowBasedHitFinderProcessor")
  {
    _description = "RowBasedHitFinderProcessor calculates TrackerHits from TrackerPulses" ;

    // register steering parameters: name, description, class-variable, default value
    registerInputCollection(LCIO::TRACKERPULSE,
                            "InputTrackerPulses" ,
                            "Name of the input TrackerPulses collection"  ,
                            _inputTrackerPulsesCollectionName ,
                            string("TPCPulses")) ;

    registerOutputCollection(LCIO::TRACKERHIT,
                             "OutputTrackerHits" ,
                             "Name of the output TrackerHits collection"  ,
                             _outputTrackerHitsCollectionName ,
                             string("TPCHits")) ;

    registerOptionalParameter("InputTPCConditions",
                              "Optional: Name of the input collection containing the TPC conditions data",
                              _inputTPCConditionsCollectionName,
                              TPCConditions::getDefaultColName()) ;

    registerProcessorParameter("WriteOutputToStorage" ,
                               "if true, the output hits collection is written to storage (default: true)"  ,
                               _outputHitsPersistent,
                               true);

    registerProcessorParameter("ChannelCorrectionCollectionName" ,
                               "Name of the conditions collection in which e stored",
                               _channelCorrectionCollectionName,
                               tpcconddata::ChannelCorrection::getDefaultColName());

    registerOptionalParameter("MaxEmptyPadsOverride" ,
                              "This parameter is deprecated and will be removed in the next release."  ,
                              _deprecated ,
                              1) ;

    registerProcessorParameter("MaxDeadChannelsOverride" ,
                               "Maximum number of empty consecutive pads in hit (default: 1)"  ,
                               _maxDeadChannelsOverride ,
                               1) ;

    registerProcessorParameter("MinHitSizeOverride" ,
                               "Minimum size of hit / Minimum number of Pads (default: 1)"  ,
                               _minHitSizeOverride ,
                               1) ;

    registerProcessorParameter("MinMaxPulseSliceHeightOverride",
                               "Minimum size of the maximum slice value of the maximal pulse (in charge) in a hit.",
                               _minMaxPulseSliceHeightOverride,
                               float(12.)) ;

    registerProcessorParameter("MaxTimeSpreadOverride" ,
                               "Maximum time between pulses in a hit in ns (default: 200.)"  ,
                               _maxTimeSpreadOverride ,
                               float(200.)) ;

    registerOptionalParameter("VDriftOverride" ,
                              "Optional: Set drift velocity in case there is no conditions data in mm/mu s"  ,
                              _driftVelocityOverride ,
                              float(0.)) ;

    registerOptionalParameter("DeadPadsOverride",
                              "Optional: the gear pad indices of dead channels",
                              _deadChannelIndicesOverride,
                              _deadChannelIndicesOverride) ;

    registerOptionalParameter("ChargeConversionFactorOverride" ,
                              "Optional: The charge conversion factor from ADC values to primary electrons",
                              _chargeConversionFactorOverride ,
                              float(1.)) ;

    registerOptionalParameter("UseGlobalCoordinates",
                              "Optional: If set to true, the output coordinate system is changed to be the global coordinates (default is false).",
                              _useGlobalCoord ,
                              false);

    _channelCorrectionListener = 0; // initialize pointer to zero/NULL
    _tpcConditionsListener     = 0; // initialize pointer to zero/NULL
  }



  void RowBasedHitFinderProcessor::init()
  {
    printParameters();

    // get drift velicity, either measured value from calibration
    // or nominal value given as process parameter
    if (parameterSet("VDriftOverride"))
      {
        m_out(WARNING) << "the drift velocity is taken from the steering file. please change" << endl;
      }
    else // CAUTION! this is not only true for this value, but currently it's the only one in use
      {
	try
	  {
	    _tpcConditionsListener = new TPCConditionsListener(_inputTPCConditionsCollectionName);
	  }
	catch( lccd::LCCDException & )
	  {
	    m_out(ERROR) << "Collection named \"" << _inputTPCConditionsCollectionName<< "\" with TPCConditions is not available."<< std::endl;
	    m_out(ERROR) << "Use the ConditionsProcessor to provide it, or set the VDriftOverride parameter if this is not possible."<< std::endl;	    
	    throw;
	  }
      }

    if (parameterSet("DeadPadsOverride"))
      {
        m_out(WARNING) << "the indices of dead pads are taken from the steering file. please change" << endl;
      }
    else
      {
	try
	  {
	    _channelCorrectionListener = new ChannelCorrectionListener(_channelCorrectionCollectionName);
	  }
	catch( lccd::LCCDException & )
	  {
	    m_out(ERROR) << "Collection named \"" << _channelCorrectionCollectionName<< "\" with ChannelCorrections is not available."<< std::endl;
	    m_out(ERROR) << "Use the ConditionsProcessor to provide it, or set the DeadPadsOverride parameter if this is not possible."<< std::endl;	    
	    throw;
	  }
      }

    if (parameterSet("ChargeConversionFactorOverride"))
      {
        m_out(WARNING) << "the conversion factor from adc counts to charge is taken from the steering file. please change" << endl;
      }
    else if (_tpcConditionsListener == 0)
      {
 	try
	  {
	    _tpcConditionsListener = new TPCConditionsListener(_inputTPCConditionsCollectionName);
 	  }
	catch( lccd::LCCDException & )
	  {
	    m_out(ERROR) << "Collection named \"" << _inputTPCConditionsCollectionName<< "\" with TPCConditions is not available."<< std::endl;
	    m_out(ERROR) << "Use the ConditionsProcessor to provide it, or set the ChargeConversionFactorOverride parameter if this is not possible."<< std::endl;	    
	    throw;
	  }
     }

    if (parameterSet("UseGlobalCoordinates"))
      {
        m_out(MESSAGE) << "The cartesian coordinate system is used." << endl;
      }

    if (parameterSet("MaxEmptyPadsOverride"))
      {
        m_out(WARNING) << "The parameter [MaxEmptyPadsOverride] is deprecated and has no effect! It will be removed in the next release. Please use [MaxDeadChannelsOverride]." << endl;
      }
  }



  void RowBasedHitFinderProcessor::processRunHeader(LCRunHeader* run)
  {
    run->parameters().setValue(_processorName + "_revision", "$Rev: 1842 $");

    for (ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++)
      {
        if (!i->second->isOptional() || i->second->valueSet())
          run->parameters().setValue(_processorName + "_" + i->second->name(), i->second->value());
      }
  }



  void RowBasedHitFinderProcessor::processEvent(LCEvent* evt)
  {
    /** the processor is supposed to find "hits" in "rows"
    input are pulses: (charge,time,moduleID,cellID + qualityFlag)
    output are hits: ( [x,y,z] + covariance matrix, charge + qualityFlag )

    steering input: - minimal hit size
                    - max number of dead/empty pads within a hit
                    - max distance in z/time between pulses
    external input: - dead pads
                    - drift velocity
                    - charge conversion factor


        structurally this happens in a decomposition of:
        (A) sort the pulses by module & row
        (B) LOOP: look for pulses that belong together in a row (for all rows in a module, for all modules)
        (C) create the "Hit" by calculating all numbers

    **/

    // get the input pulse collection
    LCCollection* inputPulses;
    try
      {
        inputPulses = evt->getCollection(_inputTrackerPulsesCollectionName);
      }
    catch (DataNotAvailableException &e)
      {
        // if no pulses are available, skip to next event
        m_out(DEBUG) << "No pulses available in event " << evt->getEventNumber() << endl;
        return;
      }

    // create output hit collection and set the persistency flag
    LCCollectionVec* tpcHitCollection = new LCCollectionVec(LCIO::TRACKERHIT);
    tpcHitCollection->setTransient(!_outputHitsPersistent);

    // the main parameter for the geometry description is the gear parameter:
    const gear::TPCParameters& tpcParameters = Global::GEAR->getTPCParameters();

    /** ********************************************************************************
    the basic data structure for the hit search is a single row;
    the row is a simple vector of the Pulse pointers:
    vector<TrackerPulse*>

    this is repeated for every (unique) row on a module -> use a map of these
    repeat this for every possible module, since the module id is unique use a map

    so the final data structure for the hit search is:
    map<int moduleID, map<int rowID, vector<TrackerPulse*>* >
    (no typedef is introduced, although arguably this might improve the readability)

    a helper class is instantiated to hold the hit candidates: "hitCandidate"
    the temporary result of the hit candidate finding is stored in a vector of these
    the key is again a map -- since it is stored per /unique/ module


    ********************************************************************************
     **/

    map<int, map<int, vector<TrackerPulse*>*>*> theRowSortedPulsesPerModule;
    map<int, vector<hitCandidate> >             theHitCandidatesPerModule;

    // (A): sort pulses
    _sortPulses(inputPulses, tpcParameters, theRowSortedPulsesPerModule);

    // (B)
    _findHitCandidates(theRowSortedPulsesPerModule, tpcParameters, theHitCandidatesPerModule);

    // Clean up the maps
    map<int, map<int, vector<TrackerPulse*>* >* >::iterator aModuleWithPulses;
    map<int, vector<TrackerPulse*>* >::iterator aRowWithPulses;
    // outer loop:: the modules
    for (aModuleWithPulses = theRowSortedPulsesPerModule.begin();
         aModuleWithPulses != theRowSortedPulsesPerModule.end(); ++aModuleWithPulses)
      {
        // inner loop:: the rows
        for (aRowWithPulses  = aModuleWithPulses->second->begin();
             aRowWithPulses != aModuleWithPulses->second->end(); ++aRowWithPulses)
          {
            delete aRowWithPulses->second;
          }
        delete aModuleWithPulses->second;
      }

    // (C)
    _createHitCollection(theHitCandidatesPerModule, tpcParameters, tpcHitCollection);

    evt->addCollection(tpcHitCollection, _outputTrackerHitsCollectionName);

    m_out(DEBUG4) << "Found " << tpcHitCollection->getNumberOfElements() << " hits in the event." << endl;

    m_out(DEBUG3) << "There are " << inputPulses->getNumberOfElements()
    << " pulses to begin with, and a total of "
    << tpcHitCollection->getNumberOfElements() << " hits found." << endl;
  }



  void RowBasedHitFinderProcessor::check(LCEvent* evt)
  {
  }



  void RowBasedHitFinderProcessor::end()
  {
    if (_channelCorrectionListener != 0) // release memory, if needed
      delete _channelCorrectionListener;
  }



  void RowBasedHitFinderProcessor::_sortPulses(LCCollection* inputPulses, const gear::TPCParameters& tpcParameters,
                                               map<int, map<int, vector<TrackerPulse*>* >* >& rowSortedPulsesPerModule)
  {
    /**  put the pulses in a real container
     in principle the idea is easy:
       every unique row on a unique module gets another pulse added
       keep in mind that more than one pulse can occur on a single pad!

     practically this tends to be a bit /lengthy/ to write in stl containers,
     please refer to the stl documentation
    **/

    pair<map<int, map<int, vector<TrackerPulse*>*>*>::iterator, bool> moduleInserter;
    pair<map<int, vector<TrackerPulse*>*>::iterator, bool> rowInserter;

    for (int aPulse = 0; aPulse < inputPulses->getNumberOfElements(); ++aPulse)
      {
        TrackerPulse* inputPulse = dynamic_cast<TrackerPulse*>(inputPulses->getElementAt(aPulse));
        if (inputPulse) // the cast worked
          {
            const int moduleID         = inputPulse->getCellID1();
            const int padID            = inputPulse->getCellID0();

            const TPCModule& theModule = tpcParameters.getModule(moduleID);
            const int rowID            = theModule.getRowNumber(padID);

            // first create/use the module insertion iterator
            map<int, vector<TrackerPulse*>*>* rmap = new map<int, vector<TrackerPulse*>*>;
            moduleInserter =
              rowSortedPulsesPerModule.insert(pair<int,  map<int, vector<TrackerPulse*>* >* >
                                              (moduleID, rmap));
            if (!moduleInserter.second)
              delete rmap;
            // use the newly created/already existing map pointer to insert the row
            vector<TrackerPulse*>* pmap = new vector<TrackerPulse*>;
            rowInserter = ((moduleInserter.first)->second)->insert
                          (pair<int, vector<TrackerPulse*>* >(rowID, pmap));
            if (!rowInserter.second)
              delete pmap;

            // and finally add the Pulse itself
            ((rowInserter.first)->second)->push_back(inputPulse);
          }
      } // end input pulse loop
  }



  void RowBasedHitFinderProcessor::_findHitCandidates
  (map<int, map<int, vector<TrackerPulse*>* >* >& rowPulsesPerModule, const gear::TPCParameters& tpcParameters,
   map<int, vector<hitCandidate> >& theHitCandidateMap)
  {
    /**  now search in each row (of each module)
     a) for the highest pulse
     b) look for neighboring pulses
     c) store it in a candidate
     d) remove the stored pulses and redo from a)
    **/

    // two iterator types are needed: for the modules and for the rows
    map<int, map<int, vector<TrackerPulse*>* >* >::iterator aModuleWithPulses;
    map<int, vector<TrackerPulse*>* >::iterator aRowWithPulses;

    // outer loop:: the modules
    for (aModuleWithPulses = rowPulsesPerModule.begin(); aModuleWithPulses != rowPulsesPerModule.end(); ++aModuleWithPulses)
      {
        int moduleID = 0;
        vector<hitCandidate> theHitCandidates;

        // inner loop:: the rows
        for (aRowWithPulses  = aModuleWithPulses->second->begin(); aRowWithPulses != aModuleWithPulses->second->end(); ++aRowWithPulses)
          {
            m_out(DEBUG) << "Starting search of neighbouring pulses in in row: " << aRowWithPulses->first << endl;

            // third loop: now row wise, find hit candidates and remove the pulses consecutively
            while (aRowWithPulses->second->size() != 0)
              {
                // first find and remove the pulse with the highest charge from the list
                TrackerPulse* theSeedPulse = _findHighestPulseInRow(*(aRowWithPulses->second));
                moduleID = theSeedPulse->getCellID1();

                m_out(DEBUG) << "Found a highest pulse with charge = " << theSeedPulse->getCharge()
                << " on pad " << theSeedPulse->getCellID0() << " and with time " << theSeedPulse->getTime() << endl;
                // time saver, conclude the search,
                // if the maximal pulse height is below the minimum requirement
                if (!_maxPulseHeightCheck(theSeedPulse))
                  {
                    m_out(DEBUG) << "Pulse is rejected, since it is too small. Continue in next row. " << endl;
                    break;
                  }

                // take the found pulse as prototype hit
                hitCandidate theProtoHit(theSeedPulse);

                // look in the adjacent pads for further pulses, store them in a vector
                // this function modifies the initial list by removing pulses!
                // check if it's needed before
                if (aRowWithPulses->second->size() != 0)
                  _addNeighbourPulses(theProtoHit, tpcParameters, *(aRowWithPulses->second));

                theHitCandidates.push_back(theProtoHit);
              }
          }
        theHitCandidateMap.insert(pair<int, vector<hitCandidate> >(moduleID, theHitCandidates));

        m_out(DEBUG) << "On module " << aModuleWithPulses->first
        << " there are " << theHitCandidates.size() << " hits found." << endl;

      } // end outer loop  of modules
  }



  TrackerPulse* RowBasedHitFinderProcessor::_findHighestPulseInRow(vector<TrackerPulse*>& pulsesInARow)
  {
    TrackerPulse* highestPulse = pulsesInARow.at(0);

    vector<TrackerPulse*>::iterator eraser = pulsesInARow.begin();
    for (vector<TrackerPulse*>::iterator iter = pulsesInARow.begin(); iter < pulsesInARow.end(); ++iter)
      {
        if (fabs((*iter)->getCharge()) > fabs(highestPulse->getCharge()))
          {
            highestPulse = *iter;
            eraser = iter;
          }
      }
    pulsesInARow.erase(eraser);
    return highestPulse;
  }



  bool RowBasedHitFinderProcessor::_maxPulseHeightCheck(TrackerPulse* theSeedPulse)
  {
    EVENT::TrackerData*  trackerData = theSeedPulse->getTrackerData();
    vector<float> theADCSpectrum = trackerData->getChargeValues();

    return *max_element(theADCSpectrum.begin(), theADCSpectrum.end()) >= _getMinMaxPulseSliceHeight();
  }



  void RowBasedHitFinderProcessor::_addNeighbourPulses
  (hitCandidate& theCandidate,  const gear::TPCParameters& tpcParameters, vector<TrackerPulse*>& pulsesInARow)
  {
    // the initial values for searching
    const TPCModule& theModule      = tpcParameters.getModule((theCandidate.getMaximumPulse())->getCellID1());
    const int theStartPadIndex      = (theCandidate.getMaximumPulse())->getCellID0();
    const int theModuleIndex        = (theCandidate.getMaximumPulse())->getCellID1();
    const float theStartTime        = (theCandidate.getMaximumPulse())->getTime();
    const float theHighestCharge    = theCandidate.getMaxPulseHeight();

    // the temporary vars for searching
    unsigned int numberOfDeadChannels = 0;
    float tempCharge            = theHighestCharge;
    int tempPadIndex            = theStartPadIndex;
    float tempTime              = theStartTime;
    bool finishedLeft           = false; // arbitrary choice: start with searching "left", used to change search direction
    bool stopSearch             = false; // loop control

    // start loop until no further pulses are found
    while (!stopSearch)
      {
        try // to get the index of the next pad
          {
            if (!finishedLeft)
              tempPadIndex = theModule.getLeftNeighbour(tempPadIndex);
            else
              tempPadIndex = theModule.getRightNeighbour(tempPadIndex);
          }
        catch (...) // FIXME: what kind of exception is thrown?
          {
            m_out(DEBUG) << "The hit candidate is at the border!" << endl;
            hitflag::setAtModuleBorder(theCandidate.qualityFlag);

            if (!finishedLeft)
              {
                finishedLeft  = true;
                tempPadIndex  = theStartPadIndex;
              }
            else
              stopSearch = true;
          }

        // if more than one pulse matches the requirements, the best match needs to be found
        // store this in a temporary vector that is searched for the best match
        vector<vector<TrackerPulse*>::iterator> tempPulses;
        tempPulses.clear();

        // check the pulses if (a) they are on the right pad and (b) in the right time window
        for (vector<TrackerPulse*>::iterator iter = pulsesInARow.begin(); iter < pulsesInARow.end(); ++iter)
          {
            if (((*iter)->getCellID0() == tempPadIndex) && (fabs((*iter)->getTime() - tempTime) <  _getMaxTimeSpread()))
              tempPulses.push_back(iter);
          }

        // now make distinction on the number of matched pulses
        if (tempPulses.size() > 0) // if several pulses can be matched, find the best match!
          {
            if (tempPulses.size() > 1) // possibly a multiple hit in z !?
              hitflag::setMultipleHitCandidate(theCandidate.qualityFlag);

            // this function returns the iterator directly, if only one pulse is passed to it
            vector<TrackerPulse*>::iterator bestPulseMatch = _findBestNeighbourPulseMatch(tempPulses, tempTime);

            theCandidate.addPulse(*bestPulseMatch); // store the pulse into hit candidate
            tempTime = (*bestPulseMatch)->getTime(); // set the new time reference
            pulsesInARow.erase(bestPulseMatch); // delete the stored pulse from list to disallow double-counting
            tempPulses.clear();

            // do a quick check if there is a rise in charge -> indication for multiple hits
            float thisCharge = (*bestPulseMatch)->getCharge();
            if (fabs(thisCharge) > fabs(tempCharge))
              hitflag::setMultipleHitCandidate(theCandidate.qualityFlag);
            tempCharge = thisCharge;

            if (numberOfDeadChannels > 0) // there was at least one dead channel in between
              {
                hitflag::setDeadChannel(theCandidate.qualityFlag);
                numberOfDeadChannels = 0; // reset the counter, if needed
              }
          }
        else // nothing found
          {
            tempCharge = 0;
            if (_padIsDead(tempPadIndex, theModuleIndex)) // no pulse found because of dead pad
              numberOfDeadChannels++; // increase the counter
            /* in two cases the search will be abandoned:
            /            a) the pad isn't dead
            /            b) the allowed number of consecutive dead channels is exceeded */
            else // case a) the pad isn't dead -> "end"
              {
                if (numberOfDeadChannels > 0)
                  hitflag::setNextToDeadChannel(theCandidate.qualityFlag);

                if (!finishedLeft) // but if the other direction isn't searched yet -> start from scratch
                  {
                    finishedLeft   = true;
                    tempTime       = theStartTime;
                    numberOfDeadChannels = 0;
                    tempPadIndex   = theStartPadIndex;
                    tempCharge            = theHighestCharge;
                  }
                else // and end
                  stopSearch = true;
              }
          }
        // cover case b); unfortunately this can only be done *after* evaluation if the pad is dead
        // the "end" code is duplicated; the only other way is a goto (as far as i can think now)
        if (numberOfDeadChannels > _getMaxDeadChannels()) // this is a dead end!
          {
            hitflag::setNextToDeadChannel(theCandidate.qualityFlag);
            if (!finishedLeft) // but if the other direction isn't searched yet -> start from scratch
              {
                finishedLeft   = true;
                tempTime       = theStartTime;
                numberOfDeadChannels = 0;
                tempPadIndex   = theStartPadIndex;
                tempCharge      = theHighestCharge;
              }
            else // and end
              stopSearch = true;
          }
      }// search loop
    m_out(DEBUG) << "A hit candidate with a width of " << theCandidate.getWidth() << " is found." << endl;
  }



  vector<TrackerPulse*>::iterator RowBasedHitFinderProcessor::_findBestNeighbourPulseMatch
  (vector<vector<TrackerPulse*>::iterator>& pulseMatches, float theTime)
  {
    if (pulseMatches.size() == 1)
      return pulseMatches.at(0);
    else
      {
        vector<TrackerPulse*>::iterator bestMatch = *(pulseMatches.begin());
        float minTime = fabs((*bestMatch)->getTime() - theTime);

        for (vector<vector<TrackerPulse*>::iterator>::iterator searchIter = pulseMatches.begin();
             searchIter < pulseMatches.end(); ++searchIter)
          {
            if (fabs((**searchIter)->getTime() - theTime) < minTime)
              {
                minTime = fabs((**searchIter)->getTime() - theTime);
                bestMatch = *searchIter;
              }
          }
        m_out(DEBUG) << "In the hit more than one pulse was found to match to the protohit,"
        "the best match in time was chosen." << endl;
        return bestMatch;
      }
  }



  void RowBasedHitFinderProcessor::_createHitCollection
  (map<int, vector<hitCandidate> >& theHitCandidateMap, const gear::TPCParameters& tpcParameters, LCCollectionVec* tpcHitCollection)
  {
    /** now loop over all hits on all modules and:
    1) check their validity
    2) calculate the geometry dependent position of the fixed coordinate given by the row (y or r)
    3) calculate the geometry dependent position of the free coordinate (x or phi)
    4) calculate the z coordinate
    5) calculate the error matrix, which is diagonal in the local coordinate system
    6) store the TrackerHit in the output collection
    **/

    map<int, vector<hitCandidate> >::iterator hitCandidatesPerModule;

    // loop over modules
    for (hitCandidatesPerModule = theHitCandidateMap.begin();
         hitCandidatesPerModule != theHitCandidateMap.end(); ++hitCandidatesPerModule)
      {
        const gear::TPCModule &module =
          marlin::Global::GEAR->getTPCParameters().getModule((*hitCandidatesPerModule).first);
        //marlin::Global::GEAR->getTPCParameters().getModule(currentPulse->first.getModuleID () );

        //loop over hitCandidates:
        for (vector<hitCandidate>::iterator candidate = hitCandidatesPerModule->second.begin();
             candidate < hitCandidatesPerModule->second.end(); ++candidate)
          {
            // check validity
            if (_hitIsValid(*candidate) == false)
              continue;

            // change output coordinate system
            double position[3];
            // make the error calculation easier:
            double meanFreeCoordinate;
            if (_useGlobalCoord)
              {
                gear::Vector2D positionGlobal2D = module.localToGlobal(
                                                    _calculateFixedCoordinate(*candidate, tpcParameters),
                                                    _calculateFreeCoordinate(*candidate, tpcParameters));
                position[0] = positionGlobal2D[0];
                position[1] = positionGlobal2D[1];
                position[2] = _calculateZCoordinate(*candidate);
                meanFreeCoordinate = position[1];
              }
            else
              {
                meanFreeCoordinate = _calculateFreeCoordinate(*candidate, tpcParameters);
                position[0] = meanFreeCoordinate;
                position[1] = _calculateFixedCoordinate(*candidate, tpcParameters);
                position[2] = _calculateZCoordinate(*candidate);
              }



            float covMatrix[6] = {pow(_calculateFreeError(*candidate, meanFreeCoordinate, tpcParameters), 2) , 0.,
                                  pow(_calculateFixedError(*candidate, tpcParameters), 2) , 0., 0.,
                                  pow(_calculateZError(*candidate), 2)
                                 };

            TrackerHitImpl* theHit = new TrackerHitImpl();
            //      theHit->setType(int type) ???
            theHit->setPosition(position);
            theHit->setCovMatrix(covMatrix);
            theHit->setEDep(_convertCharge(*candidate));
            theHit->setEDepError(sqrt(_convertCharge(*candidate)));


            vector<TrackerPulse*> hitPulses = candidate->getPulses();
            for (vector<TrackerPulse*>::const_iterator aHitPulse = hitPulses.begin();
                 aHitPulse < hitPulses.end(); ++aHitPulse)
              {
                theHit->rawHits().push_back(*aHitPulse);

                // determine the quality bits from the pulses to let hit inherit them
                int pulseQuality = (*aHitPulse)->getQuality();
                if (pulseflag::isSplit(pulseQuality)) // split pulse?
                  hitflag::setSplitPulse(candidate->qualityFlag);
                if (pulseflag::isOverflow(pulseQuality)) // overflow
                  hitflag::setOverflowPulse(candidate->qualityFlag);
                if (pulseflag::isAnomalousShape(pulseQuality)) // anomalous shape
                  hitflag::setAnomalousPulseShape(candidate->qualityFlag);

              }
            theHit->setQuality(candidate->qualityFlag);

            // the time is inherited from the pulse with the largest charge
            theHit->setTime(candidate->getMaximumPulse()->getTime());

            m_out(DEBUG) << "The hit coordinates are [" << position[0] << "|" << position[1] << "|" << position[2] << "]" << endl
            << " with variances [" << theHit->getCovMatrix().at(0) << "|" << theHit->getCovMatrix().at(3) << "|" <<  theHit->getCovMatrix().at(5)  << "]" << endl
            << ", the charge is " << theHit->getEDep() << " +- " << theHit->getEDepError() << " and has the quality byte: ["  << theHit->getQuality() << "]" << endl;
            m_out(DEBUG) << "The detailed list of hit flags: " <<  " has dead channel [" <<  hitflag::hasDeadChannel(theHit->getQuality())
            << "], has noisy channel [" <<  hitflag::hasNoisyChannel(theHit->getQuality())
            << "], is at the border of a module [" <<  hitflag::isAtModuleBorder(theHit->getQuality())
            << "], has an overflow pulse [" <<  hitflag::hasOverflowPulse(theHit->getQuality())
            << "], contains a split pulse [" <<  hitflag::hasSplitPulse(theHit->getQuality())
            << "], has an anomalous shaped pulse [" <<  hitflag::hasAnomalousPulseShape(theHit->getQuality())
            << "], is next to a dead channel [" <<  hitflag::isNextToDeadChannel(theHit->getQuality())
            << "], and is a multiple hit candidate [" <<  hitflag::isMultipleHitCandidate(theHit->getQuality()) << "]." << endl;



            tpcHitCollection->addElement(theHit);
          }
      }
    return;
  }



  bool RowBasedHitFinderProcessor::_hitIsValid(const hitCandidate& cand)
  {
    return (int)cand.getWidth() >= _getMinHitSize();
  }



  double RowBasedHitFinderProcessor::_calculateFixedCoordinate(const hitCandidate& cand, const gear::TPCParameters& tpcParameters)
  {
    const TrackerPulse* maxPulse = cand.getMaximumPulse();

    const gear::TPCModule &module = tpcParameters.getModule(maxPulse->getCellID1());
    const gear::PadRowLayout2D &localPadPlane = module.getLocalPadLayout();
    Vector2D padCenterLocal = localPadPlane.getPadCenter(maxPulse->getCellID0());

    if (localPadPlane.getCoordinateType() == PadRowLayout2D::CARTESIAN)
      return padCenterLocal[1]; // this is truly y
    else if (localPadPlane.getCoordinateType() == PadRowLayout2D::POLAR)
      return padCenterLocal[0]; // why has this to be a different index in polar... -> this is r, anyway
    else
      throw gear::Exception("Coordinate type unknown!");
  }



  double RowBasedHitFinderProcessor::_calculateFreeCoordinate(const hitCandidate& cand, const gear::TPCParameters& tpcParameters)
  {
    const TrackerPulse* maxPulse = cand.getMaximumPulse();
    const gear::TPCModule &module = tpcParameters.getModule(maxPulse->getCellID1());
    const gear::PadRowLayout2D &localPadPlane = module.getLocalPadLayout();

    vector<TrackerPulse*> thePulses = cand.getPulses();

    double chargeSum             = 0.;
    double weightedSum           = 0.;
    double position              = 0.;
    double weightedCorrectionSum = 0.;
    const double localPI         = 3.41416; // i know, this is just a local Pi version, I don't know which math lib is available
    Vector2D padCenterLocal;
    for (vector<TrackerPulse*>::const_iterator aPulse = thePulses.begin(); aPulse < thePulses.end(); ++aPulse)
      {
        chargeSum += (*aPulse)->getCharge();
        padCenterLocal = localPadPlane.getPadCenter((*aPulse)->getCellID0());

        if (localPadPlane.getCoordinateType() == PadRowLayout2D::CARTESIAN)
          position = padCenterLocal[0]; // cartesian x
        else if (localPadPlane.getCoordinateType() == PadRowLayout2D::POLAR)
          {
            position = padCenterLocal[1]; // polar phi
            // possible trouble: phi definition range is [-Pi,+Pi]
            // for the sake of computation: first shift to [Pi,3*Pi] and then back again
            position += 3.*localPI;
          }
        else
          throw gear::Exception("Coordinate type unknown!");

        weightedSum += (*aPulse)->getCharge() * position;
        weightedCorrectionSum += (*aPulse)->getCharge() * 3.*localPI;
      }

    // now correct back for polar coordinates:
    if (localPadPlane.getCoordinateType() == PadRowLayout2D::POLAR)
      weightedSum -= weightedCorrectionSum;

    if (chargeSum > 0.)
      return weightedSum / chargeSum;
    else
      return 0.;
  }



  double RowBasedHitFinderProcessor::_calculateZCoordinate(const hitCandidate& cand)
  {
    //  return _getDriftVelocity() *(cand.getMaximumPulse())->getTime() / 1000.; // time is in [ns], drift velocity in [mm/mus]

    //fg: the z-position should be: zAnode (max drift length) - driftTime * driftVelocity
    return  Global::GEAR->getTPCParameters().getMaxDriftLength()
            -  _getDriftVelocity() *(cand.getMaximumPulse())->getTime() / 1000.; // time is in [ns], drift velocity in [mm/mus]
  }



  double RowBasedHitFinderProcessor::_calculateFreeError(const hitCandidate& cand, double freeMean, const gear::TPCParameters& tpcParameters)
  {
    const TrackerPulse* maxPulse = cand.getMaximumPulse();
    const gear::TPCModule &module = tpcParameters.getModule(maxPulse->getCellID1());
    const gear::PadRowLayout2D &localPadPlane = module.getLocalPadLayout();

    // returns automatically the right coordinate system
    // The width of the pad at padIndex in mm (CARTESIAN) or radians (POLAR).
    double padWidth = module.getPadWidth(maxPulse->getCellID0());

    const double SQRT12 = 3.46410161513775458705;
    double position     = 0.;
    double freeError    = 0.;

    vector<TrackerPulse*> thePulses = cand.getPulses();
    Vector2D padCenterLocal;

    if (cand.getWidth() == 1)
      return padWidth / SQRT12; // it's the pad width divided by sqrt(12)
    else
      {
        for (vector<TrackerPulse*>::const_iterator aPulse = thePulses.begin(); aPulse < thePulses.end(); ++aPulse)
          {
            padCenterLocal = localPadPlane.getPadCenter((*aPulse)->getCellID0());

            if (localPadPlane.getCoordinateType() == PadRowLayout2D::CARTESIAN)
              position = padCenterLocal[0]; // cartesian x
            else if (localPadPlane.getCoordinateType() == PadRowLayout2D::POLAR)
              {
                position = padCenterLocal[1]; // polar phi
              }
            else
              throw gear::Exception("Coordinate type unknown!");
            freeError += pow((position - freeMean) , 2);
          }
      }
    const double theTotalCharge = cand.getCharge();

    return sqrt(freeError) / theTotalCharge;
  }



  double RowBasedHitFinderProcessor::_calculateFixedError(const hitCandidate& cand, const gear::TPCParameters& tpcParameters)
  {
    const TrackerPulse* maxPulse = cand.getMaximumPulse();
    const gear::TPCModule& module = tpcParameters.getModule(maxPulse->getCellID1());

    const double SQRT12 = 3.46410161513775458705;

    int rowNumber    = module.getRowNumber(maxPulse->getCellID0());
    double rowHeight = module.getRowHeight(rowNumber);

    return rowHeight / SQRT12;
  }



  double RowBasedHitFinderProcessor::_calculateZError(const hitCandidate& cand)
  {
    // for now the z position is calculated simply from the time of the maximum pulse
    // therefore the error on z is the same as for the pulseMatches

    const TrackerPulse* maxPulse = cand.getMaximumPulse();
    double pulseTimeError = sqrt((maxPulse->getCovMatrix()).at(2));

    return _getDriftVelocity() * pulseTimeError / 1000.;
  }



  double RowBasedHitFinderProcessor::_convertCharge(const hitCandidate& cand)
  {
    return _getChargeConversionFactor() *(cand.getCharge());
  }



  RowBasedHitFinderProcessor::hitCandidate::hitCandidate(TrackerPulse* aPulse)
      : qualityFlag(0), _maxPulse(NULL), _maxPulseCharge(0.0), _numberOfDeadChannels(0)
  {
    _thePulses.clear();
    addPulse(aPulse);
  }



  void RowBasedHitFinderProcessor::hitCandidate::addPulse(TrackerPulse* aPulse)
  {
    _thePulses.push_back(aPulse);
    if (fabs(aPulse->getCharge()) > fabs(_maxPulseCharge))
      {
        _maxPulseCharge = aPulse->getCharge();
        _maxPulse       = aPulse;
      }
  }


  void RowBasedHitFinderProcessor::hitCandidate::addDeadChannel()
  {
    _numberOfDeadChannels++;
  }



  double RowBasedHitFinderProcessor::hitCandidate::getCharge() const
  {
    double tmpCharge = 0.0;
    for (vector<TrackerPulse*>::const_iterator aPulse = _thePulses.begin(); aPulse < _thePulses.end(); ++aPulse)
      tmpCharge += (*aPulse)->getCharge();
    return tmpCharge;
  }



  double RowBasedHitFinderProcessor::hitCandidate::getMaxPulseHeight()
  {
    return _maxPulseCharge;
  }



  unsigned int RowBasedHitFinderProcessor::hitCandidate::getWidth() const
  {
    return _thePulses.size();
  }



  bool RowBasedHitFinderProcessor::_pulseComparator(TrackerPulse* p0, TrackerPulse* p1)
  {
    return fabs(p0->getCharge()) < fabs(p1->getCharge());
  }



  unsigned int RowBasedHitFinderProcessor::_getMaxDeadChannels()
  {
    if (parameterSet("MaxDeadChannelsOverride"))
      {
        return _maxDeadChannelsOverride;
      }
    else
      {
        throw tpcconddata::ConditionsObjectException(string("[RowBasedHitFinderProcessor]::_getDeadChannels()  Exception: number can't be found!"));
      }
  }



  int RowBasedHitFinderProcessor::_getMinHitSize()
  {
    if (parameterSet("MinHitSizeOverride"))
      {
        return _minHitSizeOverride;
      }
    else
      {
        throw tpcconddata::ConditionsObjectException(string("[RowBasedHitFinderProcessor]::_getMinHitSize()  Exception: Number can't be found!"));
      }
  }



  double RowBasedHitFinderProcessor::_getMinMaxPulseSliceHeight()
  {
    if (parameterSet("MinMaxPulseSliceHeightOverride"))
      {
        return _minMaxPulseSliceHeightOverride;
      }
    else
      {
        throw tpcconddata::ConditionsObjectException(string("[RowBasedHitFinderProcessor]::_getMinMaxPulseSliceHeight()  Exception: Number can't be found!"));
        return 0.;
      }
  }



  double RowBasedHitFinderProcessor::_getMaxTimeSpread()
  {
    if (parameterSet("MaxTimeSpreadOverride"))
      {
        return _maxTimeSpreadOverride;
      }
    else
      {
        throw tpcconddata::ConditionsObjectException(string("[RowBasedHitFinderProcessor]::_getMaxTimeSpread()  Exception: Number can't be found!"));
        return 0.;
      }
  }



  double RowBasedHitFinderProcessor::_getDriftVelocity()
  {
    if (parameterSet("VDriftOverride"))
      {
        return _driftVelocityOverride;
      }
    else
      {
        try
          {
            return _tpcConditionsListener->getDriftVelocity();
          }
        catch (...)
          {
            m_out(ERROR) << "[getDriftVelocity failed] The Listener for TPCConditions could not be used. Maybe the conditions object is not available?" << endl;
            m_out(ERROR) <<  "                          Possibly you need to use the override parameterin the steering file called (\"VDriftOverride\")." << endl;
            abort();
          }
      }
  }



  bool RowBasedHitFinderProcessor::_padIsDead(const int padID, const int moduleID)
  {
    if (parameterSet("DeadPadsOverride"))
      return (_deadChannelIndicesOverride.end() != find(_deadChannelIndicesOverride.begin(), _deadChannelIndicesOverride.end(), padID));
    else
      {
        pair<int, int> tmpPair;
        tmpPair.first  = padID;    // CellID0
        tmpPair.second = moduleID; // CellID1
        return _channelCorrectionListener->isDeadFromGearID(tmpPair);
      }
  }



  bool RowBasedHitFinderProcessor::_padIsNoisy(const int moduleID, const int padID)
  {
    return false;
  }



  double RowBasedHitFinderProcessor::_getChargeConversionFactor()
  {
    if (parameterSet("ChargeConversionFactorOverride"))
      {
        return _chargeConversionFactorOverride;
      }
    else
      {
        return _tpcConditionsListener->getADCtoPrimaryElectronsFactor();
      }
  }
}// namespace marlintpc
