#ifndef TimePixPixelInterpolationProcessor_h
#define TimePixPixelInterpolationProcessor_h 1

// C++
#include <string>

// LCIO
#include "lcio.h"

// Marlin
#include "marlin/Processor.h"
#include "TimePixPixelMode.h"
#include "TimePixMapHandlerProcessor.h"
#include "lccd/ConditionsMap.hh"

//AIDA
#include <AIDA/AIDA.h>

namespace marlintpc
{

/* the following comment is tranfered to doxygen documentation
 * which begins with a second asterix, or in case of on line comment an extra third slash
 */
/**  Analysis Processor of MarlinTPC for TimePix Events with extraordinary high TOT values
 *
 *   finds pixel with extremely high TOT values and sets their mode to ONEHIT for this event only.
 *
 *
 *  <h4>Input - Prerequisites</h4>
 *  a collection of TrackerHits
 *
 *  <h4>Output</h4>
 *  Collection containing zero suppressed TrackerRawData.
 *
 * @param InputCollectionName Name of input collection, TrackerRawData
 * @param OutputCollectionName Name of output collection, zero suppressed TrackerRawData
 * @param SetOutputTransient If not 0 the output collection is set transient
 * @param NColumn number of columns of the chip
 * @param NRow number of rows of the chip
 * @param _onlyRegardNonZeroPixels For interpolation only Pixels with ADC-Values larger than 0 are taken into account
 *
 * @author Simone Zimmermann
 */


class TimePixPixelInterpolationProcessor : public marlin::Processor
{

    public:

        /* This method will be called by the marlin package
         * It returns a processor of the currend type (here TimePixPixelInterpolationProcessor)
         */
        virtual Processor*  newProcessor()
        {
            return new TimePixPixelInterpolationProcessor ;
        }

        /* the default constructor
         * here the processor parameters are registered to the marlin package
         * other initialisation should be placed in the init method
         */
        TimePixPixelInterpolationProcessor() ;

        /* Called at the beginning of the job before anything is read.
         * Use to initialize the processor, e.g. book histograms
         */
        virtual void init() ;

        /* Called for every run.
         * in this processor it is not used
         */
        virtual void processRunHeader(lcio::LCRunHeader* run) ;

        /* Called for every event - the working horse.
         * Here the real conversion take place
         */
        virtual void processEvent(lcio::LCEvent * evt) ;

        /* This method is only called if the check flag is set (default) in the main processor
         * (where the input files are given)
         * It should be used to create check plots which are not needed in a complete analysis
         * (where normaly the check flag is deactivated)
         */
        virtual void check(lcio::LCEvent * evt) ;


        /* Called after data processing for clean up.
         * e.g. saving control histogramms, delete created objects
         * (which are not stored in a collection/event), etc.
         */
        virtual void end() ;

        virtual float interpolatePixel(int chipID, int pixelID);



    protected:

        /* normally the doxygen documentation is palced before the method/variable
         * but for one line comments in case of variables it may be more convenied for you
         * to place it after the variable. In this case you use an arrow to tell doxygen
         * to which variable the commet belongs to.
         */

        std::string _inputColName ; ///<- the name of the input collection
        std::string _outputColName ; ///<- the name of the output collection

        int _nColumn ; ///<-Number of columns of the chip
        int _nRow ; ///<-Number of rows of the chip
        int _chipID; ///<- number of current chip, 0 if single chip
        int _minNoPixelToInterpolate;///<- minimum number of pixels needed to interpolate
        int _onlyRegardNonZeroPixels; ///<-For interpolation only Pixels with ADC-Values larger than 0 are taken into account

        double _totThreshold; ///<-threshold on relation of TOT value to neighboring TOT values

        /** give the status which is set to the transient flag of the output collection
         *  it is an int instead of a boolean, because the processor parameter cannot be a boolean
         */
        int _outputIsTransient;

        /** is true till the first output collection is wirten to an event
         *  used to to store parameters only in the first collection
         *  (also done, if collection is transient)
         */
        bool isFirstCollection;


        tpcconddata::TimePixMap* _timePixMap;
        std::map< tpcconddata::TimePixPixelMode::PixelKey, float> _inputMap;

} ;

}

#endif

