#include "TimePixHitCenterCalculatorProcessor.h"
#include "TimePixMapHandlerProcessor.h"
#include "TimePixPixelMode.h"

// C++
#include <iostream>
#include <queue>
#include <string>
#include <sstream>
#include <cmath>
#include <stdlib.h>

// LCIO
#include <EVENT/LCCollection.h>
#include <EVENT/TrackerData.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerHitImpl.h>
#include <IMPL/TrackerDataImpl.h>
#include <IMPL/LCRelationImpl.h>
#include "UTIL/LCTOOLS.h"
#include "UTIL/LCRelationNavigator.h"

// Aida
#include <AIDA/AIDA.h>
#include <marlin/AIDAProcessor.h>
#include <AIDA/IPlotter.h>
#include <AIDA/IPlotterFactory.h>
#include <RAIDA/RAIDAUtil.h>
#include <RAIDA/IBrushStyleROOT.h>
#include <RAIDA/IPlotterFactoryROOT.h>
#include <AIDA/IMarkerStyle.h>

//Root
#include "TROOT.h"
#include "TCanvas.h"
#include "TFile.h"
#include "TH2F.h"
#include "THStack.h"
#include "TLine.h"

//marlin
#include <marlin/ConditionsProcessor.h>
#include <marlin/Global.h>

//GEAR
#include <gear/GEAR.h>
#include <gear/TPCModule.h>
#include <gear/TPCParameters.h>
#include <gear/PadRowLayout2D.h>

using std::string;
using namespace tpcconddata;
using namespace lcio ;
using namespace marlin ;

namespace marlintpc
{


TimePixHitCenterCalculatorProcessor aTimePixHitCenterCalculatorProcessor ;


TimePixHitCenterCalculatorProcessor::TimePixHitCenterCalculatorProcessor() :
        Processor("TimePixHitCenterCalculatorProcessor")
{
    /* the constructor call first the super constructor which the name
     * of the processor
     * this must by the same as the class name.
     */

    /* description of the processor which will be displayed in the
     * steering file autometicly generated by
     * marlin (use MarlinTPC -x (XML-format) or MarlinTPXC -l (text format))
     */
  _description = "TimePixHitCenterCalculatorProcessor calculates the center of gravity of each cluster" ;


    /* register steering parameters: name, description, class-variable, default value
     * the type will by definded by the type of the default value
     * string, double, float and int are possible
     * plaese note to get a float you must cast in C++: (float) 1.3
     */
  registerInputCollection(LCIO::TRACKERHIT,
                         "InputCollectionName" ,
                         "Name of the input collection"  ,
                         _inputColName ,
                         std::string("TimePixSepHitCandidates")) ;

  registerOutputCollection(LCIO::TRACKERHIT,
                          "OutputCollectionName" ,
                          "Name of the output collection"  ,
                          _outputColName ,
                          std::string("TimePixHits")) ;

  registerInputCollection(LCIO::LCGENERICOBJECT,
                         "ModeMapCollectionName" ,
                         "Name of the mode map collection"  ,
                         _modeMapColName ,
                         std::string("TimePixModeMap")) ;

  registerInputCollection(LCIO::LCGENERICOBJECT,
                         "ModeMapCorrCollectionName" ,
                         "Name of the collection containing the correction to the mode map"  ,
                         _modeMapCorrName ,
                         std::string("TimePixModeMapCorr")) ;

  registerProcessorParameter( "max_zDiffusion_perCluster",
                         "pixels further than _max_zDiffusion_perCluster away from the mean cluster time are ignored during cluster center calculation [clock cycles]"  ,
                         _max_zDiffusion_perCluster,
                         (int) 30 ) ;

  registerProcessorParameter("Shutter" ,
                         "length of shutter signal in us"  ,
                         _shutter ,
                         (double) 108.) ;

  registerProcessorParameter("MultipleShutter" ,
                         "Allow multiple shutters during one acquisition" ,
                         _multiShutter ,
                         false) ;

  registerProcessorParameter( "keepClusterHistoFlag" ,
                         "decide whether you want to keep a histo of z-Distribution for each Cluster - default is FALSE" ,
			 _keepClusterHistoFlag,
                         false ) ;
				    
  registerProcessorParameter( "_minNrOfPixelsToCalcTime" , 
                         "if there are less than _minNrOfPixelsToCalcTime with 4 or more neighbors with charge all time pixels are taken into account for mean time calculation"  ,
                         _minNrOfPixelsToCalcTime,
                         (int) 4 ) ;				    
				    
  registerProcessorParameter( "_rmsCut" , 
                         "pixels further away in z than rmsCut * RMSofDistributon are ignorred for Meancalculation"  ,
                         _rmsCut,
                         (int) 3 ) ;	

  registerProcessorParameter("ReadoutFrequency" ,
                         "readout frequency of TimePix chip in MHz"  ,
                         _readoutFrequency ,
                         (double) 100.) ;

  registerProcessorParameter("DriftVelocity" ,
                         "drift velocity of gas in mm/us"  ,
                         _driftVelocity ,
                         (double) 26.7) ;


  registerProcessorParameter("SetOutputTransient" ,
                         "if not 0 the output collection is set transient"  ,
                         _outputIsTransient,
                         0);

   _timePixMap = 0;
}


void TimePixHitCenterCalculatorProcessor::init()
{

  // usually a good idea to
  printParameters() ;

  _timePixMap = TimePixMapHandlerProcessor::getTimePixMap();	


  _zDistancesPixelsInClusters_beforeCleaning =  AIDAProcessor::histogramFactory(this)->
    createHistogram1D( "zDistancesPixelsInClusters_beforeCleaning", 
                         "zDistancesPixelsInClusters_beforeCleaning", 
                         2001, -1000, 1000);	
			     
			     
  _zDistancesPixelsInClusters_afterCleaning =  AIDAProcessor::histogramFactory(this)->
    createHistogram1D( "zDistancesPixelsInClusters_afterCleaning", 
                         "zDistancesPixelsInClusters_afterCleaning", 
		         2001, -1000, 1000);			     
      
      
  _zclusterCenterBeforeMinusAfterCleaning =  AIDAProcessor::histogramFactory(this)->
    createHistogram1D( "zclusterCenterBeforeMinusAfterCleaning", 
                         "zclusterCenterBeforeMinusAfterCleaning", 
		         201, -100, 100);

}


void TimePixHitCenterCalculatorProcessor::processRunHeader(LCRunHeader* run)
{
  run->parameters().setValue(_processorName + "_revision", "$Rev$");

  for (ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++)
  {
    if (! i->second->isOptional() || i->second->valueSet())
    {
      run->parameters().setValue(_processorName + "_" + i->second->name(), i->second->value());
    }
  }

}


void TimePixHitCenterCalculatorProcessor::processEvent(LCEvent * evt)
{

//      streamlog_out(MESSAGE1)<<"Event "<<evt->getEventNumber()<<std::endl;
//
//      streamlog_out(DEBUG2)<<"TimePixHitCenterCalculatorProcessor::processEvent( "
//             <<evt->getEventNumber()<<" ) "<<std::endl;

  // get the input collection from the event
  LCCollection* inputCol = 0;
  try
  {
    inputCol = evt->getCollection(_inputColName) ;
  } //end of try {
  catch (DataNotAvailableException)
  {
    // if there is not input collection give a warning to the user
    streamlog_out(MESSAGE2) << "processEvent: input collection "
      << _inputColName << " cannot be found in event " << evt->getEventNumber() << std::endl;
    return;
  }

  // create the output collection with a string defining the TYPE (see lcio.h)
  LCCollectionVec* outputCol = new LCCollectionVec(LCIO::TRACKERHIT);

  // set the transient flag as definded by the processor parameter
  outputCol->setTransient((_outputIsTransient != 0));

//      LCCollection* oldRelationsCol = 0;
//      try {
//    oldRelationsCol = evt->getCollection( "TimePixRelationSepHit" ) ;
//      }
//      catch(DataNotAvailableException){
//    return;
//      }
//
//      LCRelationNavigator* relationNavigator = 0;
//      LCCollectionVec* RelationsCol = 0;
//
//      if ( oldRelationsCol != 0 ) {
//    relationNavigator = new LCRelationNavigator( oldRelationsCol );
//            RelationsCol = new LCCollectionVec( LCIO::LCRELATION );
//      }

  streamlog_out(DEBUG3) << "Number of TrackerHits in event: " << inputCol->getNumberOfElements() << std::endl;
  
  for (int iHit = 0; iHit < inputCol->getNumberOfElements() ; iHit++)
  {
    TrackerHit* rawHit = 0;

    try
    {
      rawHit = dynamic_cast<TrackerHit*>(inputCol->getElementAt(iHit));
    }
    catch(DataNotAvailableException &)
    {
      streamlog_out(ERROR) << "There was no RawHit in Hit " << iHit << std::endl;
      continue;
    }

    streamlog_out(DEBUG3) << "RawData in hit " << iHit << " is: " << rawHit->getRawHits().size() << std::endl;


    // determine the chip ID from the rawHit
    LCObjectVec const & rawDataVec = rawHit->getRawHits();
    
    if (rawDataVec.size() == 0)
    {
      throw DataNotAvailableException("No TPCData stored in Timepix hit, cannot proceed.");
    }
    
    TrackerData *rawData = dynamic_cast<TrackerData *>(rawDataVec[0]);
    if (rawData == 0)
    {
      throw DataNotAvailableException("Data stored as raw hit in Timepix hit is not TrackerData, cannot proceed.");
    }
    
    int chipID = rawData->getCellID1();
    // Phew, a lot of code for just one number..


    // get the corresponding module from gear
    gear::TPCModule const &module = Global::GEAR->getTPCParameters().getModule(chipID);


    // get some geometry information from gear
    // This processor implicitly assumes a rectangular matrix of pads/pixels
    int nRows = module.getNRows();
    int nColumns = module.getNPads() / nRows;
    double pitchX = module.getPadPitch(0); // there always should be a pad 0
    double pitchY = module.getRowHeight(0); // there always should be a row 0

    streamlog_out(DEBUG2) << "nColums = " << nColumns << " ; nRows =" << nRows << std::endl;
    streamlog_out(DEBUG2) << "pitchX = " << pitchX << " ; pitchY =" << pitchY << std::endl;

    if (rawHit != 0)
    {
      TrackerHitImpl* hitCandidate = new TrackerHitImpl();
//
//      //get original Cluster to this one
//      EVENT::LCObjectVec rawHitRelatedObjects = relationNavigator->getRelatedFromObjects(rawHit);
//
//      if ( rawHitRelatedObjects.size() != 0 ) {
//
//      LCRelationImpl* relation = new LCRelationImpl( rawHit, hitCandidate );
//      RelationsCol->addElement( relation );
//      }

      //hit position x, y in mm, z in time counts
      double pos[3] = { 0, 0, 0 };

      //weighting of pixels to calculate center of gravity
      double weightingPixel   = 0;
      double weightingCluster = 0;

      // sum of x,y-coordinates weighted by TOT-value or 1 respectively
      double weightedSumX = 0;
      double weightedSumY = 0;

      // x,y- position of Hit in pixel
      double posX = 0;
      double posY = 0;

      //sum of TOT/TIME-values in cluster
      double sumTOTinCluster  = 0.;
      double sumTIMEinCluster = 0.;

      //number of TOT/TIME-pixel in cluster
      double noTOTPixel  = 0.;
      double noTIMEPixel = 0.;

      //TOT value of cluster extrapolated for all pixel in cluster
      double  TOTCluster = 0;

      //mean drift time of the cluster in microseconds
      double meanDriftTimeCluster = 0;

      //number of pixel in cluster
      int clustersize = 0;

      std::vector<LCObject*> DataVec = rawHit->getRawHits();
      std::vector<LCObject*>::iterator j;

      std::vector<float> timeValVec;
      std::vector<float> timeValVec2;	    
	    
      std::stringstream EventNumberString;
      EventNumberString << evt->getEventNumber();
      std::string HistoName;
      HistoName = "EventNr" + EventNumberString.str()  + "ClusterNr";
      //changes into directory of this processor in aida-output-file 
      AIDAProcessor::tree(this);
      _clusterHisto = new TH1F(HistoName.c_str(), HistoName.c_str(), 12000, -0.5, 11999.5);

      for (j = DataVec.begin(); j != DataVec.end(); j++)
      {
        TrackerData* Data = dynamic_cast<TrackerData*>(*j) ;

        if (Data != 0)
        {
          int cellid0 = Data->getCellID0();

          //pixel position in pixel
          double x = cellid0 % nColumns;
          double y = static_cast<int>((cellid0 - x) / nColumns);

          FloatVec ChargeValVec = Data->getChargeValues();

          FloatVec::iterator iChargeVal;

          for (iChargeVal = ChargeValVec.begin(); iChargeVal != ChargeValVec.end(); iChargeVal++)
          {
            int mode = ((_timePixMap->find(TimePixPixelMode::PixelKey(chipID, cellid0)))
              ->second)->getPixelMode();

            //count all pixels belonging to this cluster
            clustersize++;

            if (mode == TimePixPixelMode::TOT)
            {
              weightingPixel = *iChargeVal;
              streamlog_out(DEBUG0) << "TOT = " << *iChargeVal << std::endl;
              sumTOTinCluster += *iChargeVal;
              noTOTPixel++;
            }
            else weightingPixel = 1;
            
            if (mode == TimePixPixelMode::TIMEPIX)
	    {
              sumTIMEinCluster += *iChargeVal;
              // modified by C.Brezina: building vector containing time values used for enhanced mean 
              //calculation (noise suppression)
              timeValVec.push_back(*iChargeVal);
              noTIMEPixel++;
              
             if (countNrOfNeighboringPixelsWithHit(chipID, cellid0, DataVec) >= 4) 
               timeValVec2.push_back(*iChargeVal); // >4 only possible if mode pattern is not checkerboard
            }

            weightedSumX += x * weightingPixel;
            weightedSumY += y * weightingPixel;

            weightingCluster += weightingPixel;

            cellid0++;
            x++;
 
            if (x == nColumns)
            {
              x = 0;
              y++;
            }
	  } //end of for ( iChargeVal = ChargeValVec.begin(); iChargeVal != ChargeValVec.end();



          hitCandidate->rawHits().push_back(Data);
          ChargeValVec.clear();
        } //end of if ( Data != 0 ) {
      } //end of for ( j = DataVec.begin(); j != DataVec.end(); j++	

      for (std::vector<float>::iterator itTimeVal = timeValVec.begin(); itTimeVal != timeValVec.end(); 
        itTimeVal++)
      {
        _clusterHisto->Fill(*itTimeVal);	
      }

      int meanDriftTimeClusterWOcuts = _clusterHisto->GetMean();// mean Clustertime from HIsto before any cut
      
      // if the Cluster has enough Pixels with 4 Neighbors we regard only these for Meancalculation
      if ( timeValVec2.size() >= static_cast <unsigned int> ( abs(_minNrOfPixelsToCalcTime) ) )
        timeValVec = timeValVec2;

      // timeValVec now holds available timing information, while noTIMEPixel gives the number of entries
      // begin to find and exclude pixels with bad timing information (noise hits) - using sumTIMEinCluster
      _clusterHisto->Clear();
      
      for (std::vector<float>::iterator itTimeVal = timeValVec.begin(); 
        itTimeVal != timeValVec.end(); itTimeVal++)
      {
        _clusterHisto->Fill(*itTimeVal);	
      }

      meanDriftTimeCluster = _clusterHisto->GetMean();
	
      for (std::vector<float>::iterator itTimeVal = timeValVec.begin(); 
        itTimeVal != timeValVec.end(); itTimeVal++)
      {
        _zDistancesPixelsInClusters_beforeCleaning->fill( meanDriftTimeCluster - *itTimeVal ); 
      }

      // removal of outliers - cuts on  (meanDriftTimeCluster - pixelTime) > _max_zDiffusion_perCluster
 
      bool timeOutlierFlag = true;  
      int eraseCounter = 0;
      float maxDifference = _max_zDiffusion_perCluster;
      std::vector<float>::iterator maxOutlierMarker;

      while (timeOutlierFlag == true && timeValVec.size() > 0) 
      { 
        timeOutlierFlag = false;
	maxDifference = _max_zDiffusion_perCluster;
		       
	for (std::vector<float>::iterator itTimeVal = timeValVec.begin(); 
          itTimeVal != timeValVec.end(); itTimeVal++)
	{ 
          if ( fabs(meanDriftTimeCluster - *itTimeVal) > maxDifference )  // values in Clockcycles!!
	  {
            timeOutlierFlag = true;
	    maxOutlierMarker = itTimeVal; 
	    maxDifference = fabs(meanDriftTimeCluster - *itTimeVal);
	  }
	}
	// if there is at least one outlier, remove the worst one and check again
	if (timeOutlierFlag == true) 
	{ 
          timeValVec.erase( maxOutlierMarker );
	  eraseCounter++;		   
	  // now calculate time from remaining pixels
	  _clusterHisto->Clear();		   
	
          for (std::vector<float>::iterator itTimeVal = timeValVec.begin(); 
            itTimeVal != timeValVec.end(); itTimeVal++)
          {
	    _clusterHisto->Fill(*itTimeVal);	
	  }
	
          meanDriftTimeCluster = _clusterHisto->GetMean();
	}
      }	

      // end removal of outliers
      
      // apply cut on few RMS of distribution
      _clusterHisto->GetXaxis()->SetRangeUser( meanDriftTimeCluster - _rmsCut * _clusterHisto->GetRMS(), 
        meanDriftTimeCluster + _rmsCut * _clusterHisto->GetRMS() );
      meanDriftTimeCluster = _clusterHisto->GetMean();
	
      if (_keepClusterHistoFlag == true) _clusterHisto->Write();
      delete _clusterHisto;
	
      // create hisogramm showing zDistances after removal of bad pixels
      for (std::vector<float>::iterator itTimeVal = timeValVec.begin(); 
        itTimeVal != timeValVec.end(); itTimeVal++) 
      {
        _zDistancesPixelsInClusters_afterCleaning->fill( meanDriftTimeCluster - *itTimeVal ); 
      }

      _zclusterCenterBeforeMinusAfterCleaning->fill(meanDriftTimeClusterWOcuts - meanDriftTimeCluster);

      if (eraseCounter > 1) streamlog_out(MESSAGE3) << "Removed "<<eraseCounter<<" Pixels in Event Nr.: "
        << evt->getEventNumber() << std::endl;
     
      eraseCounter = 0;
	
      // end of time information calculation 
     
      posX = weightedSumX / weightingCluster;
      posY = weightedSumY / weightingCluster;
      
      // now translating clock counts to drift time
      if ( timeValVec.size() > 0)
      {
        if( !_multiShutter ) meanDriftTimeCluster = _shutter - (meanDriftTimeCluster)/_readoutFrequency; 
	else meanDriftTimeCluster = _shutter - fmod( ( (meanDriftTimeCluster)/_readoutFrequency ), _shutter);
      }
      // if not enogh (>1) time pixel can be used the cluster has to be disregardet from now on
      else meanDriftTimeCluster = -1.0;

      TOTCluster = sumTOTinCluster * static_cast<double>(clustersize) / noTOTPixel;

      streamlog_out(DEBUG1) << "meanDriftTimeCluster = "
        << _shutter << " - (" << sumTIMEinCluster << " / " << noTIMEPixel << ") / "
        << _readoutFrequency << " = " << meanDriftTimeCluster << std::endl;
      streamlog_out(DEBUG1) << "TOTCluster = " << sumTOTinCluster << " * " << clustersize << " / " << noTOTPixel
        << " = " << TOTCluster << std::endl;

      //x-position in mm, the pixels on the chip are in the wrong direction
      pos[0] = (nColumns - (posX + 0.5)) * pitchX;
      pos[1] = (posY + 0.5) * pitchY; //y-position in mm
      pos[2] = meanDriftTimeCluster * _driftVelocity; //drift length in [us * mm/us =  mm]

      streamlog_out(DEBUG1) << "pos[2] = " << meanDriftTimeCluster << " * "
        << _driftVelocity << " = " << pos[2] << std::endl;

      streamlog_out(DEBUG1) << "Local hit position: " << pos[0] << " , " << pos[1] 
        << " , " << pos[2] << std::endl;

      // not translate to global coordinates
      gear::Vector2D globalPos = module.localToGlobal(pos[0] , pos[1]);
      pos[0] = globalPos[0];
      pos[1] = globalPos[1];

      hitCandidate->setPosition(pos);
      hitCandidate->setEDep(TOTCluster);
      hitCandidate->setTime(meanDriftTimeCluster*1000);   //ns
      
      if ( meanDriftTimeCluster != -1 ) outputCol->addElement(hitCandidate);
      else delete hitCandidate;

      DataVec.clear();
    } //end of if ( rawHit != 0 )
  } //end of for(int iHit=0; iHit< nHit ; iHit++){

  evt->addCollection(outputCol, _outputColName);

// if ( RelationsCol != 0 )
// evt->addCollection( RelationsCol, "TimePixRelationHit");

}


void TimePixHitCenterCalculatorProcessor::check(LCEvent * evt)
{

    // the complete check has to be rewritten because the local hit positions can be
    // translated / rotated to arbitrary positions in global coordinates

//      LCCollection* HitCol;
//
//      try{
//    HitCol = evt->getCollection( _outputColName );
//      }
//      catch(DataNotAvailableException)
//      {
//    return;
//      }
//
//      std::stringstream sEvt;
//      sEvt << evt->getEventNumber();
//
//      std::string histoname = "clustersevt" + sEvt.str();
//      std::string histonameSP = "clustersevt" + sEvt.str() + "SP";
//      std::string histoname3D = "clustersevt" + sEvt.str() + "3D";
//
//      int n = HitCol->getNumberOfElements();
//
//      AIDAProcessor::histogramFactory(this)->destroy( histoCluster );
//      histoCluster = AIDAProcessor::histogramFactory(this)->
//  createHistogram2D( histoname.c_str(), histoname.c_str(),
//             256, 0, 256, 256, 0, 256);
//
//      AIDAProcessor::histogramFactory(this)->destroy( histoSP );
//      histoSP = AIDAProcessor::histogramFactory(this)->
//  createHistogram2D( histonameSP.c_str(), histonameSP.c_str(),
//             256, 0, 256, 256, 0, 256);
//
//      AIDAProcessor::histogramFactory(this)->destroy( histo3D );
//      histo3D = AIDAProcessor::histogramFactory(this)->
//  createHistogram3D( histoname3D.c_str(), histoname3D.c_str(),
//             256, 0, 14.08, 256, 0, 14.08, 40, -20, 20);
//
//      for( int i = 0; i < n; i++ ) {
//  TrackerHit* Hit = dynamic_cast<TrackerHit*>( HitCol->getElementAt( i ) ) ;
//
//  //color functionality not yet implemented in RAIDA!
//  //=================================================
//
////    int color = i%10+1;
////    if ( color == 0 || color == 10 || color == 18 ) color = 46;
////    histoCluster->SetLineColor(color);
////    histoSP->SetMarkerStyle( 10 );
////    histoSP->SetMarkerSize(10);
////    histoSP->SetMarkerColor( color + 1 );
//
//  if ( Hit != 0 ) {
//
//    int xsp = static_cast<int>(Hit->getPosition()[0]/_pitch + 0.5 ); //+0.5 for correct rounding
//    int ysp = static_cast<int>(Hit->getPosition()[1]/_pitch + 0.5 ); //+0.5 for correct rounding
//
//    histoSP->fill( xsp, ysp );
//    histo3D->fill( Hit->getPosition()[0], Hit->getPosition()[1], Hit->getPosition()[2] );
//    std::vector<LCObject*> DataVec = Hit->getRawHits();
//    std::vector<LCObject*>::iterator j;
//    for ( j = DataVec.begin(); j != DataVec.end(); j++) {
//
//      TrackerData* Data = dynamic_cast<TrackerData*>( *j ) ;
//
//      if ( Data!=0 ) {
//        int cellid0 = Data->getCellID0();
//        int x = cellid0%_nColumn;
//        int y = static_cast<int>((cellid0-x)/_nColumn);
//
//        FloatVec ChargeVec  = Data->getChargeValues();
//        FloatVec::iterator sv;
//        for ( sv = ChargeVec.begin(); sv != ChargeVec.end(); sv++ ) {
//
//        int mode =
//            ((_timePixMap->find( TimePixPixelMode::PixelKey( _chipID, cellid0 ) ))
//             ->second)->getPixelMode();
//
//        float weigh = 0;
//        if ( mode == TimePixPixelMode::TOT ) weigh = *sv;
//
//        histoCluster->fill( x, y, weigh );
//        x++;
//
//        if ( x == _nColumn ) {
//            x = 0;
//            y++;
//        }
//
//        } //end of for ( sv = ChargeVec.begin(); sv != ChargeVec.end(); sv++ ) {
//
//        ChargeVec.clear();
//
//      } //end of if ( Data!=0 ) {
//
//
//    } //end of for ( j = DataVec.begin(); j != DataVec.end(); j++) {
//
//  } //end of if ( Hit != 0 ) {
//
//
//
//
//      } //end of for( int i = 0; i < n; i++ ) {
//
//
}


int TimePixHitCenterCalculatorProcessor::countNrOfNeighboringPixelsWithHit(int chipID, int pixelID, std::vector<LCObject*> DataVec)
{ 
  // get the corresponding module from gear
  gear::TPCModule const &module = Global::GEAR->getTPCParameters().getModule( chipID );
	
  // get some geometry information from gear
  // This processor implicitly assumes a rectangular matrix of pads/pixels
  int nRows = module.getNRows();
  int nColumns = module.getNPads() / nRows;
   
  std::vector<int> NeighborVec;
    
  //fill with the PixelIDs of the surrounding pixel:
  NeighborVec.push_back( pixelID - nColumns - 1 );  // |-----------|
  NeighborVec.push_back( pixelID - nColumns     );  // | 0 | 1 | 2 |
  NeighborVec.push_back( pixelID - nColumns + 1 );  // |-----------|
  NeighborVec.push_back( pixelID - 1            );  // | 3 | x | 4 |
  NeighborVec.push_back( pixelID + 1            );  // |-----------|
  NeighborVec.push_back( pixelID + nColumns - 1 );  // | 5 | 6 | 7 |
  NeighborVec.push_back( pixelID + nColumns     );  // |-----------|
  NeighborVec.push_back( pixelID + nColumns + 1 );  // 
        
  //set all pixelIDs to -1 that can not be used
  for ( unsigned int n = 0; n < NeighborVec.size(); n++ )
  {
    if ( (NeighborVec[n] >= 0) && (NeighborVec[n]< module.getNPads() ) )  //the pixels are on the chip ??
    {  
      if ( ((_timePixMap->find( TimePixPixelMode::PixelKey( chipID, NeighborVec[n] ) ) )->second)
        ->getPixelMode() != TimePixPixelMode::TOT ) 
      {
        NeighborVec[n] = -1;
      }// only look at Pixels in ToT mode 
     
      if ( n == 0 || n == 3 || n == 5 )
      {
        //the pixel lies on the left edge of the chip
        //the left neighbors therefore are on the other side of the chip
        //and should not be used
        if ( NeighborVec[n] % nColumns == nColumns - 1 )
	{
          NeighborVec[n] = -1;
	}
      }
      
      if ( n == 2 || n == 4 || n == 7 )
      {
        //the pixel lies on the right edge of the chip
	//the right neighbors therefore are on the other side of the chip
	//and should not be used
	if ( NeighborVec[n] % nColumns == 0 )
	{
          NeighborVec[n] = -1;
	}
      }
      
      //these pixels are not on the chip anymore
//      if ( NeighborVec[n] > ( nColumns * nRows - 1 ) || NeighborVec[n] < 0 )
// 	{
//        NeighborVec[n] = -1;
// 	}

    }  // end if
    else
    {
      NeighborVec[n] = -1;
    } // end else
  } // end for ( int n = 0; n < NeighborVec.size(); n++ )
 
  int neighborCounter = 0;
  int cellID0;
  
  for ( unsigned int n = 0; n < NeighborVec.size(); n++ )
  {
    if (NeighborVec[n] != -1)
    {
      for (std::vector<LCObject*>::iterator j = DataVec.begin(); j != DataVec.end(); j++) 
      {
        TrackerData* Data = dynamic_cast<TrackerData*>( *j ) ;
	FloatVec ChargeValVec = Data->getChargeValues();
	cellID0 = Data->getCellID0();
	
	for (FloatVec::iterator iChargeVal = ChargeValVec.begin(); 
          iChargeVal != ChargeValVec.end(); iChargeVal++ ) 
        {
	  if ( ( cellID0 == NeighborVec[n] ) && ( *iChargeVal > 0 )  ) neighborCounter++;
         	  
          cellID0++;
	}
      }
    }
  }
  
  streamlog_out(DEBUG1)  << "Row " << pixelID % nColumns << " Colummn " 
    << static_cast<int>((pixelID)/nColumns) << " has "<< neighborCounter 
    << " Neighbors in ToT Mode with Charge"<< std::endl;
  
  return (neighborCounter);
  
} // end  "countNrOfNeighboringPixelsWithHit"


void TimePixHitCenterCalculatorProcessor::end()
{

    /*
    streamlog_out(MESSAGE4) << "TimePixHitCenterCalculatorProcessor::end()  " << name()
      << " processed " << evt->getEventNumber() << " events in " << _nRun << " runs "
      << std::endl ;
    */
}




}
