#ifndef LINEARREGRESSIONPROCESSOR_H
#define LINEARREGRESSIONPROCESSOR 1

#include "marlin/Processor.h"
//#include "lcio.h"
//#include <string>
#include <vector>

#include <EVENT/TrackerHit.h>

namespace AIDA
{

class IHistogram1D;
class IHistogram2D;


}


namespace marlintpc
{

/** LinearRegressionProcessor performs a linear regression on the hits in seed tracks.
 *  It calculates the track parameters for a straight line. The linear regression is an
 *   analytical calculation, but equivalent to a chi^2 fit.
 *  Does not need seed tracks, so it can directly work on track candidates.
 *  Orientation of the track can be either along x- or y-axis.\n
 *  \li case1:\f$\displaystyle |\phi | < \frac{\pi}{4}\f$\n
 *  y = ax + b\n
 *  z = cx + d\n
 *  \f$\displaystyle \phi = atan(a) \f$\n
 *  \f$\displaystyle  d_0 = b\cdot cos(\phi) \f$\n
 *  \n
 *  calculation of \f$\lambda\f$:\n
 *  (angle betwenn track and x-y-plane, that is between track and projection on x-y-plane)\n
 *  track: \f$\displaystyle \vec{g}(x) = \left(\begin{array}{c} x \\ ax + b \\ cx + d \end{array} \right)\f$\n
 *  projection: \f$\displaystyle \vec{f}(x) = \left(\begin{array}{c} x \\ ax + b \\ 0 \end{array} \right)\f$\n
 *  \f$\displaystyle  cos( \lambda ) = \frac{(\vec{g}(1) - \vec{g}(0))\cdot(\vec{f}(1) - \vec{f}(0))}{|(\vec{g}(1) - \vec{g}(0))| \cdot |(\vec{f}(1) - \vec{f}(0))| } \f$\n
 *  \f$\displaystyle  \Rightarrow  cos( \lambda ) = \frac{ \sqrt{ a^2 + 1 } }{ \sqrt{ a^2 + 1 + c^2 } }\f$ \n
 *  \f$\displaystyle  z_0 = c\cdot x_0 + d = c\cdot sin( \phi )\cdot d_0 \f$
 *  \n
 *  \n
 *  \li case2:\f$\displaystyle |\phi | > \frac{\pi}{4}\f$\n
 *  x = ay + b\n
 *  z = cy + d\n
 *  \f$\displaystyle \phi = \frac{\pi}{2}-atan(a) \f$\n
 *  \f$\displaystyle  d_0 = b\cdot sin(\phi) \f$\n
 *  same as above:\n
 * \f$\displaystyle  cos( \lambda ) = \frac{ \sqrt{ a^2 + 1 } }{ \sqrt{ a^2 + 1 + c^2 } }\f$\n
 * \f$\displaystyle  z_0 = c\cdot y_0 + d = c\cdot sin( \frac{\pi}{2} - \phi)\cdot d_0 \f$\n
 * \n
 * dEdx is the charge deposition per track length. The procedure is the following:\n
 * 1) one sorts the hits along the track (parametrization:
 *    \f$\displaystyle \vec{P}=\vec{d_0}+s\begin{array}{c} cos\phi \\sin\phi \end{array} \f$
 *     \f$\Rightarrow\f$ sorting in s) \n
 * 2) calculate distance from first to last point along track \f$\displaystyle l=|s_{last}-s_{first}|\f$ \n
 * 3) \f$\displaystyle  l_{eff} = l\cdot\frac{n}{n-1}\f$ where n is the number of hits on the track
 * 4) \f$\displaystyle dEdx = \frac{\sum Q}{l_{eff}} \f$
 *  @author M. Killenberg(University of Bonn), S.Zimmermann(University of Bonn)
 *
 *  <h4>Input - Prerequisites</h4>
 *  \li A collection of track candidates or seen tracks (lcio::Tracks) containing the TrackerHits.<br>
 *  Note: The LCIO::TRBIT_HITS flag has to be set when this collection is created so
 *      the hits are stored in the track.
 *
 *  <h4>Output</h4>
 *  \li Collection of straight tracks.
 *
 *  @param InputTrackCandidates The the name of the input collection of track candidates (default: TPCTrackCandidates)
 *  @param OutputTracks The name of the collection the straight tracks be stored under (default: TPCTracks)
 *  @param SetOutputTransient If not 0 the output collection is set transient (default: 0)
 *
 */
class LinearRegressionProcessor : public marlin::Processor
{

    public:

        /** Default constructor
        */
        LinearRegressionProcessor();


        /** Destructor
        */
        ~LinearRegressionProcessor();


        /** Return a new instance of this processor
        */
        Processor * newProcessor();


        /** Called at the begin of the job before anything is read.
        */
        void init();


        /** Called once per run to process the event's header
        */
        void processRunHeader(EVENT::LCRunHeader *run);


        /** Called for every event - the working horse.
        */
        virtual void processEvent(EVENT::LCEvent *evt);


        /** Called for every event - right after processEvent()
         *has been called for all processors.
        */
        virtual void check(EVENT::LCEvent *evt);


        /** Called after data processing for clean up in the
         *  inverse order of the init() method so that resources
         *  allocated in the first processor also will be available for
         *  all following processors.
        */
        virtual void end();

        /* Sort function to sort the Hits along the track
         */
        static bool compareHit(const std::pair<TrackerHit*, double> & pair1,
                               const std::pair<TrackerHit*, double> & pair2)
        {
            return pair1.second < pair2.second;
        }

    private:

        /** the name of the input collection of track candidates
        */
        std::string _input_track_candidates_collection_name;

        /** the name of the collection the seed tracks be stored under
        */
        std::string _output_seed_tracks_collection_name;

        /** Give the status which is set to the transient flag of the output collection.
         *  It is an int instead of a boolean, because the processor parameter
         *  cannot be a boolean.
         *  Hits and tracks can be set separately.
         */
        int _outputIsTransient; //< if not 0 the hits collection is set transient
        //(default 0)


        std::vector<AIDA::IHistogram2D *> HistoLineVec;
        std::vector<AIDA::IHistogram2D *> HistoTrackVec;
};

}// namespace marlintoc

#endif // TRACKSEEDERPROCESSOR_H



