#ifndef SIMPLETRACKSEEDERPROCESSOR_H
#define SIMPLETRACKSEEDERPROCESSOR_H 1

#include "marlin/Processor.h"
#include "lcio.h"
#include <string>
#include <vector>

#include <EVENT/TrackerHit.h>

namespace marlintpc
{

/** SimpleTrackSeederProcessor takes the TrackerHits of each track and
 *  calculates initial track parameters to have a seed for fitting algorithms.
 *    @author M. Killenberg(University of Bonn)
 *
 *    Based on the disk layout algorithm by
 *    @author J. Abernathy (UVic)
 *
 *  <h4>Input - Prerequisites</h4>
 *  \li A collection of track candidates (lcio::Tracks) containing the TrackerHits from
 *      which the seed parameters are calculated.<br>
 *  Note: The LCIO::TRBIT_HITS flag has to be set when this collection is created so
 *      the hits are stored in the track.
 *
 *  <h4>Output</h4>
 *  \li Collection of seed tracks
 *
 *  @param InputTrackCandidates The the name of the input collection of track candidates (default: TPCTrackCandidates)
 *  @param OutputSeedTracks The name of the collection the seed tracks be stored under (default: TPCSeedTracks)
 *  @param SetOutputTransient If not 0 the output collection is set transient (default: 0)
 *
 * <h4>How the algorithm works:</h4>
 *     - xy-plane: The hits are sorted in y. The first, the last and the point in
 *       the middle of this list are taken to calculate the curvature Omega, d_0 and phi.
 *       The reference point is the origin.
 *     - sz-plane: Now that the parameters in the xy plane are known on can calculate the
 *       arc length in the xy-plane (s) for each point.
 *       Afterwards the points are sorted in z. In principle it would be enough to
 *       calculate \f$\frac{z_2 - z_1}{s_2 - s_1}\f$, where \f$s_i\f$ and \f$z_i\f$ are the
 *       s and the z of the first and the second point, respectivley (or any other combination
 *       with \f$z_2 > z_2\f$, preferably \f$\Delta z\f$ being as large as possible to have a
 *       good lever arm). <br>
 *       Unfortunately this does not work, since s can only be determined modulo \f$2\pi R\f$
 *       in the interval \f$\left[-\pi R..+\pi R\right]\f$ from the x and y coordinates.
 *       To solve this ambiguity one could map z in to the range
 *       \f$\left[-\pi R \tan\lambda .. +\pi R \tan\lambda \right]\f$.<br>
 *       <img src="../images/sz.png" alt="szplot" align="middle" border="0">
 *       In the example the point at z=6.5 is wrapped from s/z = -3.24 to +3.04, so the first
 *       slope is not correct. This jump can occur at any position in the list, not only
 *       between the fist and the second point. It is assumed that two of the first three slopes
 *       m1, m2, m3 are correct, while one might be wrong due to such a jump. So if one of the
 *       ratios m1/m2, m1/m3 or m2/m3 is near 1 (FIXME 0.9..1.1?, or within the errors? which errors?)
 *       the mean of these two slopes is taken as estimate for \f$\tan\lambda\f$.
 *
 */
class SimpleTrackSeederProcessor : public marlin::Processor
{

    public:

        /** Default constructor
        */
        SimpleTrackSeederProcessor();


        /** Destructor
        */
        ~SimpleTrackSeederProcessor();


        /** Return a new instance of this processor
        */
        Processor * newProcessor();


        /** Called at the begin of the job before anything is read.
        */
        void init();


        /** Called once per run to process the event's header
        */
        void processRunHeader(EVENT::LCRunHeader *run);


        /** Called for every event - the working horse.
        */
        virtual void processEvent(EVENT::LCEvent *evt);


        /** Called for every event - right after processEvent() has been called for all processors.
        */
        virtual void check(EVENT::LCEvent *evt);


        /** Called after data processing for clean up in the inverse order of the init() method so that resources allocated in the first processor also will be available for all following processors.
        */
        virtual void end();

    private:

        /** the name of the input collection of track candidates
        */
        std::string _input_track_candidates_collection_name;

        /** the name of the collection the seed tracks be stored under
        */
        std::string _output_seed_tracks_collection_name;

        /** Give the status which is set to the transient flag of the output collection.
         *  It is an int instead of a boolean, because the processor parameter cannot be a boolean.
         *  Hits and tracks can be set separately.
         */
        int _outputIsTransient;    //< if not 0 the hits collection is set transient (default 0)

        /** A nested helper class for sorting
         *  pointers of TrackerHits to the Hits' y position
         */
        class sort_by_y
        {
            public:
                bool operator()(EVENT::TrackerHit* hit1, EVENT::TrackerHit* hit2)
                {
                    return (hit1->getPosition()[1] < hit2->getPosition()[1]);
                }
        };

        /** A nested helper class for sorting
         *  sz pairs by z
         */
        class sort_by_z
        {
            public:
                bool operator()(std::pair<double, double> sz1, std::pair<double, double> sz2)
                {
                    return (sz1.second < sz2.second);
                }
        };

//  /// The pad layout as defined in GEAR
//  gear::PadRowLayout2D const *_padLayout;
//  int _padLayoutType; ///< Pad layout type (PadRowLayout2D::CARTESIAN or PadRowLayout2D::POLAR)

};

}// namespace marlintoc

#endif // SIMPLETRACKSEEDERPROCESSOR_H



