
// STL
//#include <vector>
//#include <map>
//#include <algorithm>
//#include <iomanip>
//#include <cassert>
//
//// LCIO
#include "EVENT/LCIO.h"
#include "EVENT/LCCollection.h"
#include "EVENT/TrackerHit.h"
#include <IMPL/LCCollectionVec.h>
#include "IMPL/TrackImpl.h"
#include <IMPL/LCFlagImpl.h>

#include <iostream>
#include <iomanip>
#include <cmath>
#include <algorithm>

// Marlin
#include "marlin/Global.h"
#include "LinearRegressionProcessor.h"
#include "LinearTrackRegression.h"

// Aida
#ifdef MARLIN_USE_AIDA
#include <AIDA/AIDA.h>
#include <marlin/AIDAProcessor.h>

//marlin
#include <marlin/ConditionsProcessor.h>


#endif

using namespace marlin;
using namespace lcio;

namespace marlintpc
{

LinearRegressionProcessor aLinearRegressionProcessor;

LinearRegressionProcessor::LinearRegressionProcessor() : Processor("LinearRegressionProcessor")
{

    _description = "Calculates seed track parameters from  in the TrackerHits in the track candidates collection!";

    registerInputCollection(LCIO::TRACK,
                            "InputTrackCandidates",
                            "The the name of the input collection of track candidates (default: TPCTrackCandidates)",
                            _input_track_candidates_collection_name,
                            std::string("TPCTrackCandidates"));

    registerOutputCollection(LCIO::TRACK,
                             "OutputSeedTracks",
                             "The name of the collection the seed tracks be stored under (default: TPCSeedTracks)",
                             _output_seed_tracks_collection_name,
                             std::string("TPCSeedTracks"));

    registerProcessorParameter("SetOutputTransient" ,
                               "if not 0 the output hits collection is set transient (default: 0)",
                               _outputIsTransient,
                               int(0));

}


LinearRegressionProcessor::~LinearRegressionProcessor()
{

}


void LinearRegressionProcessor::init()
{

    printParameters();

}


void LinearRegressionProcessor::processRunHeader(LCRunHeader *run)
{

    run->parameters().setValue(_processorName + "_revision", "$Rev: 2628 $");

    for (ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++)
        {
            if (! i->second->isOptional() || i->second->valueSet())
                {
                    run->parameters().setValue(_processorName + "_" + i->second->name(), i->second->value());
                }
        }

}

void LinearRegressionProcessor::processEvent(LCEvent *evt)
{

    streamlog_out(DEBUG4) << "LinearRegressionProcessor" << std::endl;

    IMPL::LCCollectionVec* trackCandidates = 0;
    try
        {
            // the tracker hit collection
            trackCandidates =
                dynamic_cast<IMPL::LCCollectionVec*>
                (evt->getCollection(_input_track_candidates_collection_name));
        }
    catch (DataNotAvailableException &)
        {
            streamlog_out(MESSAGE1) << "No seed tracks collection \"" << _input_track_candidates_collection_name
            << "\" in event " << evt->getEventNumber() << std::endl;
            return;
        }
    catch (std::bad_cast)
        {
            streamlog_out(MESSAGE1) << "Could not convert collection \""
            << _input_track_candidates_collection_name
            << "\" to LCCollectionVec, skipping event "
            << evt->getEventNumber() << std::endl;
            return;
        }

//      const double innermost_electronics_z = tpc_parameters.getMaxDriftLength();



    // a collection to hold the new track
    IMPL::LCCollectionVec* outputTrackCollection = new IMPL::LCCollectionVec(EVENT::LCIO::TRACK);

    LCFlagImpl trkFlag(0) ;
    trkFlag.setBit(LCIO::TRBIT_HITS) ;
    outputTrackCollection->setFlag(trkFlag.getFlag());

    //loop all tracks
    for (IMPL::LCCollectionVec::iterator trackIter = trackCandidates->begin();
            trackIter < trackCandidates->end() ; trackIter++)
        {
            EVENT::Track *inputTrack = dynamic_cast<EVENT::Track *>(*trackIter);

            if (inputTrack == 0)
                {
                    streamlog_out(WARNING2) << "bad_cast: Track candidates collection"
                    << _input_track_candidates_collection_name
                    << " does not contain EVENT::Tracks!" << std::endl;
                    throw DataNotAvailableException("No track candidates found.");
                }

            // The TrackerHitVec is a vector of pointers to TrackerHits:  std::vector< TrackerHit * >
            // trackerHits is a copy (!) of the TrackerHitVec in the track, so we can resort the pointers
            // using the helpe classes defined in the header file
            EVENT::TrackerHitVec inputTrackerHits = inputTrack->getTrackerHits();


            // check whether there are at least 4 hits stored in the track, otherwise continue with next track
            if (inputTrackerHits.size() < 4)
                {
                    streamlog_out(WARNING2) << "Less than 4 hits in track candidates. "
                    << "Check your settings of the track finder!" << std::endl;
                    continue;
                }

            // the actual track estimate
            IMPL::TrackImpl* outputTrack = new IMPL::TrackImpl();
            outputTrack->setTypeBit(LCIO::TRBIT_HITS);

            // fill all hits into the seed track:
            for (EVENT::TrackerHitVec::iterator hit = inputTrackerHits.begin();
                    hit < inputTrackerHits.end(); hit++)
                {
                    outputTrack->addHit(*hit);
                }


            //if track is parameterized by y ( x = ay + b )
            if (fabs(inputTrack->getPhi()) > M_PI / 4)
                {


                    LinearTrackRegression::StraightLine straightlineX =
                        LinearTrackRegression::doRegressionX(outputTrack);

                    streamlog_out(DEBUG4) << "straigtline: a = " << straightlineX.slope1
                    << " b = " << straightlineX.intersept1 << std::endl;

                    // now set the estimated values in the seed track
                    outputTrack->setOmega(0);

                    outputTrack->setPhi(M_PI / 2 - atan(straightlineX.slope1));

                    outputTrack->setD0(- straightlineX.intersept1 * sin(outputTrack->getPhi()));

                    //Lambda has the sign of  the slope in z! sign = straightlineX.slope2 / fabs( straightlineX.slope2 ) ;
                    outputTrack->setTanLambda
                    (tan(straightlineX.slope2 / fabs(straightlineX.slope2) *
                         acos(sqrt(straightlineX.slope1*straightlineX.slope1 + 1) /
                              sqrt(straightlineX.slope1*straightlineX.slope1 + 1
                                   + straightlineX.slope2*straightlineX.slope2))));


                    //d^2 = x^2 + (slope1*x + intersept1)^2
                    //minimalize => x0 = (-slope1*intersept1)/(slope1^2 + 1)
                    //z_0 = x0 * slope2 + intersept2

                    outputTrack->setZ0
                    (- (straightlineX.slope1*straightlineX.intersept1*straightlineX.slope2) /
                     (straightlineX.slope1*straightlineX.slope1 + 1)
                     + straightlineX.intersept2);



                }

            //if track is parameterized by x ( y = ax + b )
            else
                {

                    LinearTrackRegression::StraightLine straightlineY =
                        LinearTrackRegression::doRegressionY(outputTrack);


                    // now set the estimated values in the seed track
                    outputTrack->setOmega(0);

                    outputTrack->setPhi(atan(straightlineY.slope1));

                    outputTrack->setD0(straightlineY.intersept1 * cos(outputTrack->getPhi()));

                    outputTrack->setTanLambda
                    (tan(straightlineY.slope2 / fabs(straightlineY.slope2) *
                         acos(sqrt(straightlineY.slope1*straightlineY.slope1 + 1) /
                              sqrt(straightlineY.slope1*straightlineY.slope1 + 1
                                   + straightlineY.slope2*straightlineY.slope2))));

                    //d^2 = x^2 + (slope1*x + intersept1)^2
                    //minimalize => x0 = (-slope1*intersept1)/(slope1^2 + 1)
                    //z_0 = x0 * slope2 + intersept2

                    outputTrack->setZ0(- (straightlineY.slope1*straightlineY.intersept1*straightlineY.slope2) /
                                       (straightlineY.slope1*straightlineY.slope1 + 1)
                                       + straightlineY.intersept2);

                }

            //determine dEdx:
            //===============


            TrackerHitVec HitVec = outputTrack->getTrackerHits();

            double phi =  outputTrack->getPhi();
            double d0 =  outputTrack->getD0();
            float Q_sum = 0;

            typedef std::vector< std::pair<TrackerHit*, double> > HitDouble;
            HitDouble HitPairVec;

            //fill pair of Hit and s and sort them by s
            for (TrackerHitVec::iterator i = HitVec.begin(); i < HitVec.end(); i++)
                {

                    double s = cos(phi) * (*i)->getPosition()[0] + sin(phi) * (*i)->getPosition()[1]
                               - 2 * d0 * cos(phi) * sin(phi);   // track length in xy Projection

                    s = s *   std::sqrt(1 + outputTrack->getTanLambda() * outputTrack->getTanLambda());    // FIXME here comes the third dimension

                    TrackerHit* Hit = *i;
                    std::pair<TrackerHit*, double> HitPair(Hit , s);
                    HitPairVec.push_back(HitPair);

                    //sum of charge on track
                    streamlog_out(DEBUG2) << "q_i = " << Hit->getEDep() << std::endl;
                    Q_sum += Hit->getEDep();

                }

            sort(HitPairVec.begin(), HitPairVec.end(), compareHit);

            //effective length of track is distance between outer hits times n/(n-1):
            //distance between outer points on track
            float l = fabs((*HitPairVec.begin()).second - (*(HitPairVec.end() - 1)).second);
            //effective length
            float l_eff = l * HitVec.size() / (HitVec.size() - 1);

            streamlog_out(MESSAGE1) << "l = " << l << "  l_eff = " << l_eff << std::endl;
            streamlog_out(MESSAGE1) << "Q_sum = " << Q_sum << "  dEdx = " << Q_sum / l_eff << std::endl;

            //set dEdx
            outputTrack->setdEdx(Q_sum / l_eff);

            outputTrackCollection->push_back(outputTrack);


            streamlog_out(MESSAGE0) << std::setprecision(5) << "\toutputTrack phi:"
            << std::setw(8) << outputTrack->getPhi() << " rads ,\tOmega: "
            << std::setw(8) << outputTrack->getOmega() << " mm^-1 ,\td0: "
            << std::setw(8) << outputTrack->getD0()
            << " mm ,\ttan(lambda): " << std::setw(8) << outputTrack->getTanLambda()
            << " ,\tz0: " << std::setw(8) << outputTrack->getZ0() << "mm, \tdEdx: "
            << outputTrack->getdEdx() << " TOT/mm "
            << std::endl;;

        } // loop tracks

    // now add the collection to the event!
    try
        {

            evt->addCollection(outputTrackCollection, _output_seed_tracks_collection_name);
        }
    catch (EVENT::EventException e)
        {

            std::cout << "LinearRegressionProcessor::processEvent(). " << e.what()
                      << "Collection " << _output_seed_tracks_collection_name
                      << " is being deleted." << std::endl;

            std::vector<EVENT::LCObject*>::iterator iter = outputTrackCollection->begin();

            for (; iter != outputTrackCollection->end(); iter++)
                {

                    if (*iter)
                        {

                            delete(*iter);
                            (*iter) = NULL;
                        }
                }

            outputTrackCollection->clear();
            delete outputTrackCollection;
        }
}// processEvent


Processor* LinearRegressionProcessor::newProcessor()
{

    return new LinearRegressionProcessor();
}


void LinearRegressionProcessor::check(LCEvent *evt)
{

    LCCollection* outputCol = 0;
    try
        {
            outputCol = evt->getCollection(_output_seed_tracks_collection_name) ;
        }

    catch (DataNotAvailableException)
        {
            // if there is not output collection give a warning to the user
            streamlog_out(WARNING2)
            << "check: output collection "
            << _output_seed_tracks_collection_name << " cannot be found in event "
            << evt->getEventNumber() << std::endl;
            return;
        }

    for (std::vector<AIDA::IHistogram2D*>::iterator i = HistoLineVec.begin();
            i < HistoLineVec.end(); i++)
        {

            AIDAProcessor::histogramFactory(this)->destroy(*i);

        }
    for (std::vector<AIDA::IHistogram2D*>::iterator i = HistoTrackVec.begin();
            i < HistoTrackVec.end(); i++)
        {

            AIDAProcessor::histogramFactory(this)->destroy(*i);

        }

    HistoLineVec.clear();
    HistoTrackVec.clear();

    std::stringstream sEvt;
    sEvt << evt->getEventNumber();

    double _pitch = 0.055;

    for (int iTrack = 0; iTrack < outputCol->getNumberOfElements() ; iTrack++)
        {

            std::stringstream siTrack;
            siTrack << iTrack;

            std::string histoname3 = "HistoLine" + sEvt.str() + "_Track" + siTrack.str();
            std::string histoname4 = "HistoTrack" + sEvt.str() + "_Track" + siTrack.str();

            streamlog_out(DEBUG0) << "creating histo : " << histoname3 << std::endl;

            HistoLineVec.push_back(AIDAProcessor::histogramFactory(this)->
                                   createHistogram2D(histoname3.c_str(), "track in mm",
                                                     256 , 0, 14.08, 256 , 0, 14.08));

            AIDA::IHistogram2D *HistoLine = *(HistoLineVec.end() - 1);


            streamlog_out(DEBUG0) << "creating histo : " << histoname4 << std::endl;

            HistoTrackVec.push_back(AIDAProcessor::histogramFactory(this)->
                                    createHistogram2D(histoname4.c_str(), "track in mm",
                                                      256 , 0, 14.08, 256 , 0, 14.08));

            AIDA::IHistogram2D *HistoTrack = *(HistoTrackVec.end() - 1);


            Track* RawTrack =
                dynamic_cast<Track*>(outputCol->getElementAt(iTrack)) ;

            if (RawTrack != 0)
                {


                    std::vector< TrackerHit * > HitVec = RawTrack->getTrackerHits();
                    for (std::vector< TrackerHit * >::iterator i = HitVec.begin();
                            i < HitVec.end(); i ++)
                        {

                            TrackerHit* Hit = *i;
                            const double *pos = Hit->getPosition();
                            HistoTrack->fill(pos[0], pos[1]);

                        }

                    double a = 0, b = 0, c = 0, d = 0;

                    if (fabs(RawTrack->getPhi()) < M_PI / 4)     //y=ax+b
                        {

                            a = tan(RawTrack->getPhi());
                            b = RawTrack->getD0() / cos(RawTrack->getPhi());
                            c = 0;
                            d = 0;

                            if (a != 0)
                                {

                                    c = 1 / a;
                                    d = -b / a;

                                }

                        }

                    else  //x=cy+d
                        {


                            a = tan(RawTrack->getPhi());
                            b = - RawTrack->getD0() / cos(RawTrack->getPhi());
                            c = 0;
                            d = 0;

                            if (a != 0)
                                {

                                    c = 1 / a;
                                    d = -b / a;

                                }


                        }
                    streamlog_out(DEBUG4) << " a = " << a << "  b = " << b << std::endl;
                    streamlog_out(DEBUG4) << " c = " << c << "  d = " << d << std::endl;


                    for (int i = 0; i < 256; i++)
                        {

                            HistoLine->fill(i*_pitch, a*i*_pitch  + b);
                            HistoLine->fill(c*i*_pitch  + d  , i*_pitch);

                        }

                } //end if( RawTrack != 0 )

        }//end for(int iTrack=0; iTrack < outputCol->getNumberOfElements()


}


void LinearRegressionProcessor::end()
{


}



} // namespace marlintpc
