#ifdef USE_TRACK_FITTER_KALMAN
#include "StraightKalTrack.h"

#include <marlin/Global.h>
#include <gear/PadRowLayout2D.h>
#include <gearimpl/RectangularPadRowLayout.h>
#include <gear/TPCParameters.h>
#include <gear/TPCModule.h>
#include "TPCHitMap.h"

#include <streamlog/streamlog.h>

using namespace gear;
using namespace std;

namespace marlintpc
{

StraightKalTrack::StraightKalTrack(const TPCParameters* tp)
        : AttKalTrack(tp), _sv(2, 1), _sv1(2, 1), _H(1, 2), _H_t(2, 1),
        _K(2, 1), _P(2, 2), _P1(2, 2), _R(1, 1), _M_unit(2, 2), _deltaChi2(0.0)
{
    reset();
}


StraightKalTrack::StraightKalTrack(const StraightKalTrack& orig)
        : AttKalTrack(orig), _sv(orig._sv), _sv1(orig._sv1), _H(orig._H), _H_t(orig._H_t),
        _K(orig._K), _P(orig._P), _P1(orig._P1), _R(orig._R), _M_unit(orig._M_unit), _deltaChi2(orig._deltaChi2)
{
}

void StraightKalTrack::reset()
{
    const double P1_init[4] = {1.0e10, 0, 0, 1.0e10};
    const double sv1_init[2] = {0, 0};
    const double PHI_init[4] = {1, 0, 0, 1};

    _P1(0, 0) = P1_init[0];
    _P1(0, 1) = P1_init[1];
    _P1(1, 0) = P1_init[2];
    _P1(1, 1) = P1_init[3];

    _sv1(0, 0) = sv1_init[0];
    _sv1(1, 0) = sv1_init[1];

    _M_unit(0, 0) = PHI_init[0];
    _M_unit(0, 1) = PHI_init[1];
    _M_unit(1, 0) = PHI_init[2];
    _M_unit(1, 1) = PHI_init[3];

    _H[0][1] = 1;
    _H_t[1][0] = 1;

    _R(0, 0) = 0.4;
    _hits.clear();
}


AttKalTrack * StraightKalTrack::clone()
{
    return new StraightKalTrack(*this);
}


bool StraightKalTrack::searchTrack(ModuleRowHit firstHit)
{
    streamlog_out(DEBUG3) << "---------------------------" << endl
    << "In SearchTrack()"            << endl
    << "---------------------------" << endl;

    bool foundTrack = false;

    _domain = new TPCHitDomain(_tpcParameters, _maxSkipRows);
    if (!_domain->setDomain(firstHit, 1))return false;//next hit domain

    const double *pos = TPCHitMap::getHitPosition(firstHit);

    streamlog_out(DEBUG3) << "hit1_x=" << pos[0] << " hit1_y=" << pos[1] << endl;

    list<ModuleRowHit> hitList = _tpcHitMap->getHitsInDomain(*_domain);
    list<ModuleRowHit>::iterator secondHit;

    for (secondHit = hitList.begin(); secondHit != hitList.end(); secondHit++)
        {
            if (foundTrack)break;

            //reset track
            reset();

            doAddAndFilter(firstHit);  //add the 1st hit
            doAddAndFilter(*secondHit);//add the 2nd hit

            if (!_domain->setDomain(*secondHit, 1))return false;
            searchTrackHit();//it's recursive

            if (getNumberOfHits() >= _minTrackHits)foundTrack = true;
        }

    delete _domain;
    return foundTrack;
}


bool StraightKalTrack::doAddAndFilter(ModuleRowHit hit)
{
    ///////FIXME///////////
    const double SIGMA = 0.3;
    const double N_SIGMA = 5;
    const double MAX_TIME = 200;   //max time difference
    const double SIGMA_TIME = 50;
    ///////FIXME///////////

    streamlog_out(DEBUG3) << "--------------------------" << endl
    << "In AddAndFilter()"          << endl
    << "--------------------------" << endl;
    if (_hits.size() < 2)
        {
            _hits.push_back(hit);
            _lastBestHit = hit;

            if (_hits.size() == 2)//if the track has two hits, calculate the state vector
                {
                    const double* hp1 = TPCHitMap::getHitPosition(*(_hits.begin()));
                    const double* hp2 = TPCHitMap::getHitPosition(*(_hits.rbegin()));

                    streamlog_out(DEBUG3) << "hit1_x=" << hp1[0] << " hit1_y=" << hp1[1] << endl;
                    streamlog_out(DEBUG3) << "hit2_x=" << hp2[0] << " hit2_y=" << hp2[1] << endl;

                    _sv1[0][0] = (hp2[1] - hp1[1]) / (hp2[0] - hp1[0]);//slope
                    _sv1[1][0] = hp2[1] - _sv1[0][0] * hp2[0];//intercept

                    _sv = _sv1;

                    streamlog_out(DEBUG3) << "Initial state vector: " << "slope=" << _sv[0][0] << "  intercept="  << _sv[1][0] << endl;
                }

            return true;//a hit is added,retrun true
        }
    else
        {
            const double* hitPos1 = TPCHitMap::getHitPosition(*_hits.rbegin());
            const double* hitPos2 = TPCHitMap::getHitPosition(hit);

            double predHitPos = _sv[0][0] * hitPos2[0] + _sv[1][0];//predicted hit position
            double delta = fabs(hitPos2[1] - predHitPos);

            streamlog_out(DEBUG3) << "hit_x=" << hitPos2[0] << " hit_y=" << hitPos2[1];
            streamlog_out(DEBUG3) << "  prediction_y=" << predHitPos  << " delta=" << delta << endl;

            if (delta < N_SIGMA*SIGMA)
                {
                    double timeDiff = fabs(hitPos2[2] - hitPos1[2]);//time difference
                    if (timeDiff > MAX_TIME)return false;

                    //update state vector
                    double z = hitPos2[1];
                    _H[0][0] = _H_t[0][0] = hitPos2[0];
                    _K = _P1 * _H_t;
                    TMatrixD t = (_H * _P1 * _H_t + _R).Invert();

                    _K[0][0] = _K[0][0] * t[0][0];
                    _K[1][0] = _K[1][0] * t[0][0];

                    _sv = _sv1 + _K * (z - _H * _sv1);
                    _P = (_M_unit - _K * _H) * _P1;
                    _sv1 = _sv;
                    _P1 = _P;

                    _deltaChi2 = pow(delta / SIGMA, 2) + pow(timeDiff / SIGMA_TIME, 2);
                    streamlog_out(DEBUG3) << "k=" << _sv1[0][0] << "   b=" << _sv1[1][0] << endl;

                    return true;
                }
            else
                return false;
        }
}

void StraightKalTrack::printStateVector()
{
    streamlog_out(DEBUG4) << "slope=" << _sv[0][0]
    << " intercept=" << _sv[1][0]
    << endl;
}

}
#endif
