#include "TrackFitterBase.h"
#include <gear/GEAR.h>
#include <gear/TPCParameters.h>
#include <gear/PadRowLayout2D.h>
#include <CLHEP/Vector/ThreeVector.h>

#include <cmath>

#include <streamlog/streamlog.h>

//#include "IntersectionCalculator.h"

namespace marlintpc
{

// the static variables
//static const int BASE = 0;
const unsigned char TrackFitterBase::FitterTypes::LIKELIHOOD; ///< Global likelihood method
const unsigned char TrackFitterBase::FitterTypes::SIMPLECHISQUARE; ///< Simple \f$\chi^2\f$ without geometry information
const unsigned char TrackFitterBase::FitterTypes::SIMPLECHISQUAREPADS; ///< Simple \f$\chi^2\f$ using pad geometry information
const unsigned char TrackFitterBase::FitterTypes::LINEARREGRESSION; ///< Linear regression for a straight line
const unsigned char TrackFitterBase::FitterTypes::KALMAN;  ///< Kalman filtering of the track
const unsigned char TrackFitterBase::FitterTypes::SIMPLEMINIMIZER; ///< Fitter which does not need parameters
const unsigned char TrackFitterBase::FitterTypes::CHISQUARE; ///< The \f$\chi^2\f$ fitter


TrackFitterBase::TrackFitterBase() : _d0Start(0.), _useD0Start(false),
        _phiStart(0.), _usePhiStart(false), _omegaStart(0.),  _useOmegaStart(false),
        _tanLambdaStart(0.), _useTanLambdaStart(false), _z0Start(0.), _useZ0Start(false),
        _noZFit(false)
{}

EVENT::DoubleVec  TrackFitterBase::calculateResiduals(EVENT::Track const * testTrack,
                                                      unsigned int const testHitNumber,
                                                      EVENT::Track const * referenceTrack) const
{
    // get hit testHitNumber from track
    EVENT::TrackerHitVec const &hitVector = testTrack->getTrackerHits();

    if (testHitNumber >= hitVector.size())
        {
            streamlog_out(ERROR1) << "TrackFitterBase: Cannot calulate residual, not hit number "
            << testHitNumber << " in track." << std::endl;
            throw EVENT::DataNotAvailableException("Could not find hit in track.");
        }

    EVENT::TrackerHit *hit = hitVector[testHitNumber];

    // now simply call calculateResidual(testHit, trackWithoutTestHit),
    // functionality is the same for this basic implementation

    if (referenceTrack) // residual wrt. referenceTrack has been requested
        return calculateResiduals(hit, referenceTrack);

    else // return residual wrt. track which contains the hit
        return calculateResiduals(hit, testTrack);

}

EVENT::DoubleVec TrackFitterBase::calculateResiduals(EVENT::TrackerHit const * testHit,
                                                     EVENT::Track const * trackWithoutTestHit) const
{
    // just for convenience: alias for trackWithoutTestHit
    EVENT::Track const *track = trackWithoutTestHit;

    // first calculate some track parameters from track:
    double phi =  track->getPhi(); // just to improve readability

    // unit vector from reference point to point of closest approach (pca) (xy plane only)
    CLHEP::Hep3Vector n_pca(-sin(phi), cos(phi));

    // reference point in xy projection
    CLHEP::Hep3Vector reference_point(track->getReferencePoint()[0],
                                      track->getReferencePoint()[1],
                                      track->getReferencePoint()[2]);

    // point of closest approach (3D point which is closest in the xy plane)
    CLHEP::Hep3Vector pca =  reference_point
                             +  track->getD0() *  n_pca  // the xy part
                             +  track->getZ0() * CLHEP::Hep3Vector(0, 0, 1); // the z part

    streamlog_out(DEBUG4) << "PCA is " << pca[0] << " , " << pca[1] << " , " << pca[2] << std::endl;

    // A Hep3Vector for the test hit
    CLHEP::Hep3Vector hit(testHit->getPosition()[0],
                          testHit->getPosition()[1],
                          testHit->getPosition()[2]);

    // psi is the angle of the pca wrt the helix center
    // (Note: in the paper from Lasiuk et al. from 1998 this is Phi_0,
    //  while Psi is the angle in track direktion.
    //  As Phi is the angle in track direction in LCIO,
    //  Phi and psi are exchanged wrt. the Lasiuk paper.)
    double psi = track->getPhi() + (track->getOmega() < 0 ? -1 : 1) * M_PI_2;

    // wrap psi to -2pi .. 2pi to prevent the while loop running for ages
    // in case there are nonsense values for the track's phi
    if (std::fabs(psi) > 20)
        {
            streamlog_out(WARNING3) << "Psi is " << psi << ", check phi of the input tracks!"
            << std::endl;
        }
    psi = std::fmod(psi, 2 * M_PI);

    // Calculate pca wrt. the hit by using s, which is the length on the track
    double s; // the track length in the xy-plane (projection)
    if (track->getOmega() != 0.)
        {
            double co = std::fabs(track->getOmega()) * (hit[0] - pca[0]) + cos(psi);
            double si = std::fabs(track->getOmega()) * (hit[1] - pca[1]) + sin(psi);
            streamlog_out(DEBUG2) << "TrackFitterBase: co " << co << " : si " << si
            << " : psi " << psi << std::endl;


            double phi = (co == 0 ? (si < 0 ? - M_PI_2    : M_PI_2)
                                  : (co > 0 ? - atan(si / co) : M_PI - atan(si / co)));

            // wrap phi to be in +- pi wrt. -psi (minus because the pca is at -psi wrt helix centre)
            while (-psi - phi >  M_PI) phi += 2 * M_PI ;
            while (-psi - phi < -M_PI) phi -= 2 * M_PI ;

            s = (-psi - phi) / track->getOmega();
            //      s = ( atan(si/co) + ( co < 0 ? M_PI : 0 ) - psi ) / track->getOmega();

//    //This is the method from Lasiuk et al. It does not work because
//    // the atan only works for +- pi/2, but we need a full circle
//    s = atan( ( (hit[1] - pca[1] ) * sin(phi) + (hit[0] -pca[0]) * cos(phi) ) /
//          ( 1/std::fabs( track->getOmega() ) +
//            (hit[0] - pca[0]) * sin(phi) -
//            (hit[1] - pca[1]) * cos(phi)
//          )
//        )
//        /  track->getOmega() ;
            streamlog_out(DEBUG3) << "TrackFitterBase: Hit " << hit[0] << " : " << hit[1] << " : " << hit[2]
            << " ; s is " << s << std::endl;
        }
    else
        {
            s = (hit[1] - pca[1]) * sin(phi) + (hit[0] - pca[0]) * cos(phi);
        }

    CLHEP::Hep3Vector pcaWrtHit;
    // the tangent at the pcaWrtHit (unit vector)

    CLHEP::Hep3Vector tangent_pcaWrtHit;
    // calculate x y and z from s
    if (track->getOmega() != 0.)
        {
            // for a helix we need the angle psi, which is the angle of the
            // pca measured from the centre of the helix.
            pcaWrtHit[0] = pca[0] +
                           (cos(psi + track->getOmega() * s) - cos(psi)) / std::fabs(track->getOmega()) ;
            pcaWrtHit[1] = pca[1] +
                           (sin(psi + track->getOmega() * s) - sin(psi)) / std::fabs(track->getOmega()) ;
            pcaWrtHit[2] = pca[2] - s * track->getTanLambda();

            double phi_pcaWrtHit =  psi + track->getOmega() * s + (track->getOmega() < 0 ? -1 : 1) * M_PI_2;
            tangent_pcaWrtHit = CLHEP::Hep3Vector(cos(phi_pcaWrtHit) , sin(phi_pcaWrtHit));
        }
    else
        {
            pcaWrtHit[0] = pca[0] + s * cos(phi);
            pcaWrtHit[1] = pca[1] + s * sin(phi);
            pcaWrtHit[2] = pca[2] + s * track->getTanLambda();

            // the tangent is the direction of the track itself (it's a straight line, omega=0 )
            tangent_pcaWrtHit = CLHEP::Hep3Vector(cos(phi) , sin(phi));

        }

    // now calculate the (signed!) residuals

    // the z-component of the cross  product of (pcaWrtHit - hit) with tangent_pcaWrtHit,
    // which is a unit vector,
    // gives the magnitude of (pcaWrtHit - hit) since the vectors are perpendicular,
    // including the correct sign
    CLHEP::Hep3Vector pcaWrtHit_minus_hit = (pcaWrtHit - hit);
    pcaWrtHit_minus_hit[2] = 0; // we do not considder the z coordinate;

    double residualInReadoutPlane = pcaWrtHit_minus_hit[0] * tangent_pcaWrtHit[1]
                                    - pcaWrtHit_minus_hit[1] * tangent_pcaWrtHit[0];
    streamlog_out(DEBUG4) << "TrackFitterBase: residual xy is " <<  residualInReadoutPlane
    << "(pcaWrtHit - hit).mag() is " << (pcaWrtHit - hit).mag() << std::endl;

    double residualZ = hit[2]  - pcaWrtHit[2];
    streamlog_out(DEBUG4) << "TrackFitterBase: zpos is " <<  hit[2]
    << " ; pcaWrtHit[2] is " << pcaWrtHit[2]  << std::endl;

    // s is only known mod(2 pi R), so it might be on the wrong loop of the curler
    // maximum distance in z to the next loop: pi*r*tan(lanbda)
    if (track->getOmega() != 0.)
        {
            // first wrap to [- 2pi r tan(lambda) : 2pi r tan(lambda)]  (!not to  pi*r*tan(lambda) !)
            residualZ = fmod(residualZ , 2 * M_PI * fabs(track->getTanLambda() / track->getOmega())) ;


            // now ensure that the results are whithin the correct loop...
            if (residualZ > M_PI*fabs(track->getTanLambda() / track->getOmega()))
                {
                    streamlog_out(DEBUG1) << "#plus" << residualZ << std::endl;
                    residualZ -= 2 * M_PI * fabs(track->getTanLambda() / track->getOmega());
                }

            if (residualZ < - M_PI*fabs(track->getTanLambda() / track->getOmega()))
                {
                    streamlog_out(DEBUG1) << "#minus" << residualZ << std::endl;
                    residualZ += 2 * M_PI * fabs(track->getTanLambda() / track->getOmega());
                }
        }

    streamlog_out(DEBUG2) << "TrackFitterBase: residualZ = " <<  residualZ  << std::endl;

    EVENT::DoubleVec retVal;
    retVal.push_back(residualInReadoutPlane);
    retVal.push_back(residualZ);

    return retVal;

}// calculateResidual( testHit, trackWithoutTestHit )

std::string TrackFitterBase::getRevision() const
{
    return std::string("$Rev: 2253 $");
}

// calculate angular distance on a circle, max angle can be pi
double TrackFitterBase::distanceOnCircle(double phi1, double phi2) const
{
    double deltaPhi = std::fmod(std::fabs(phi1 - phi2) , 2 * M_PI);
    if (deltaPhi > M_PI)
        deltaPhi = 2 * M_PI - deltaPhi;

    return deltaPhi;
}

void TrackFitterBase::setTrackFitterType(unsigned char type, IMPL::TrackImpl *track)
{
    for (int index = 0; index < 8; index++)
        {
            // set the bit in the tracks type word if the bit in the char is set (whew, this IS ugly)
            track->setTypeBit(index + 16, (type & (1 << index)));
        }
}

unsigned char TrackFitterBase::getFitterTypeFromTrack(int trackTypeWord)
{
    return static_cast<unsigned char>((trackTypeWord >> 16) & 0xFF);
}

void TrackFitterBase::setD0Start(double d0)
{
    _d0Start = d0;
    _useD0Start = true;
}

void TrackFitterBase::setPhiStart(double phi)
{
    _phiStart = phi;
    _usePhiStart = true;
}

void TrackFitterBase::setOmegaStart(double omega)
{
    _omegaStart = omega;
    _useOmegaStart = true;
}

void TrackFitterBase::setTanLambdaStart(double tanLambda)
{
    _tanLambdaStart = tanLambda;
    _useTanLambdaStart = true;
}

void TrackFitterBase::setZ0Start(double z0)
{
    _z0Start = z0;
    _useZ0Start = true;
}

void TrackFitterBase::setNoZFitFlag(bool noZFit)
{
    _noZFit = noZFit;
}

bool TrackFitterBase::getNoZFitFlag() const
{
    return _noZFit;
}

void TrackFitterBase::fixD0(bool fix)
{
    _fixD0 = fix;
}

void TrackFitterBase::fixPhi(bool fix)
{
    _fixPhi = fix;
}

void TrackFitterBase::fixOmega(bool fix)
{
    _fixOmega = fix;
}

void TrackFitterBase::fixTanLambda(bool fix)
{
    _fixTanLambda = fix;
}

void TrackFitterBase::fixZ0(bool fix)
{
    _fixZ0 = fix;
}
}//namespace marlintpc
