
// LCIO
#include <EVENT/LCIO.h>
#include <EVENT/LCCollection.h>
#include <EVENT/MCParticle.h>
#include <EVENT/SimTrackerHit.h>
#include <IMPL/LCCollectionVec.h>

#include <cassert>

// Marlin
#include <marlin/Exceptions.h>

// MarlinTPC
#include "ElectronCloud.h"
#include "PrimaryCloudIonisationProcessor.h"

using namespace std;
using namespace EVENT;
using namespace IMPL;
using namespace marlin;

namespace marlintpc {

// create one to register it with Marlin
PrimaryCloudIonisationProcessor aPrimaryCloudIonisationProcessor;


PrimaryCloudIonisationProcessor::PrimaryCloudIonisationProcessor():
	Processor("PrimaryCloudIonisationProcessor"),
	_input_collection_name("TPCSimTrackerHits"),
	_output_collection_name("TPCElectronClouds"),
	_simulate_only_primary_particle(false) {

	registerInputCollection(
		LCIO::SIMTRACKERHIT,
		"InputTPCSimTrackerHits",
		"Name of the input SimTrackerHits collection from the TPC simulation",
		_input_collection_name,
		string("TPCCollection") );

	registerOutputCollection(
		LCIO::LCGENERICOBJECT,
		"OutputTPCElectronClouds",
		"Name of the output LCGeneric collection full of electron clouds",
		_output_collection_name,
		string("TPCElectronClouds") );

	registerOptionalParameter(
		"OnlySimulatePrimaryParticle?",
		"Controls if just the energy deposits from the primary particle are considered (true) or all deposits are simulated (false - default)",
		_simulate_only_primary_particle,
		false );
}


PrimaryCloudIonisationProcessor::~PrimaryCloudIonisationProcessor() {}


void PrimaryCloudIonisationProcessor::processRunHeader( LCRunHeader* run) {

	// store the processor parameters in the event
    for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {

        if ( ! i->second->isOptional() || i->second->valueSet() ) {

            run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
        }
    }
}


void PrimaryCloudIonisationProcessor::processEvent( LCEvent * evt ) {

	// grab the collection of energy deposits from the simulation
	LCCollection *simtrackerhit_collection = NULL;
	try {

		simtrackerhit_collection = evt->getCollection( _input_collection_name );
	}
	catch (DataNotAvailableException &e) {

		// the collection of SimTrackerHits with that name wasn't in the LCIO event file, skip simulating this event
		throw SkipEventException( this );
	}

	// create a collection to store the electron clouds (don't write this collection at the end)
	LCCollectionVec *electroncloud_collection = new LCCollectionVec( LCIO::LCGENERICOBJECT );
	electroncloud_collection->setTransient();

	streamlog_out( MESSAGE3 ) << "Event# " << evt->getEventNumber() << ". Processing " << simtrackerhit_collection->getNumberOfElements() << " SimTrackerHits\n";

	// keep track of the number of clouds produced for output
	unsigned int number_of_electron_clouds = 0;

	// now loop through the energy deposits creating clouds of electrons
	for( int hit_number = 0; hit_number < simtrackerhit_collection->getNumberOfElements(); hit_number++ ) {

		SimTrackerHit *current_hit = dynamic_cast<SimTrackerHit*>( simtrackerhit_collection->getElementAt( hit_number ) );
		assert(current_hit);

		// check if only requested to simulate some of the event
		if( _simulate_only_primary_particle ) {

			if( current_hit->getMCParticle()->getParents().size() != 0 ) {

				continue;
			}
		}

		// get some information about this hit
		const double energy = current_hit->getEDep();

		const int number_of_electrons = (int) ( (energy * 1.E09) / 26. );	// 26eV per electron

		if( number_of_electrons <= 0 ) {

			continue;
		}

		// create an electron cloud
		ElectronCloud *new_cloud = new ElectronCloud();

		// set up the deviation in the position of the electrons in the cloud
		new_cloud->set_transverse_stdev(0.01);
		new_cloud->set_longitudinal_stdev(0.01);

		new_cloud->set_cloud_id( hit_number );

		// set up the mean position of the electrons in the cloud
		const double *sim_position = current_hit->getPosition();
		new_cloud->set_x_position( sim_position[0] );
		new_cloud->set_y_position( sim_position[1] );
		new_cloud->set_z_position( sim_position[2] );

		streamlog_out( DEBUG2 ) << "Produced a cloud with " << number_of_electrons << " electrons at \t" << sim_position[0] << "mm, \t" << sim_position[1] << "mm, \t" << sim_position[2] << "mm\n";

		// set up the other parameters
		// geant4 outputs the time in ns. put the time in microseconds, that's what the drift velocity is parameterised as
		new_cloud->set_time( current_hit->getTime()*1.E-03 );
		new_cloud->set_electron_count( number_of_electrons );

		// now add the cloud to the collection
		electroncloud_collection->addElement( new_cloud );

		number_of_electron_clouds += 1;
	}

	streamlog_out( MESSAGE3 ) << "Produced " << number_of_electron_clouds << " electron clouds.\n";

	// now add the collection to the event so other processors may access it
	try {

		evt->addCollection( electroncloud_collection, _output_collection_name );
	}
	catch( ... ) {

		throw SkipEventException( this );
	}
	
}


void PrimaryCloudIonisationProcessor::init() {

	// usually a good idea to
	printParameters();
}


void PrimaryCloudIonisationProcessor::check( LCEvent * evt ) { 

}


void PrimaryCloudIonisationProcessor::end() { 

}

}
