
// LCIO
#include <EVENT/LCIO.h>
#include <EVENT/LCCollection.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/TrackerDataImpl.h>
#include <IMPL/LCFlagImpl.h>

// GEAR
#include <gear/GEAR.h>
#include <gear/TPCParameters.h>

// Marlin
#include <marlin/Global.h>
#include <marlin/Exceptions.h>

// MarlinTPC
#include "ChannelCorrection.h"
#include "ChargeSignal.h"
#include "SignalShaperGaussianProcessor.h"

#include <cassert>

using namespace std;
using namespace EVENT;
using namespace IMPL;
using namespace marlin;
using namespace tpcconddata;

namespace marlintpc {

// create one to register it with Marlin
SignalShaperGaussianProcessor aSignalShaperGaussianProcessor;


SignalShaperGaussianProcessor::SignalShaperGaussianProcessor():
	Processor("SignalShaperGaussianProcessor"),
	_input_collection_name("TPCChargeSignals"),
	_output_collection_name("TPCShapedSignals"),
	_tpc_channel_correction_collection_name(ChannelCorrection::getDefaultColName()),
	_gain(1.),
	_polarity(true),
	_mV_per_electron(1.),
	_induction_time(0.),
	_time_resolution(50.) {

	registerInputCollection(
		LCIO::LCGENERICOBJECT,
		"InputTPCChargeSignals",
		"Name of the input LCGenericObject collection containing ChargeSignals",
		_input_collection_name,
		string("TPCChargeSignals") );

	registerOutputCollection(
		LCIO::TRACKERDATA,
		"OutputTPCShapedSignals",
		"Name of the output TrackerData collection containing shaped signals.",
		_output_collection_name,
		string("TPCShapedSignals") );

	registerOptionalParameter(
		"TPCChannelCorrectionCollectionName",
		"The name to use for the tpc channel correction collection.",
		_tpc_channel_correction_collection_name,
		ChannelCorrection::getDefaultColName());

	registerProcessorParameter(
		"Gain",
		"The amplification of this electronic shaper.",
		_gain,
		1.);

	registerProcessorParameter(
		"CurrentPerCharge",
		"The linear voltage per charge constant of this shaper (mV/e).",
		_mV_per_electron,
		1.);

	registerOptionalParameter(
		"Polarity",
		"The polarity of this electronic shaper (false = negative signal).",
		_polarity,
		true);

	registerOptionalParameter(
		"InductionTime",
		"The amount of time the signal has been induced (ns).",
		_induction_time,
		0.);

	registerOptionalParameter(
		"TimeResolution",
		"The width of a signal sample (1/Hz) in ns.",
		_time_resolution,
		50.);
}


SignalShaperGaussianProcessor::~SignalShaperGaussianProcessor() {}


void SignalShaperGaussianProcessor::processRunHeader( LCRunHeader* run) {

	// store the processor parameters in the event
    for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {

        if ( ! i->second->isOptional() || i->second->valueSet() ) {

            run->parameters().setValue( _processorName + "_" + i->second->name(), i->second->value() );
        }
    }
}


void SignalShaperGaussianProcessor::processEvent( LCEvent * evt ) {

	// grab the collection of energy deposits from the simulation
	LCCollection *charge_signal_collection = NULL;
	try {

		charge_signal_collection = evt->getCollection( _input_collection_name );
	}
	catch (DataNotAvailableException &e) {

		// the collection of charge signals with that name wasn't in the LCIO event file, skip simulating this event
		throw SkipEventException( this );
	}

	streamlog_out( MESSAGE0 ) << "Shaping " << charge_signal_collection->getNumberOfElements() << " signals\n";

	// create a collection for the channel corrections
	LCCollectionVec *channel_corrections_collection = new LCCollectionVec( LCIO::LCGENERICOBJECT );
	channel_corrections_collection->setTransient(false);
	channel_corrections_collection->setSubset(false);

	// get the width of a time bin in ns
	if( !parameterSet("TimeResolution") ) {

		_time_resolution = 1.E09 / Global::GEAR->getTPCParameters().getReadoutFrequency();
	}

	// create a collection to store the TrackerPulses
	LCCollectionVec *trackerdata_collection = new LCCollectionVec( LCIO::TRACKERDATA );
	trackerdata_collection->setTransient(true);
	trackerdata_collection->setSubset(false);

	// set the flags that cellID1 should be stored
	lcio::LCFlagImpl trkFlag(0) ;
	trkFlag.setBit( lcio::LCIO::TRAWBIT_ID1 ) ;
	trackerdata_collection->setFlag( trkFlag.getFlag() );

	// now loop through the energy deposits creating clouds of electrons
	for( int charge_signal_number = 0; charge_signal_number < charge_signal_collection->getNumberOfElements(); charge_signal_number++ ) {

		ChargeSignal *current_charge_signal = dynamic_cast<ChargeSignal*>( charge_signal_collection->getElementAt( charge_signal_number ) );
		assert(current_charge_signal);

		// amplitude is proportional to the number of electrons
		const double total_electrons_collected = current_charge_signal->get_electron_count();

		if( total_electrons_collected <= 0. ) {

			continue;
		}

		// the arrival time is mean arrival time of the electrons (in nanoseconds)
		const double mean_arrival_time = current_charge_signal->get_mean_time()*1.E03;

		// the deviation of the arrival time (in nanoseconds)
		const double arrival_time_deviation = sqrt( current_charge_signal->get_time_variance() )*1.E03;

		TrackerDataImpl *new_data = new TrackerDataImpl();

		new_data->setCellID0( current_charge_signal->get_pad_id() );
		new_data->setCellID1( current_charge_signal->get_module_id() );

		streamlog_out( DEBUG3 ) << "shaping signal on pad: " << new_data->getCellID0() << " with " << total_electrons_collected << " electrons, mean time: " << mean_arrival_time << "ns, arrival deviation: " << arrival_time_deviation << "ns." << endl;

		// get 3 sigma worth of the signal
		const double signal_start_time = mean_arrival_time - (3 * arrival_time_deviation) - _induction_time;
		const double signal_end_time = signal_start_time + (6 * arrival_time_deviation);

		int signal_start_bin = (int)(signal_start_time / _time_resolution);

		double current_bin_time = signal_start_bin * _time_resolution;

		// a constant for shaping
		const double amplitude = total_electrons_collected * _gain * _mV_per_electron;

		double debug_total = 0.;

		// FIXME allow negative signals
		double max_bin = 0.;
		while( current_bin_time <= signal_end_time ) {

			const double left  = current_bin_time;

			const double left_erf_param = (mean_arrival_time - left) * M_SQRT1_2 / arrival_time_deviation;
			const double right_erf_param = (mean_arrival_time - left - _time_resolution) * M_SQRT1_2 / arrival_time_deviation;

			const double distribution = 0.5 * ( erf( left_erf_param ) - erf( right_erf_param ) );

			double voltage = amplitude * distribution;
			max_bin = max( max_bin, voltage );

			debug_total += voltage;

			new_data->chargeValues().push_back(voltage);

			current_bin_time += _time_resolution;
		}

		// create the calibration factor to get the number of electrons from the height of the maximum bin
		const double total_gain = max_bin / total_electrons_collected;

		// create a correction for this channel to restore the number of electrons
        ChannelCorrection *channel_correction = new ChannelCorrection( new_data->getCellID0(), new_data->getCellID1(), -1, -1  );
		channel_correction->setCalibrationFactor( total_gain );

		streamlog_out( DEBUG2 ) << *channel_correction << endl;

		channel_corrections_collection->addElement( channel_correction );

		streamlog_out( DEBUG3 ) << "start time: " << signal_start_bin*_time_resolution << ", total: " << debug_total << ", should be: " << amplitude << ", total gain: " << total_gain << endl;

		new_data->setTime(signal_start_bin*_time_resolution);

		trackerdata_collection->addElement( new_data );
	}

	streamlog_out( DEBUG4 ) << "Added " << channel_corrections_collection->getNumberOfElements() << " channel corrections to the event." << endl;
	streamlog_out( DEBUG4 ) << "Added " << trackerdata_collection->getNumberOfElements() << " pulses to the event." << endl;
	// now add the collection of tracker data and channel corrections to the event so other processors may access it
	try {

		evt->addCollection( trackerdata_collection, _output_collection_name );
		evt->addCollection( channel_corrections_collection, _tpc_channel_correction_collection_name );
	}
	catch( ... ) {

		throw SkipEventException( this );
	}
	
}


void SignalShaperGaussianProcessor::init() {

    // usually a good idea to
    printParameters();
}


void SignalShaperGaussianProcessor::check( LCEvent * evt ) { 

}


void SignalShaperGaussianProcessor::end() { 

}

}
