/*
 * TIEventBlock.hh
 *
 *  Created on: Jul 10, 2012
 *      Author: caiazza
 */

#ifndef TIEVENTBLOCK_HH_
#define TIEVENTBLOCK_HH_

#include "TIAltroBlock.hh"

//LCIO
#include "lcio.h"
#include "UTIL/LCTime.h"
#include "EVENT/LCEvent.h"
#include "EVENT/LCCollection.h"

namespace marlintpc
{
/**
 * Interface for the class decoding the ALTRO Event block
 */
class TIEventBlock : public TIAltroBlock
{
public:
    /**
     * Virtual destructor
     */
    virtual ~TIEventBlock(){}

    /**
     * Creates an LC Event with the data of the block.
     * @return A newly created LCEvent
     */
    virtual lcio::LCEvent* CreateLCEvent(std::string& DataCollectionName) = 0;

    /**
     * Timestamp of the event
     * @return Timestamp of the event
     */
    virtual lcio::LCTime GetTimestamp() const = 0;

    /**
     * Gets the event number
     * @return Event Number
     */
    virtual uint32_t GetNEvent() const = 0;

    /**
     * Gets a collection containing the General Electronics Parameters for this event.
     * @return lcio collection of generic electronic parameters
     */
    virtual lcio::LCCollection* GetGeneralElectronicPar() const = 0;

    /**
     * Gets a collection containing the Extended Electronics Parameters
     * @return lcio collection of extended electronics parameters
     */
    virtual lcio::LCCollection* GetExtendedElectronicPar() const = 0;
};// end interface

/**
 * Implementation of the event block decoder for V4.2-4.5
 */
class TEventBlock_V42: public TIEventBlock
{
public:
    TEventBlock_V42(std::vector<uint32_t>& Content);

    virtual ~TEventBlock_V42();

    /**
     * Get the block lenght in number of 32 bit words
     * @return number of 32 bit words in the block
     */
    virtual uint32_t GetLength() const;

    /**
     * Verifies that the content of the block matches the expected content of the type of block we build.
     * @return True if the content of the block matches the expectation, False otherwise
     */
    virtual bool CheckIntegrity() const;

    /**
     * Returns a constant reference to the content of the block
     * @return vector of 32 bit data words
     */
    virtual const std::vector<uint32_t>& Content() const;

    /**
     * Returns a data word that identifies the block type
     * @return 32 bit data word identifying the block
     */
    virtual uint32_t GetIdentifier() const;


    /**
     * Creates an LC Event with the data of the block.
     * @return A newly created LCEvent
     */
    virtual lcio::LCEvent* CreateLCEvent(std::string& DataCollectionName);

    /**
     * Timestamp of the event
     * @return Timestamp of the event
     */
    virtual lcio::LCTime GetTimestamp() const;

    /**
     * Gets the event number
     * @return Event Number
     */
    virtual uint32_t GetNEvent() const;

    /**
     * Gets a collection containing the General Electronics Parameters for this event
     * @return lcio collection of generic electronic parameters
     */
    virtual lcio::LCCollection* GetGeneralElectronicPar() const;

    /**
     * Gets a collection containing the Extended Electronics Parameters
     * @return lcio collection of extended electronics parameters
     */
    virtual lcio::LCCollection* GetExtendedElectronicPar() const;

    /**
     * Allows to set additional parameters to the RunHeader.
     * @param RunHeader
     */
    virtual void SetAdditionalParameters(lcio::LCRunHeader* RunHeader);

protected:
    /**
     * Sets the word index where the RCU data is supposed to start
     * @param Word Index
     */
    virtual void SetRCUDataStart(uint32_t Word);

private:
    std::vector<uint32_t>* _content;

    uint32_t _RCUDataStart;

};//end implementation for v4.6+

/**
 * Implementation of the event block decoding for V4.6+
 */
class TEventBlock_V46: public TEventBlock_V42
{
public:
    TEventBlock_V46(std::vector<uint32_t>& Content);

    virtual ~TEventBlock_V46();

    /**
     * Timestamp of the event
     * @return Timestamp of the event
     */
    virtual lcio::LCTime GetTimestamp() const;
};

}

#endif /* TICORBLOCK_HH_ */
