/*
 * TAltroBlockFactory.cpp
 *
 *  Created on: Jul 9, 2012
 *      Author: caiazza
 */

//STD
#include <string>
#include <iostream>
#include <sstream>

//Converter
#include "TAltroBlockFactory.hh"
#include "AltroConverterGlobals.hh"
#include "AltroConverterExceptions.hh"
#include "TISORBlock.hh"
#include "TIEORBlock.hh"
#include "TIPORBlock.hh"
#include "TICORBlock.hh"
#include "TIBOFBlock.hh"
#include "TIEOFBlock.hh"
#include "TIEventBlock.hh"

namespace marlintpc
{

TAltroBlockFactory::TAltroBlockFactory(std::pair<int, int> DataVersion)
{
    SetDataVersion(DataVersion);
}

TAltroBlockFactory::TAltroBlockFactory(const TAltroBlockFactory& other):
        _DataVersion(other._DataVersion)
{

}

TAltroBlockFactory::~TAltroBlockFactory()
{

}

TAltroBlockFactory&
TAltroBlockFactory::operator=(const TAltroBlockFactory& other)
{
    if(this != &other)
    {
        _DataVersion = other._DataVersion;
    }
    return *this;
}

void
TAltroBlockFactory::SetDataVersion(std::pair<int, int> DataVersion) throw (TAltroConverterVersionException)
{
    if(DataVersion.second != 0 && DataVersion.second != 1)
    {
        std::stringstream message;
        message << "Data Format not recognized. It must be 0 or 1. Data format detected "
                << DataVersion.second << std::endl;
        throw TAltroConverterVersionException("TAltroBlockFactory", message.str());
    }
    if(DataVersion.first < 420)
    {
        std::stringstream message;
        message << "This program only supports DAQ version 4.2+. There should be no test beam data taken "
                << "with previous versions of the DAQ. Detected DAQ version: "<< DataVersion.first << std::endl;
        throw TAltroConverterVersionException("TAltroBlockFactory", message.str());
    }
    _DataVersion = DataVersion;
}

TIAltroBlock*
TAltroBlockFactory::GetBlock(std::ifstream* input) throw (TAltroConverterException, std::exception)
{
    if(input == 0)
    {
        std::stringstream message;
        message << "The input file stream provided for the block creation is empty." << std::endl;
        throw TAltroConverterInvalidInputException("TAltroBlockFactory", message.str());
    }

    input->peek();
    if(input->eof())
    {
        std::stringstream message;
        message << "Requiring a block after the end of file." << std::endl;
        throw TAltroConverterEOFException("TAltroBlockFactory", message.str());
    }

    uint32_t BlockLength = ReadWord32(input);
    uint32_t TempCursor = input->tellg();
    input->seekg(0, std::ifstream::end);
    uint32_t Remaining = static_cast<uint32_t>(input->tellg()) - TempCursor;
    if(Remaining < (BlockLength*WORD32))
    {
        std::stringstream message;
        message << "Corrupted input file. The next block has a length higher than the remaining "
                << "file length"<< std::endl;
        message << "Block Length: " << BlockLength*WORD32 << "\t Remaining bytes: " << Remaining << std::endl;
        throw TAltroConverterException("TAltroBlockFactory", message.str());
    }
    input->seekg(TempCursor, std::ifstream::beg);

    bool doDelete = true;
    std::vector<uint32_t>* content = new std::vector<uint32_t>();
    content->reserve(BlockLength+1);
    content->push_back(BlockLength);
    for(uint32_t i = 0; i < BlockLength; ++i)
    {
        content->push_back(ReadWord32(input));
    }

    TIAltroBlock* result = 0;
    switch(content->at(2))
    {
    case BLOCK_SOR:
    {
        if(_DataVersion.first >= 460)
        {
            result = new TSORBlock_V46(*content);
        }
        else if (_DataVersion.first >= 420)
        {
            result = new TSORBlock_V42(*content);
        }
        else
        {
            std::stringstream message;
            message << "This program only supports DAQ version 4.2+. There should be no test beam data taken "
                    << "with previous versions of the DAQ. Detected DAQ version: "<< _DataVersion.first << std::endl;
            throw TAltroConverterVersionException("TAltroBlockFactory", message.str());
        }
        break;
    }
    case BLOCK_EOR:
    {
        if(_DataVersion.first >= 420)
        {
            result = new TEORBlock_V42(*content);
        }
        else
        {
            std::stringstream message;
            message << "This program only supports DAQ version 4.2+. There should be no test beam data taken "
                    << "with previous versions of the DAQ. Detected DAQ version: "<< _DataVersion.first << std::endl;
            throw TAltroConverterVersionException("TAltroBlockFactory", message.str());
        }
        break;
    }
    case BLOCK_POR:
    {
        if(_DataVersion.first >= 420)
        {
            result = new TPORBlock_V42(*content);
        }
        else
        {
            std::stringstream message;
            message << "This program only supports DAQ version 4.2+. There should be no test beam data taken "
                    << "with previous versions of the DAQ. Detected DAQ version: "<< _DataVersion.first << std::endl;
            throw TAltroConverterVersionException("TAltroBlockFactory", message.str());
        }
        break;
    }
    case BLOCK_COR:
    {
        if(_DataVersion.first >= 420)
        {
            result = new TCORBlock_V42(*content);
        }
        else
        {
            std::stringstream message;
            message << "This program only supports DAQ version 4.2+. There should be no test beam data taken "
                    << "with previous versions of the DAQ. Detected DAQ version: "<< _DataVersion.first << std::endl;
            throw TAltroConverterVersionException("TAltroBlockFactory", message.str());
        }
        break;
    }
    case BLOCK_EVENT:
    {
        if(_DataVersion.first >= 460)
        {
            result = new TEventBlock_V46(*content);
        }
        else if (_DataVersion.first >= 420)
        {
            result = new TEventBlock_V42(*content);
        }
        else
        {
            std::stringstream message;
            message << "This program only supports DAQ version 4.2+. There should be no test beam data taken "
                    << "with previous versions of the DAQ. Detected DAQ version: "<< _DataVersion.first << std::endl;
            throw TAltroConverterVersionException("TAltroBlockFactory", message.str());
        }
        doDelete = false;
        break;
    }
    case BLOCK_BOF:
    {
        if(_DataVersion.first >= 420)
        {
            result = new TBOFBlock_V42(*content);
        }
        else
        {
            std::stringstream message;
            message << "This program only supports DAQ version 4.2+. There should be no test beam data taken "
                    << "with previous versions of the DAQ. Detected DAQ version: "<< _DataVersion.first << std::endl;
            throw TAltroConverterVersionException("TAltroBlockFactory", message.str());
        }
        break;
    }
    case BLOCK_EOF:
    {
        if(_DataVersion.first >= 420)
        {
            result = new TEOFBlock_V42(*content);
        }
        else
        {
            std::stringstream message;
            message << "This program only supports DAQ version 4.2+. There should be no test beam data taken "
                    << "with previous versions of the DAQ. Detected DAQ version: "<< _DataVersion.first << std::endl;
            throw TAltroConverterVersionException("TAltroBlockFactory", message.str());
        }
        break;
    }
    case BLOCK_RCU:
    {
        std::stringstream message;
        message << "The RCU Block should only be opened by the Event block" << std::endl;
        throw TAltroConverterDecodeException("TAltroBlockFactory", message.str());
        break;
    }
    default:
    {
        std::stringstream message;
        message << "Unrecognized block."<< std::endl;
        throw TAltroConverterDecodeException("TAltroBlockFactory", message.str());
        break;
    }
    }

    if(doDelete)
    {
        delete content;
    }
    return result;
}

bool
TAltroBlockFactory::isSOR(TIAltroBlock* Block) const
{
    return (Block->GetIdentifier() == BLOCK_SOR);
}

bool
TAltroBlockFactory::isEvent(TIAltroBlock* Block) const
{
    return (Block->GetIdentifier() == BLOCK_EVENT);
}

bool
TAltroBlockFactory::isEOR(TIAltroBlock* Block) const
{
    return (Block->GetIdentifier() == BLOCK_EOR);
}

bool
TAltroBlockFactory::isBOF(TIAltroBlock* Block) const
{
    return (Block->GetIdentifier() == BLOCK_BOF);
}

bool
TAltroBlockFactory::isEOF(TIAltroBlock* Block) const
{
    return (Block->GetIdentifier() == BLOCK_EOF);
}

bool
TAltroBlockFactory::isPOR(TIAltroBlock* Block) const
{
    return (Block->GetIdentifier() == BLOCK_POR);
}

bool
TAltroBlockFactory::isCOR(TIAltroBlock* Block) const
{
    return (Block->GetIdentifier() == BLOCK_COR);

}

bool
TAltroBlockFactory::isRCU(TIAltroBlock* Block) const
{
    return (Block->GetIdentifier() == BLOCK_RCU);
}

}
