/*
 * TRCUBlock_V42.cpp
 *
 *  Created on: Jul 11, 2012
 *      Author: caiazza
 */

#include "TIRCUBlock.hh"
#include "DataFormatDecoder.hh"
#include <string>
#include <sstream>

//MarlinTPC
#include "GenericADCElectronicsParameters.h"
#include "ExtendedAltroParameters.h"

namespace marlintpc
{

TRCUBlock_V42::TRCUBlock_V42(std::vector<uint32_t>& Content, const std::vector<uint32_t>::iterator beginRCU ):
        _content(&Content),
        _BlockStart(beginRCU)
{
    if(!CheckIntegrity())
    {
        std::stringstream message;
        message << "Failed integrity check while analyzing the RCU Block" << std::endl;
        if(_content != 0)
        {
            message << "Dumping the block Content" << std::hex << std::endl;
            std::vector<uint32_t>::iterator itContent = _content->begin();
            for(; itContent != _content->end(); ++itContent)
            {
                message << *itContent << std::endl;
            }
            message << std::dec;
        }
        throw TAltroConverterDecodeException("TRCUBlock_V42", message.str());
    }
}

TRCUBlock_V42::~TRCUBlock_V42()
{

}

uint32_t
TRCUBlock_V42::GetLength() const
{
    return (*_BlockStart) + 1;
}

bool
TRCUBlock_V42::CheckIntegrity() const
{
    bool result = true;
    result = result && (_content != 0)
            && (*(_BlockStart+2) == 0xffffffff);

    return result;
}

const std::vector<uint32_t>&
TRCUBlock_V42::Content() const
{
    return (*_content);
}

uint32_t
TRCUBlock_V42::GetIdentifier() const
{
    return *(_BlockStart+2);
}

void
TRCUBlock_V42::SetAdditionalParameters(lcio::LCRunHeader* RunHeader)
{

}

void
TRCUBlock_V42::AddData(lcio::LCCollection* Collection) const
{
    uint32_t RCUID = *(_BlockStart+1);
    std::vector<uint32_t>::iterator BlockEnd = GetBlockEnd();
    uint32_t TrailerLength = *(BlockEnd - 1) & TRAILERLENGTHMASK;
    uint32_t Payload = *(BlockEnd - TrailerLength) & WORDVALUEMASK;

    if(Payload != 0)
    {
        std::vector<uint32_t>::iterator DataStart = _BlockStart+10;
        DataFormatDecoder_V1 Decoder(DataStart, RCUID);
        Decoder.AddData(Collection);
    }
}

void
TRCUBlock_V42::AddGeneralElectronicInfo(lcio::LCCollection* Collection) const
{
    int RCUID = static_cast<int>(*(_BlockStart+1));
    std::vector<uint32_t>::iterator BlockEnd = GetBlockEnd();
    uint32_t TrailerLength = *(BlockEnd - 1) & TRAILERLENGTHMASK;
    std::vector<uint32_t>::iterator TrailerStart = BlockEnd - TrailerLength;

    //Decoding the polarity. The RCU register stores whether the ADC value were inverted.
    //The signal is actually negative so that the inversion makes it positive, and vice versa.
    int polarity = static_cast<int>((*(TrailerStart + 6) & POLARITYMASK) >> 4);
    if(polarity == 1)
    {
        polarity = -1;
    }
    else if (polarity == 0)
    {
        polarity = 1;
    }

    int MaxADCValue = 1023;

    uint32_t FrequencyCode = (*(TrailerStart + 7) & FREQUENCYCODEMASK) >> 5;
    float Frequency;
    switch(FrequencyCode)
    {
    case 0:
    {
        Frequency = 20.0;
        break;
    }
    case 1:
    {
        Frequency = 10.0;
        break;
    }
    case 2:
    {
        Frequency = 5.0;
        break;
    }
    default:
    {
        std::stringstream message;
        message << "Unrecognized frequency code in the RCU Data" << std::endl;
        message << "The decoded value is " << FrequencyCode << std::endl;
        message << "The Data word is " << std::hex << *(TrailerStart + 7) << std::endl;
        throw TAltroConverterDecodeException("TRCUBlock_V42", message.str());
    }
    }

    bool PedestalSubtraction = true;
    bool ZeroSuppression = true;

    tpcconddata::GenericADCElectronicsParameters* theConditions =
            new tpcconddata::GenericADCElectronicsParameters(
                    RCUID, polarity, MaxADCValue, Frequency, PedestalSubtraction, ZeroSuppression);
    Collection->addElement(theConditions);
}

void
TRCUBlock_V42::AddExtendedElectronicsInfo(lcio::LCCollection* Collection) const
{
    std::vector<uint32_t>::iterator BlockEnd = GetBlockEnd();
    uint32_t TrailerLength = *(BlockEnd - 1) & TRAILERLENGTHMASK;
    std::vector<uint32_t>::iterator TrailerStart = BlockEnd - TrailerLength;

    tpcconddata::ExtendedAltroParameters::ExtendedAltroData Data;
    Data.ZSGlitchFilter = (*(TrailerStart + 6) & GLITCHFILTERMASK) >> 12;
    Data.PreSamples = (*(TrailerStart + 6) & PRESAMPLESMASK) >> 17;
    Data.PostSamples = (*(TrailerStart + 6) & POSTSAMPLESMASK) >> 14;
    Data.PreTriggerSamples = (*(TrailerStart + 7) & PRETRIGGERSAMPLESMASK) >> 20;
    Data.RCUAddress = (*(TrailerStart + 8) & RCUADDRESSMASK) >> 7;
    Data.RCUFWVersion = (*(TrailerStart + 8) & RCUFWVERMASK) >> 16;
    Data.RCUActiveFECA = *(TrailerStart + 4) & WORDVALUEMASK;
    Data.RCUActiveFECB = *(TrailerStart + 5) & WORDVALUEMASK;
    Data.RCUConfigRegister1 = *(TrailerStart + 6) & WORDVALUEMASK;
    Data.RCUConfigRegister2 = *(TrailerStart + 7) & WORDVALUEMASK;
    Data.RCUStatusErrors = (*(_BlockStart + 6) & RCUERRORSTATUSMASK) >> 12;

    tpcconddata::ExtendedAltroParameters* theConditions =
            new tpcconddata::ExtendedAltroParameters(Data);
    Collection->addElement(theConditions);
}

std::vector<uint32_t>::iterator
TRCUBlock_V42::GetBlockEnd() const
{
    return (_BlockStart+GetLength());
}

}


