/*
 * TSORBlock_V42.cpp
 *
 *  Created on: Jul 10, 2012
 *      Author: caiazza
 */

#include "TISORBlock.hh"
#include <map>
#include <string>
#include <sstream>

#include "IMPL/LCRunHeaderImpl.h"

namespace marlintpc
{

TSORBlock_V42::TSORBlock_V42(std::vector<uint32_t>& Content):
        _content(Content)
{
    if(!CheckIntegrity())
    {
        std::stringstream message;
        message << "Failed integrity check while analyzing the Start of Run Block" << std::endl;
        message << "Dumping the block Content" << std::hex << std::endl;
        std::vector<uint32_t>::iterator itContent = _content.begin();
        for(; itContent != _content.end(); ++itContent)
        {
            message << *itContent << std::endl;
        }
        message << std::dec;
        throw TAltroConverterDecodeException("TSORBlock_V42", message.str());
    }
    _runNumber = _content.at(4);

    int day = _content[5] & 0x000000ff;
    int month = (_content[5] & 0x0000ff00) >> 8;
    int year = (_content[5] & 0xffff0000) >> 16;
    int second = _content[6] & 0x000000ff;
    int minute = (_content[6] & 0x0000ff00) >> 8;
    int hour = (_content[6] & 0xffff0000) >> 16;
    _start = UTIL::LCTime(year, month, day, hour, minute, second);

    _PCA16DAC = (_content[7] & 0x0000ffff);

    _registers._preampEnable = (_content[7] & 0x00010000) >> 16;

    _registers._gain = gainSettings((_content[7] & 0x00060000) >> 17);

    _registers._shaper = shaperSettings(
            (_content[7] & 0x00380000) >> 19);

    _registers._shutdown = (_content[7] & 0x00400000) >> 22;

    _registers._polarity = (_content[7] & 0x00800000) >> 23;

    _runSets._zeroSuppression = (_content[8] & 0x00000001);

    _runSets._pedestalSubtraction = (_content[8] & 0x00000002) >> 1;

    _EUDAQRunNumber = _content[9];
}

TSORBlock_V42::~TSORBlock_V42()
{

}

uint32_t
TSORBlock_V42::GetLength() const
{
    return _content.size();
}

bool
TSORBlock_V42::CheckIntegrity() const
{
    bool result = true;
    result = result && (GetLength() == _content.at(0) + 1)
                     && (GetLength() >= 10)
                     && (_content.at(2) == 0x11111111)
                     && ((_content.at(3)& 0x0000ffff) >= 420 )
                     && (((_content.at(3)& 0xffff0000)>>16) == 1 );

    return result;
}

const std::vector<uint32_t>&
TSORBlock_V42::Content() const
{
    return _content;
}

uint32_t
TSORBlock_V42::GetIdentifier() const
{
    return _content.at(2);
}

lcio::LCRunHeader*
TSORBlock_V42::CreateRunHeader() const
{
    IMPL::LCRunHeaderImpl* runHeader = new IMPL::LCRunHeaderImpl();
    runHeader->setRunNumber(GetRunNumber());
    runHeader->setDetectorName("LPTPC");
    runHeader->setDescription("Test beam run");

    std::map<gainSettings, std::string> PCA16GainDescription;
    PCA16GainDescription[MV12] = "12 mV/fC";
    PCA16GainDescription[MV15] = "15 mV/fC";
    PCA16GainDescription[MV19] = "19 mV/fC";
    PCA16GainDescription[MV27] = "27 mV/fC";

    std::map<shaperSettings, std::string> PCA16ShaperDescription;
    PCA16ShaperDescription[NS30] = "30 ns";
    PCA16ShaperDescription[NS60] = "60 ns";
    PCA16ShaperDescription[NS90] = "90 ns";
    PCA16ShaperDescription[NS120] = "120 ns";

    runHeader->parameters().setValue("RUN Start Date", GetRunStart().getDateString());
    runHeader->parameters().setValue("RUN Start Timestamp", GetRunStart().unixTime());
    runHeader->parameters().setValue("DAQ Version", static_cast<int>(_content.at(3)& 0x0000ffff));
    runHeader->parameters().setValue("ALTRO Data Format", static_cast<int>((_content.at(3)& 0xffff0000) >> 16) );
    runHeader->parameters().setValue("PCA16DAC", static_cast<int>(_PCA16DAC));
    runHeader->parameters().setValue("preampEnable", _registers._preampEnable);
    runHeader->parameters().setValue("Gain", _registers._gain);
    runHeader->parameters().setValue("Gain Description", PCA16GainDescription[_registers._gain]);
    runHeader->parameters().setValue("Shaper", _registers._shaper);
    runHeader->parameters().setValue("Shaping Time Description", PCA16ShaperDescription[_registers._shaper]);
    runHeader->parameters().setValue("polarity", _registers._polarity);
    runHeader->parameters().setValue("shutdown", _registers._shutdown);
    runHeader->parameters().setValue("Pedestal Subtraction", _runSets._pedestalSubtraction);
    runHeader->parameters().setValue("Zero Suppression", _runSets._zeroSuppression);
    runHeader->parameters().setValue("EUDAQ Run Number", static_cast<int>(_EUDAQRunNumber));

    return runHeader;
}

uint32_t
TSORBlock_V42::GetRunNumber() const
{
    return _runNumber;
}

lcio::LCTime
TSORBlock_V42::GetRunStart() const
{
    return _start;
}

void
TSORBlock_V42::SetAdditionalParameters(lcio::LCRunHeader* RunHeader)
{

}

} //End namespace


