#ifndef GLOBALFIELDPROCESSOR_H
#define GLOBALFIELDPROCESSOR_H 1

#include <string>
#include <map>

// CLHEP
#include <CLHEP/Matrix/Vector.h>
#include <CLHEP/Vector/ThreeVector.h>

// LCCD
#include <lccd/IConditionsChangeListener.hh>

// Marlin
#include "Field.h"
#include "MagneticField.h"
#include "marlin/Processor.h"

namespace EVENT {

	class LCCollection;
	class LCEvent;
	class LCGenericObject;
	class LCRunHeader;
}

namespace marlintpc {

class MagneticField;

/**
 * GlobalFieldProcessor : This processor manages the fields used in MarlinTPC
 * 
 *    @author J. Abernathy (UVic)<br>
 */
class GlobalFieldProcessor: public marlin::Processor {

public:

	/** constructor
	 */
	GlobalFieldProcessor();

	/** destructor
	 */
	virtual ~GlobalFieldProcessor();

	/** used by Marlin to create a new processor
	 */
	virtual Processor* newProcessor();

	/**
	 * Called at the begin of the job before anything is read.
	 * Use to initialize the processor, e.g. book histograms.
	 */
	virtual void init();

	/** Called for every run.
	 */
	virtual void processRunHeader( EVENT::LCRunHeader* run );

	/**
	 * Called for every event - the working horse.
	 */
	virtual void processEvent( EVENT::LCEvent * evt );

	/**
	 * Called after the event has been processed
	 */
	virtual void check( EVENT::LCEvent * evt );

	/**
	 * Called after data processing for clean up.
	 */
	virtual void end();

	/** registers a field with the field manager
	 */
	bool register_field( Field* );

	/** Returns the magnetic field (in T) at a given point and time in the detector
	 * @deprecated replaced by get_global_magnetic_field(const CLHEP::Hep3Vector&)
	 */
	CLHEP::HepVector get_global_magnetic_field( const CLHEP::HepVector& x ) const;

	/** Returns the magnetic field (in T) at a given point and time in the detector
	 */
	CLHEP::Hep3Vector get_global_magnetic_field( const CLHEP::Hep3Vector& x ) const;

	/** Returns the combined symmetric field errors (in T) of measured magnetic field maps
	 */
	CLHEP::Hep3Vector get_global_magnetic_field_error( ) const;

	/** Returns the electric field (in V) at a given point and time in the detector
	 * @deprecated replaced by get_global_electric_field(const CLHEP::Hep3Vector&)
	 */
	CLHEP::HepVector get_global_electric_field( const CLHEP::HepVector& x ) const;

	/** Returns the electric field (in V) at a given point and time in the detector
	 */
	CLHEP::Hep3Vector get_global_electric_field( const CLHEP::Hep3Vector& x ) const;

	/** Returns the combined symmetric field errors (in T) of measured electric field maps
	 */
	CLHEP::Hep3Vector get_global_electric_field_error( ) const;

	/** return the instance of the global field processor
	 */
	static GlobalFieldProcessor &get_instance();

	/** returns the main crossing angle of the ILC model
	 */
	double get_main_crossing_angle() const;

protected:

	class FieldNotifier: public lccd::IConditionsChangeListener {

	public:

		FieldNotifier( Field& field );

		/** from IConditionsChangeListener. this function is called whenever a condition has changed in this collection
		*/
		virtual void conditionsChanged( lcio::LCCollection* col );

		private:

			Field &_field;
	};

	class CoilPositionNotifier: public lccd::IConditionsChangeListener {

	public:

		CoilPositionNotifier( std::vector<Field*> _active_magnetic_fields );

		/** from IConditionsChangeListener. this function is called whenever a condition has changed in this collection
		*/
		virtual void conditionsChanged( lcio::LCCollection* col );

	private:

		std::vector<MagneticField*> _magnetic_fields;

	};

private:
	
	/** Controls the source of the magnetic field. If this is set to true then it overrides any other magnetic field information.
	 */
	bool _use_gear_bfield;

	#ifdef USE_CGA
	/** Controls the source of the electric and magnetic fields. If this is set to true then CGA is used and it overrides any other field information.
	 */
	bool _use_cga_fields;
	#endif // USE_CGA

	/** an array of pairs of strings of the format: (field type, lcio collection)
	 */
	std::vector<std::string> _input_magnetic_fields;

	/** an array of pairs of strings of the format: (field type, lcio collection)
	 */
	std::vector<std::string> _input_electric_fields;

	/** the singleton instance of the processor
	 */
	static GlobalFieldProcessor *_instance;

	/** the collection of fields mapped by their typename
	 */
	std::map<std::string, Field*> _field_by_typename;

	/** the collection of active magnetic fields
	 */
	std::vector<Field*> _active_magnetic_fields;

	/** the collection of magnetic fields mapped by their typename
	 */
	std::vector<Field*> _active_electric_fields;

	/** the collections of FieldNotifiers which tell the fields when they have been changed in LCCD
	 */
	std::vector<FieldNotifier*> _field_notifiers;

	/** the main crossing angle of the collider
	 */
	double _main_crossing_angle;

	/** the name of the CoilPosition collection
	 */
	std::string _coil_position_collection;

	/** the CoilPositionNotifier which tells the magnetic fields the relative position of the coil has changed
	 */
	CoilPositionNotifier* _coil_position_notifier;

};

} // namespace marlintpc

#endif // GLOBALFIELDPROCESSOR_H
