/*
 * GenericADCElectronicsParameters2LCCD.cpp
 *
 *  Created on: Nov 1, 2011
 *      Author: caiazza
 */
/**
 * This tool is to be used to insert Generic ADC electronics data in the LCCD database
 * or in an slcio file for offline use.
 * It will take its input from a text file listing the validity time and the values of
 * the condition data to insert.
 * This tool is specific for the GenericADCElectronicsParameters object as defined in revision 2814
 */

//standard includes
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <vector>
#include <stdexcept>
#define NDEBUG
#include "assert.h"

//LCCD
#include "lccd.h"
#include "lccd/DBInterface.hh"

//LCIO
#include "lcio.h"
#include "Exceptions.h"
#include "IO/LCWriter.h"
#include "IMPL/LCEventImpl.h"
#include "IMPL/LCCollectionVec.h"
#include "UTIL/LCTime.h"
#include "IMPL/LCTOOLS.h"

//Marlin

//MarlinTPC
#include "GenericADCElectronicsParameters.h"

//Structure containing the elements to fill in the db or file to be read from the different lines of the input file
struct TPCCondrecord
{
    tpcconddata::GenericADCElectronicsParameters* theConditions;
    lcio::LCTime validSince;
    lcio::LCTime validTill;
};

void printUsage()
{
    std::cout<<"Usage:"<<std::endl;
    std::cout<<"To write data to a lcio simple file:"<<std::endl;
    std::cout<<"  GenericADCElectronicsParameters2LCCD.exe -S outfile \\"<<std::endl;
    std::cout<<"  conditionfile.txt [conditionfile.txt [conditionfile.txt ... ]]"<<std::endl;
    std::cout<<"To write data to a database:"<<std::endl;
    std::cout<<"  GenericADCElectronicsParameters2LCCD.exe -D connectionString folder \\"<<std::endl;
    std::cout<<"  conditionfile.txt [conditionfile.txt [conditionfile.txt ... ]]"<<std::endl;
    std::cout<<"Formats:"<<std::endl;
    std::cout<<" Text file: validSinceDate validSinceTime validTillDate validTillTime readoutGroupID polarity maxADCValue frequency pedsub_onoff zerosuppr_onoff" << std::endl;
    std::cout<<" Date format: YYYY-MM-DD" << std::endl;
    std::cout<<" Time Format: HH:MM:SS where the number of seconds can be a decimal value" << std::endl;
    std::cout<<" Connection String: hostname:database:login:password[:port]"<<std::endl;
    std::cout<<" folder: /database/folder"<<std::endl;
}

//Function that converts the date and time strings from the input file in an LCTime object from which to obtain the timestamp
lcio::LCTime convert2LCTime(std::string Date, std::string Time)
{
    std::vector<std::string> tokens;
    int Year;
    int Month;
    int Day;
    int Hour;
    int Minute;
    int Second;
    EVENT::long64 NanoSecond;
    lcio::LCTime timestamp;

    if(Date.size() != 10)
    {
        std::string error = std::string("The date string is not correctly formatted");
        std::invalid_argument e(error);
        throw  e;
    }
    std::stringstream datestream(Date);
    std::string token;
    while(std::getline(datestream, token, '-'))
    {
        tokens.push_back(token);
    }

    if(Time.size() < 8)
    {
        std::string error = std::string("The time string is not correctly formatted");
        std::invalid_argument e(error);
        throw  e;
    }

    std::stringstream timestream(Time);
    while(std::getline(timestream, token, ':'))
    {
        tokens.push_back(token);
    }

    std::stringstream convStream;
    double tempSec;
    convStream << tokens[0] << " " << tokens[1] << " " << tokens[2] << " " << tokens[3] << " " << tokens[4] << " " << tokens[5];
    convStream >> Year >> Month >> Day >> Hour >> Minute >> tempSec;
    Second = static_cast<int>(tempSec);
    NanoSecond = static_cast<EVENT::long64>((tempSec - static_cast<EVENT::long64>(tempSec)) * 10E9);

    timestamp=LCTime(Year,Month,Day,Hour,Minute,Second);
    timestamp+= NanoSecond;
    return timestamp;
}

int main(int argc, char **argv)
{
    //Program variable declaration
    //Command line options
    bool writeSimpleFile=false;
    bool write2DB=false;
    std::string dbInitString;
    std::string dbFolder;
    std::string outFileName;
    std::vector<std::string> inFileNames;
    int  optionCounter=0;

    // Other variables
    lccd::DBInterface* dbConnection = 0;
    lcio::LCCollectionVec* conditionsColl = 0;
    const std::string COLLNAME = std::string("GenericADCElectronicsParameters");
    lcio::LCWriter* lcWrt = 0;
    lcio::LCEventImpl*  evt = 0;

    //Vector containing the elements to fill in the db or file
    std::vector<TPCCondrecord> recordVector;

    //Manage Command Line and Fill variables
    //Check minimum number of variables
    if ( argc < 2 )
    {
        printUsage();
        return(1);
    }


    //Check that the second argument is a valid option
    std::string option = argv[1];
    if ( (option == "-S") && (argc > 3))    //If option -S (simple file)
    {
        writeSimpleFile=true;    //Set write file to true
        outFileName=argv[2];
        optionCounter=3;
        try
        {
            std::cout << "Creating the event writer to write the file" << std::endl;
            lcWrt = lcio::LCFactory::getInstance()->createLCWriter() ;    //Create LCWriter
            lcWrt->open( outFileName, lcio::LCIO::WRITE_NEW ) ;
            evt = new lcio::LCEventImpl() ;
        }
        catch(lcio::Exception &e)
        {
            std::cerr << "GenericADCElectronicsParameters2LCCD: ERROR while creating the LCWriter and the event objects" <<std::endl;
            std::cerr << e.what() << std::endl;
            if(lcWrt != 0)
            {
                lcWrt->close();
                delete lcWrt;
                lcWrt = 0;
            }
            if(evt != 0)
            {
                delete evt;
                evt = 0;
            }
            return 1;
        }assert(lcWrt != 0);
    }
    else if ( option == "-D" && argc > 4)  //If option -D (database)
    {
        write2DB=true;    //Set write 2 db to true
        //Fill the DB variables
        dbInitString=argv[2];
        dbFolder=argv[3];
        //Create DB Connection
        try
        {
            dbConnection = new lccd::DBInterface(dbInitString, dbFolder, write2DB);
        }
        catch(std::exception &e)
        {
            std::cerr << "GenericADCElectronicsParameters2LCCD: ERROR while creating the DB connection" << std::endl;
            std::cerr << e.what() << std::endl;
            if(dbConnection != 0)
            {
                delete dbConnection;
                dbConnection = 0;
            }
            return 1;
        } assert(dbConnection != 0);

        optionCounter=4;
    }
    else
    {
        printUsage();
        return(1);
    } assert(dbConnection != 0 || lcWrt != 0);
    assert(writeSimpleFile ^ write2DB);


    //Check that there is at least on input file
    if(argc >= optionCounter)
    {
        //Insert the filenames in the vector
        for ( int i=optionCounter; i<argc ; ++i)
        {
            inFileNames.push_back(argv[i]);
        }
    }
    else
    {
        printUsage();
        if(dbConnection != 0)
        {
            delete dbConnection;
            dbConnection = 0;
        }
        if(lcWrt != 0)
        {
            lcWrt->close();
            delete lcWrt;
            lcWrt = 0;
        }
        if(evt != 0)
        {
            delete evt;
            evt = 0;
        }
        return(1);
    }assert(inFileNames.size() != 0);

    int fileNum=1;
    // loop over the input files
    for (std::vector<std::string>::iterator inFile = inFileNames.begin(); inFile != inFileNames.end(); ++inFile)
    {
        //Open the input file and connects it to a read buffer
        std::filebuf infile;
        infile.open(inFile->c_str(),std::ios::in);
        std::istream inputstream(&infile);
        std::cout<<" Reading file #"<<fileNum<<" : "<<*inFile<<std::endl;
        fileNum++;
        int linenumber = -1;

        //Read all the lines of the file until the end
        while(!inputstream.eof())
        {
            //Take a line from the stream and feeds it to a string
            std::string line;
            line.clear();
            std::getline(inputstream, line);
            linenumber++;

            // check if the next line is a comment or if it's an empty line
            if (*line.begin() != '#' && line.size() != 0)
            {
                // associate the line to a stream so we can stream into the variables
                std::stringstream linestream(line);
                //TPCCondition variables with default values
                tpcconddata::GenericADCElectronicsParameters* theConditions = 0;
                lcio::LCTime validSince;
                lcio::LCTime validTill;
                std::string validSinceDate;
                std::string validSinceTime;
                std::string validTillDate;
                std::string validTillTime;
                int readoutGroupID;
                int polarity;
                int maxADCValue;
                float frequency;
                bool pedsub_onoff;
                bool zerosuppr_onoff;

                linestream >> validSinceDate >> validSinceTime >> validTillDate >> validTillTime
                    >> readoutGroupID >> polarity >> maxADCValue >> frequency
                    >> pedsub_onoff >> zerosuppr_onoff;
                // test if everything went fine
                if (linestream.fail())
                {
                    std::cout << "GenericADCElectronicsParameters2LCCD: ERROR reading input file at line " << linenumber << " : " <<line << std::endl;

                    if(dbConnection != 0)
                    {
                        delete dbConnection;
                        dbConnection = 0;
                    }
                    if(lcWrt != 0)
                    {
                        lcWrt->close();
                        delete lcWrt;
                        lcWrt = 0;
                    }
                    if(evt != 0)
                    {
                        delete evt;
                        evt = 0;
                    }
                    return 1;
                }

                //Create General ADC Electronics Object Object
                try
                {
                    //Create the condition object
                    theConditions = new tpcconddata::GenericADCElectronicsParameters(readoutGroupID, polarity, maxADCValue, frequency, pedsub_onoff, zerosuppr_onoff);

                    validSince = convert2LCTime(validSinceDate, validSinceTime);
                    validTill = convert2LCTime(validTillDate, validTillTime);
                    std::cout << "Valid From: " << validSince.timeStamp() << " to " << validTill.timeStamp() << std::endl;
                    TPCCondrecord thisRecord = {theConditions, validSince, validTill};
                    recordVector.push_back(thisRecord);
                }
                catch(std::exception &e)
                {
                    std::cerr << "GenericADCElectronicsParameters2LCCD: ERROR while creating the LCIO objects" <<std::endl;
                    std::cerr << e.what() << std::endl;
                    if(dbConnection != 0)
                    {
                        delete dbConnection;
                        dbConnection = 0;
                    }
                    if(lcWrt != 0)
                    {
                        lcWrt->close();
                        delete lcWrt;
                        lcWrt = 0;
                    }
                    if(evt != 0)
                    {
                        delete evt;
                        evt = 0;
                    }
                    if(theConditions != 0)
                    {
                        delete theConditions;
                        theConditions = 0;
                    }
                    if(recordVector.size() != 0)//Cleanup the vector
                    {
                        for(std::vector<TPCCondrecord>::iterator it = recordVector.begin(); it != recordVector.end(); it++)
                        {
                            delete it->theConditions;
                        }
                    }
                    return 1;
                }assert(theConditions != 0); assert(recordVector.size() != 0);
            }
            else
            {
                std::cout << "  DEBUG: skipping comment line " << linenumber << " : "  << line <<  std::endl;
            }
        }
        //Close the input file
        infile.close();
    }


    //If write to file I create a single collection containing multiple TPC Condition objects
    if(writeSimpleFile)
    try
    {
        //Create collection
        conditionsColl = new lcio::LCCollectionVec(lcio::LCIO::LCGENERICOBJECT);
        //Fill the collection
        for(std::vector<TPCCondrecord>::iterator it = recordVector.begin(); it != recordVector.end(); it++)
        {
            conditionsColl->addElement(it->theConditions);
            it->theConditions = 0;
        }
        //Add the collection to the event
        evt->addCollection(conditionsColl,COLLNAME);
        //Dump the event to screen
        lcio::LCTOOLS::dumpEvent( evt ) ;
        // write event to file
        lcWrt->writeEvent( evt ) ;
    }
    catch(lcio::Exception & e)
    {
        std::cerr << "GenericADCElectronicsParameters2LCCD: ERROR while creating the LCIO collection" <<std::endl;
        std::cerr << e.what() << std::endl;
        if(lcWrt != 0)
        {
            delete lcWrt;
            lcWrt = 0;
        }
        if(evt != 0)
        {
            delete evt;
            evt = 0;
        }
        if(recordVector.size() != 0)//Cleanup the vector
        {
            for(std::vector<TPCCondrecord>::iterator it = recordVector.begin(); it != recordVector.end(); it++)
            {
                if(it->theConditions != 0) delete it->theConditions;
            }
        }
        return 1;
    }
    else if (write2DB)
    {
        try
        {
            //Create one collection containing one element for each vector element and store it between the given timestamps
            for(std::vector<TPCCondrecord>::iterator it = recordVector.begin(); it != recordVector.end(); it++)
            {
                //TODO Add an output for each line written to the DB
                conditionsColl = new lcio::LCCollectionVec(lcio::LCIO::LCGENERICOBJECT);
                conditionsColl->addElement(it->theConditions);
                dbConnection->storeCollection(it->validSince.timeStamp(), it->validTill.timeStamp(), conditionsColl, COLLNAME);
            }
        }
        catch(std::exception &e)
        {
            std::cerr << "GenericADCElectronicsParameters2LCCD: ERROR while storing the condition object in the DB" <<std::endl;
            std::cerr << e.what() << std::endl;
            if(dbConnection != 0)
            {
                delete dbConnection;
                dbConnection = 0;
            }
            if(recordVector.size() != 0)//Cleanup the vector
            {
                for(std::vector<TPCCondrecord>::iterator it = recordVector.begin(); it != recordVector.end(); it++)
                {
                    if(it->theConditions != 0) delete it->theConditions;
                }
            }
            if(conditionsColl != 0)
            {
                delete conditionsColl;
                conditionsColl = 0;
            }
            return 1;
        }
    }

    //Close DB and clean up
    if(dbConnection != 0)
    {
        delete dbConnection;
        dbConnection = 0;
    }
    if(lcWrt != 0)
    {
        lcWrt->close();
    }
    if(evt != 0)
    {
        delete evt;
    }

    return 0;
}//End Main
