// STL
#include <vector>

//standard stuff
#include <iostream>
#include <fstream>

// GEAR
#include <gear/GEAR.h>
#include <gear/TPCParameters.h>
#include <gear/PadRowLayout2D.h>
#include <gearimpl/Vectors.h>
#include <gear/TPCModule.h>

// Marlin
#include <marlin/Global.h>
#include "PadGeometryChecker.h"

using namespace std;
using namespace lcio;
using namespace marlin;

PadGeometryChecker aPadGeometryChecker;

PadGeometryChecker::PadGeometryChecker(): Processor("PadGeometryChecker") 
{
  _description = "Dumps the pad geometry input in a human readable form to the console.";

  registerProcessorParameter( "OutfileName" ,
			      "Name of the output file",
			      _outfileName,
			      string("padGeometryCheck.txt")
			      );

}


PadGeometryChecker::~PadGeometryChecker() 
{
}


void PadGeometryChecker::init() 
{
  printParameters();

  try 
    {
      // the TPC stuff
      const gear::TPCParameters& tpc_parameters = Global::GEAR->getTPCParameters();
      
      const double max_drift_length = tpc_parameters.getMaxDriftLength();
      cout << "[PadGeometryChecker] Max drift length is " << max_drift_length << endl;
      
      switch (tpc_parameters.getCoordinateType())
	{
	case gear::PadRowLayout2D::POLAR : 
	  cout << "[PadGeometryChecker] Pad coordinate type is POLAR" << endl;
	  break;
	case gear::PadRowLayout2D::CARTESIAN : 
	  cout << "[PadGeometryChecker] Pad coordinate type is CARTESIAN" << endl;
	  break;
	default:
	  throw Exception("[PadGeometryChecker]: unknown geometry type in TPC layout");
	}
      // end TPC
      
   
      // the MODULES
      const vector<gear::TPCModule*> &modules = tpc_parameters.getModules();

      //open the output stream, this is unprotected
      ofstream theOutfile;
      theOutfile.open(_outfileName.c_str());

      for (vector<gear::TPCModule*>::const_iterator module = modules.begin(); module < modules.end(); module++)
        {
	  
	  int moduleID = (**module).getModuleID();
	  theOutfile <<  "[MODULE] on module with ID " << moduleID << " of a total of " << modules.size() << " modules" << endl;

	  // check shape
	  if ( (*module)->getPadShape() != gear::PadRowLayout2D::RECTANGLE)
	    {
	      throw Exception("[PadGeometryChecker] Wrong type of pad shape" );
	    }
	  
	  const int numberOfRowsInModule = (*module)->getNRows();
	  theOutfile << "[MODULE] There are " << numberOfRowsInModule << " rows on module ID " << moduleID << endl;
	  const vector<double>& thePadPlane = (*module)->getPlaneExtent();

	  if( thePadPlane.size() == 4)
	    {
	      theOutfile << "[MODULE] the active readout area of module ID " << moduleID << " has the boundaries"
			 << " xMin = " << thePadPlane.at(0) 
			 << " xMax = " << thePadPlane.at(1) 
			 << " yMin = " << thePadPlane.at(2) 
			 << " yMax = " << thePadPlane.at(3) 
			 << endl;		
	    }

	  // go through rows and pads
	  // first the rows ....
	  for (int rowNumber = 0; rowNumber < numberOfRowsInModule; rowNumber++)
            {
	      const vector<int>& padsInRow = (*module)->getPadsInRow (rowNumber);
	      theOutfile << " [ROW] in row " << rowNumber << " there are " << padsInRow.size() << " pads:" << endl;
	      theOutfile << "  [PADINDICES]";

	      // ... then the pads
	      for (vector<int>::const_iterator itPad = padsInRow.begin(); itPad < padsInRow.end(); itPad++)
                {
		  const int padNumber = (*module)->getPadNumber (*itPad);
		  const int padIndex  = (*module)->getPadIndex (rowNumber, padNumber);
		  theOutfile << " [i:" << padIndex << "]";
		  // 		  const double padWidth = (*module)->getPadWidth(padIndex);
		  // 		  const double padPitch = (*module)->getPadPitch(padIndex);
		} // end pad loop
	      theOutfile << endl;
	      
	      theOutfile << "  [PADCENTERS]";
	      for (vector<int>::const_iterator itPad = padsInRow.begin(); itPad < padsInRow.end(); itPad++)
                {
		  const int padNumber = (*module)->getPadNumber (*itPad);
		  const int padIndex  = (*module)->getPadIndex (rowNumber, padNumber);

		  const gear::Vector2D padCenterVector = (*module)->getPadCenter( padIndex);
		  theOutfile << " [" << padCenterVector[0] << "/" << padCenterVector[1] << "]";
		  //<< " , pitch=" << padPitch << " and width=" << padWidth << endl;
		} // end pad loop
	      theOutfile << endl;
		  
	      
	    } // end row loop
	} // end module loop
    } // end TRY
    catch( ... ) 
      {
      }
}


void PadGeometryChecker::processRunHeader(LCRunHeader *run) 
{
}


void PadGeometryChecker::processEvent(LCEvent *evt) 
{
}



Processor* PadGeometryChecker::newProcessor() 
{
  return new PadGeometryChecker();
}


void PadGeometryChecker::check(LCEvent *evt) 
{
}


void PadGeometryChecker::end() 
{
}

