/*
 * TPCCondition2LCCD.cc
 *
 *  Created on: Oct 21, 2011
 *      Author: caiazza
 */

/**
 * This tool is to be used to insert TPC Conditions data in the LCCD database
 * or in an slcio file for offline use.
 * It will take its input from a text file listing the validity time and the values of
 * the condition data to insert.
 * This tool is specific for the TPCConditions object as defined in revision 2814
 *
 * USAGE:
 * To write data to a lcio simple file:
 * TPCConditions2LCCD.exe -S outfile conditionfile.txt [conditionfile.txt [conditionfile.txt ... ]]
 *
 * To write data to a database:
 * TPCConditions2LCCD.exe -D connectionString CondDBfolder conditionfile.txt [conditionfile.txt [conditionfile.txt ... ]]
 *
 * FORMATS:
 * Text file:
 * validSinceDate validSinceTime validTillDate validTillTime driftVel amplification LDiff LDefoc TDiff TDefoc ADC2PrimaryFactor
 * Date format: YYYY-MM-DD
 * Time Format: HH:MM:SS where the number of seconds can be a decimal value
 * Connection String: hostname:database:login:password[:port]
 * ConDBfolder: /database/folder
 */

/*
 * TODO Add -h option to print the usage information
 */

//standard inclusions
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <vector>
#include <stdexcept>
#define NDEBUG
#include "assert.h"

//LCCD
#include "lccd.h"
#include "lccd/DBInterface.hh"

//LCIO
#include "lcio.h"
#include "Exceptions.h"
#include "IO/LCWriter.h"
#include "IMPL/LCEventImpl.h"
#include "IMPL/LCCollectionVec.h"
#include "UTIL/LCTime.h"
#include "IMPL/LCTOOLS.h"

//MarlinTPC
#include "TPCConditions.h"

/*
 * Structure encapsulating the TPCCondition object and the LCTime object to separate the phase of
 * file reading and parsing from the writing phase
 */

struct TPCCondrecord
{
    tpcconddata::TPCConditions* theConditions;
    lcio::LCTime validSince;
    lcio::LCTime validTill;
};

/*
 * Prints the usage informations
 */
void printUsage()
{
    std::cout<<"Usage:"<<std::endl;
    std::cout<<"To write data to a lcio simple file:"<<std::endl;
    std::cout<<"  TPCConditions2LCCD.exe -S outfile \\"<<std::endl;
    std::cout<<"  conditionfile.txt [conditionfile.txt [conditionfile.txt ... ]]"<<std::endl;
    std::cout<<"To write data to a database:"<<std::endl;
    std::cout<<"  TPCConditions2LCCD.exe -D connectionString folder \\"<<std::endl;
    std::cout<<"  conditionfile.txt [conditionfile.txt [conditionfile.txt ... ]]"<<std::endl;
    std::cout<<"Formats:"<<std::endl;
    std::cout<<" Text file: validSinceDate validSinceTime validTillDate validTillTime driftVel amplification LDiff LDefoc TDiff TDefoc ADC2PrimaryFactor" << std::endl;
    std::cout<<" Date format: YYYY-MM-DD" << std::endl;
    std::cout<<" Time Format: HH:MM:SS where the number of seconds can be a decimal value" << std::endl;
    std::cout<<" Connection String: hostname:database:login:password[:port]"<<std::endl;
    std::cout<<" folder: /database/folder"<<std::endl;
}

//Function that converts the date and time strings from the input file in an LCTime object from which to obtain the timestamp
lcio::LCTime convert2LCTime(std::string Date, std::string Time)
{
    //Variable delcaration
    std::vector<std::string> tokens;
    int Year;
    int Month;
    int Day;
    int Hour;
    int Minute;
    int Second;
    EVENT::long64 NanoSecond;
    lcio::LCTime timestamp;

    //TODO More meaningful, if any format checking for date and time
    //Date format: YYYY-MM-DD
    if(Date.size() != 10)
    {
        std::string error = std::string("The date string is not correctly formatted");
        std::invalid_argument e(error);
        throw  e;
    }
    std::stringstream datestream(Date);
    std::string token;
    while(std::getline(datestream, token, '-'))
    {
        tokens.push_back(token);
    }

    //TODO More meaningful, if any format checking for date and time
    //Time Format: HH:MM:SS[.ddd]
    if(Time.size() < 8)
    {
        std::string error = std::string("The time string is not correctly formatted");
        std::invalid_argument e(error);
        throw  e;
    }

    std::stringstream timestream(Time);
    while(std::getline(timestream, token, ':'))
    {
        tokens.push_back(token);
    }

    //Create a new stream using the strings that I got from the parsing then using the stream I/O
    //mechanism to feed the data in the variables using the internal operator(<<) for teh conversion
    std::stringstream convStream;
    double tempSec;
    convStream << tokens[0] << " " << tokens[1] << " " << tokens[2] << " " << tokens[3] << " " << tokens[4] << " " << tokens[5];
    convStream >> Year >> Month >> Day >> Hour >> Minute >> tempSec;
    //Retrieve the integral and the decimal part of the variable indicating the seconds
    Second = static_cast<int>(tempSec);
    NanoSecond = static_cast<EVENT::long64>((tempSec - static_cast<EVENT::long64>(tempSec)) * 10E9);

    timestamp=LCTime(Year,Month,Day,Hour,Minute,Second);
    timestamp+= NanoSecond;
    return timestamp;
}

//Main program
int main(int argc, char **argv)
{
    //Program variable declaration
    //Program steering options
    bool writeSimpleFile=false;
    bool write2DB=false;
    std::string dbInitString;
    std::string dbFolder;
    std::string outFileName;
    std::vector<std::string> inFileNames;

    // Other variables
    int  optionCounter=0;
    lccd::DBInterface* dbConnection = 0;
    lcio::LCCollectionVec* conditionsColl = 0;
    const std::string COLLNAME = std::string("TPCConditions");
    lcio::LCWriter* lcWrt = 0;
    lcio::LCEventImpl*  evt = 0;
    std::vector<TPCCondrecord> recordVector; //Vector containing the elements to fill in the db or file

    //Manage Command Line and Fill variables
    //Check minimum number of variables
    if ( argc < 2 )
    {
        printUsage();
        return(1);
    }

    //Check that the second argument is a valid option
    //TODO? Change the elseif to a switch structure (more readable)
    //TODO? Single try...catch block for the entire switch structure
    std::string option = argv[1];
    if ( (option == "-S") && (argc > 3))    //If option -S (simple file)
    {
        writeSimpleFile=true;    //Set write file to true
        outFileName=argv[2];
        optionCounter=3;
        try
        {
            std::cout << "Creating the event writer to write the file" << std::endl;
            lcWrt = lcio::LCFactory::getInstance()->createLCWriter() ;    //Create LCWriter
            lcWrt->open( outFileName, lcio::LCIO::WRITE_NEW ) ;
            evt = new lcio::LCEventImpl() ;
        }
        catch(lcio::Exception &e)
        {
            std::cerr << "TPCConditions2LCCD: ERROR while creating the LCWriter and the event objects" <<std::endl;
            std::cerr << e.what() << std::endl;
            if(lcWrt != 0)
            {
                lcWrt->close();
                delete lcWrt;
                lcWrt = 0;
            }
            if(evt != 0)
            {
                delete evt;
                evt = 0;
            }
            return 1;
        }assert(lcWrt != 0);
    }
    else if ( option == "-D" && argc > 4)  //If option -D (database)
    {
        write2DB=true;    //Set write 2 db to true
        //Fill the DB variables
        dbInitString=argv[2];
        dbFolder=argv[3];
        //Create DB Connection
        try
        {
            dbConnection = new lccd::DBInterface(dbInitString, dbFolder, write2DB);
        }
        catch(std::exception &e)
        {
            std::cerr << "TPCConditions2LCCD: ERROR while creating the DB connection" << std::endl;
            std::cerr << e.what() << std::endl;
            if(dbConnection != 0)
            {
                delete dbConnection;
                dbConnection = 0;
            }
            return 1;
        } assert(dbConnection != 0);

        optionCounter=4;
    }
    else
    {
        printUsage();
        return(1);
    } assert(dbConnection != 0 || lcWrt != 0);
    assert(writeSimpleFile ^ write2DB);

    //Check that there is at least on input file
    if(argc >= optionCounter)
    {
        //Insert the file names in a vector for future access
        for ( int i=optionCounter; i<argc ; ++i)
        {
            inFileNames.push_back(argv[i]);
        }
    }
    else
    {
        printUsage();
        if(dbConnection != 0)
        {
            delete dbConnection;
            dbConnection = 0;
        }
        if(lcWrt != 0)
        {
            lcWrt->close();
            delete lcWrt;
            lcWrt = 0;
        }
        if(evt != 0)
        {
            delete evt;
            evt = 0;
        }
        return(1);
    }assert(inFileNames.size() != 0);

    //loop over the input files
    //TODO Add a try...catch clause around the file reading and parsing procedure
    for (std::vector<std::string>::iterator inFile = inFileNames.begin(); inFile != inFileNames.end(); ++inFile)
    {
        //Open the input file and connects it to a read buffer
        std::filebuf infile;
        infile.open(inFile->c_str(),std::ios::in);
        std::istream inputstream(&infile);
        std::cout << " Reading file #" << (inFile - inFileNames.begin() + 1) <<" : "<< *inFile << std::endl;
        int linenumber = 0; //<- Spy variables for debugging info

        //Read all the lines of the file until the end
        while(!inputstream.eof())
        {
            //Take a line from the stream and feeds it to a string
            std::string line;
            line.clear();
            std::getline(inputstream, line);

            // check if the next line is a comment or if it's an empty line
            // If the line is valid associate it to a stream so we can use the >> operator
            // to do the string conversions into the variables
            if (*line.begin() != '#' && line.size() != 0)
            {
                //Local variables creation
                std::stringstream linestream(line);
                tpcconddata::TPCConditions* theConditions = 0;
                lcio::LCTime validSince;
                lcio::LCTime validTill;
                std::string validSinceDate;
                std::string validSinceTime;
                std::string validTillDate;
                std::string validTillTime;
                float driftVelocity;
                float amplification;
                float longDiffusionCoeff;
                float longDefocussing;
                float transDiffusionCoeff;
                float transDefocussing;
                float adcToPrimaryElectronsFactor;

                linestream >> validSinceDate >> validSinceTime >> validTillDate >> validTillTime
                    >> driftVelocity >> amplification >> longDiffusionCoeff >> longDefocussing
                    >> transDiffusionCoeff >> transDefocussing >> adcToPrimaryElectronsFactor;

                // test if everything went fine
                if (linestream.fail())
                {
                    std::cout << "TPCConditions2LCCD: ERROR reading input file at line " << linenumber << " : " <<line << std::endl;

                    if(dbConnection != 0)
                    {
                        delete dbConnection;
                        dbConnection = 0;
                    }
                    if(lcWrt != 0)
                    {
                        lcWrt->close();
                        delete lcWrt;
                        lcWrt = 0;
                    }
                    if(evt != 0)
                    {
                        delete evt;
                        evt = 0;
                    }
                    return 1;
                }

                //Create TPC Condition Object
                try
                {
                    //Create the condition object
                    theConditions = new tpcconddata::TPCConditions();
                    theConditions->setDriftVelocity(driftVelocity);
                    theConditions->setAmplification(amplification);
                    theConditions->setLongDiffusionCoef(longDiffusionCoeff);
                    theConditions->setLongDefocussing(longDefocussing);
                    theConditions->setTransDiffusionCoef(transDiffusionCoeff);
                    theConditions->setTransDefocussing(transDefocussing);
                    theConditions->setADCtoPrimaryElectronsFactor(adcToPrimaryElectronsFactor);

                    validSince = convert2LCTime(validSinceDate, validSinceTime);
                    validTill = convert2LCTime(validTillDate, validTillTime);
                    std::cout << "Valid From: " << validSince.timeStamp() << " to " << validTill.timeStamp() << std::endl;
                    TPCCondrecord thisRecord = {theConditions, validSince, validTill};
                    recordVector.push_back(thisRecord);
                }
                catch(std::exception &e)
                {
                    std::cerr << "TPCConditions2LCCD: ERROR while creating the LCIO objects" <<std::endl;
                    std::cerr << e.what() << std::endl;
                    if(dbConnection != 0)
                    {
                        delete dbConnection;
                        dbConnection = 0;
                    }
                    if(lcWrt != 0)
                    {
                        lcWrt->close();
                        delete lcWrt;
                        lcWrt = 0;
                    }
                    if(evt != 0)
                    {
                        delete evt;
                        evt = 0;
                    }
                    if(theConditions != 0)
                    {
                        delete theConditions;
                        theConditions = 0;
                    }
                    if(recordVector.size() != 0)//Cleanup the vector
                    {
                        for(std::vector<TPCCondrecord>::iterator it = recordVector.begin(); it != recordVector.end(); it++)
                        {
                            delete it->theConditions;
                        }
                    }
                    return 1;
                }assert(theConditions != 0); assert(recordVector.size() != 0);
            }
            else
            {
                std::cout << "  DEBUG: skipping comment or empty line " << linenumber << " : "  << line <<  std::endl;
            }
            linenumber++;
        }
        //Close the input file
        infile.close();
    }

    //If write to file I create a single collection containing multiple TPC Condition objects
    if(writeSimpleFile)
    try
    {
        //Create collection
        conditionsColl = new lcio::LCCollectionVec(lcio::LCIO::LCGENERICOBJECT);
        //Fill the collection
        for(std::vector<TPCCondrecord>::iterator it = recordVector.begin(); it != recordVector.end(); it++)
        {
            conditionsColl->addElement(it->theConditions);
            it->theConditions = 0;
        }
        //Add the collection to the event
        evt->addCollection(conditionsColl,COLLNAME);
        //Dump the event to screen
        lcio::LCTOOLS::dumpEvent( evt ) ;
        // write event to file
        lcWrt->writeEvent( evt ) ;
    }
    catch(lcio::Exception & e)
    {
        std::cerr << "TPCConditions2LCCD: ERROR while creating the LCIO collection" <<std::endl;
        std::cerr << e.what() << std::endl;
        if(lcWrt != 0)
        {
            delete lcWrt;
            lcWrt = 0;
        }
        if(evt != 0)
        {
            delete evt;
            evt = 0;
        }
        if(recordVector.size() != 0)//Cleanup the vector
        {
            for(std::vector<TPCCondrecord>::iterator it = recordVector.begin(); it != recordVector.end(); it++)
            {
                if(it->theConditions != 0) delete it->theConditions;
            }
        }
        return 1;
    }
    else if (write2DB)
    {
        try
        {
            //Create one collection containing one element for each vector element and store it between the given timestamps
            for(std::vector<TPCCondrecord>::iterator it = recordVector.begin(); it != recordVector.end(); it++)
            {
                //TODO Add an output for each line written to the DB
                conditionsColl = new lcio::LCCollectionVec(lcio::LCIO::LCGENERICOBJECT);
                conditionsColl->addElement(it->theConditions);
                dbConnection->storeCollection(it->validSince.timeStamp(), it->validTill.timeStamp(), conditionsColl, COLLNAME);
            }
        }
        catch(std::exception &e)
        {
            std::cerr << "TPCConditions2LCCD: ERROR while storing the condition object in the DB" <<std::endl;
            std::cerr << e.what() << std::endl;
            if(dbConnection != 0)
            {
                delete dbConnection;
                dbConnection = 0;
            }
            if(recordVector.size() != 0)//Cleanup the vector
            {
                for(std::vector<TPCCondrecord>::iterator it = recordVector.begin(); it != recordVector.end(); it++)
                {
                    if(it->theConditions != 0) delete it->theConditions;
                }
            }
            if(conditionsColl != 0)
            {
                delete conditionsColl;
                conditionsColl = 0;
            }
            return 1;
        }
    }

    //Close DB and clean up
    if(dbConnection != 0)
    {
        delete dbConnection;
        dbConnection = 0;
    }
    if(lcWrt != 0)
    {
        lcWrt->close();
    }
    if(evt != 0)
    {
        delete evt;
    }

    return 0;
}//End main


