//##############################################################################
//      programm to convert data taken by the TimePix Chip to LCIO
//
//                    input arguments are a .txt file 
//        giving the path of all .dat files containing the data
//
//                        Simone Zimmermann 07/09/06
//
//##############################################################################

#include "lcio.h"

#include "IO/LCWriter.h"
#include "EVENT/LCIO.h"

#include "IMPL/LCEventImpl.h" 
#include "IMPL/LCRunHeaderImpl.h" 
#include "IMPL/LCCollectionVec.h"
#include "IMPL/LCTOOLS.h"
#include "IMPL/TrackerRawDataImpl.h"

#include "gearimpl/RectangularPadRowLayout.h"
#include "gearimpl/Util.h"
#include "gearxml/GearXML.h"
#include "gear/GearMgr.h"
#include "gear/GEAR.h"

#include "TimePixPixelMode.h"
//#include "TimePixModeMap.h"
#include "Pedestal.h"

#include <iostream>
#include <sstream>
#include <string>
#include <cstdio>
#include <fstream>

using namespace std ;
using namespace lcio ;
using namespace gear;
using namespace tpcconddata;

//global parameters
int _nRow = 256; //number of rows on chip
int _nColumn = 256; //number of columns on chip

void printUsage()
{
  cout << "TimePixToLCIOConverter paths.txt [outfile.slcio]" <<endl;
  cout << "paths.txt gives the paths to .dat files \n " 
       << "outfile name is optional, "
       << "default is TimePixRawData.slcio"<< endl;
}

void createMixedMode1Map( LCCollectionVec* modeMap ) 
{
  //creates map of chip operated in MixedMode starting with TIMEPIX
  int ncells = _nRow*_nColumn;
  int chipid = 0;
  int TOT = 1;
  int TIMEPIX = 3;

  for ( int cellid = 0; cellid < ncells; cellid++ ){

    int x = cellid%_nColumn;
    int y = static_cast<int>((cellid-x)/_nColumn);

    if ( ( x%2 == 0 && y%2 ==0 ) || (x%2 != 0 && y%2 != 0 ) ) {

      modeMap->addElement( new TimePixPixelMode( chipid, cellid, TIMEPIX) );  

    }
    else {

      modeMap->addElement( new TimePixPixelMode( chipid, cellid, TOT) );  


    }


    }

}

void createMixedMode2Map( LCCollectionVec* modeMap ) 
{
  //creates map of chip operated in MixedMode starting with TOT
  int ncells = _nRow*_nColumn;
  int chipid = 0;
  int TOT = 1;
  int TIMEPIX = 3;

  for ( int cellid = 0; cellid < ncells; cellid++ ){
    
    int x = cellid%_nColumn;
    int y = static_cast<int>((cellid-x)/_nColumn);

    if ( ( x%2 == 0 && y%2 ==0 ) || (x%2 != 0 && y%2 != 0 ) ) {

      modeMap->addElement( new TimePixPixelMode( chipid, cellid, TOT) );  

    }
    else {

      modeMap->addElement( new TimePixPixelMode( chipid, cellid, TIMEPIX) );  

    }


    }

}

void createTOTMap( LCCollectionVec* modeMap ) 
{
  //creates map of chip operated in TOT mode alone
  int ncells = _nRow*_nColumn;
  int chipid = 0;
  int TOT = 1;

  for ( int cellid = 0; cellid < ncells; cellid++ ){
    
    modeMap->addElement( new TimePixPixelMode( chipid, cellid, TOT) );  

    }

}

void createTIMEPIXMap( LCCollectionVec* modeMap ) 
{

  //creates map of chip operated in TIMEPIX mode alone
  int ncells = _nRow*_nColumn;
  int chipid = 0;
  int TIMEPIX = 3;

  for ( int cellid = 0; cellid < ncells; cellid++ ){
    
    modeMap->addElement( new TimePixPixelMode( chipid, cellid, TIMEPIX) );  

    }

}





int main(int argc, char **argv)
{

  if (argc!=2 && argc!=3 )
  {
    printUsage();
    return(1);
  }

  int cellid0 = 0;
  bool isFirstEvent = true;

  //get infile name
  string inFileName;
  if (argc > 1)
    inFileName = argv[1];
  else 
    {
      cout<<"No input data given"<<endl;
      printUsage();
      return 0;					
    }

  // get outfile name
  string outFileName;
  if (argc==3)
    outFileName=argv[2];
  else
    outFileName="TimePixRawData.slcio";

  // open outfile			     
  try{
    // create sio writer
    LCWriter* lcWrt = LCFactory::getInstance()->createLCWriter() ;
    
    try{  lcWrt->open( outFileName, LCIO::WRITE_NEW ) ;
    }
    catch( IOException& e ){    
      cout << e.what() << endl ;
      return 0 ;
    }
    
    // create runheader and write it
    // the information should come from a file
    // or be given interactively in a future version
    LCRunHeaderImpl* runHdr = new LCRunHeaderImpl ; 
    int runNr=0;
    runHdr->setRunNumber( runNr) ;
    
    string detName("TimePix Bonn")  ;
    runHdr->setDetectorName( detName ) ;
    
    stringstream description ; 
    description << "TimePix data taken at LCTPC Lab Bonn";
    runHdr->setDescription( description.str()  ) ;
    
    string tpcName("Version 0") ;
    runHdr->addActiveSubdetector( tpcName ) ;
    
    lcWrt->writeRunHeader( runHdr ) ;
    
    //read in .dat files  
    ifstream infile;
    try { infile.open( inFileName.c_str() ); }
    catch (IOException& e )
      {
	cout<< e.what() <<endl;
	return 0;
      }
    if ( !infile.good() ) {
      infile.close();
      cout<<"!"<<inFileName<<".good()"<<endl;
      return 0;
    }
    
    int eventNr = 0;
    while (1) {

      char datafile[7200];
      
      infile >> datafile;
      if (!infile.good() ) break;
      
      // create new lcio event
      LCEventImpl*  evt = new LCEventImpl() ;
	
      evt->setRunNumber(  runNr   ) ;
      evt->setEventNumber( eventNr ) ;
      evt->setDetectorName( detName ) ;
      

      // we want to store a collection of trackerRawData
      LCCollectionVec* trackerRawDataVec = new LCCollectionVec( LCIO::TRACKERRAWDATA )  ;
      
      ifstream eventfile;
      eventfile.open( datafile );
      
      if (!eventfile.good()) {
	  cout<<"!"<<datafile<<".good()"<<endl;
	  eventfile.close();
	  continue;
      }
      
      TrackerRawDataImpl* rawData = new TrackerRawDataImpl; 
      rawData->setCellID0(cellid0);
      EVENT::ShortVec ADCValues;
      
      while (1) {
	
	short data;
	eventfile >> data;
	if (!eventfile.good()) break;
	
	ADCValues.push_back( data );
	//	rawData->setADCValues(ADCValues);
      }

      rawData->setADCValues(ADCValues);

      // add the  rawData to the collection
      trackerRawDataVec->push_back(rawData);
      
      // add the rawData collection to the event
      evt->addCollection( trackerRawDataVec , "TimePixRawData" ) ;



     
      // add the conditions data to the first event
      if ( isFirstEvent ) {
	
	LCCollectionVec* modeMap = new LCCollectionVec( LCIO::LCGENERICOBJECT );

	//create the modeMap
	createTOTMap( modeMap );

	evt->addCollection( modeMap, "TimePixModeMap" );
	
	isFirstEvent = false;
      }


      // write the event to the file
      lcWrt->writeEvent( evt ) ;
      
      // dump the event to the screen 
      LCTOOLS::dumpEvent( evt ) ;

      // ------------ IMPORTANT ------------- !
      // we created the event so we need to delete it ...
      delete evt ;
      // -------------------------------------
      
      eventNr++;
      
    } // for eventNr

    delete runHdr ;
    
    lcWrt->close() ;
      
  } catch( std::exception& ex){
    
    cout << " an excpetion occured: " << endl ;
    cout << "  " << ex.what() << endl ;
    return 1 ;
  }

     
return 0;

}
