#include "TimePixMaskCreaterProcessor.h"
#include "TimePixPixelMode.h"

#include "marlin/ProcessorMgr.h"

#include "IO/LCWriter.h"
#include "IMPL/LCEventImpl.h"
#include "IMPL/LCRunHeaderImpl.h"
#include "IMPL/TrackerRawDataImpl.h"
#include "IMPL/LCCollectionVec.h"

//#include "UTIL/LCStdHepRdr.h"
#include "UTIL/LCTOOLS.h"

#include <TROOT.h>
#include <TFile.h>
#include <TH2D.h>

using namespace lcio;
using namespace tpcconddata;

namespace marlin{



  TimePixMaskCreaterProcessor aTimePixMaskCreaterProcessor ;
  
  
  TimePixMaskCreaterProcessor::TimePixMaskCreaterProcessor() : DataSourceProcessor("TimePixMaskCreaterProcessor") {
    
    _description = "Creates 1 LCIO event with LCGenericObjectCollection containing the dead or noisy pixel "
      " Make sure to not specify any LCIOInputFiles in the steering in order to read the occupancy histogram." ;


   registerOutputCollection( LCIO::LCGENERICOBJECT,
			     "ModeMapCollectionName" , 
			     "Name of the output mode map collection"  ,
			      _statusMapColName,
			     std::string("TimePixStatusMap") ) ;
    
    registerProcessorParameter( "HistoFileName" , 
				"input file with occupancy histogram"  ,
				_histoFileName ,
				std::string("TimePixControlPlots.root") ) ;

   registerProcessorParameter( "HistoDirectoryName" , 
				"directory in input file with occupancy histogram"  ,
				_histoDirectoryName ,
				std::string("MyTimePixOccupancyProcessor") ) ;

   registerProcessorParameter( "HistoName" , 
				"name of the occupancy histogram"  ,
				_histoName ,
				std::string("TimePix_occupancy_chip0") ) ;

    registerProcessorParameter( "DetectorName" , 
				"Name of the detector"  ,
				_detName ,
				std::string("TPC" ) );

    registerProcessorParameter( "Description" , 
				"Description of data"  ,
				_description ,
				std::string("no description given" ) ) ;
    
    registerProcessorParameter( "SubDetectorName" , 
				"Name of the subdetector"  ,
				_subDetName ,
				std::string("TimePix" ) );

    registerProcessorParameter( "ChipID" , 
				"ID number of the chip "  ,
				_chipID,
				(int) 0 ) ;

    registerProcessorParameter( "NColumn" , 
				"Number of columns on the chip "  ,
				_nColumn,
				(int) 256 ) ; 

    registerProcessorParameter( "NoiseThreshold" , 
				"Threshold above which pixel is declared noisy "  ,
				_noiseThreshold,
				(double) 1000. ) ; 

    registerProcessorParameter( "SetOutputTransient" , 
				"if not 0 the output collection is set transient"  ,
				_outputIsTransient,
				0);
    
  }
  
  TimePixMaskCreaterProcessor*  TimePixMaskCreaterProcessor::newProcessor() { 
    return new TimePixMaskCreaterProcessor ;
  }
  
  void TimePixMaskCreaterProcessor::init() {    
    printParameters() ;    
    
    _isFirstEvent = true; 
    _runNr=0;
    _evtNr=0;

    _file = new TFile( _histoFileName.c_str() );

    std::string histopath = _histoDirectoryName + "/" + _histoName;

    _histoOccupancy = (TH2D*)_file->Get( histopath.c_str() );

  }


  void TimePixMaskCreaterProcessor::readDataSource( int numEvents ) {
 
 
    streamlog_out(MESSAGE1)<<"readDataSource"<<std::endl;


    // create new lcio event
    LCEventImpl*  evt = new LCEventImpl() ;
    evt->setRunNumber(  _runNr   ) ;
    evt->setEventNumber( _evtNr ) ;
    evt->setDetectorName( _detName ) ;
    
    streamlog_out(DEBUG0)<<"created event"<<std::endl;
    
    // we want to store a collection of LCGenericObject
    LCCollectionVec* statusMapVec = new LCCollectionVec( LCIO::LCGENERICOBJECT )  ;
 
    streamlog_out(DEBUG0)<<"created statusMapVec"<<std::endl;
   
    // set the transient flag as definded by the processor parameter
    statusMapVec->setTransient( (_outputIsTransient != 0) );
    
    // add the conditions data to the first event  
    LCCollectionVec* statusMap = new LCCollectionVec( LCIO::LCGENERICOBJECT );
  
    streamlog_out(DEBUG0)<<"created statusMap"<<std::endl;
  
    int nbinsx = _histoOccupancy->GetXaxis()->GetNbins();
    int nbinsy = _histoOccupancy->GetYaxis()->GetNbins();
    
    double bincontent;
    int cellid = 0;
    //root binning convention
    //bin = 0;       underflow bin
    //bin = 1;       first bin with low-edge xlow INCLUDED
    //bin = nbins;   last bin with upper-edge xup EXCLUDED
    //bin = nbins+1; overflow bin
    
    for ( int binx = 1; binx < nbinsx+1; binx++ ) {
      
      for ( int biny = 1; biny < nbinsy+1; biny++ ) {
	
	bincontent = _histoOccupancy->GetBinContent( binx, biny );
	
	//	if ( bincontent == 0. ) {
	if ( binx == 42 && biny > 90 ) {

	  streamlog_out(DEBUG0)<<"dead: "<<bincontent<<std::endl;

	  cellid = (binx-1) + _nColumn*(biny-1);
	  statusMap->addElement( new TimePixPixelMode
				 ( _chipID, cellid, 
				   TimePixPixelMode::ONEHIT, 
				   TimePixPixelMode::DEAD ) );
	  
	}
	
	//	if ( bincontent >= _noiseThreshold ) {
	if ( binx == 42 && biny <= 90 ) {  

	  streamlog_out(DEBUG0)<<"noisy: "<<bincontent<<std::endl;

	  cellid = (binx-1) + _nColumn*(biny-1);
	  statusMap->addElement( new TimePixPixelMode
				 ( _chipID, cellid, 
				   TimePixPixelMode::ONEHIT, 
				   TimePixPixelMode::NOISY ) );
	  
	}
	
      } //for ( int biny 
      
    } //for ( int binx = 1;
    
    
    streamlog_out(DEBUG0)<<"statusMap->getNumberOfElements() = "
			 <<statusMap->getNumberOfElements()<<std::endl;
  
    evt->addCollection( statusMap, _statusMapColName );
    
    
    //create RunHeader
    streamlog_out(DEBUG0)<<"create RunHeader"<<std::endl;
    
    LCRunHeaderImpl* runHdr = new LCRunHeaderImpl ; 
    
    runHdr->setRunNumber( _runNr) ;
    runHdr->setDetectorName( _detName ) ;
    runHdr->setDescription( _description ) ;
    runHdr->addActiveSubdetector( _subDetName ) ;
    
    ProcessorMgr::instance()->processRunHeader( runHdr ) ;
    delete runHdr;
    
    // dump the event to the screen 
    if( streamlog::out.write< streamlog::DEBUG2 >() )
      LCTOOLS::dumpEvent( evt ) ;
      
    ProcessorMgr::instance()->processEvent( evt ) ;     
    
    
    // ------------ IMPORTANT ------------- !
    // we created the event so we need to delete it ...
    delete evt ;
    // -------------------------------------
    
    streamlog_out(DEBUG0)<<"in readDataSource"<<std::endl;
    
    streamlog_out(DEBUG0)<<"\nend of readDataSource"<<std::endl;

  }
  


  void TimePixMaskCreaterProcessor::end() {



  }


}
