#include "TimePixXYReaderProcessor.h"
#include "TimePixPixelMode.h"

// Marlin
#include "marlin/ProcessorMgr.h"

// LCIO
#include <IO/LCWriter.h>
#include <IMPL/LCEventImpl.h>
#include <IMPL/LCRunHeaderImpl.h>
#include <IMPL/TrackerRawDataImpl.h>
#include <IMPL/LCCollectionVec.h>
#include <IMPL/LCFlagImpl.h>
#include <UTIL/LCTOOLS.h>


namespace marlin{



  TimePixXYReaderProcessor aTimePixXYReaderProcessor ;
  
  
  TimePixXYReaderProcessor::TimePixXYReaderProcessor() : DataSourceProcessor("TimePixXYReaderProcessor"),
							 _evtNr(0),
							 _activatedPixels(0),
							 _nPixels(256),
							 _isFirstEvent(true){

    
    _description = std::string("Reads TimePix raw data files as input and creates LCIO events with TrackerRawData collections.")
      + std::string(" Make sure to not specify any LCIOInputFiles in the steering in order to read TimePix raw data files.") ;


    registerProcessorParameter( "FileListName" , 
				"Input file with list of TimePix files (default: TimePixData.txt)"  ,
				_listName ,
				std::string("TimePixData.txt") ) ;
  
    registerOutputCollection( LCIO::TRACKERRAWDATA,
			      "OutputCollectionName" , 
			      "Name of the output collection (default: TimePixRawData)"  ,
			      _outputColName,
			      std::string("TimePixRawData") ) ;
  
    registerProcessorParameter( "NChipRows" , 
				"Number of chip rows (default: 1) "  ,
				_nChipRow,
				int(1) ) ; 
  
    registerProcessorParameter( "NChipColumns" , 
				"Number of chip columns (default: 1)"  ,
				_nChipColumn,
				int(1) ); 
				
    registerProcessorParameter( "SetOutputTransient" , 
				"If true the output collection is set transient and not added to the file output (default: false)"  ,
				_outputIsTransient,
				bool(false) );

    registerProcessorParameter( "MinPixels",
				"Minimum number of activated pixels in one event. Events with less pixels are skipped (default: 10)",
				_minPixels,
				int(10) );				

    registerOptionalParameter( "RunNumber",
			       "Set RunNumber (default: 0)",
			       _runNr,
			       int(0) );
    
    registerOptionalParameter( "Description" , 
			       "Description of data (default: No description given)"  ,
			       _dataDescription ,
			       std::string( "No description given" ) ) ;
    
    registerOptionalParameter( "DetectorName" , 
			       "Name of the detector (default: TPC)"  ,
			       _detName ,
			       std::string("TPC") );
  
    registerOptionalParameter( "SubDetectorName" , 
			       "Name of the subdetector (default: TimePix)"  ,
			       _subDetName ,
			       std::string("TimePix") );
  
    registerOptionalParameter( "ChipIDs" , 
			       "ID number(s) of the chip(s) (default: 0) "  ,
			       _chipIDs,
			       std::vector<int>(1,0) ) ;

  
  } // end of C'tor


  
  TimePixXYReaderProcessor*  TimePixXYReaderProcessor::newProcessor() { 

    return new TimePixXYReaderProcessor ;
  }


  
  void TimePixXYReaderProcessor::init() {    


    // Print parameters to screen
    printParameters() ;    
    

    // Try to read in the file containing the data file list.
    try{ 

      _inFile.open( _listName.c_str() ); 

    }catch (IOException& e ){
      
      streamlog_out(ERROR) << e.what() <<std::endl;
    }


    if( !_inFile.good() ){
      
      _inFile.close();
      streamlog_out(ERROR) << "!" << _listName << ".good()" << std::endl;
    }

    if( !parameterSet( "ChipIDs" ) ){

      const unsigned nChips = _nChipColumn * _nChipRow;
      _chipIDs.resize( nChips );

      for( unsigned iChip = 0; iChip < nChips; ++iChip ){

	_chipIDs[iChip] = iChip;
      }
      
    } // end of if( !parameterSet( "ChipIDs" ) )


  } // end of void TimePixXYReaderProcessor::init()





  void TimePixXYReaderProcessor::readDataSource( int numEvents ) {


    while(true){

      char dataFile[7200];
      _inFile >> dataFile;

      if( _inFile.fail() || _inFile.eof() ){

	  break;
      }

      // Create new event
      LCEventImpl*  evt = new LCEventImpl() ;
      evt->setRunNumber(  _runNr   );
      evt->setEventNumber( _evtNr );
      evt->setDetectorName( _detName );
    
      // Create collection from data file
      IMPL::LCCollectionVec* outCol = this->readXYFile( dataFile );   


      if( outCol == NULL || (_activatedPixels < static_cast<unsigned>(_minPixels)) ){

	delete outCol;
	delete evt;
	continue;
      }

      evt->addCollection( outCol , _outputColName ) ;

      if( _isFirstEvent ) {
	
	LCRunHeaderImpl* runHdr = new LCRunHeaderImpl ; 
	
	runHdr->setRunNumber( _runNr );
	runHdr->setDetectorName( _detName );
	runHdr->setDescription( _dataDescription );
	runHdr->addActiveSubdetector( _subDetName );

	// Log revision
	runHdr->parameters().setValue( _processorName + "_revision", "$Rev: 3306 $" );

	// Log parameters
	for( ProcParamMap::iterator iPara = _map.begin(); iPara != _map.end(); ++iPara ){
    
	  if( ! iPara->second->isOptional() || iPara->second->valueSet() ){
    	
	    runHdr->parameters().setValue( _processorName + "_" + iPara->second->name(), iPara->second->value() );
	  }
	}
	
	ProcessorMgr::instance()->processRunHeader( runHdr ) ;
	delete runHdr;
	
	_isFirstEvent = false;

      } // end of if( _isFirstEvent )
    

      if( streamlog::out.write< streamlog::DEBUG2 >() )
	LCTOOLS::dumpEvent( evt ) ;

      ProcessorMgr::instance()->processEvent(evt) ;     
      
      delete evt ;
      
      ++_evtNr;

    } // end of while(true)

  } // end of readDataSource( int numEvents )





  IMPL::LCCollectionVec* TimePixXYReaderProcessor::readXYFile( const std::string& fileName ){

    _activatedPixels = 0;

    std::ifstream dataFileStream;
    dataFileStream.open( fileName.c_str() );

    if( !dataFileStream.good() ){

      streamlog_out(ERROR) << "!" << fileName << ".good()" << std::endl;
      dataFileStream.close();

      return NULL;
    }

    IMPL::LCCollectionVec* trackerRawDataCol = new IMPL::LCCollectionVec( LCIO::TRACKERRAWDATA );


    // Set the flags to store cellID1
    lcio::LCFlagImpl trkFlag(0) ;
    trkFlag.setBit( lcio::LCIO::TRAWBIT_ID1 ) ;
    trackerRawDataCol->setFlag( trkFlag.getFlag() );

    // Set the transient flag as defined by the processor parameter
    trackerRawDataCol->setTransient( _outputIsTransient );


    int data[3];
    dataFileStream >> data[0];
    dataFileStream >> data[1];
    dataFileStream >> data[2];


    while( !dataFileStream.fail() ){

      ++_activatedPixels;

      TrackerRawDataImpl* rawData = new TrackerRawDataImpl; 

      const int pixelID = (data[0] % _nPixels) + (data[1] * _nPixels);

      
      rawData->setCellID0( pixelID );
      rawData->setCellID1( getChipID( data[0], data[1] ) );

      EVENT::ShortVec adcValues;
	
      adcValues.push_back( data[2] );

      int prevCol = data[0], prevRow = data[1];
      while(true){

	dataFileStream >> data[0];
	dataFileStream >> data[1];
	dataFileStream >> data[2];

	if( dataFileStream.fail()  || 
	    data[0] != (prevCol+1) ||
	    data[1] != (prevRow)   ||
	    getChipID( data[0], data[1] ) != getChipID( prevCol, prevRow ) ){

	  break;
	}

	++_activatedPixels;

	prevCol = data[0];
	prevRow = data[1];

	adcValues.push_back( data[2] );

      } // end of while(true)

      rawData->setADCValues( adcValues);

      // Add the TrackerRawData object to the collection
      trackerRawDataCol->push_back( rawData );

    } // end of while( !dataFileStream.fail() )

    
    if( trackerRawDataCol->empty() ){

      delete trackerRawDataCol;
      trackerRawDataCol = NULL;
    }
    
    return trackerRawDataCol;

  } // end of IMPL::LCCollectionVec* TimePixXYReaderProcessor::getDataCollection()








  int TimePixXYReaderProcessor::getChipID( const int& col, const int& row ) const {

    const int chipInRow = row / _nPixels;
    const int chipInCol = col / _nPixels;

    const int entry = ( chipInRow * _nChipColumn ) + chipInCol ;

    return _chipIDs[entry];

  } // end of int TimePixXYReaderProcessor::getChipID( const int& col, const int& row ) const 

  



} // end of namespace marlin
