#ifndef ADCChannelMapping_h
#define ADCChannelMapping_h 1

//C++
#include "iostream"
#include "string"

//LCIO
#include "lcio.h"
#include "EVENT/LCGenericObject.h"
#include "IMPL/LCGenericObjectImpl.h"

namespace tpcconddata
{

    class ADCChannelMapping;

    /** class to store the mapping pad (position) to the channel ID (electronic)
     *  also the type of the used electronic can be stored.
     *  The class is also able to handle the module ID. It is fully backward compatible
     *  to the version without module ID, although the implementation has changed from
     *  LCFixedObject to LCGenerigObject to provide the required felixibility.
     *
     *  Like the LCFixedObject this thing contains an LCGenericObject, but is not derrived from it.
     *
     *  @param channelID      Hardware channel number
     *  @param padID          The GEAR pad index
     *  @param type           The type of readout electronics, optional, default -1
     *  @param moduleID       The GEAR module ID, optional, default 0
     *  @param readoutGroup   The hardware readout group number. Used
     *                         if there are several readout units which all provide
     *                         the same range of channel numbers (altro RCUs for instance),
     *                        optional, default 0
     *
     */
    class ADCChannelMapping : public  EVENT::LCGenericObject
    {

        public:

            /** The constructor with all the member functions given.
             *  Use this to create on object on the heap.
             */
            ADCChannelMapping(int channelID, int padIndex, int type = -1, int moduleID = 0,
                int readoutGroup = 0);

            /** The constructor to be used with LCObjects from a collection.
             */
            ADCChannelMapping(EVENT::LCObject* obj) ;

            /** The desctructor.
             */
            virtual ~ADCChannelMapping();

            int getChannelID() const;
            int getPadID() const;
            int getType() const ;
            int getModuleID() const;
            int getReadoutGroup() const;

            static std::string getRevision();
            static std::string getDefaultColName();

            IMPL::LCGenericObjectImpl* obj() const { return _myObj; }
            virtual int id() const { return _myObj->id(); }

            void print( std::ostream& os = std::cout) const;

            // -------- need to implement abstract methods from LCGenericObject
            virtual const std::string getTypeName() const
            {
                return std::string("ADCChannelMapping");
            }

            virtual const std::string getDataDescription() const
            {
                return std::string("i:ChannelIDs,PadID,Type,ModuleID,ReadoutGroup");
            }

            // the other functions inherited from LCGenericObject
            // just pass them to the implementation
            virtual int getNInt()    const { return _myObj->getNInt(); }
            virtual int getNFloat()  const { return _myObj->getNFloat(); }
            virtual int getNDouble() const { return _myObj->getNDouble(); }

            virtual int    getIntVal(int i)    const { return _myObj->getIntVal(i); }
            virtual float  getFloatVal(int i)  const { return _myObj->getFloatVal(i); }
            virtual double getDoubleVal(int i) const { return _myObj->getDoubleVal(i); }

            virtual bool isFixedSize() const { return _myObj->isFixedSize(); }

        #ifdef USE_LCCD
            // add some helper functions for using the native LCCD conditions map
            typedef std::pair<int, int> key_type;

            key_type softwareToHardwareKey() const;
            key_type hardwareToSoftwareKey() const;
        #endif

        protected:
            bool _createObj;
            IMPL::LCGenericObjectImpl* _myObj;

    };

    std::ostream &operator<<(std::ostream &os, const ADCChannelMapping &acm);
}
#endif
