/*
 * ExtendedAltroParameters.h
 *
 *  Created on: Jul 13, 2012
 *      Author: caiazza
 */

#ifndef EXTENDEDALTROPARAMETERS_H_
#define EXTENDEDALTROPARAMETERS_H_

//STD
#include <iostream>
#include <string>

//LCIO
#include "lcio.h"
#include "UTIL/LCFixedObject.h"

namespace tpcconddata
{
/**
 * Class storing to the condDB additional electronics information specific to the ALTRO electronics
 *  PreampEnable: Set true if the PCA16 amplifier is enables
 *  Gain: Gain settings of the PCA16 amplifier. Can be 12, 15, 19 or 27 mV/fC
 *  ShapingTime: The shaping time of the PCA16 amplifier. Can be 30, 60, 90 or 120 ns
 *  Shutdown: Value of the shutdown bit of the PCA16 preamp
 *  ZSThreshold: Value of the Zero Suppression Threshold in ADC counts. A negative value stands for data not available.
 *  ZSGlitchFilter: Minimum number of consecutive samples over threhold to trigger the pulse storage [0-3]
 *  PreSamples: Number of samples to store before the zero suppression threshold [0-3]
 *  PostSamples: Number of samples to store after the zero suppression threshold [0-7]
 *  PreTriggerSamples: Number of samples to evaluate before the trigger signal [0-15]
 *  RCUAddress: Hardware address of the RCU
 *  RCUFWVersion: RCU Firmware version
 *  RCUActiveFECA: Bit map of the active FECs in the branch A
 *  RCUActiveFECB: Bit map of the active FECs in the branch B
 *  RCUConfigRegister1: Copy of the configuration register RDO_CFG1 of the RCU
 *  RCUConfigRegister2: Copy of the configuration register RDO_CFG2 of the RCU
 *  RCUStatusErrors: Copy of the Status/Error data word of the RCU
 *  @author S.Caiazza (DESY)
 */

static const int EXTENDEDALTROPARSNINT = 16;
static const int EXTENDEDALTROPARSNFLOAT = 0;
static const int EXTENDEDALTROPARSNDOUBLE = 0;

class ExtendedAltroParameters : public lcio::LCFixedObject<EXTENDEDALTROPARSNINT, EXTENDEDALTROPARSNFLOAT ,EXTENDEDALTROPARSNDOUBLE>
{
public:
    /**
     * Enumerator to store the available settings of the gain and match them with
     * an human readable name
     */
    enum GainSettings {MV12 = 0, MV15 = 1, MV19 = 2, MV27 = 3};

    /**
     * Enumerator to store the available settings of the Shaping Time and match them with
     * an human readable name
     */
    enum ShapingSettings {NS30 = 7, NS60 = 3, NS90 = 1, NS120 = 0};

    struct ExtendedAltroData
    {
        bool PreampEnable;
        GainSettings Gain;
        ShapingSettings ShapingTime;
        bool Shutdown;
        int ZSThreshold;
        int ZSGlitchFilter;
        int PreSamples;
        int PostSamples;
        int PreTriggerSamples;
        int RCUAddress;
        int RCUFWVersion;
        int RCUActiveFECA;
        int RCUActiveFECB;
        int RCUConfigRegister1;
        int RCUConfigRegister2;
        int RCUStatusErrors;

        ExtendedAltroData():
            PreampEnable(0),
            Gain(GainSettings(0)),
            ShapingTime(ShapingSettings(0)),
            Shutdown(0),
            ZSThreshold(0),
            ZSGlitchFilter(0),
            PreSamples(0),
            PostSamples(0),
            PreTriggerSamples(0),
            RCUAddress(0),
            RCUFWVersion(0),
            RCUActiveFECA(0),
            RCUActiveFECB(0),
            RCUConfigRegister1(0),
            RCUConfigRegister2(0),
            RCUStatusErrors(0)
        {

        }
    };

    /**
     * Convenient constructor to set all the data using the ExtendedAltroData structure to collect the data
     */
    ExtendedAltroParameters(ExtendedAltroData& Data);

    /**
     * Database copy constructor. Allows the upcasting from the generic LCObject retrieved from
     * the database to the specific type
     * @param obj
     */
    ExtendedAltroParameters(EVENT::LCObject* obj);

    /**
     * Destructor. (Not virtual I don't want this class to be extended)
     */
    ~ExtendedAltroParameters();

    /**
     * Sets all the data at the same time
     * @param Data
     */
    void SetData(ExtendedAltroData& Data);

    /**
     * Retrieves all the data in a single structure
     * @return
     */
    ExtendedAltroData GetData() const;

    /**
     * Single element setters. The description of the different data is in the
     * general class description
     */
    void SetPreampEnable(bool& Enable);
    void SetGainSettings(GainSettings& Gain);
    void SetShapingTime(ShapingSettings& ShapingTime);
    void SetShutdown(bool& Shutdown);
    void SetZSThreshold(int& Threshold);
    void SetZSGlitchFilter(int &MinSamples);
    void SetPreSamples(int& PreSamples);
    void SetPostSamples(int& PostSamples);
    void SetPreTriggerSamples(int& PreTriggerSamples);
    void SetRCUAddress(int& RCUAddress);
    void SetRCUFWVersion(int& RCUFWVersion);
    void SetRCUActiveFECA(int& RCUActiveFECA);
    void SetRCUActiveFECB(int& RCUActiveFECB);
    void SetRCUConfigRegister1(int& RCUConfigRegister1);
    void SetRCUConfigRegister2(int& RCUConfigRegister2);
    void SetRCUStatusErrors(int& RCUStatusErrors);

    /**
     * Single element Getters.
     * The description of the data is in the general class description
     */
    bool GetPreampEnable() const ;
    GainSettings GetGainSettings() const ;
    ShapingSettings GetShapingTime() const ;
    bool GetShutdown() const ;
    int GetZSThreshold() const ;
    int GetZSGlitchFilter() const;
    int GetPreSamples() const ;
    int GetPostSamples() const ;
    int GetPreTriggerSamples() const ;
    int GetRCUAddress() const ;
    int GetRCUFWVersion() const ;
    int GetRCUActiveFECA() const ;
    int GetRCUActiveFECB() const ;
    int GetRCUConfigRegister1() const ;
    int GetRCUConfigRegister2() const ;
    int GetRCUStatusErrors() const ;

    /**
     * Retrieves the version of this condition object
     * @return
     */
    static std::string getRevision();

    /**
     * Default name for the collection of this type
     * @return
     */
    static std::string getDefaultColName();

    /**
     * Prints a summary of the data
     * @param os Output stream where to write the summary
     */
    void print(  std::ostream& os = std::cout ) const;

    /**
     * Defines the type name used for the Condition Database
     * @return
     */
    const std::string getTypeName() const;

    /**
     * Defines the data description in the standard condDB format
     * @return
     */
    const std::string getDataDescription() const;
};

/**
 * Writes to stream the summary of the data
 * @param os output stream
 * @param p Condition object
 * @return same stream as input to chain the streams
 */
std::ostream &operator<<(std::ostream &os, const ExtendedAltroParameters &p);

/**
 * Comparison operator that compares 1-1 all the internal data
 * @param first First object to compare
 * @param second Second Object to compare
 * @return Comparison result
 */
bool operator==(const ExtendedAltroParameters &first, const ExtendedAltroParameters &second);
bool operator!=(const ExtendedAltroParameters &first, const ExtendedAltroParameters &second);
}


#endif /* EXTENDEDALTROPARAMETERS_H_ */
