#include "TimePixEventHistoProcessor.h"
#include "TimePixMapHandlerProcessor.h"

// C++
#include <iostream>


// Aida
#ifdef MARLIN_USE_AIDA
//AIDA
#include <AIDA/AIDA.h>
#include <marlin/AIDAProcessor.h>
#include <AIDA/IHistogram2D.h>
#endif

// LCIO
#include <EVENT/TrackerData.h>

//Marlin
#include "TimePixMapListener.h"
#include "TimePixPixelMode.h"
#include <marlin/AIDAProcessor.h>

using namespace lcio ;
using namespace marlin ;
using namespace tpcconddata;

namespace marlintpc{


    TimePixEventHistoProcessor aTimePixEventHistoProcessor ;
    

    TimePixEventHistoProcessor::TimePixEventHistoProcessor() : 
      Processor("TimePixEventHistoProcessor") {
	/* the constructor call first the super constructor which the name 
	 * of the processor
	 * this must by the same as the class name.
	 */
	
	/* description of the processor which will be displayed in the
	 * steering file autometicly generated by 
	 * marlin (use MarlinTPC -x (XML-format) or MarlinTPXC -l (text format))
	 */
	_description = "TimePixEventHistoProcessor creates one aida histogramm of each chip and event" ;	
	
	/* register steering parameters: name, description, class-variable, default value
	 * the type will by definded by the type of the default value
	 * string, double, float and int are possible
	 * please note to get a float you must cast in C++: (float) 1.3 
	 */
	
	registerInputCollection( LCIO::TRACKERDATA,
				 "InputCollectionName" , 
				 "Name of the input collection"  ,
				 _inputColName ,
				 std::string("TimePixZeroSuppressedData") ) ;

	registerProcessorParameter( "HistoBaseName" , 
				    "Base name of the histograms (default TimePixEvent)"  ,
				    _histoBaseName ,
				    std::string("TimePixEvent") ) ;
	
	registerProcessorParameter( "nColumn" , 
				    "number of columns of the chip"  ,
				    _nColumn ,
				    (int) 256 ) ;
	
	registerProcessorParameter( "nRow" , 
				    "number of rows of the chip"  ,
				    _nRow ,
				    (int) 256 ) ;

	registerProcessorParameter( "Mode" , 
				    "mode that should be plotted (TOT, TIMEPIX, MEDIPIX, ONEHIT or ANY), default: ANY"  ,
				    _mode ,
				    std::string("ANY") ) ;
	
	
	_timePixMap = 0;

    }
    
    
    void TimePixEventHistoProcessor::init() { 
	
	_timePixMap = TimePixMapHandlerProcessor::getTimePixMap();
	
    }
    
    void TimePixEventHistoProcessor::processRunHeader( LCRunHeader* run) { 
	
      
      run->parameters().setValue( _processorName + "_revision", "$Rev: 2729 $");

      for ( ProcParamMap::iterator i = _map.begin(); i != _map.end(); i++ ) {
	if ( ! i->second->isOptional() || i->second->valueSet() ) {
	  run->parameters().setValue( _processorName + "_" + i->second->name(), 
				      i->second->value() );
	}
      }

    } 
    
    void TimePixEventHistoProcessor::processEvent( LCEvent * evt )
    { 

	streamlog_out(MESSAGE1)<<"processEvent( "<<evt->getEventNumber()<<" )"<<std::endl;

	LCCollection* TrackerDataCol;
	try
	{
	    TrackerDataCol = evt->getCollection( _inputColName );
	}
	catch(lcio::DataNotAvailableException)
	{
	    streamlog_out(WARNING0) << "Empty event: no zero suppressed TimePix data  in event " 
				    <<evt->getEventNumber()<<std::endl;
	    return;
	}
	
	int plotMode;

	if ( _mode == "TOT" ) 
	    plotMode = TimePixPixelMode::TOT;
	else if ( _mode == "TIMEPIX" )
	    plotMode = TimePixPixelMode::TIMEPIX;
	else if ( _mode == "MEDIPIX" )
	    plotMode = TimePixPixelMode::MEDIPIX;
	else if ( _mode == "ONEHIT" )
	    plotMode = TimePixPixelMode::ONEHIT;
	else
	    plotMode = -1; // plot all modes
	    
	streamlog_out(DEBUG4)<<"plotmode =  "<<plotMode<<std::endl;
	
//	std::stringstream sEvt;
//	sEvt << evt->getEventNumber();
//	
//	std::string histoname = "TimePixEvt" + sEvt.str();
//	
	//clean up all histograms of the previous event.
	// The destructor of the AIDAHistogram does this automatically.
	_eventHistos.clear();
//	    
//	eventHisto = AIDAProcessor::histogramFactory(this)->
//	    createHistogram2D( histoname.c_str(), histoname.c_str(), 
//			       _nColumn, 0, _nColumn, _nRow, 0, _nRow );
//
	streamlog_out(DEBUG4)<<"loop over "<<TrackerDataCol->getNumberOfElements()
			     <<" Elements in inputCol"<<std::endl;
	
	for( int i = 0; i < TrackerDataCol->getNumberOfElements(); i++ )
	{
	    streamlog_out(DEBUG4)<<i<<" . "<<std::flush;

	    TrackerData* Data = 
		dynamic_cast<TrackerData*>( TrackerDataCol->getElementAt( i ) ) ;
	    
	    if ( Data != 0 ) {
		
		int cellid0 = Data->getCellID0();
		int chipID = Data->getCellID1();

		// Try to get a histogram for this chip from the map. If it does not exist
		// create it.
		AIDA::IHistogram2D* eventHisto = _eventHistos[chipID].histo;

		if ( eventHisto == 0 )
		{
		    // create the histogram name from the base name, the event number, cellID0 and chipID
		    std::stringstream histoNameStream;
		    histoNameStream << _histoBaseName <<"_"<< evt->getEventNumber()
				    << "_"<<chipID;

		    // request a histogram from the factory
		    eventHisto = AIDAProcessor::histogramFactory(this)->
			createHistogram2D( histoNameStream.str() ,  histoNameStream.str(), 
					   _nColumn, 0, _nColumn, _nRow, 0, _nRow );

		    // assign pointer to the map for propper cleanup
		    _eventHistos[chipID].histo=eventHisto;
		}

		int x = cellid0%_nColumn;
		int y = static_cast<int>((cellid0-x)/_nColumn)+1;
	  
		FloatVec ChargeVec  = Data->getChargeValues();
		
		for ( FloatVec::iterator sv = ChargeVec.begin(); sv != ChargeVec.end(); sv++ ) 
		{
		    TimePixMap::const_iterator modeIter = 
			_timePixMap->find( TimePixPixelMode::PixelKey( chipID, cellid0 ) );

		    if ( modeIter == _timePixMap->end() )
		    {
			// The pixel was not found in the mode map
			// Give an error message and throw an exception
			streamlog_out(ERROR) << "Pixel key (chipID "<< chipID 
					     << ", pixelID "<< cellid0 
					     << " could not be found in Timepix mode map!"
					     << std::endl;

			throw lcio::DataNotAvailableException("Pixel key not found in mode map");
		    }

		    int mode = (modeIter->second)->getPixelMode();
		    
		    if ( ( mode == plotMode ) || ( plotMode == -1 ) )
			eventHisto->fill( x, y, *sv );
		    
		    cellid0++;
		    x++;
		    if ( x == _nColumn ) {
			x = 0;
			y++;
		    }
		}
		
		
	    } //if ( Data != 0 )

	
    }

    streamlog_out(DEBUG4)<<std::endl;
	
	

    streamlog_out(DEBUG4)<<"end of processEvent( "<<evt->getEventNumber()<<" )"<<std::endl;
}
    
    
void TimePixEventHistoProcessor::check( LCEvent * evt ) { 



    }

    
    
    void TimePixEventHistoProcessor::end(){ 
      /*	
      streamlog_out(MESSAGE4) << "TimePixEventHistoProcessor::end()  " << name() 
		<< " processed " << evt->getEventNumber() 
		<< " events in " << _nRun << " runs "
		<< std::endl ;
      */	

    }
    
    TimePixEventHistoProcessor::AIDAHistogram::AIDAHistogram()
    : histo(0) // this is the main reason for this class: the pointer is initialised to 0
    { }

    TimePixEventHistoProcessor::AIDAHistogram::~AIDAHistogram()
    {
	// delete the corresponding histogram
	// Use the static instance of TimePixEventHistoProcessor to tell the 
	// AIDAProcessor which type of processor to use
	AIDAProcessor::histogramFactory(&aTimePixEventHistoProcessor)->destroy( histo );
    }
    
}
